# This module implements classes that represent force fields
# for bonded interactions.
#
# Written by Konrad Hinsen
# last revision: 1997-12-12
#

from ForceField import ForceField, ForceFieldData, ForceFieldEvaluator
import Utility
import Vector
import Numeric

#
# The base class BondedForceField provides the common
# functionality for all bonded interactions. The derived
# classes have to deal with determining functional forms
# and parameters and providing the evaluation code
#
class BondedForceField(ForceField):

    def __init__(self, name):
	ForceField.__init__(self, name)
	self.type = 'bonded'

    def evaluator(self, universe, subset1, subset2, global_data):
	data = ForceFieldData()
	if subset1 is not None:
	    label1 = Utility.uniqueAttribute()
	    label2 = Utility.uniqueAttribute()
	    for atom in subset1.atomList():
		setattr(atom, label1, None)
	    for atom in subset2.atomList():
		setattr(atom, label2, None)
	for o in universe:
	    for bu in o.bondedUnits():
		if not hasattr(bu, 'bonds'): continue
		options = {'bonds': 1, 'bond_angles': 1,
			   'dihedrals': 1, 'impropers': 1}
		self.getOptions(bu, options)
		if options['bonds']:
		    if subset1 is None:
			for bond in bu.bonds:
			    self.addBondTerm(data, bond, bu, global_data)
		    else:
			for bond in bu.bonds:
			    atoms = [bond.a1, bond.a2]
			    if _checkSubset(atoms, label1, label2):
			        self.addBondTerm(data, bond, bu, global_data)
		if options['bond_angles']:
		    if subset1 is None:
			for angle in bu.bonds.bondAngles():
			    self.addBondAngleTerm(data, angle, bu, global_data)
		    else:
			for angle in bu.bonds.bondAngles():
			    atoms = [angle.a1, angle.a2, angle.ca]
			    if _checkSubset(atoms, label1, label2):
			        self.addBondAngleTerm(data, angle, bu,
						      global_data)
		d = options['dihedrals']
		i = options['impropers']
		if d or i:
		    if subset1 is None:
			for angle in bu.bonds.dihedralAngles():
			    if angle.improper and i:
				self.addImproperTerm(data, angle, bu,
						     global_data)
			    elif not angle.improper and d:
				self.addDihedralTerm(data, angle, bu,
						     global_data)
		    else:
			for angle in bu.bonds.dihedralAngles():
			    atoms = [angle.a1, angle.a2, angle.a3, angle.a4]
			    if _checkSubset(atoms, label1, label2):
			        if angle.improper and i:
				    self.addImproperTerm(data, angle, bu,
							 global_data)
				elif not angle.improper and d:
				    self.addDihedralTerm(data, angle, bu,
							 global_data)
	if subset1 is not None:
	    for atom in subset1.atomList():
		delattr(atom, label1)
	    for atom in subset2.atomList():
		delattr(atom, label2)
	self.data = data
	data.universe = universe
	global_data.add('initialized', 'bonded')
	return self.evalObject(global_data)

    def bondedForceFields(self):
	return [self]

    # The following methods must be overridden by derived classes.
    def addBondTerm(self, data, bond, object, global_data):
	raise AttributeError
    def addBondAngleTerm(self, data, angle, object, global_data):
	raise AttributeError
    def addDihedralTerm(self, data, dihedral, object, global_data):
	raise AttributeError
    def addImproperTerm(self, data, improper, object, global_data):
	raise AttributeError
    def evalObject(self, global_data):
	raise AttributeError

    def defaultBondLength(self,bond,object):
	raise AttributeError
    def defaultBondAngle(self,bond1,bond2,object):
	raise AttributeError

    # The following methods are recommended for derived classes.
    # They allow to read out the force field specification, e.g. for
    # interfacing to other programs.
    def bonds(self, global_data):
	raise AttributeError
    def angles(self, global_data):
	raise AttributeError
    def dihedrals(self, global_data):
	raise AttributeError

# Check if an energy term matches the specified atom subset
def _checkSubset(atoms, label1, label2):
    s1 = 0
    s2 = 0
    for a in atoms:
	flag = 0
	if hasattr(a, label1):
	    s1 = 1
	    flag = 1
	if hasattr(a, label2):
	    s2 = 1
	    flag = 1
	if not flag:
	    return 0
    return s1 and s2

#
# A simple bonded ff for demonstrational purposes.
# It uses "default" (i.e. nonsense) parameters.
#
class DemoBondedFF(BondedForceField):

    def __init__(self):
	BondedForceField.__init__(self, 'demo_bonded')
    
    def addBondTerm(self, data, bond, object, global_data):
	data.add('bonds', (bond.a1.index, bond.a2.index, 1., 1.))
	global_data.add('excluded_pairs', (bond.a1.index, bond.a2.index))
    def addBondAngleTerm(self, data, angle, object, global_data):
	data.add('angles', (angle.a1.index, angle.ca.index,
				 angle.a2.index, 1.57079, 1.))
	global_data.add('excluded_pairs', (angle.a1.index, angle.a2.index))
    def addDihedralTerm(self, data, dihedral, object, global_data):
	pass
    def addImproperTerm(self, data, improper, object, global_data):
	pass

    def evalObject(self, global_data):
	return BondedFFEvaluator(self.data)


#
# Force field evaluator in Python. For efficiency, force field
# evaluators are normally written in C. But a simple Python version
# using automatic derivatives is a great help in debugging the C
# version!
#
from ForceField import addToGradients, addToForceConstants

class BondedFFTestEvaluator(ForceFieldEvaluator):

    def __init__(self, data, global_data):
	self.data = data
	self.last_energy = 0.

    name = 'bonded'

    def __call__(self, coordinates, gradients, force_constants,small_change=0):
	energy = 0.
	if force_constants is not None:
	    print 'test force_constants'
	    for b in self.data.get('bonds'):
		e, g, fc = bond_e_g_fc(Vector.Vector(coordinates[b[0]]),
				       Vector.Vector(coordinates[b[1]]),
				       b[2], b[3])
		energy = energy + e
		addToGradients(gradients, b[:2], g)
		addToForceConstants(force_constants, b[:2], fc)
	    for a in self.data.get('angles'):
		e, g, fc = angle_e_g_fc(Vector.Vector(coordinates[a[0]]),
					Vector.Vector(coordinates[a[1]]),
					Vector.Vector(coordinates[a[2]]),
					a[3], a[4])
		energy = energy + e
		addToGradients(gradients, a[:3], g)
		addToForceConstants(force_constants, a[:3], fc)
	    for d in self.data.get('dihedrals'):
		e, g, fc = dihedral_e_g_fc(Vector.Vector(coordinates[d[0]]),
					   Vector.Vector(coordinates[d[1]]),
					   Vector.Vector(coordinates[d[2]]),
					   Vector.Vector(coordinates[d[3]]),
					   d[4], d[5], d[6])
		energy = energy + e
		addToGradients(gradients, d[:4], g)
		addToForceConstants(force_constants, d[:4], fc)
	elif gradients is not None:
	    print 'test gradients'
	    for b in self.data.get('bonds'):
		e, g = bond_e_g(Vector.Vector(coordinates[b[0]]),
				Vector.Vector(coordinates[b[1]]),
				b[2], b[3])
		energy = energy + e
		addToGradients(gradients, b[:2], g)
	    for a in self.data.get('angles'):
		e, g = angle_e_g(Vector.Vector(coordinates[a[0]]),
				 Vector.Vector(coordinates[a[1]]),
				 Vector.Vector(coordinates[a[2]]), a[3], a[4])
		energy = energy + e
		addToGradients(gradients, a[:3], g)
	    for d in self.data.get('dihedrals'):
		e, g = dihedral_e_g(Vector.Vector(coordinates[d[0]]),
				    Vector.Vector(coordinates[d[1]]),
				    Vector.Vector(coordinates[d[2]]),
				    Vector.Vector(coordinates[d[3]]),
				    d[4], d[5], d[6])
		energy = energy + e
		addToGradients(gradients, d[:4], g)
	else:
	    print 'test other'
	    for b in self.data.get('bonds'):
		energy = energy + bond_energy(Vector.Vector(coordinates[b[0]]),
					      Vector.Vector(coordinates[b[1]]),
					      b[2], b[3])
	    for a in self.data.get('angles'):
		energy = energy + angle_energy(Vector.Vector(coordinates[a[0]]),
					       Vector.Vector(coordinates[a[1]]),
					       Vector.Vector(coordinates[a[2]]),
					       a[3], a[4])
	    for d in self.data.get('dihedrals'):
		energy = energy + \
			 dihedral_energy(Vector.Vector(coordinates[d[0]]),
					 Vector.Vector(coordinates[d[1]]),
					 Vector.Vector(coordinates[d[2]]),
					 Vector.Vector(coordinates[d[3]]),
					 d[4], d[5], d[6])
	self.last_energy = energy
	return energy

#
# Potential energy terms.
#
from Potential import PotentialWithGradients, \
     PotentialWithGradientsAndForceConstants

def bond_energy(r1, r2, l0, k):
    dr = r2 - r1
    return k*(dr.length()-l0)**2

def angle_energy(r1, r2, r3, phi0, k):
    n1 = (r1-r2).normal()
    n2 = (r3-r2).normal()
    return k*(Numeric.arccos(n1*n2)-phi0)**2

def dihedral_energy(r1, r2, r3, r4, n, delta, k):
    r12 = r2-r1
    r23 = r3-r2
    r34 = r4-r3
    n1 = r12.cross(r23).normal()
    n2 = r23.cross(r34).normal()
    phi = Numeric.arccos(n1*n2)
    return k*(1.+Numeric.cos(n*phi-delta))

bond_e_g = PotentialWithGradients(bond_energy)
angle_e_g = PotentialWithGradients(angle_energy)
dihedral_e_g = PotentialWithGradients(dihedral_energy)

bond_e_g_fc = PotentialWithGradientsAndForceConstants(bond_energy)
angle_e_g_fc = PotentialWithGradientsAndForceConstants(angle_energy)
dihedral_e_g_fc = PotentialWithGradientsAndForceConstants(dihedral_energy)

#
# Interface to the fast force field evaluator (written in C).
# Currently only force field terms used in the Amber potential are
# implemented.
#

try:

    from _forcefield import HarmonicBondEvaluator, HarmonicAngleEvaluator, \
	 CosineDihedralEvaluator, CombinedEvaluator

    def BondedFFEvaluator(data, global_data):
	f, p = data.universe.CdistanceFunction()
	if f is None:
	    raise TypeError, "Universe definition incomplete"
	eval_list = []
	bonds = data.get('bonds')
	if bonds:
	    indices = Numeric.array(map(lambda b: b[:2], bonds))
	    parameters = Numeric.array(map(lambda b: b[2:], bonds))
	    eval_list.append(HarmonicBondEvaluator(f, p, indices, parameters))
	angles = data.get('angles')
	if angles:
	    indices = Numeric.array(map(lambda a: a[:3], angles))
	    parameters = Numeric.array(map(lambda a: a[3:], angles))
	    eval_list.append(HarmonicAngleEvaluator(f, p, indices, parameters))
	dihedrals = data.get('dihedrals')
	if dihedrals:
	    indices = Numeric.array(map(lambda d: d[:4], dihedrals))
	    parameters = Numeric.array(map(_dihedral_parameters, dihedrals))
	    eval_list.append(CosineDihedralEvaluator(f, p, indices, parameters))
	return CombinedEvaluator(Numeric.array(eval_list))

    def _dihedral_parameters(p):
	return [p[4], Numeric.cos(p[5]), Numeric.sin(p[5]), p[6]]

except ImportError:

    BondedFFEvaluator = BondedFFTestEvaluator
