# This module implements classes that represent lists of bonds,
# bond angles, and dihedral angles.
#
# Written by Konrad Hinsen
# last revision: 1997-11-1
#

from UserList import UserList
from DictWithDefault import DictWithDefault
import Database, Utility, Vector
from Units import deg
from umath import pi

BondOrderNotSpecified = 0

def findCommonAtom(bond1, bond2):
    if bond1.hasAtom(bond2.a1):
	return bond2.a1
    if bond1.hasAtom(bond2.a2):
	return bond2.a2
    raise ValueError,'no common atom between ' + `bond1` + ' and ' + `bond2`

_bondDatabase = {}

#
# Bonds
#
# Bond objects are created from the specifications in the data base.
#
class Bond:

    def __init__(self, blueprint, memo = None):
	if blueprint is not None:
	    if type(blueprint) is type(()):
		self.a1 = blueprint[0]
		self.a2 = blueprint[1]
		try:
		    self.order = blueprint[2]
		except IndexError:
		    self.order = BondOrderNotSpecified
	    else:
		self.a1 = Database.instantiate(blueprint.a1, memo)
		self.a2 = Database.instantiate(blueprint.a2, memo)
		self.order = blueprint.order
	    if id(self.a2) < id(self.a1):
		self.a1, self.a2 = self.a2, self.a1
	    key = (id(self.a1), id(self.a2))
	    if _bondDatabase.has_key(key):
		print 'dup',hex(id(self.a1)),hex(id(self.a2)),self.a1,self.a2
		Utility.warning('Bond(%s, %s) duplicated' % (`self.a1`, `self.a2`))
		raise RuntimeError
	    _bondDatabase[key] = self
	    if self.a1 is not None and self.a2 is not None and self.a1.name == 'C3' and self.a2.name == 'H1':
		print 'strange bond',self

    blueprintclass = Database.BlueprintBond
    is_bond = 1

    def __getinitargs__(self):
	return (None,)

    def __repr__(self):
	if self.order != BondOrderNotSpecified:
	    ostr = ', ' + `self.order`
	else:
	    ostr = ''
	return 'Bond(' + `self.a1` + ', ' + `self.a2` + ostr + ')'
    __str__ = __repr__

#    def __cmp__(self,other):
#	if self.a1 is other.a1:
#	    if self.a2 is other.a2:
#		return 0
#	    return 1
#	return 1

    def delete(self):
	key = (id(self.a1), id(self.a2))
	if _bondDatabase.has_key(key):
	    del _bondDatabase[key]
	else: print 'Bond.delete not found',self

    def removeAndConvertToDangling(self, remove_atom):
	self.direction = self.asVector(self.otherAtom(remove_atom))
	self.delete()
	if remove_atom is self.a1:
	    self.a1 = self.a2
	elif not (remove_atom is self.a2):
	    raise ValueError
	self.a2 = None
	self.__class__ = DanglingBond
	self.constraint = None
	_bondDatabase[self.key()] = self

    def hasAtom(self, a):
	return a is self.a1 or a is self.a2

    def otherAtom(self, a):
	if a is self.a1:
	    return self.a2
	elif a is self.a2:
	    return self.a1
	else:
	    raise ValueError, 'atom not in bond'

    def otherAtomFromId(self, a_id):
	if a_id == id(self.a1):
	    return self.a2
	elif a_id == id(self.a2):
	    return self.a1
	else:
	    msg = 'atom %x not in bond' % a_id
	    raise ValueError, msg

    def setOrder(self, order):
	self.order = order

    def replaceAtom(self, new_atoms):
	new_a1 = new_atoms[id(self.a1)]
	if new_a1 is None:
	    new_a1 = self.a1
	new_a2 = new_atoms[id(self.a2)]
	if new_a2 is None:
	    new_a2 = self.a2
	return generateBond((new_a1, new_a2, self.order), self.__class__)

    def asVector(self,start_atom = None,configuration = None):
	if start_atom == None:
	    start_atom = self.a1	# ambiguous direction, pick one
	pos1 = start_atom.position(configuration)
	if pos1 is None:
	    raise ValueError, `start_atom`+' has no position defined'
	pos2 = self.otherAtom(start_atom).position(configuration)
	if pos2 is None:
	    raise ValueError, `self.otherAtom(start_atom)`+' has no position defined'
	return pos2 - pos1

    def _graphics(self, conf, distance_fn, model, module, options):
	objects = []
	if model != 'wireframe':
	    return []
	p1 = self.a1.position(conf)
	p2 = self.a2.position(conf)
	if p1 is not None and p2 is not None:
	    bond_vector = 0.5*distance_fn(self.a1, self.a2, conf)
	    cut = bond_vector != 0.5*(p2-p1)
	    material1 = material = module.DiffuseMaterial(self.a1.color)
	    material2 = material = module.DiffuseMaterial(self.a2.color)
	    if material1 == material2 and not cut:
		objects.append(module.Line(p1, p2, material = material1))
	    else:
		objects.append(module.Line(p1, p1+bond_vector,
					   material = material1))
		objects.append(module.Line(p2, p2-bond_vector,
					   material = material2))
	return objects

def isBond(obj):
    return hasattr(obj, 'is_bond')

Database.registerInstanceClass(Bond.blueprintclass, Bond)


def lookUpBond(atom1, atom2):		# doesn't handle DanglingBond's yet
    if id(atom2) < id(atom1) and atom2 is not None:
	atom1, atom2 = atom2, atom1
    key = (id(atom1), id(atom2))
    if _bondDatabase.has_key(key):
	return _bondDatabase[key]
    return None

def generateBond(key_tuple, gen_class = None):
    if len(key_tuple) >= 2:
	(atom1, atom2) = (key_tuple[0], key_tuple[1])
	if None in key_tuple:
	    gen_class = DanglingBond
	    if key_tuple[0] is None:
		atom1, atom2 = atom2, atom1
	elif id(atom2) < id(atom1):
	    atom1, atom2 = atom2, atom1
	key = (id(atom1), id(atom2))
    else:
	key = key_tuple
    #print gen_class,atom1,atom2,'has_key?',_bondDatabase.has_key(key)
    if _bondDatabase.has_key(key):
	return _bondDatabase[key]
    try:
	order = key_tuple[2]
    except IndexError:
	order = BondOrderNotSpecified
    if gen_class is None:
	return Bond((atom1, atom2, order))
    return gen_class((atom1, atom2, order))

class DanglingBond(Bond):
    def __init__(self, blueprint, memo = None):
	if blueprint is not None:
	    if type(blueprint) is type(()):
		self.a1 = blueprint[0]
		self.a2 = None
		if len(blueprint) > 1:
		    self.constraint = blueprint[1]
		else: self.constraint = None
		if len(blueprint) > 2:
		    self.direction = blueprint[2]
		else: self.direction = None
	    else:
		self.a1 = Database.instantiate(blueprint.a1, memo)
		self.a2 = None
		self.constraint = Database.instantiate(blueprint.constraint, memo)
		self.direction = Database.instantiate(blueprint.direction, memo)
	    assert(self.a1 is not None)
	    if self.constraint is not None:
		key1 = self.key()
		if _bondDatabase.has_key(key1):
		    Utility.warning('DanglingBond(%s, %s) duplicated' % (`self.a1`, `self.constraint`))
		    raise RuntimeError
		_bondDatabase[key1] = self

    blueprintclass = Database.BlueprintDanglingBond

    def setDirection(self, v):
	self.direction = v

    def setConstraint(self, c):
	self.constraint = c
	key1 = self.key()
	if _bondDatabase.has_key(key1):
	    Utility.warning('DanglingBond(%s, %s) duplicated' % (`self.a1`, `self.constraint`))
	    raise RuntimeError
	_bondDatabase[key1] = self

    def fill(self, atom):
	self.a2 = atom

    def __repr__(self):
	return 'DanglingBond(' + `self.a1` + ', ' + `self.constraint` + ')'
    __str__ = __repr__

    def __cmp__(self,other):
	if self.a1 is other.a1:
	    if self.a2 is other.a2:
		if (self.constraint and not other.constraint) or\
		   (not self.constraint and other.constraint):
		    return 1
		if self.direction is other.direction and\
		   self.constraint is other.constraint:
		    return 0
	return 1

    def key(self, new_atom = None, c = None):
	if c is None:
	    c = self.constraint
	if new_atom is not None:
	    a = new_atom
	else: a = self.a1
	if isDihedral(c):
	    return (id(a), id(c.a1), id(c.a2), id(c.a3), id(c.a4), `c.angle`)
	return (id(a),`c`)

    def replaceAtom(self, new_atoms):
	c = self.constraint
	if isDihedral(c) and (new_atoms[id(c.a1)] or new_atoms[id(c.a2)] or\
			      new_atoms[id(c.a3)] or new_atoms[id(c.a4)]):
	    c = c.replaceAtom(new_atoms)
	new_a1 = new_atoms[id(self.a1)]
	if new_a1:
	    key2 = self.key(new_a1, c)
	else: key2 = self.key(self.a1, c)
	if _bondDatabase.has_key(key2):
	    r = _bondDatabase[key2]
	    r.constraint = c
	elif new_a1:
	    r = DanglingBond((new_a1,c))
	else:
	    r = DanglingBond((self.a1,c))
	key1 = r.key()
	r.direction = self.direction
	return r

    def DanglingPosition(self):
	try:
	    return self.a1.position() + self.direction
	except (AttributeError, TypeError):
	    return None

    def HookLabel(self):
	if hasattr(self, 'hook_label'):
	    #print 'get label',hex(id(self)),self.hook_label
	    return self.hook_label
	return ""

    def setHookLabel(self, label):
	self.hook_label = label
	#print ' '*len(label),'set label',hex(id(self)),self.hook_label

    def reconcile(self, other):	# ugly hack to deal with inability
				# to prevent dup instances of a DanglingBond
	print 'reconcile',hex(id(self)),hex(id(other)),hasattr(self,'hook_label'),other.hook_label
	#self.hook_label = other.hook_label

    def asVector(self, start_atom = None, configuration = None):
	if Vector.isVector(self.direction):
	    return self.direction
	return None

Database.registerInstanceClass(DanglingBond.blueprintclass, DanglingBond)

def isDanglingBond(obj):
    return hasattr(obj, 'blueprintclass') and obj.__class__ is DanglingBond

#
# Bond angles
#
class BondAngle:

    def __init__(self, b1, b2, ca = None):
	self.b1 = b1 # bond 1
	self.b2 = b2 # bond 2
	if ca is None:
	    ca = findCommonAtom(b1, b2)
	self.ca = ca # common atom
	if id(self.b2) < id(self.b1):
	    self.b1, self.b2 = self.b2, self.b1
	self.a1 = b1.otherAtom(ca)
	self.a2 = b2.otherAtom(ca)

    def __repr__(self):
	return 'BondAngle(' + `self.a1` +',' + `self.ca` +','+ `self.a2` +')'

    def radians(self):
	return self.b1.asVector(self.ca).angle(self.b2.asVector(self.ca))

    def hasBond(self, b):
	return b is self.b1 or a is self.b2

    def hasAtom(self, a):
	return a is self.a1 or a is self.a2 or a is self.ca

    def otherAtom(self, bond):
	if bond is self.b1:
	    return self.b2.otherAtom(self.ca)
	elif bond is self.b2:
	    return self.b1.otherAtom(self.ca)
	else:
	    raise ValueError, 'bond not in bond angle'

    def otherBond(self, bond):
	if bond == self.b1:
	    return self.b2
	elif bond == self.b2:
	    return self.b1
	else:
	    print hex(id(bond)),'isnt',hex(id(self.b1)),hex(id(self.b2))
	    print bond,'isnt',self.b1,self.b2
	    raise ValueError, 'bond not in bond angle'

    def replaceAtom(self, new_atoms):
	new_ca = new_atoms[id(self.ca)]
	if new_ca is None:
	    new_ca = self.ca
	b1 = self.b1
	if new_ca or new_atoms[id(b1.a1)] or new_atoms[id(b1.a2)]:
	    b1 = b1.replaceAtom(new_atoms)
	b2 = self.b2
	if new_ca or new_atoms[id(b2.a1)] or new_atoms[id(b2.a2)]:
	    b2 = b2.replaceAtom(new_atoms)
	return BondAngle(b1, b2, new_ca)

    def replaceBond(self, new_bonds):
	b1 = new_bonds[id(self.b1)]
	b2 = new_bonds[id(self.b2)]
	if b1 is None:
	    b1 = self.b1
	if b2 is None:
	    b2 = self.b2
	return BondAngle(b1, b2, self.ca)

def calcDihedralAngle(v1, v3, cv):
    normal1 = v1.cross(cv)
    normal2 = cv.cross(v3)
    try:
	angle = normal2.angle(normal1) - pi
    except ZeroDivisionError:
	print 'ZeroDivisionError',normal1,normal2,cv,v3
	return None
    #print 'calcDihedralAngle',normal1.cross(normal2)*cv,\
    # normal1.cross(normal2).angle(cv)/deg,angle/deg
    if normal1.cross(normal2) * cv > 0:
	angle = -angle
    return angle

#
# Dihedral angles
#
class DihedralAngle:

    is_dihedral = 1

    def __init__(self, ba1, ba2, cb, angle = None):
	self.ba1 = ba1 # bond angle 1
	self.ba2 = ba2 # bond angle 2
	# cb is the common bond, i.e. the central bond for a proper dihedral
	if id(self.ba2) < id(self.ba1):
	    self.ba1, self.ba2 = self.ba2, self.ba1
	self.angle = angle
	self.improper = (self.ba1.ca is self.ba2.ca)
	if self.improper:
	    self.b1 = self.ba1.otherBond(cb)
	    self.b2 = cb
	    self.b3 = self.ba2.otherBond(cb)
	    self.a1 = self.ba1.ca # central atom
	    self.a2 = self.b1.otherAtom(self.ba1.ca)
	    self.a3 = cb.otherAtom(self.ba1.ca)
	    self.a4 = self.b3.otherAtom(self.ba2.ca)
	    # each improper dihedral will come in three versions;
	    # identify an arbitrary unique one for constructing the list
	    self.normalized = (id(cb) < id(self.b1) and id(cb) < id(self.b3))
	else:
	    self.b1 = self.ba1.otherBond(cb)
	    self.b2 = cb
	    self.b3 = self.ba2.otherBond(cb)
	    self.a1 = self.b1.otherAtom(self.ba1.ca)
	    self.a2 = self.ba1.ca # these two are
	    self.a3 = self.ba2.ca #   on the common bond
	    self.a4 = self.b3.otherAtom(self.ba2.ca)
	    self.normalized = 1

    def __repr__(self):
	if self.improper:
	    return 'ImproperDihedral(' + `self.a1` +','+ `self.a2` +','+ \
		   `self.a3` +','+ `self.a4` +')'
	else:
	    return 'Dihedral(' + `self.a1` +','+ `self.a2` +','+ \
		   `self.a3` +','+ `self.a4` +')'

    def otherBond(self, bond1, bond2):
	blist = (self.b1 ,self.b2, self.b3)
	ilist = [0,0,0]
	for i in range(0, 3):
	    if bond1 == blist[i] or bond2 == blist[i]: ilist[i] = 1
	if ilist.count(1) < 2:
	    msg = 'bond ' + repr(bond1) + ' or ' + repr(bond2) + ' not in dihedral' + repr(blist)
	    raise ValueError,msg
	if ilist.count(1) > 2:		# should be impossible
	    msg = 'DihedralAngle.otherBond too many matches ',ilist
	    raise ValueError,msg
	return blist[ilist.index(0)]

    def endsIn(self, a):
	return a is self.a4 or a is self.a1

    def hasAllBonds(self):
	return 1

    def calcAngle(self):
	try:
	    v1 = self.b1.asVector(self.a2)
	    cv = self.b2.asVector(self.a2)
	    v3 = self.b3.asVector(self.a3)
	except:
	    return None
	return calcDihedralAngle(v1, v3, cv)

    def setAngle(self):
	self.angle = self.calcAngle()
	return self.angle

    def constrains(self, new_bond, require_positions = 1):
	if self.angle is None:
	    return None
	try:
	    if new_bond.hasAtom(self.a4) and new_bond.hasAtom(self.a3):
		if not require_positions:
		    return self
		if Utility.isDefinedPosition(self.a1.position())\
		   and Utility.isDefinedPosition(self.a2.position())\
		   and Utility.isDefinedPosition(self.a3.position()):
		    #and not Utility.isDefinedPosition(self.a4.position()):
		    return self
	    elif new_bond.hasAtom(self.a1) and new_bond.hasAtom(self.a2):
		if not require_positions:
		    return self
		if Utility.isDefinedPosition(self.a4.position())\
		   and Utility.isDefinedPosition(self.a2.position())\
		   and Utility.isDefinedPosition(self.a3.position()):
		    #and (self.a1 is None or not Utility.isDefinedPosition(self.a1.position())):
		    return self
	except AttributeError: pass
	return None

    def hasAtom(self, a):
	return self.a1 is a or self.a2 is a or self.a3 is a or self.a4 is a

    def hasBond(self, b):
	return self.b1 is b or self.b2 is b or self.b3 is b

    def replaceAtom(self, new_atoms):
	return DihedralAngle(self.ba1.replaceAtom(new_atoms), \
			     self.ba2.replaceAtom(new_atoms), \
			     self.b2.replaceAtom(new_atoms), self.angle)

    def replaceBond(self, new_bonds):
	new_b = new_bonds[id(self.b2)]
	if new_b is None:
	    new_b = self.b2
	return DihedralAngle(self.ba1.replaceBond(new_bonds),\
			     self.ba2.replaceBond(new_bonds), new_b, self.angle)

    def fillAtom(self, atom):
	if self.a1 is None:
	    self.a1 = atom
	elif self.a4 is None:
	    self.a4 = atom
	else:
	    raise ValueError,'Neither a1 nor a4 are empty'

def isDihedral(x):
    return hasattr(x,'is_dihedral')

class DihedralConstraint(DihedralAngle):

    def __init__(self, b1, b3, cb, angle):
	if b1 is not None and cb is not None:
	    ba1 = BondAngle(b1, cb, findCommonAtom(b1, cb))
	    ba2 = BondAngle(cb, b3, findCommonAtom(cb, b3))
	    DihedralAngle.__init__(self, ba1, ba2, cb)
	else:	# b3 should always be a Bond, b1 and/or cb can be None here
	    self.b1 = None
	    self.b3 = b3
	    self.ba1 = None
	    self.b2 = cb
	    if cb is not None:
		if cb.hasAtom(b3.a1):
		    ca = b3.a1
		else:
		    ca = b3.a2
		self.ba2 = BondAngle(cb, b3, ca)
		self.a2 = cb.otherAtom(ca)
		self.a3 = ca
		self.a4 = b3.otherAtom(ca)
	    else:
		self.ba2 = None
		self.a2 = None
		self.a3 = b3.a1
		self.a4 = b3.a2
	    self.a1 = None
	    self.normalized = None
	    self.improper = 1
	self.angle = angle

    def hasAllBonds(self):
	return self.b1 is not None and self.b2 is not None and self.b3 is not None

    def __repr__(self):
	try:
	    constraint_str = '%f*deg' % self.angle/deg
	except:
	    constraint_str = `self.angle`
	return 'DihedralConstraint(' + `self.a1` +','+ `self.a2` +','+ \
	       `self.a3` +','+ `self.a4` +','+ constraint_str +')'
    __str__ = __repr__

#
# Dihedral is mainly just different way to construct a DihedralAngle
#

class Dihedral(DihedralAngle):
    def __init__(self, blueprint, memo = None):
	if blueprint is not None:
	    if type(blueprint) is type(()) or type(blueprint) is type([]):
		if len(blueprint) == 4 and hasattr(blueprint[0],'a1'):
		    cb = blueprint[2]
		    ca1 = findCommonAtom(blueprint[0], cb)
		    ba1 = BondAngle(blueprint[0], cb, ca1)
		    ca2 = findCommonAtom(blueprint[1], cb)
		    ba2 = BondAngle(blueprint[1], cb, ca2)
		    angle = blueprint[3]
		else:
		    a1 = blueprint[0]
		    a2 = blueprint[1]
		    a3 = blueprint[2]
		    a4 = blueprint[3]
		    angle = blueprint[4]
		    cb = generateBond((a2, a3))
		    ba1 = BondAngle(generateBond((a1, a2)), cb, a2)
		    ba2 = BondAngle(cb, generateBond((a3, a4)), a3)
	    else:
		a2 = Database.instantiate(blueprint.a1, memo)
		a1 = Database.instantiate(blueprint.a2, memo)
		a3 = Database.instantiate(blueprint.a3, memo)
		a4 = Database.instantiate(blueprint.a4, memo)
		angle = Database.instantiate(blueprint.angle, memo)
		cb = generateBond((a1, a3))
		ba1 = BondAngle(generateBond((a1, a2)), cb, a1)
		ba2 = BondAngle(cb,generateBond((a3, a4)), a3)
	    DihedralAngle.__init__(self, ba1, ba2, cb, angle)

    def __repr__(self):
	try:
	    angle = `self.angle/deg` + '*deg'
	except:
	    angle = 'None'
	if self.improper:
	    return 'ImproperDihedral(' + `self.a1` +','+ `self.a2` +','+ \
		   `self.a3` +','+ `self.a4` + ',' + angle + ')'
	else:
	    return 'Dihedral(' + `self.a1` +','+ `self.a2` +','+ \
		   `self.a3` +','+ `self.a4` + ',' + angle + ')'


    blueprintclass = Database.BlueprintDihedral

Database.registerInstanceClass(Dihedral.blueprintclass, Dihedral)

#
# Bond lists
#
# Bond lists can create bond angle and dihedral angle lists
# for themselves. These are cached for efficiency. The cached
# copy is deleted whenever the bond list is modified.
#
class BondList(UserList):

    def __init__(self, list):
	if list is not None:
	    UserList.__init__(self, list)
	    self._clearCache()
    def __getinitargs__(self):
	return (None,)
    def __getstate__(self):
	self._clearCache()
	return self.__dict__

    def __setitem__(self, i, item):
	self.data[i] = item
	self._clearCache()
    def __setslice__(self, i, j, list):
	if type(list) == type(self.data):
	    self.data[i:j] = list
	else:
	    self.data[i:j] = list.data
	self._clearCache()
    def __delslice__(self, i, j):
	del self.data[i:j]
	self._clearCache()
    def append(self, item):
	self.data.append(item)
	self._clearCache()
    def insert(self, i, item):
	self.data.insert(i, item)
	self._clearCache()
    def remove(self, item):
	self.data.remove(item)
	self._clearCache()

    def _clearCache(self):
	self.bond_angles = None
	self.dihedral_angles = None

    def bondAngles(self):
	if self.bond_angles is None:
	    # find all atoms that are involved in more than one bond
	    atoms = DictWithDefault([])
	    for bond in self.data:
		atoms[bond.a1] = atoms[bond.a1] + [bond]
		atoms[bond.a2] = atoms[bond.a2] + [bond]
	    angles = []
	    for atom, bonds in atoms.items():
		# each pair of bonds at the same atom defines a bond angle
		for p in Utility.pairs(bonds):
		    angles.append(BondAngle(p[0], p[1], atom))
	    self.bond_angles = BondAngleList(angles)
	return self.bond_angles

    def dihedralAngles(self):
	if self.dihedral_angles is None:
	    self.dihedral_angles = self.bondAngles().dihedralAngles()
	return self.dihedral_angles

    def bondedTo(self, atom):
	list = []
	for b in self.data:
	    if b.hasAtom(atom):
		list.append(b.otherAtom(atom))
	return list

    def bondsOf(self, atom):
	list = []
	for b in self.data:
	    if b.hasAtom(atom):
		list.append(b)
	return list

    def setBondAttributes(self):
	for b in self.data:
	    b.a1.setBondAttribute(b.a2)
	    b.a2.setBondAttribute(b.a1)

#
# Bond angle lists
#
class BondAngleList:

    def __init__(self, angles):
	self.data = angles

    def __repr__(self):
	return repr(self.data)
    def __len__(self):
	return len(self.data)
    def __getitem__(self, i):
	return self.data[i]

    def dihedralAngles(self):
	# find all bonds that are involved in more than one bond angle
	bonds = DictWithDefault([])
	for angle in self.data:
	    bonds[id(angle.b1)] = bonds[id(angle.b1)] + [angle]
	    bonds[id(angle.b2)] = bonds[id(angle.b2)] + [angle]
	dihedrals = []
	for key, angles in bonds.items():
	    if key == id(angles[0].b1):
		bond = angles[0].b1
	    else:
		bond = angles[0].b2
	    # each pair of bond angles with a common bond defines a dihedral
	    for p in Utility.pairs(angles):
		d = DihedralAngle(p[0], p[1], bond)
		if d.normalized:
		    dihedrals.append(d)
	return DihedralAngleList(dihedrals)

#
# Dihedral angle lists
#
class DihedralAngleList:

    def __init__(self, dihedrals):
	self.data = dihedrals

    is_dihedralanglelist = 1

    # cache entries by bond for faster searches
    def generateMap(self):
	bond_map = {}
	for d in self.data:
	    for b in (d.b1, d.b2, d.b3):
		if not bond_map.has_key(id(b)):
		    bond_map[id(b)] = [d]
		else:
		    bond_map[id(b)].append(d)
	self.bond_map = bond_map

    def __repr__(self):
	return repr(self.data)
    def __len__(self):
	return len(self.data)
    def __getitem__(self, i):
	return self.data[i]
    def __setitem__(self, i, val):
	self.data[i] = val
    def __getslice__(self, first, last):
	return DihedralAngleList(self.data[first:last])

    def searchList(self, search_bond):
	if search_bond is None:
	    return []
	try:
	    try:
		return self.bond_map[id(search_bond)]
	    except AttributeError:
		return self.data
	except KeyError:
	    return []

    def findCenteredOn(self, bond):
	for da in self.data:
	    if bond is da.b2:
		return da
	return None

    def findConstraints(self, bond, require_positions = 1):
	result = []
	for da in self.searchList(bond):
	    if da.constrains(bond, require_positions):
		result.append(da)
	return result

    def findRotatesAbout(self, cmbond, bonds_to_atoms = None):
	for da in self.data: #self.searchList(cmbond):
	    if not hasattr(da, 'b2'):
		print 'warning - findRotatesAbout bad dihedral',da
		continue
	    if da.b2 is cmbond:
		try:
		    if bonds_to_atoms is None or \
		       ((da.b1 in bonds_to_atoms[id(da.b1.a1)]) \
			and (da.b2 in bonds_to_atoms[id(da.b2.a1)]) \
			and (da.b3 in bonds_to_atoms[id(da.b3.a1)])):
			return da
		except KeyError: pass
	return None

    def __add__(self, other):
	return DihedralAngleList(self.data+other.data)
    __radd__ = __add__

    def index(self, object):
	return self.data.index(object)

    def remove(self, object):
	self.data.remove(object)

    def append(self, object):
	self.data.append(object)

def isDihedralAngleList(obj):
    return hasattr(obj, 'is_dihedralanglelist')

from time import clock
stimes = 0

def findShortestPath(bonds_to_atoms, a_start, end,\
		     visited = None, path = None, last_atom = None, depth = 0):
    global stimes
    stimes = stimes - clock()
    depth = depth + 1
    if not visited:
	visited = {}
    if not path:
	#print a_start,end, a_start.name,end.name
	path = []
    if last_atom is not None:
	path = path + [lookUpBond(last_atom, a_start)]
    if a_start is end:
	stimes = stimes + clock()
	return path
    if not bonds_to_atoms.has_key(id(a_start)):
	stimes = stimes + clock()
	return []
    shortest = []
    for b in bonds_to_atoms[id(a_start)]:
	a = b.otherAtom(a_start)
	if not visited.has_key(id(a)) and not visited.has_key(id(b))\
	   and a not in path:
	    visited[id(a)] = 1
	    visited[id(b)] = 1
	    stimes = stimes + clock()
	    newpath = findShortestPath(bonds_to_atoms, a, end, visited,
				       path, a_start, depth)
	    stimes = stimes - clock()
	    if newpath:
		if not shortest or len(newpath) < len(shortest):
		    shortest = newpath
    stimes = stimes + clock()
    #if depth == 1: print 'findShortestPath time',stimes
    return shortest
