# This module implements classes that represent atoms, molecules, and
# complexes. They are made as copies from blueprints in the database.
#
# Written by Konrad Hinsen
# last revision: 1998-3-30
#

import Vector
from Tensor import Tensor, delta
import Bonds, Collection, Database, Utility, Visualization
from CompositeObject import CompositeChemicalObject
from Transformation import Rotation, Translation
import copy, Numeric, operator, string, types
import Joiner
import DictWithDefault
import Units
from BondDeducer import deduceBonds

#
# The base class for all chemical objects.
#
class ChemicalObject(Collection.GroupOfAtoms, Visualization.Viewable):

    def __init__(self, blueprint, memo):
	if type(blueprint) == types.StringType:
	    if not hasattr(self,'name'):
		self.name = blueprint
	    blueprint = self.blueprintclass(blueprint)
	self.type = blueprint.type
	if hasattr(blueprint, 'name'):
	    self.name = blueprint.name
	if memo is None: memo = {}
	memo[id(blueprint)] = self
	for attr in blueprint.instance:
	    setattr(self, attr,
		    Database.instantiate(getattr(blueprint, attr), memo))

    is_chemical_object = 1
    is_incomplete = 0
    is_modified = 0

    def __getinitargs__(self):
	return (None,)

    def __getattr__(self, attr):
	if attr[:1] == '_' or attr[:3] == 'is_':
	    raise AttributeError
	else:
	    return getattr(self.type, attr)

    def isSubsetModel(self):
	return 0

    def addProperties(self, properties):
	if properties:
	    for item in properties.items():
		if hasattr(self, item[0]) and item[0] != 'name':
		    raise TypeError, 'attribute '+item[0]+' already defined'
		setattr(self, item[0], item[1])

    def binaryProperty(self, properties, name, default):
	value = default
	try:
	    value = properties[name]
	    del properties[name]
	except KeyError:
	    pass
	return value

    def topLevelChemicalObject(self):
	if self.parent is None or not isChemicalObject(self.parent):
	    return self
	else:
	    return self.parent.topLevelChemicalObject()

    def universe(self):
	if self.parent is None:
	    return None
	else:
	    return self.parent.universe()

    def bondedUnits(self):
	return [self]

    def fullName(self):
	if self.parent is None or not isChemicalObject(self.parent):
	    return self.name
	else:
	    if hasattr(self,'parent_subgroup'):
		parent = self.parent_subgroup
		pathname = ''
		while parent is not None:
		    pathname = parent.name + '.' + pathname
		    try:
			parent = parent.parent
		    except AttributeError: break
		return pathname + self.name
	    if self.parent is self:
		print 'fullName',hex(id(self)),self.name
		raise RuntimeError,'fullName self-parent'
	    return self.parent.fullName() + '.' + self.name

    def description(self, index_map = None):
	tag = Utility.uniqueAttribute()
	s = self._description(tag, index_map, 1)
	for a in self.atomList():
	    delattr(a, tag)
	return s

    def __repr__(self):
	return self.__class__.__name__ + ' ' + self.fullName()
    __str__ = __repr__

    def __copy__(self):
	if self.is_incomplete:
	    raise TypeError, "Can't copy incomplete object"
	return copy.deepcopy(self, {id(self.parent): None})

# Type check

def isChemicalObject(object):
    return hasattr(object, 'is_chemical_object')

#
# The classes for atoms, groups, molecules, and complexes.
#
atime0 = 0
atime1 = 0
atime2 = 0

class Atom(ChemicalObject):

    def __init__(self, blueprint, _memo = None, **properties):
	self.name = ''
	if blueprint is not None:
	    ChemicalObject.__init__(self, blueprint, _memo)
	    self._mass = self.type.average_mass
	    self.array = None
	    self.index = None
	    if properties.has_key('position'):
		self.setPosition(properties['position'])
		del properties['position']
	    self.addProperties(properties)

    blueprintclass = Database.BlueprintAtom

    def __getstate__(self):
	state = copy.copy(self.__dict__)
	if self.array is not None:
	    state['array'] = None
	    state['pos'] = Vector.Vector(self.array[self.index,:])
	return state

    def atomList(self):
	return [self]

    def bindList(self):
	return []

    def danglingBondList(self):
	return []

    def setPosition(self, position):
	if position is None:
	    if self.array is None:
		try: del self.pos
		except AttributeError: pass
	    else:
		self.array[self.index,0] = Utility.undefined
		self.array[self.index,1] = Utility.undefined
		self.array[self.index,2] = Utility.undefined
	else:
	    if self.array is None:
		self.pos = position
	    else:
		self.array[self.index,0] = position[0]
		self.array[self.index,1] = position[1]
		self.array[self.index,2] = position[2]

    translateTo = setPosition

    def position(self, conf = None):
	if conf is None:
	    if self.array is None:
		try:
		    return self.pos
		except AttributeError:
		    return None
	    else:
		if Numeric.logical_or.reduce(
		    Numeric.greater(self.array[self.index,:],
				    Utility.undefined_limit)):
		    return None
		else:
		    return Vector.Vector(self.array[self.index,:])
	else:
	    return conf[self]

    centerOfMass = position

    def setMass(self, mass):
	self._mass = mass

    def getAtom(self, atom):
	return self

    def translateBy(self, vector):
	if self.array is None:
	    self.pos = self.pos + vector
	else:
	    self.array[self.index,0] = self.array[self.index,0] + vector[0]
	    self.array[self.index,1] = self.array[self.index,1] + vector[1]
	    self.array[self.index,2] = self.array[self.index,2] + vector[2]

    def numberOfCartesianCoordinates(self):
	return 1

    def setIndex(self, index):
	if self.index is not None and self.index != index:
	    raise ValueError, 'Wrong atom index'
	self.index = index

    def setArray(self, array, indices):
	if len(indices) == 1:
	    index = indices[0]
	else:
	    if self.index is None or self.index not in indices:
		return 0
	    index = self.index
	    indices.remove(index)
	if array is None:
	    self.index = index
	    self.array = None
	    return 1
	if self.array is None:
	    try:
		array[index,0] = self.pos[0]
		array[index,1] = self.pos[1]
		array[index,2] = self.pos[2]
	    except AttributeError:
		array[index,0] = Utility.undefined
		array[index,1] = Utility.undefined
		array[index,2] = Utility.undefined
	else:
	    array[index,0] = self.array[self.index,0] 
	    array[index,1] = self.array[self.index,1] 
	    array[index,2] = self.array[self.index,2] 
	self.array = array
	self.index = index
	try:
	    del self.pos
	except AttributeError: pass
	return 1

    def getArray(self):
	return self.array

    def setBondAttribute(self, atom):
	try:
	    self.bonded_to__.append(atom)
	except AttributeError:
	    self.bonded_to__ = [atom]

    def clearBondAttribute(self):
	try:
	    del self.bonded_to__
	except AttributeError: pass
	try:
	    del self.dangling_bonds_to__
	except AttributeError: pass

    def bondedTo(self):
	try:
	    return self.bonded_to__
	except AttributeError:
	    if self.parent is None or not isChemicalObject(self.parent):
		return []
	    else:
		return self.parent.bondedTo(self)

    def numDanglingBondsTo(self):
	try:
	    return len(self.dangling_bonds_to__)
	except AttributeError:
	    if self.parent is None or not isChemicalObject(self.parent):
		return 0
	    else:
		return self.parent.numDanglingBondsTo(self)

    def delete(self):
	if self.parent is not None:
	    self.parent._deleteAtom(self)	    

    def replaceWith(self, new_atom):
	if self.parent is not None:
	    self.parent._replaceAtom(self, new_atom)

    def cloneAndRename(self, parent, new_name):
	new_atom = Atom(None)
	new_atom.type = self.type
	new_atom._mass = self._mass
	new_atom.name = new_name
	new_atom.parent = parent
	new_atom.index = None
	new_atom.array = None
	new_atom.setPosition(self.position())
	return new_atom

    def getAtomProperty(self, atom, property, levels = None):
	if self != atom:
	    raise ValueError, "Wrong atom"
	return getattr(self, property)

    def _description(self, tag, index_map, toplevel):
	setattr(self, tag, None)
	if index_map is None:
	    index = self.index
	else:
	    index = index_map[self.index]
	if toplevel:
	    return 'A(' + `self.name` + ',' + `index` + ',' + \
		   `self.symbol` + ')'
	else:
	    return 'A(' + `self.name` + ',' + `index` + ')'

    def _graphics(self, conf, distance_fn, model, module, options):
	return []

import Hooks

class Group(CompositeChemicalObject, ChemicalObject):

    def __init__(self, blueprint, _memo = None, **properties):
	if blueprint is not None:
	    ChemicalObject.__init__(self, blueprint, _memo)
	    if hasattr(self, 'configurations'):
		properties = copy.copy(properties)
		CompositeChemicalObject.__init__(self, properties)
	    else:
		CompositeChemicalObject.init_limited(self)
		self.addProperties(properties)
	    if hasattr(self, 'dihedrals'):
		self.dihedrals = Bonds.DihedralAngleList(self.dihedrals)
		self.dihedrals.generateMap()
	    if hasattr(self, 'hooks'):
		for i in range(len(self.hooks)):
		    self.hooks[i].setGroup(self)
	else:
	    self.type = Database.ConstructedGroupType()
	    CompositeChemicalObject.init_limited(self)
	if hasattr(self, 'bonds'):
	    self.bonds = Bonds.BondList(self.bonds)
	if hasattr(self, 'danglingbonds'):
	    self.danglingbonds = Bonds.BondList(self.danglingbonds)

    blueprintclass = Database.BlueprintGroup
    is_incomplete = 1

    def clone(self):			# crude hack
	new_object = Group(None)
	new_atoms1 = DictWithDefault.DictWithDefault(None)
	for a in self.atoms:
	    new_atom = a.cloneAndRename(new_object, a.fullName())
	    new_object.addAtom(new_atom)
	    new_atoms1[id(a)] = new_atom
	new_object.bonds = self.bonds[:]
	Joiner.replaceAtom(new_object.bonds, new_atoms1)
	new_object.setBondAttributes()
	return new_object

    def bondedTo(self, atom):
	if self.parent is None or not isChemicalObject(self.parent):
	    try:
		return self.bonds.bondedTo(atom)
	    except AttributeError:
		return []
	else:
	    return self.parent.bondedTo(atom)

    def numDanglingBondsTo(self, atom):
	if self.parent is None or not isChemicalObject(self.parent):
	    try:
		return len(self.danglingbonds.bondedTo(atom))
	    except AttributeError:
		return 0
	else:
	    return self.parent.numDanglingBondsTo(atom)

    def _subunits(self):
	return self.groups

    def _descriptionSpec(self):
	return "G", None

class Molecule(CompositeChemicalObject, ChemicalObject):

    def __init__(self, blueprint, _memo = None, **properties):
	if blueprint is not None:
	    ChemicalObject.__init__(self, blueprint, _memo)
	    properties = copy.copy(properties)
	    CompositeChemicalObject.__init__(self, properties)
	else:
	    self.type = Database.ConstructedGroupType()
	    CompositeChemicalObject.init_limited(self)
	if hasattr(self, 'bonds'):
	    self.bonds = Bonds.BondList(self.bonds)

    blueprintclass = Database.BlueprintMolecule

    def bondedTo(self, atom):
	return self.bonds.bondedTo(atom)

    def numDanglingBondsTo(self, atom):
	try:
	    return len(self.danglingbonds.bondedTo(atom))
	except AttributeError:
	    return 0

    def setBondAttributes(self):
	self.bonds.setBondAttributes()

    def clearBondAttributes(self):
	for a in self.atoms:
	    a.clearBondAttribute()

    def _subunits(self):
	return self.groups

    def _descriptionSpec(self):
	return "M", None

class PDBMolecule(Molecule):
    def __init__(self, conf, **properties):
	blueprintclass = Database.BlueprintMolecule
	Molecule.__init__(self, 'dummy', None)
	self.atoms = []
	self.bonds = []
	self.dihedrals = []
	self.danglingbonds = []
	for residue in conf:
	    for a in residue:
		a1 = Atom(Database.BlueprintAtom(a.name))
		a1.name = '%s%d' % (a.name, a.number)
		a1.setPosition(a.position*Units.Ang)
		self.atoms.append(a1)
	    #self.bonds = self.bonds + residue.bonds
	    #for i in range(len(self.groups)-1):
		#self.bonds.append(Bonds.Bond((CBond, NBond)))
	if properties.has_key('forcefield'):
	    if not self.bonds:
		self.bonds = deduceBonds(self.atoms, properties['forcefield'])
	    del properties['forcefield']
	self.bonds = Bonds.BondList(self.bonds)
	self.parent = None
	self.type = None
	self.configurations = {}

	try:
	    self.name = properties['name']
	    del properties['name']
	except KeyError:
	    self.name = ''
	if properties.has_key('configuration'):
	    conf = properties['configuration']
	    conf = ConfigIO.PDBFile(conf).readSequenceWithConfiguration()
	    ConfigIO.PDBConfiguration(conf).applyTo(self, 1)
	#elif conf:
	    #ConfigIO.PDBConfiguration(conf).applyTo(self, 1)
	if properties.has_key('position'):
	    self.translateTo(properties['position'])
	    del properties['position']
	self.addProperties(properties)

class Crystal(CompositeChemicalObject, ChemicalObject):

    def __init__(self, blueprint, _memo = None, **properties):
	if blueprint is not None:
	    ChemicalObject.__init__(self, blueprint, _memo)
	    properties = copy.copy(properties)
	    CompositeChemicalObject.__init__(self, properties)
	    self.bonds = Bonds.BondList(self.bonds)

    blueprintclass = Database.BlueprintCrystal

    def _subunits(self):
	return self.groups

    def _descriptionSpec(self):
	return "X", None

class Complex(CompositeChemicalObject, ChemicalObject):

    def __init__(self, blueprint, _memo = None, **properties):
	if blueprint is not None:
	    ChemicalObject.__init__(self, blueprint, _memo)
	    properties = copy.copy(properties)
	    CompositeChemicalObject.__init__(self, properties)

    blueprintclass = Database.BlueprintComplex

    def bondedUnits(self):
	return self.molecules

    def _subunits(self):
	return self.molecules

    def _descriptionSpec(self):
	return "C", None

Database.registerInstanceClass(Atom.blueprintclass, Atom)
Database.registerInstanceClass(Group.blueprintclass, Group)
Database.registerInstanceClass(Molecule.blueprintclass, Molecule)
Database.registerInstanceClass(Crystal.blueprintclass, Crystal)
Database.registerInstanceClass(Complex.blueprintclass, Complex)
