# This module defines collections of chemical objects.
#
# Written by Konrad Hinsen
# last revision: 1998-5-30
#

import ConfigIO, Geometry, Utility, Units, Variables, Visualization
import Vector, Tensor, Quaternion, Transformation
import copy, Numeric, operator, types

#
# This class defines groups of atoms. It is used as a base class
# for anything containing atoms, including chemical objects, collections,
# universes etc., but it can't be used directly. All its subclasses
# must define a method atomList() that returns a list of all their atoms.
#
class GroupOfAtoms:

    def numberOfAtoms(self):
	return len(self.atomList())

    def degreesOfFreedom(self):
	return 3*self.numberOfAtoms()

    def atomCollection(self):
	return Collection(self.atomList())

    def atomsWithDefinedPositions(self, conf = None):
	return Collection(filter(lambda a, c=conf:
				 Utility.isDefinedPosition(a.position(c)),
				 self.atomList()))

    def mass(self):
	return reduce(operator.add, map(lambda a: a._mass, self.atomList()))

    def centerOfMass(self, conf = None):
	m = 0.
	mr = Vector.Vector(0.,0.,0.)
	for a in self.atomList():
	    m = m + a._mass
	    mr = mr + a._mass * a.position(conf)
	return mr/m

    position = centerOfMass

    def centerAndMomentOfInertia(self, conf = None):
	m = 0.
	mr = Vector.Vector(0.,0.,0.)
	t = Tensor.Tensor(3*[3*[0.]])
	for a in self.atomList():
	    ma = a._mass
	    r = a.position(conf)
	    m = m + ma
	    mr = mr + ma*r
	    r = r.asTensor()
	    t = t + ma*(r*r)
	cm = mr/m
	cm_t = cm.asTensor()
	t = t - m*(cm_t*cm_t)
	t = t.trace()*Tensor.delta - t
	return cm, t

    def boundingBox(self, conf = None):
	atoms = self.atomList()
	min = atoms[0].position(conf).array
	max = min
	for a in atoms[1:]:
	    r = a.position(conf).array
	    min = Numeric.minimum(min, r)
	    max = Numeric.maximum(max, r)
	return Vector.Vector(min), Vector.Vector(max)

    def boundingSphere(self, conf = None):
	atoms = self.atomList()
	r = Vector.Vector(0., 0., 0.)
	for a in atoms:
	    r = r + a.position(conf)
	center = r/len(atoms)
	r = 0.
	for a in self.atomList():
	    r = max(r, (a.position(conf)-center).length())
	return Geometry.Sphere(center, r)

    def rmsDifference(self, conf1, conf2 = None):
	universe = conf1.universe
	m = 0.
	rms = 0.
	for a in self.atomList():
	    ma = a._mass
	    dr = universe.distanceVector(a.position(conf1), a.position(conf2))
	    m = m + ma
	    rms = rms + ma*dr*dr
	return Numeric.sqrt(rms/m)

    def findTransformation(self, conf1, conf2 = None):
	import LinearAlgebra
	if conf2 is None:
	    conf1, conf2 = conf2, conf1
	cm1 = self.centerOfMass(conf1)
	cm2 = self.centerOfMass(conf2)
	m = 0.
	k = Numeric.zeros((4,4))
	for a in self.atomList():
	    ma = a._mass
	    r1 = a.position(conf1) - cm1
	    r2 = a.position(conf2) - cm2
	    d = (r1-r2).array
	    w = Tensor.epsilon.dot((r1+r2).asTensor()).array
	    w = Numeric.concatenate((Numeric.reshape(d, (1,3)), w))
	    d = -Numeric.concatenate((Numeric.zeros((1,1), Numeric.Float),
				      Numeric.reshape(d, (3,1))))
	    w = Numeric.concatenate((d, w), -1)
	    k = k + ma*Numeric.dot(Numeric.transpose(w), w)
	    m = m + ma
	k = k/m
	e, v = LinearAlgebra.eigenvectors(k)
	i = Numeric.argmin(e)
	if e[i] <= 0.:
	    rms = 0.
	else:
	    rms = Numeric.sqrt(e[i])
	return Transformation.Translation(cm2) * \
	       Quaternion.Quaternion(v[i]).asRotation() * \
	       Transformation.Translation(-cm1), \
	       rms

    def translateBy(self, vector):
	for a in self.atomList():
	    a.translateBy(vector)

    def translateTo(self, position):
	self.translateBy(position-self.centerOfMass())

    def normalizePosition(self):
	self.translateTo(Vector.Vector(0,0,0))

    def applyTransformation(self, t):
	for a in self.atomList():
	    a.setPosition(t(a.position()))

    def displacementUnderTransformation(self, t):
	d = Variables.ParticleVector(self.universe())
	for a in self.atomList():
	    r = a.position()
	    d[a] = t(r)-r
	return d

    def rotateAroundCenter(self, axis, angle):
	cm = self.centerOfMass()
	t = Transformation.Translation(cm) * \
	    Transformation.Rotation(axis, angle) * \
	    Transformation.Translation(-cm)
	self.applyTransformation(t)

    def rotateAroundPoint(self, axis, angle, point):
	t = Transformation.Translation(point) * \
	    Transformation.Rotation(axis, angle) * \
	    Transformation.Translation(-point)
	self.applyTransformation(t)

    def rotateAroundOrigin(self, axis, angle):
	self.applyTransformation(Transformation.Rotation(axis, angle))

    def writeToFile(self, filename, configuration = None, format = None):
	file = ConfigIO.OutputFile(filename, format)
	file.write(self, configuration)
	file.close()

    def view(self, configuration = None, format = 'pdb'):
	Visualization.viewConfiguration(self, configuration, format)

    def temperature(self, velocities = None):
	if velocities is None:
	    velocities = self.atomList()[0].universe().velocities()
	kinetic_energy = 0.
	atoms = self.atomList()
	natoms = 0
	for a in atoms:
	    v = velocities[a]
	    m = a._mass
	    if m > 0.:
		kinetic_energy = kinetic_energy + m*(v*v)
		natoms = natoms + 1
	return kinetic_energy/(3*natoms*Units.k_B)

    def momentum(self, velocities = None):
	if velocities is None:
	    velocities = self.atomList()[0].universe().velocities()
	p = Vector.Vector(0., 0., 0.)
	for a in self.atomList():
	    p = p + a._mass*velocities[a]
	return p

    def angularMomentum(self, velocities = None, conf = None):
	if velocities is None:
	    velocities = self.atomList()[0].universe().velocities()
	cm = self.centerOfMass(conf)
	l = Vector.Vector(0., 0., 0.)
	for a in self.atomList():
	    l = l + a._mass*a.position(conf).cross(velocities[a])
	return l

    def universe(self):
	atoms = self.atomList()
	if not atoms:
	    return None
	universe = atoms[0].universe()
	for a in atoms[1:]:
	    if a.universe() is not universe:
		return None
	return universe

    def forcefield(self):
	ff = self.universe().forcefield()
	if ff is None:
	    raise ValueError, "no force field defined"
	return ff

    def charge(self):
	return self.forcefield().charge(self)

    def dipole(self, reference = None):
	return self.forcefield().dipole(self, reference)

#
# This class defines a general collection that can contain
# chemical objects and other collections.
#
import ChemicalObjects

class Collection(GroupOfAtoms, Visualization.Viewable):

    def __init__(self, *args):
	self.objects = []
	self.addObject(args)

    is_collection = 1

    def addObject(self, object):
	if ChemicalObjects.isChemicalObject(object):
	    self.addChemicalObject(object)
	elif isCollection(object):
	    self.addChemicalObjectList(object.objectList())
	elif isSequenceObject(object):
	    if object and ChemicalObjects.isChemicalObject(object[0]):
		self.addChemicalObjectList(list(object))
	    else:
		for o in object:
		    self.addObject(o)
	else:
	    raise TypeError, 'Wrong object type in collection'

    def addChemicalObject(self, object):
	self.objects.append(object)

    def addChemicalObjectList(self, list):
	self.objects = self.objects + list

    def removeObject(self, object):
	if ChemicalObjects.isChemicalObject(object):
	    self.removeChemicalObject(object)
	elif isCollection(object) or isSequenceObject(object):
	    for o in object:
		self.removeObject(o)
	else:
	    raise ValueError, 'Object not in this collection'

    def removeChemicalObject(self, object):
	try:
	    self.objects.remove(object)
	except ValueError:
	    raise ValueError, 'Object not in this collection'

    def selectShell(self, point, r1, r2=0.):
	if r1 > r2:
	    r1, r2 = r2, r1
	in_shell = []
	for o in self.objects:
	    r =  (o.position()-point).length()
	    if r >= r1 and r <= r2:
		in_shell.append(o)
	return Collection(in_shell)

    def selectBox(self, p1, p2):
	x1 = Numeric.minimum(p1.array, p2.array)
	x2 = Numeric.maximum(p1.array, p2.array)
	in_box = []
	for o in self.objects:
	    r = o.position().array
	    if Numeric.logical_and.reduce( \
		Numeric.logical_and(Numeric.less_equal(x1, r),
				    Numeric.less(r, x2))):
		in_box.append(o)
	return Collection(in_box)

    def objectList(self):
	return self.objects

    def atomList(self):
	lists = map(lambda o: o.atomList(), self.objectList())
	if lists:
	    return reduce(operator.add, lists)
	else:
	    return []

    def bondList(self):
	lists = map(lambda o: o.bondList(), self.objectList())
	if lists:
	    return reduce(operator.add, lists)
	else:
	    return []
	
    def danglingBondList(self):
	lists = map(lambda o: o.danglingBondList(), self.objectList())
	if lists:
	    return reduce(operator.add, lists)
	else:
	    return []

    def universe(self):
	if not self.objects:
	    return None
	universe = self.objects[0].universe()
	for o in self.objects[1:]:
	    if o.universe() is not universe:
		return None
	return universe

    def __len__(self):
	return len(self.objects)

    def __getitem__(self, item):
	return self.objects[item]

    def __add__(self, other):
	return Collection(self.objectList(), other.objectList())

    def map(self, function):
	list = map(function, self.objectList())
	if list and ChemicalObjects.isChemicalObject(list[0]):
	    return Collection(list)
	else:
	    return list

    def _graphics(self, conf, distance_fn, model, module, options):
	lists = []
	for o in self.objects:
	    lists.append(o._graphics(conf, distance_fn, model, module, options))
	return reduce(operator.add, lists)

    def __copy__(self):
	return self.__class__(copy.copy(self.objects))

# type check for sequence objects

def isSequenceObject(object):
    t = type(object)
    return t == types.ListType or t == types.TupleType \
	   or (t == types.InstanceType and hasattr(object, '__getitem__') \
	       and hasattr(object, '__len__'))

# type check for collections

def isCollection(object):
    return hasattr(object, 'is_collection')

#
# This class defines a partitioned collection. Such collections
# divide their objects into cubic boxes according to their positions.
# It is then possible to find potential neighbours much more efficiently.
#
class PartitionedCollection(Collection):

    def __init__(self, partition_size, *args):
	self.partition_size = 1.*partition_size
	self.undefined = []
	self.partition = {}
	self.addObject(args)

    def addChemicalObject(self, object):
	p = object.position()
	if p is None:
	    self.undefined.append(object)
	else:
	    index = self.partitionIndex(p)
	    try:
		partition = self.partition[index]
	    except KeyError:
		partition = []
		self.partition[index] = partition
	    partition.append(object)
	self.all = None

    def addChemicalObjectList(self, list):
	for object in list:
	    self.addChemicalObject(object)

    def removeChemicalObject(self, object):
	p = object.position()
	if p is None:
	    self.undefined.remove(object)
	else:
	    index = self.partitionIndex(p)
	    try:
		partition = self.partition[index]
	    except KeyError:
		raise ValueError, 'Object not in this collection'
	    try:
		partition.remove(object)
	    except ValueError:
		raise ValueError, 'Object not in this collection'
	self.all = None

    def partitionIndex(self, x):
	return (int(Numeric.floor(x[0]/self.partition_size)),
		int(Numeric.floor(x[1]/self.partition_size)),
		int(Numeric.floor(x[2]/self.partition_size)))

    def objectList(self):
	return reduce(operator.add, self.partition.values()+[self.undefined])

    def __len__(self):
	return Numeric.add.reduce(map(len, self.partition.values())) + \
	       len(self.undefined)

    def __getitem__(self, item):
	if self.all is None:
	    self.all = self.objectList()
	if item >= len(self.all):
	    self.all = None
	    raise IndexError
	return self.all[item]

    def __copy__(self):
	return self.__class__(self.partition_size, copy.copy(self.objectList()))

    def partitions(self):
	list = []
	for index, objects in self.partition.items():
	    min = Vector.Vector(index)*self.partition_size
	    max = min + Vector.Vector(3*[self.partition_size])
	    list.append((min, max, objects))
	return list

    def selectCube(self, point, edge):
	x = int(round(point[0]/self.partition_size))
	y = int(round(point[1]/self.partition_size))
	z = int(round(point[2]/self.partition_size))
	d = (Vector.Vector(x, y, z)*self.partition_size-point).length()
	n = int(Numeric.ceil((edge + d)/(2.*self.partition_size)))
	objects = []
	for nx in range(-n, n):
	    for ny in range(-n, n):
		for nz in range(-n, n):
		    try:
			objects.append(self.partition[(nx+x, ny+y, nz+z)])
		    except KeyError: pass
	return Collection(objects)

    def selectShell(self, point, min, max=0):
	if min > max:
	    min, max = max, min
	objects = []
	point = Numeric.array(list(point))
	min = min**2
	max = max**2
	box = Numeric.array(3*[self.partition_size])
	for index in self.partition.keys():
	    d = self.partition_size*Numeric.array(index) - point
	    dminmax = Numeric.array([Numeric.fabs(d),
				     Numeric.fabs(d + box)])
	    dmin = Numeric.add.reduce(Numeric.minimum.reduce(dminmax)**2)
	    dmax = Numeric.add.reduce(Numeric.maximum.reduce(dminmax)**2)
	    if dmax >= min and dmin <= max:
		objects = objects + self.partition[index]
	return Collection(objects)

    def pairsWithinCutoff(self, cutoff):
	pairs = []
	positions = {}
	for index, objects in self.partition.items():
	    pos = map(lambda o: o.position(), objects)
	    positions[index] = pos
	    for o1, o2 in Utility.pairs(map(None, objects, pos)):
		if (o2[1]-o1[1]).length() <= cutoff:
		    pairs.append((o1[0], o2[0]))
	partition_cutoff = int(Numeric.floor((cutoff/self.partition_size)**2))
	ones = Numeric.array([1,1,1])
	zeros = Numeric.array([0,0,0])
	keys = self.partition.keys()
	for i in range(len(keys)):
	    p1 = keys[i]
	    for j in range(i+1, len(keys)):
		p2 = keys[j]
		d = Numeric.maximum(abs(Numeric.array(p2)-Numeric.array(p1)) -
				    ones, zeros)
		if Numeric.add.reduce(d*d) <= partition_cutoff:
		    for o1, pos1 in map(None, self.partition[p1],
					positions[p1]):
			for o2, pos2 in map(None, self.partition[p2],
					    positions[p2]):
			    if (pos2-pos1).length() <= cutoff:
				pairs.append((o1, o2))
	return pairs

#
# A special form of partitioned collection that stores the atoms
# of all objects that are added to it.
#
class PartitionedAtomCollection(PartitionedCollection):

    def __init__(*args):
	apply(PartitionedCollection.__init__, args)

    def addChemicalObject(self, object):
	for atom in object.atomList():
	    PartitionedCollection.addChemicalObject(self, atom)

    def removeChemicalObject(self, object):
	for atom in object.atomList():
	    PartitionedCollection.removeChemicalObject(self, atom)

#
# Test code
#
if __name__ == '__main__':

    from RandomStuff import randomPointInBox
    from copy import copy

    box = PartitionedCollection(1.)

    for i in xrange(100):
	x = randomPointInBox(4., 4., 4.)
	box.addObject(ChemicalObjects.Atom('c', position = x))
