# This module implements a class that represents a set of atoms
# (normally bonded together as a moiety or a molecule).
# They are made as copies from blueprints in the database or by
# construction via the Hooks module.
#
# Written by Peter McCluskey and Konrad Hinsen

import Vector
import Units
import Bonds, Database, Utility
import operator, string, types
import umath
from RandomStuff import randomDirection
import Transformation
from Geometry import rotateDirection, Cone, _intersectConeCone,\
     Plane, Line, Circle, Sphere, vectorFromPointToPlane, rotatePoint,\
     rotatePointMatrix, createRotationMatrix, findRotationAngle,\
     angleBetweenDirections
from Hooks import Hook, HookList
try:
    import hull
    from DelaunayTriangulation import DelaunayTriangulation
except ImportError: pass
import DictWithDefault
import regsub
import pprint
from time import clock

#
# The second base class for all composite chemical objects.
#
class CompositeChemicalObject:

    def __init__(self, properties):
	self.init_limited()
	if properties.has_key('configuration'):
	    conf = properties['configuration']
	    self.configurations[conf].applyTo(self)
	    del properties['configuration']
	elif self.configurations.has_key('default'):
	    self.configurations['default'].applyTo(self)
	if properties.has_key('position'):
	    self.translateTo(properties['position'])
	    del properties['position']
	self.addProperties(properties)
	if not hasattr(self, 'bonds_to_atom'):
	    self.bonds_to_atom = {}
	for b in self.bonds:
	    if Utility.isDefinedPosition(b.a1.position()) and\
	       Utility.isDefinedPosition(b.a2.position()):
		if self.bonds_to_atom.has_key(id(b.a1)):
		    self.bonds_to_atom[id(b.a1)].append(b)
		else:
		    self.bonds_to_atom[id(b.a1)] = [b]
		    'posbond',hex(id(b.a1)),b.a1,hex(id(self))
		if self.bonds_to_atom.has_key(id(b.a2)):
		    self.bonds_to_atom[id(b.a2)].append(b)
		else:
		    self.bonds_to_atom[id(b.a2)] = [b]
		    'posbond',hex(id(b.a2)),b.a2,hex(id(self))


    def init_limited(self):
	self.atom_references = {}
	self.position_map = {}

    def atomList(self):
	return self.atoms

    def bondList(self):
	try:
	    return self.bonds
	except AttributeError:
	    return []

    def danglingBondList(self):
	try:
	    return self.danglingbonds
	except AttributeError:
	    return []

    def setPosition(self, atom, position):
	if atom.__class__ is Database.AtomReference:
	    atom = self.atoms[atom.number]
	atom.setPosition(position)

	k = self.positionMapKey(position)
	if not self.position_map.has_key(k):
	    self.position_map[k] = [atom]
	else:
	    self.position_map[k].append(atom)

    def positionMapKey(self, position):
	return (int(round(position[0]*10)),
		int(round(position[1]*10)),
		int(round(position[2]*10)))

    def setIndex(self, atom, index):
	if atom.__class__ is Database.AtomReference:
	    atom = self.atoms[atom.number]
	else:
	    self.atom_references[id(atom)] = Database.AtomReference(index)
	atom.setIndex(index)

    def getAtom(self, atom):
	if atom.__class__ is Database.AtomReference:
	    atom = self.atoms[atom.number]
	return atom

    def getReference(self, atom):
	if atom.__class__ is Database.AtomReference:
	    return atom
	try:
	    return self.atom_references[id(atom)]
	except KeyError:
	    r = Database.AtomReference(self.atoms.index(atom))
	    self.atom_references[id(atom)] = r
	    return r

    def reindex(self):
	for i in range(len(self.atoms)):
	    try:
		ref = self.atom_references[id(self.atoms[i])]
	    except KeyError:
		ref = Database.AtomReference(i)
		self.atom_references[id(self.atoms[i])] = ref
	    if ref.number != i:
		attrs = []
		for attr in ('amber_atom_type', 'amber_charge', 'mm3_atom_type', 'mm3_charge'):
		    try:
			attrs.append((attr, self.getAtomProperty(ref, attr)))
		    except AttributeError: pass
		ref.increaseBy(i - ref.number)
		#print 'reindex ',self.atoms[i].index,'to',i,self.atoms[i].fullName()
		for (attr, val) in attrs:
		    self.setAtomProperty(ref, attr, val)
		self.atoms[i].index = i
	

    def getReferenceFromIndex(self, atom, index):
	if atom.__class__ is Database.AtomReference:
	    return atom
	return Database.AtomReference(index)

    def setAtomProperty(self, atom, property, val):
	try:
	    attr = getattr(self, property)
	except AttributeError:
	    attr = {}
	    setattr(self, property, attr)
	attr[self.getReference(atom)] = val

    def getAtomProperty(self, atom, property, levels = None):
	try:
	    return getattr(self, property)[self.getReference(atom)]
	except (AttributeError, KeyError):
	    if levels is None:
		object = atom
		levels = []
		while object != self:
		    levels.append(object)
		    object = object.parent
	    if not levels:
		raise KeyError, 'Property ' + property + \
		      ' not defined for  ', `atom`
	    return levels[-1].getAtomProperty(atom, property, levels[:-1])

    def addAtom(self, atom):
	i = len(self.atoms)
	atom.setIndex(i)
	self.atom_references[id(atom)] = Database.AtomReference(i)
	self.atoms.append(atom)

    def numberOfCartesianCoordinates(self):
	n = 0
	for a in self.atoms:
	    n = n + a.numberOfCartesianCoordinates()
	return n

    def deleteUndefinedAtoms(self):
	delete = []
	for a in self.atoms:
	    if a.position() is None:
		delete.append(a)
	for a in delete:
	    a.delete()

    def constructBondsToAtom(self):
	bonds_to_atom = {}
	for a in self.atoms:
	    bonds_to_atom[id(a)] = []
	for b in self.bonds:
	    bonds_to_atom[id(b.a1)].append(b)
	    bonds_to_atom[id(b.a2)].append(b)
	self.bonds_to_atom = bonds_to_atom
	return bonds_to_atom

    def _deleteAtom(self, atom):
	self.atoms.remove(atom)
	self.is_modified = 1
	self.type = None
	self._removeAtomRefs(atom)
	if hasattr(self, 'subgroups'):
	    for subgrp in self.subgroups:
		try:
		    subgrp._deleteAtom(atom)
		except ValueError: pass

    def _removeAtomRefs(self, atom):
	if self.parent is not None:
	    self.parent._deleteAtom(atom)
	if hasattr(self, 'danglingbonds'):
	    for db in self.danglingbonds:
		c = db.constraint
		if c is not None and atom in (c.a1, c.a2, c.a3, c.a4):
		    db.constraint = None
	if hasattr(self, 'dihedrals'):
	    for d in self.dihedrals[:]:
		if atom in (d.a1, d.a2, d.a3, d.a4):
		    self.dihedrals.remove(d)
	for b in self.bonds:
	    if atom in (b.a1, b.a2):
		print hex(id(b)),'contains removed',atom

    def _deleteAtoms(self, atom_dict):
	new_list = []
	for a in self.atoms:
	    if not atom_dict.has_key(a):
		new_list.append(a)
	self.atoms = new_list
	self.is_modified = 1
	self.type = None
	self._removeAtomsRefs(atom_dict)
	if hasattr(self, 'subgroups'):
	    for subgrp in self.subgroups:
		try:
		    subgrp._deleteAtoms(atom_dict)
		except ValueError: pass

    def _removeAtomsRefs(self, atom_dict):
	#if self.parent is not None:
	#    self.parent._deleteAtoms(atom_dict)
	if hasattr(self, 'danglingbonds'):
	    for db in self.danglingbonds:
		c = db.constraint
		if c is not None and \
		   (atom_dict.has_key(c.a1) or atom_dict.has_key(c.a2) or \
		    atom_dict.has_key(c.a3) or atom_dict.has_key(c.a4)):
		    db.constraint = None
	if hasattr(self, 'dihedrals'):
	    new_dihedrals = []
	    for d in self.dihedrals:
		if not (atom_dict.has_key(d.a1) or atom_dict.has_key(d.a2) or \
			atom_dict.has_key(d.a3) or atom_dict.has_key(d.a4)):
		    new_dihedrals.append(d)
	    self.dihedrals = Bonds.DihedralAngleList(new_dihedrals)

    def _replaceAtom(self, atom, new_atom):
	pos = self.atoms.index(atom)
	print '_replaceAtom',hex(id(self)),pos,atom.fullName(),new_atom.fullName()
	self.atoms[pos] = new_atom
	self.is_modified = 1
	self.type = None
	self.setIndex(new_atom, pos)
	self._removeAtomRefs(atom)
	if self.parent is not None:
	    self.parent._replaceAtom(atom, new_atom)
	if hasattr(self, 'subgroups'):
	    for subgrp in self.subgroups:
		try:
		    subgrp._replaceAtom(atom, new_atom)
		except ValueError: pass

    def findAtomByName(self, name):
	for a1 in self.atoms:
	    if a1.fullName() == name:
		return a1
	return None

		# should this be allowed on Molecules or only on Groups?
		# result doesn't fit normal expectations for Molecule
    def removeAtoms(self, remove_list, hook_names):
	deleted_already = {}
	for i in range(len(remove_list)):
	    a = remove_list[i]
	    if type(a) is type(''):
		a_new = self.findAtomByName(a)
		if a_new is None:
		    msg = "Can't find atom with name " + a
		    raise ValueError, msg
		a = a_new
	    deleted_already[a] = 1
	    #try:
		#print 'removeAtoms',hex(id(a)),a.fullName(),len(self.bonds_to_atom[id(a)]),a.bondedTo()
	    #except (KeyError, AttributeError): pass
	    if hasattr(self, 'bonds_to_atom'):
		db_list = []
		for b in self.bonds_to_atom[id(a)][:]:
		    deleted_already[id(b)] = 1
		    #print '  removeAtoms b',hex(id(b)), b
		    other_a = b.otherAtom(a)
		    if other_a is None:
			print 'danglingbond removed'
			b.delete()
		    else:
			try:
			    del self.alt_bond_dir[id(other_a)]
			except (AttributeError, KeyError): pass
			self.bonds_to_atom[id(other_a)].remove(b)
			b.removeAndConvertToDangling(a)
			db_list.append(b)
			self.danglingbonds.append(b)
			try:
			    hook_name = hook_names[i]
			except (AttributeError, IndexError): pass
			else:
					# if multiple bonds to a, all get
					# assigned to 1 hook; should support
					# alternative
			    self.addToHook(hook_name, b)
	    self.bonds_to_atom[id(a)] = []
	    #bondedTo?
	new_dblist = []
	self._deleteAtoms(deleted_already)
	if hasattr(self, 'danglingbonds'):
	    for db in self.danglingbonds:
		if not deleted_already.has_key(db.a1):
		    new_dblist.append(db)
	self.danglingbonds = new_dblist
	_removeBonds(self, deleted_already)
	self.reindex()
	self.clearBondAttributes()
	self.setBondAttributes()

    def addToHook(self, hook_name, dbond):
	if hasattr(self, hook_name):# 
	    old_hook = getattr(self, hook_name)
	    new_hook = Hook((self, old_hook.getBonds() + [dbond]))
	    pos = self.hooks.index(old_hook)
	    self.hooks[pos] = new_hook
	else:
	    new_hook = Hook((self, [dbond]))
	    if not hasattr(self, 'hooks'):
		self.hooks = []
	    self.hooks.append(new_hook)
	setattr(self, hook_name, new_hook)

    def replaceAtoms(self, atom_list, replace_with_list, forcefield = None):
	if len(atom_list) != len(replace_with_list):
	    raise ValueError, 'replaceAtoms called with unequal size lists'
	deleted_already = {}
	for i in range(len(atom_list)):
	    alist = atom_list[i]
	    replacement = replace_with_list[i]
	    if type(alist) is type(''):
		a_new = self.findAtomByName(alist)
		if a_new is None:
		    msg = "Can't find atom with name " + alist
		    raise ValueError, msg
		alist = [a_new]
	    elif type(alist) is not type([]):
		alist = [alist]
	    sum_pos = Vector.null
	    for j in range(len(alist)):
		a = alist[j]
		sum_pos = sum_pos + a.position()
		if j:
		    deleted_already[a] = 1
		else:
		    replacement.parent = a.parent
		    if hasattr(a, 'parent_subgroup'):
			replacement.parent_subgroup = a.parent_subgroup
	        #for attr in ('amber_atom_type', 'amber_charge',\
		    #	 'mm3_atom_type', 'mm3_charge'):
		    #self.setAtomProperty(?)
	    self.setPosition(replacement, sum_pos/len(alist))
		
	    alist[0].replaceWith(replacement)
	    #? type
	    print 'replaceAtoms',hex(id(a)), a.fullName()
			# for moment, only support replacement's which
			# leave no dangling bonds
	    if hasattr(self, 'danglingbonds'):
		for db in self.danglingbonds:
		    if db.a1 in alist:
			self.danglingbonds.remove(db)
			print 'replaceAtomsdb',hex(id(db)),db
			#self.removeFromHook(hook_name, db)
	    if hasattr(self, 'bonds_to_atom'):
		#assert(len(self.bonds_to_atom[id(a)] == bond_order(replacement)))
		new_bonds = []
		bonded_to = []
		for a in alist:
		    for b in self.bonds_to_atom[id(a)]:
			other_a = b.otherAtom(a)
			bonded_to.append(other_a)
			new_atom_dict = DictWithDefault.DictWithDefault(None)
			new_atom_dict[id(a)] = replacement
			new_b = b.replaceAtom(new_atom_dict)
			new_bonds.append(new_b)
			_replaceBond(self, b, new_b)
			blist = self.bonds_to_atom[id(other_a)]
			pos = blist.index(b)
			blist[pos] = new_b
			# notify dihedrals????
		    del self.bonds_to_atom[id(a)]
		self.bonds_to_atom[id(replacement)] = new_bonds
		if forcefield is not None:
		    if replacement.symbol == 'C':# ???? ugly hack, mm3 only
			self.setType(replacement, forcefield, 1)
		    try:
			t = self.getAtomProperty(replacement,\
						 forcefield.atom_type_property)
			print 'type now',t
		    except AttributeError:
			self.autoType(replacement, forcefield, bonded_to)
		    more_bonds_needed = forcefield.bondOrder(replacement, self)\
					- len(new_bonds)
		    print 'replaceAtoms more_bonds_needed',more_bonds_needed,\
			  len(new_bonds),replacement.symbol
		    for i in range(more_bonds_needed):
			db = Bonds.DanglingBond((replacement,))
			self.danglingbonds.append(db)
	if deleted_already:
	    new_dblist = []
	    self._deleteAtoms(deleted_already)
	    if hasattr(self, 'danglingbonds'):
		for db in self.danglingbonds:
		    if not deleted_already.has_key(db.a1):
			new_dblist.append(db)
	    self.danglingbonds = new_dblist
	_removeBonds(self, deleted_already)
	self.reindex()
	self.clearBondAttributes()
	self.setBondAttributes()

    def deduceBondAnglesFromConfig(self):
	self.clearBondAttributes()
	self.setBondAttributes()
	for a in self.atoms:
	    pos = a.position()
	    if not Utility.isDefinedPosition(pos):
		continue
	    bonded = a.bondedTo()
	    if len(bonded) > 1:
		a.bond_angles = DictWithDefault.DictWithDefault(None)
		for i in range(len(bonded)):
		    posi = bonded[i].position()
		    if not Utility.isDefinedPosition(posi):
			continue
		    for j in range(i+1, len(bonded)):
			posj = bonded[j].position()
			if not Utility.isDefinedPosition(posj):
			    continue
			key = (id(bonded[i]), id(bonded[j]))
			if key[0] > key[1]:
			    key = (key[1], key[0])
			a.bond_angles[key] = (posj - pos).angle(posi - pos)
	

    def deduceDihedralsFromConfig(self):
	self.clearBondAttributes()
	self.setBondAttributes()
	for b in self.bonds:
	    a0list = b.a1.bondedTo()[:]
	    a0list.remove(b.a2)
	    a3list = b.a2.bondedTo()[:]
	    a3list.remove(b.a1)
	    if len(a0list) and len(a3list):
		for i in range(min(len(a0list), len(a3list))):
		    d = Bonds.Dihedral((a0list[i], b.a1, b.a2, a3list[i], None))
		    d.setAngle()
		    print 'append Dihedral',d.a1.name,d.a2.name,d.a3.name,d.a4.name
		    self.dihedrals.append(d)

		# the volume argument to atomsWithinRegion should be a
		# GeometricalObject3D whose enclosesPoint doesn't always
		# return 0
    def atomsWithinRegion(self, volume, bonds_crossing = None,\
			  bonds_inside = None):
	atoms_inside = []
	a_is_inside = DictWithDefault.DictWithDefault(0)
	for a in self.atomList():
	    if volume.enclosesPoint(a.position()):
		atoms_inside.append(a)
		a_is_inside[a] = 1
	if bonds_crossing is not None:
	    for b in self.bonds:
		if a_is_inside[b.a1] != a_is_inside[b.a2]:
		    bonds_crossing.append(b)
		if a_is_inside[b.a1] and a_is_inside[b.a2]:
		    bonds_inside.append(b)
	return atoms_inside

    def RotateAtom(self, a, axis, angle, rotated_atom_list):
	rotated_atoms = {}
	for a1 in rotated_atom_list:
	    rotated_atoms[id(a1)] = 1
	m = createRotationMatrix(axis, angle)
	self.RotateAtomMatrix(a, m, axis.point, rotated_atoms)

    def RotateAtomMatrix(self, a, mat, axis_point, rotated_atoms):
	if not Utility.isDefinedPosition(a.position()):
	    return
	if not rotated_atoms.has_key(id(a)):
	    #print '[%.3f,%.3f,%.3f]' % (a.position()[0],a.position()[1],a.position()[2]),
	    self.setPosition(a, rotatePointMatrix(a.position(), mat, axis_point))
	    #print 'rotate',a.fullName()[6:]#,a.position()
	    rotated_atoms[id(a)] = 1
	    for b in self.bonds_to_atom[id(a)]:
		#print a.name,'nested rotate level',b.otherAtom(a)
		self.RotateAtomMatrix(b.otherAtom(a), mat, axis_point, rotated_atoms)

    rot_cnt = 0
    def RotateAtomsOnOneSide(self, a, cb, axis, angle, exclude_atoms = None):
	self.b1times0 = self.b1times0 - clock()
	m = createRotationMatrix(axis, angle)
	rotated_atoms = {id(a) : 1}
	if exclude_atoms is not None:
	    for a1 in exclude_atoms:
		rotated_atoms[id(a1)] = 1
	#print 'bonds x',rotated_atoms,exclude_atoms,angle/Units.deg,self.rot_cnt
	for b in self.bonds_to_atom[id(a)]:
	    if b is not cb:
		#print a.name,'top rotate level',cb,b,rotated_atoms
		self.RotateAtomMatrix(b.otherAtom(a), m, axis.point,\
				      rotated_atoms)
	self.rot_cnt = self.rot_cnt + len(rotated_atoms.keys())
	#print 'arot',len(rotated_atoms.keys())
	self.b1times0 = self.b1times0 + clock()

    def rotateToMatchDihedral(self, dihedral):
	cb = dihedral.b2
	a = cb.a1
	axis = Line(a.position(), cb.asVector(a))
	current_angle = dihedral.calcAngle()
	angle = dihedral.angle - current_angle
	self.RotateAtomsOnOneSide(a, cb, axis, angle)
	#print 'current_angle',current_angle/Units.deg,'want',dihedral.angle/Units.deg,'get',dihedral.calcAngle()/Units.deg

    def listAtomsWithinDistance(self, pos, max_dist):
	bbox_min = self.positionMapKey([pos[0] - max_dist, pos[1] - max_dist, pos[2] - max_dist])
	bbox_max = self.positionMapKey([pos[0] + max_dist, pos[1] + max_dist, pos[2] + max_dist])
	result = []
	if 1 or (bbox_max[0] - bbox_min[0])*(bbox_max[1] - bbox_min[1])\
	   *(bbox_max[2] - bbox_min[2]) > len(self.atoms):
	    for a in self.atoms:
		apos = a.position()
		if apos is not None and (pos - apos).length() <= max_dist:
		    result.append(a)
	else:
	    for x in range(bbox_min[0], bbox_max[0]+1):
		for y in range(bbox_min[1], bbox_max[1]+1):
		    for z in range(bbox_min[2], bbox_max[2]+1):
			if self.position_map.has_key((x, y, z)):
			    result = result + self.position_map[(x, y, z)]
			    #print 'get pos[',(x,y,z),']',len(result)
	return result

    def _findNearestAtom(self, new_atom, new_b, forcefield):
	a_start = new_b.otherAtom(new_atom)
	old_b = self.bonds_to_atom[id(a_start)][0]
	desired_length = forcefield.defaultBondLength(new_b, self)
	angle = forcefield.defaultBondAngle(old_b, new_b, self)
	d1 = umath.cos(angle)*desired_length
	d2 = umath.sin(angle)*desired_length
	dir = old_b.asVector().normal()
	circle = Circle(a_start.position() + d1*dir, dir, d2)
	closest_atom = None
	closest_distance = Units.mega
	for a in self.listAtomsWithinDistance(a_start.position(), 0.2):
	    if not Utility.isDefinedPosition(a.position()):
		continue
	    if old_b.hasAtom(a):
		continue
	    d = circle.distanceFrom(a.position())
	    if d < closest_distance:
		closest_distance = d
		closest_atom = a
	return closest_atom

    def calcRotateAngle(self, d, a1, pos2, atom3, desired_distance):
	pos1 = a1.position()
	a4 = d.a4
	if atom3 is not None:
	    if not (atom3 is d.a3):
		a4 = d.a1
	else:
	    atom3 = d.a3
	atom3pos = atom3.position()
	normal = d.b2.asVector(d.b2.otherAtom(atom3))
	line1 = Line(atom3pos, normal)
	center = line1.projectionOf(pos1)
	circle1 = Circle(center, normal, (pos1-center).length())
	if circle1.radius < 0.001:
	    return RotationParms(d)		# give up
		# circle1 is set of possible a1 positions consistent with d
	sphere2 = Sphere(pos2, desired_distance)
	result = sphere2.intersectWith(circle1)
	a1dir = pos1 - center
	warn_msg = None
	if result is None:
	    vector_to_circle = vectorFromPointToPlane(pos2, circle1.planeOf())
	    projection = pos2 + vector_to_circle
	    v_between_centers = projection - center
	    dist_center_a2 = v_between_centers.length()
	    warn_msg = 'Warning - RotateDihedral cannot move ' + `pos1`\
			   + ' close enough to ' + `pos2`
	    try:
		rotate_angle = (projection - center).angle(a1dir)
	    except ZeroDivisionError, msg:
		return RotationParms(d)	# rotation won't do anything
	    dir = normal * a1dir.cross(projection - center)
	    for view_cmd in []:#('80,10','45,45','10,80'):
		import Gnuplot
		Gnuplot.plot(Sphere(pos2,desired_distance+0.025).coordList(),
			     circle1.coordList(), [projection, pos2],
			     threedim = 1, view = view_cmd)
	else:				# tuple of 2 possible points
	    if Vector.isVector(result[0]):# otherwise single point intersect?
		result = result[0] # pick arbitrary pt
	    rotate_angle = (result - center).angle(a1dir)
	    dir = normal * a1dir.cross(result - center)
	x = (pos2 - center).cross(pos1 - center)
	axis = Line(center, normal)
	pos_plus = rotatePoint(pos1, axis, rotate_angle)
	pos_minus = rotatePoint(pos1, axis, -rotate_angle)
	#print 'dist-',(pos_minus - pos2).length(),'dist+',(pos_plus - pos2).length(),'center dist',(sphere2.center - circle1.center).length()
	if abs((pos_minus - pos2).length() - desired_distance) \
	   < abs((pos_plus - pos2).length() - desired_distance):
	    rotate_angle = -rotate_angle
	    dist = (pos_minus-pos2).length()
	else:
	    dist = (pos_plus-pos2).length()
	error = abs(dist - desired_distance)
	print 'error %.4f desire %.4f %s %s atom moved %s atom3 %s %s %s' %\
	      (error,desired_distance,d.b2.a1.fullName()[-4:],d.b2.a2.fullName()[-4:],\
	       a1.fullName()[-4:],atom3.fullName()[-4:],hex(id(d.b2.a1)),hex(id(d.b2.a2)))
	if error > 0.075:
	    print d
	    warn_msg = 'Rotating %s by %.1f degrees about %s made it %.2f Angstroms from %s. Giving up.'\
	    % (`pos1`, rotate_angle/Units.deg, `d.b2`, dist/Units.Ang, `pos2`)
	if 0:
	    import RasMol
	    ras_session = RasMol.RasMol(self, 1)
	    ras_session.AddPolyLines(circle1.coordList())
	    ras_session.AddPolyLines([pos_plus,pos1,pos_minus,circle1.center])
	    sphere2.radius = sphere2.radius + 0.02
	    ras_session.AddPolyLines(sphere2.coordList(1))
	    ras_session.RefreshScreen()
	    ras_session.Interact()
	    raise RuntimeError

	return RotationParms(d, error, rotate_angle, axis, atom3, [], warn_msg)

    def _altRotateParms(self, d, a1, a2, atom3, desired_distance, desired_vector):
	(altdir, ca) = d.angle.altDirection()
	print '_altRotateParms', hex(id(d.angle)),d.angle,'\n', d, ca
	atoms_rotating_other_way = [d.angle.bond1.a1, d.angle.bond1.a2,\
				    d.angle.bond2.a1, d.angle.bond2.a2]
	origdir = self.bonds_to_atom[id(ca)][2].asVector(ca)
	#print 'abond',a1,atom3,atoms_rotating_other_way,self.bonds_to_atom[id(ca)][2]
	axis = Line(ca.position(), d.b2.asVector(ca))
	rotate_angle = origdir.angle(altdir)
	if not ca is atom3:
	    print 'swap',atom3,ca
	    atom3 = ca
	    (a1, a2) = (a2, a1)
	alt_pos_pos = rotatePoint(a1.position(), axis, rotate_angle)
	dist_pos = (alt_pos_pos - (a2.position() + desired_vector)).length()
	error_pos = abs(dist_pos - desired_distance)
	alt_pos_neg = rotatePoint(a1.position(), axis, -rotate_angle)
	dist_neg = (alt_pos_neg - (a2.position() + desired_vector)).length()
	error_neg = abs(dist_neg - desired_distance)
	print 'error_pos',error_pos,'error_neg',error_neg
	if error_neg < error_pos:
	    error = error_neg
	    dist = dist_neg
	    alt_pos = alt_pos_neg
	else:
	    error = error_pos
	    dist = dist_pos
	    alt_pos = alt_pos_pos
	#print 'rotate',rotate_angle/Units.deg,'about',axis,'error',error,'dist',(alt_pos - a1.position()).length(),'alt_pos',alt_pos,a2.position()
	return RotationParms(d, error, rotate_angle, axis, atom3,\
			     atoms_rotating_other_way, None)

    def bendBondsToCloseCycle(self, dpair_list, a1, a2, desired_distance,\
	      desired_vector, bonds_to_bend, exclude_atoms, forcefield):
	dist = (a1.position() - (a2.position()+desired_vector)).length()
	if not bonds_to_bend:
	    visited = {}
	    for b in self.bonds_to_atom[id(a1)]:
		if b.hasAtom(a2): # prevent findShortestPath from using this
		    visited[id(b)] = 1
	    bonds_to_bend = Bonds.findShortestPath(self.bonds_to_atom,\
						   a1, a2, visited)
	    #print 'findShortestPath',bonds_to_bend
	    if not bonds_to_bend:
		return dist
	print a1,'-->',a2,'dist',dist
	atoms_at_bends = []
	atoms_used = {}
	for b in bonds_to_bend:
	    for a in (b.a1, b.a2):
		if not atoms_used.has_key(a):
		    atoms_used[a] = 1
		    atoms_at_bends.append(a)
	for pair in dpair_list:
	    for i in (0, 1):
	      for d in pair[i]:
		  if d.hasAllBonds():
		      for j in (0, 1):
			  pivot_atom = ((d.a3, d.a4), (d.a2, d.a1))[i][j]
			  if (d.b1, d.b3)[j].hasAtom(pivot_atom):
			      angle = Bonds.BondAngle(d.b2, d.b3, d.a3)
			  else:
			      angle = Bonds.BondAngle(d.b1, d.b2, d.a2)
			  if not angle.ca in atoms_at_bends\
			     or not Utility.isDefinedPosition(angle.a1.position())\
			     or not Utility.isDefinedPosition(angle.a2.position()):
			      continue
			  rotate_angle = self.Bend1BondAngle(angle, a1, a2,\
				 pivot_atom, desired_distance,\
				 desired_vector, exclude_atoms, forcefield)
			  dist = (a1.position() - (a2.position()+desired_vector)).length()
			  v1 = (angle.a1.position() - angle.ca.position())
			  v2 = (angle.a2.position() - angle.ca.position())
			  v3 = (a1.position() - angle.ca.position())
			  v4 = ((a2.position()+desired_vector) - angle.ca.position())
			  v5 = (a2.position() - angle.ca.position())
			  if rotate_angle != 0.0:
			      print '%s rotate_angle %.2f new dist %.3f v1v2 %.1f v3 %.1f v5 %.3f'\
				    % (angle.ca.fullName()[-6:],rotate_angle/Units.deg,dist,\
				       v1.angle(v2)/Units.deg, v3.angle(v4)/Units.deg,\
				       v3.angle(v5)/Units.deg)
			  if abs(dist - desired_distance) < 0.001:
			      return dist
	if abs(dist - desired_distance) > 0.005 and len(bonds_to_bend) > 1:
	    for i in range(len(bonds_to_bend)-1):
		b = bonds_to_bend[i]
		pos = atoms_at_bends.index(b.a1)
		if pos > 0 and atoms_at_bends[pos-1] is b.a2:
		    pivot_atom = b.a2
		else:
		    pivot_atom = b.a1
		angle = Bonds.BondAngle(b, bonds_to_bend[i+1])
		if not angle.ca in atoms_at_bends:
		    continue
		rotate_angle = self.Bend1BondAngle(angle, a1, a2,\
				 pivot_atom, desired_distance,\
				 desired_vector, exclude_atoms, forcefield)
		dist = (a1.position() - (a2.position()+desired_vector)).length()
		#print 'bend to %.3f, err %.3f dist %.3f' %\
		#      (angle.radians()/Units.deg,dist-desired_distance,\
		#       (a1.position() - a2.position()).length()), angle
	return dist

    def Bend1BondAngle(self, angle, a1, a2, pivot_atom, desired_distance, \
		       desired_vector, exclude_atoms, forcefield):
	a = angle.ca
	b1vect = angle.b1.asVector(a)
	b2vect = angle.b2.asVector(a)
	a_pos = a.position()
	axis = Line(a_pos, b2vect.cross(b1vect))
	a1_pos = a1.position()
	a2_pos = a2.position() + desired_vector
	ba_plane = Plane(a_pos, angle.a1.position(), angle.a2.position())
	a1_proj = ba_plane.projectionOf(a1_pos)
	a2_proj = ba_plane.projectionOf(a2_pos)
	if (a1_proj - a_pos).length() < 1.e-9 or (a2_proj - a_pos).length() < 1.e-9:
	    print 'WARN null vector in Bend1BondAngle',a_pos,a1_proj,a2_proj
	    return 0.0
	current_angle1 = (a1_proj - a_pos).angle(a2_proj - a_pos)
	current_angle2 = b2vect.angle(b1vect)
	default_angle = forcefield.defaultBondAngle(angle.b1, angle.b2, self)
	desired_angle = _findDesiredAngle(a_pos, a1_proj, a2_proj,\
					  desired_distance)
	try:
	    print 'bend %.3f %.3f  %.3f %.3f %s %s' % (current_angle1/Units.deg,current_angle2/Units.deg,\
	      default_angle/Units.deg,desired_angle/Units.deg,a,pivot_atom)
	except TypeError: print 'type error',current_angle1,current_angle2,default_angle
	desired_change = desired_angle - current_angle1
	rotate_angle = desired_change
	pos_plus = rotatePoint(a1_pos, axis, rotate_angle)
	pos_minus = rotatePoint(a1_pos, axis, -rotate_angle)
	if (pos_plus - a2_pos).length() > (pos_minus - a2_pos).length():
	    rotate_angle = -rotate_angle
	expect_pos = rotatePoint(a1_pos, axis, rotate_angle)
	if abs(current_angle2 + rotate_angle - default_angle) < umath.pi/12:
	    pass
	elif rotate_angle < 0:
	    print 'limit-'
	    rotate_angle = default_angle - umath.pi/12 - current_angle2
	else:
	    print 'limit+'
	    rotate_angle = default_angle + umath.pi/12 - current_angle2
	print 'rotate about',pivot_atom,[a] + exclude_atoms
	self.RotateAtom(pivot_atom, axis, rotate_angle, [a] + exclude_atoms)
	if desired_vector.length() > 0:
	    new_desired = rotateDirection(desired_vector, axis, rotate_angle)
	    desired_vector.__setstate__([new_desired.x(), new_desired.y(),\
					 new_desired.z()])
	return rotate_angle

    def CycleErrorMessage(self, atom1, atom2, dist):
	path = Bonds.findShortestPath(self.bonds_to_atom, atom1, atom2)
	apath = []
	a = atom1
	for b in path:
	    a = b.otherAtom(a)
	    apath.append(a)
	msg = 'Bonds influencing their position were already positioned along '
	if len(apath) < 32:
	    path_msg = msg + 'the following path: ' + `apath`
	else:
	    path_msg = msg + 'a path with ' + `len(apath)` + ' atoms'
	return 'Unable to move atoms %s and %s into expected relative positions. They ended up %.3f nanometers apart. %s'\
	       % (atom1.fullName(), atom2.fullName(), dist, path_msg)

    rtimes0 = 0
    rtimes1 = 0
    rtimes2 = 0
    rtimes3 = 0
    rtimes4 = 0

    def RotatePartsToCloseCycle(self, a1, a2, desired_distance,\
				forcefield, bonds_to_bend):
	msg = None
	if Vector.isVector(desired_distance): # move to specified position
	    desired_vector = desired_distance
	    desired_distance = 0.0
	else:		# move 2 atoms to desired_distance from each other
	    desired_vector = Vector.null
	    desired_distance = desired_distance
	dist = (a1.position() - (a2.position() + desired_vector)).length()
	if 0 and (a1.name == 'C24' or a2.name == 'C24') and (a1.name == 'C29' or a2.name == 'C29'):
	    import RasMol
	    ras_session = RasMol.RasMol(self, 1)
	    ras_session.AddPolyLines(Sphere(a2.position() + desired_vector,0.02).coordList(1))
	    ras_session.RefreshScreen()
	    ras_session.Interact()
	    raise ConformationError
	prior_dist = dist
	if abs(dist - desired_distance) < 0.001:
	    print 'dist',dist,'desired_distance',desired_distance
	    return dist
	self.rtimes1 = self.rtimes1 - clock()
	#print 'cycle:',len(bond_path)
	dihedrals_in_cycle = []
	parms_list = []
	#print 'findd',dist,hex(id(a1)),hex(id(endAtom)),a1,endAtom,pos2
	dpair_list = self._dihedrals_between.findDihedralsOnPath(a1, a2)
	for pair in dpair_list:
	    for i in (0,1):
		#print 'look through',i,'of',len(pair),'(',len(pair[0]),',',len(pair[1]),')'
		#pprint.pprint(pair[i])
		for d in pair[i]:
		    if not (d.hasAllBonds()\
		       and (d.angle is None or isSwapWithDangling(d.angle))):
			continue
		    dihedrals_in_cycle.append(d)
		    # ask if d is usefull
		    if not Utility.isDefinedPosition(d.a4.position()):
			continue
		    if not Utility.isDefinedPosition(d.a1.position()):
			continue
		    atom3 = (d.a3, d.a2)[i]
		    #print d.a3.name,atom3.name,d.a4.position(),d.a1.position()
		    if d.angle is None:
			try:
			    self.a2_debug = a2.position()
			    parms = self.calcRotateAngle(d, a1, a2.position()+desired_vector,\
							 atom3, desired_distance)
			except ConformationError, msg:
			    print msg
			else:
			    parms_list.append(parms)
		    else:
			if d.b1.hasAtom(atom3):
			    btmp = d.b1
			else:
			    btmp = d.b3
			ct = self.dihedrals.findConstraints(btmp) #slow?
			if not ct or not self.dihedrals_used.has_key(id(ct)):
			    p = self._altRotateParms(d, a1, a2, atom3, \
				     desired_distance, desired_vector)
			    parms_list.append(p)
			else: print '_altrot too constrained',d,'\nct',ct,'\n',btmp
	self.rtimes1 = self.rtimes1 + clock()
	self.rtimes2 = self.rtimes2 - clock()
	#print 'dihedrals_in_cycle:'
	#pprint.pprint(dihedrals_in_cycle)
	if not parms_list:
	    error = abs(dist - desired_distance)
	    print 'no relevant dihedral; error',error
	    if bonds_to_bend is not None and error > 0.001:
		dist = self.bendBondsToCloseCycle(dpair_list, a1, a2,\
				  desired_distance, desired_vector, \
				  bonds_to_bend, [], forcefield)
		error = abs(dist - desired_distance)
		print 'bend to',error
	    if error > 0.075:
		Utility.warning(self.CycleErrorMessage(a1, a2, dist))
		raise ConformationError, "Can't close cycle"
	    return dist
	while len(parms_list) > 0 and bonds_to_bend is not None:
	    bestparms = None
	    for parms in parms_list:
		if bestparms is None or parms.error < bestparms.error:
		    bestparms = parms
	    print 'best',bestparms.error,bestparms.atom3,\
		  bestparms.angle/Units.deg,bestparms.dihedral
	    if bestparms.error > abs(prior_dist - desired_distance) + 1.e-7:
		print 'no progress',bestparms.error, bestparms.dihedral
		if bonds_to_bend is not None and bestparms.error > 0.001:
		    dist = self.bendBondsToCloseCycle(dpair_list, a1, a2,\
				  desired_distance, desired_vector, \
				  bonds_to_bend, [], forcefield)
		    error = abs(dist - desired_distance)
		    print 'bend to',error
		break
	    olda1pos = a1.position()
	    if bestparms.error < abs(prior_dist - desired_distance):
		old_pos2 = a2.position()
		self.RotateAtomsOnOneSide(bestparms.atom3, bestparms.dihedral.b2,\
					  bestparms.axis, \
					  bestparms.angle, bestparms.atoms_to_reverse)
		if old_pos2 is not a2.position():
		    print 'rotate vector',a2.fullName(), old_pos2, a2.position(),desired_vector,
		    desired_vector = rotateDirection(desired_vector, bestparms.axis,\
						     bestparms.angle)
		    print 'to',desired_vector
	    for a in bestparms.atoms_to_reverse:
		print 'reverse rotate',a,'exclude',bestparms.atom3
		#self.RotateAtom(a, bestparms.axis, -bestparms.angle, [bestparms.atom3])
	    dist = (a1.position() - (a2.position() + desired_vector)).length()
	    if abs(bestparms.error - abs(dist - desired_distance)) > 0.01:
		print 'error',bestparms.error,'dist',dist,\
		      'desired_distance',desired_distance
		print bestparms.atom3,bestparms.dihedral.b2,(atom3 is a2),a2,a1
		#print 'a1pos',a1.position(),a2.position()
		self.RotateAtomsOnOneSide(bestparms.atom3, bestparms.dihedral.b2, bestparms.axis,\
					  -bestparms.angle, bestparms.atoms_to_reverse)
		dist = (a1.position() - (a2.position() + desired_vector)).length()
		print 'unRotateAtom produced dist',dist,a1.position(),a2.position()
		parms_list.remove(parms)
		continue
		#raise RuntimeError
	    error = abs(dist - desired_distance)
	    if error > 0.01:
		if bestparms.warn_msg is not None:
		    print bestparms.warn_msg
	    else:			# clear success
		break
	    if abs(dist - prior_dist) < 0.001:
		if bonds_to_bend is not None\
		   and abs(dist - desired_distance) > 0.001:
		    dist = self.bendBondsToCloseCycle(dpair_list, a1, a2,\
				      desired_distance, desired_vector,\
				      bonds_to_bend, [], forcefield)
		    bonds_to_bend = None
		else:
		    print 'progress too slow',dist,prior_dist
		    break		# quit due to inadequate progress
	    #if bestparms.atoms_to_reverse:
		#?? parms_list.remove(parms)
	    for i in range(len(parms_list)):  # compute new angles for rest
		parms_list[i] = self.calcRotateAngle(parms_list[i].dihedral, a1,\
						     a2.position()+desired_vector,\
						     parms_list[i].atom3, desired_distance)
	    prior_dist = dist
	    if prior_dist < 0.001:
		break

	self.rtimes2 = self.rtimes2 + clock()
	self.rtimes3 = self.rtimes3 - clock()
	#print 'dihedrals_in_cycle'
	#pprint.pprint(dihedrals_in_cycle)
	for d in dihedrals_in_cycle:
	    self._dihedrals_between.removeAbout(d.b2)
	print 'parms_list exhausted: dist',dist,'desired_distance',desired_distance,'prior_dist',prior_dist
	self.rtimes3 = self.rtimes3 + clock()
	#print 'rotatetimes',self.rtimes0,self.rtimes1,self.rtimes2,self.rtimes3
	if abs(dist - desired_distance) < 0.17: # 0.13??
	    return dist			# tolerable results
	if not msg:
	    msg = 'RotatePartsToCloseCycle failed to find good dihedral'\
		  + ' to rotate ' + `a1` + ' close to ' + `a2`
	raise ConformationError, msg

    m2times0 = 0
    m2times1 = 0
    m2times2 = 0
    m2times3 = 0

    def moveToSatisfy(self, dihedral, new_bond, forcefield, bond_to_bend = None):
	self.m2times0 = self.m2times0 - clock()
	if new_bond.hasAtom(dihedral.a4):
	    endAtom = dihedral.a4
	else: endAtom = dihedral.a1
	#print dihedral,dihedral.angle/Units.deg,'now',dihedral.calcAngle()/Units.deg
	angle = forcefield.defaultBondAngle(self.bonds_to_atom[id(endAtom)][0],
					    new_bond, self)
	pivot_atom = new_bond.otherAtom(endAtom)
	(r, constrained, cmnbond, far_bond) =\
	    self.rotateDihedral(dihedral, new_bond, pivot_atom, angle)
	self.m2times0 = self.m2times0 + clock()
	newpos = pivot_atom.position() + r
	print 'moveToSatisfy',endAtom,'from',endAtom.position(),'to',newpos,r.length(),(newpos - endAtom.position()).length()
	#print 'pivot_atom',pivot_atom,pivot_atom.position(),'+',r
	print '{',
	for btmp in self.bonds_to_atom[id(pivot_atom)]:
	    try:
		if btmp is not new_bond:
		    v = r.angle(btmp.asVector(pivot_atom))/Units.deg
		    print '%s : %5.2f' % (btmp.otherAtom(pivot_atom).name, v),
	    except ValueError,msg:
		print 'ValueError',msg,btmp,b
	print '}' #,dir
	r = self.RotatePartsToCloseCycle(endAtom, pivot_atom, r, forcefield,\
					 bond_to_bend)
	#print 'movedtosatisfy %.3f' % (dihedral.calcAngle()/Units.deg,)
	return r

    def rotateDihedral(self, dihedral, new_bond, a_start, angle):
	far_bond = dihedral.b1
	if far_bond is new_bond or not dihedral.a1:
	    far_bond = dihedral.b3
	    d_atoms = (dihedral.a4, dihedral.a3, dihedral.a2, dihedral.a1)
	else:
	    d_atoms = (dihedral.a1, dihedral.a2, dihedral.a3, dihedral.a4)
	#print dihedral,dihedral.angle/Units.deg,'constrains',new_bond,d_atoms[1].position(),d_atoms[0].position()
	d1pos = d_atoms[1].position()
	far_bond_dir = d1pos - d_atoms[0].position()
	bond0_dir = d_atoms[2].position() - d1pos
	#print 'a_start',a_start,d_atoms[3]
	if a_start is d_atoms[3]:
	    normv = bond0_dir.cross(far_bond_dir)
	else:
	    normv = far_bond_dir.cross(bond0_dir)
	    angle = umath.pi - angle
	otherAtom = self.bonds_to_atom[id(a_start)][0].otherAtom(a_start)
	#print 'otherAtom',otherAtom.fullName()
	#if not otherAtom in d_atoms:
	#    angle = -angle	# why does this kludge cause lego6.py to work??
	axis = rotateDirection(normv, bond0_dir.normal(), -dihedral.angle)
	assert(axis.length() > 1.e-6)
	constrained = dihedral.angle
	r = rotateDirection(bond0_dir, axis.normal(), angle)
	an2 = Bonds.calcDihedralAngle(-far_bond_dir, r, dihedral.b2.asVector(d_atoms[1]))
	#print 'nextBondDirectionR angle',angle/Units.deg, bond0_dir.angle(r)/Units.deg, dihedral.angle/Units.deg, an2/Units.deg
	#print a_start,'newpos',(a_start.position() + r)
	if 0 and far_bond is not None:
	    print 'expect',dihedral.angle/Units.deg,far_bond_dir.angle(bond0_dir)/Units.deg,axis
	return (r, constrained, dihedral.b2, far_bond)

    def reconcileVectorWithDihedral(self, dihedral, new_atom, new_bond, vect):
	if dihedral.b1.hasAtom(new_atom):
	    v1 = vect
	    v3 = dihedral.b3.asVector(dihedral.a3)
	else:
	    v3 = vect
	    v1 = dihedral.b1.asVector(dihedral.a2)
	v2 = dihedral.b2.asVector(dihedral.a2)
	al = Bonds.calcDihedralAngle(v1, v3, v2)
	adiff = angleBetweenDirections(al, dihedral.angle)
	#print 'al',al/Units.deg,'vs',dihedral.angle/Units.deg,'{%.2f' % (adiff/Units.deg), '%.2f}' % (abs(adiff - (al - dihedral.angle))/Units.deg),dihedral
	v = rotateDirection(vect, v2.normal(), al - dihedral.angle)
	if dihedral.b1.hasAtom(new_atom):
	    v1 = v
	else:
	    v3 = v
	al2 = Bonds.calcDihedralAngle(v1, v3, v2)
	if abs(angleBetweenDirections(al2, dihedral.angle))\
	   > abs(angleBetweenDirections(al, dihedral.angle)) + 0.001:
	    print 'reverse',(abs(al2 - dihedral.angle) % umath.pi)/Units.deg,\
		  (abs(al - dihedral.angle) % umath.pi)/Units.deg
	    v = rotateDirection(vect, v2.normal(), dihedral.angle - al)
	    if dihedral.b1.hasAtom(new_atom):
		v1 = v
	    else:
		v3 = v
	else:
	    print 'no reverse',(abs(al2 - dihedral.angle) % umath.pi)/Units.deg,\
		  (abs(al - dihedral.angle) % umath.pi)/Units.deg
	al3 = Bonds.calcDihedralAngle(v1, v3, v2)
	print 'al2',al2/Units.deg,al3/Units.deg
	if vect.angle(v) > umath.pi/8:
	    a_start = new_bond.otherAtom(new_atom)
	    Utility.warning('Warning - dihedral constraint causes %s to be %.2f degrees from where other bonds about %s suggest it should be.' %\
		(new_bond, vect.angle(v)/Units.deg, a_start.fullName()))
	# should check that rotation doesn't violate other constraints?
	#print '2.usedd',hex(id(dihedral))
	self.last_usedd = dihedral
	self.dihedrals_used[id(dihedral)] = 1
	return v

    def _selectDihedral(self, dihedrals, new_bond, a_start, show_partial = 0):
	if not dihedrals:
	    return None
	dihedral = None
	for da in dihedrals:
	    far_b = da.otherBond(new_bond, da.b2)
	    if far_b not in self.bonds_to_atom[id(far_b.a1)]:
		continue
	    if not show_partial and da.b2 not in self.bonds_to_atom[id(da.b2.a1)]:
		continue		# bond on unreconciled part of cycle
	    if hasattr(self,'last_dihedral_used') and \
	       dihedral is self.last_dihedral_used and len(dihedrals) > 1:
		dihedral = da
	    elif da.b2.hasAtom(a_start) and \
		 (dihedral is None or not dihedral.b2.hasAtom(a_start)):
		if dihedral:
		    print 'need to switch dihedral',a_start,dihedral.b2,da.b2
		dihedral = da
	if hasattr(self,'last_dihedral_used')\
	   and self.last_dihedral_used is dihedral\
	   and hasattr(self,'dont_use_last_dihedral')\
	   and self.dont_use_last_dihedral:
	    return None
	else:
	    self.last_dihedral_used = dihedral
	    try:
		angle1 = dihedral.angle/Units.deg
	    except (AttributeError,TypeError):
		angle1 = None
	    #print 'use',dihedral,angle1,'of',len(dihedrals)
	    #pprint.pprint(dihedrals)
	return dihedral

    def findPossibleDirections(self, a_start, new_bond, bonds, forcefield):
	vlist = []
	cone_list = []
	for b in bonds:
	    self.ntimesv1 = self.ntimesv1 - clock()
	    angle = self.findBondAngle(b, new_bond, a_start, forcefield)
	    new_cone = Cone(a_start.position(), b.asVector(a_start), angle)
	    self.ntimesv1 = self.ntimesv1 + clock()
	    self.ntimesv2 = self.ntimesv2 - clock()
	    for c in cone_list:
		intersects = _intersectConeCone(c, new_cone)
		#print 'intersects',intersects, c.axis.angle(new_cone.axis)/Units.deg
		if type(intersects) is type(()) and len(intersects) == 2:
		    vlist.append(intersects[0].normal())
		    vlist.append(intersects[1].normal())
		elif hasattr(intersects, '__class__') and\
		     intersects.__class__ is Line:
		    vlist.append(intersects.direction)
		else:
		    angle_between_cones = c.axis.angle(new_cone.axis)
		    if angle_between_cones - (c.angle + angle) < umath.pi/15:
			rotate_angle = (2*umath.pi - angle_between_cones)/2
			nrm = c.axis.cross(new_cone.axis).normal()
			dir = rotateDirection(new_cone.axis, nrm, rotate_angle)
			vlist.append(dir)
	    cone_list.append(new_cone)
	    self.ntimesv2 = self.ntimesv2 + clock()
	return vlist

    def findBondAngle(self, old_bond, new_bond, a_start, forcefield):
	key = (id(new_bond.otherAtom(a_start)),\
	       id(old_bond.otherAtom(a_start)))
	if key[0] > key[1]:
	    key = (key[1], key[0])
	if hasattr(a_start, 'bond_angles') and \
	   a_start.bond_angles.has_key(key):
	    angle = a_start.bond_angles[key]
	else:
	    angle = forcefield.defaultBondAngle(old_bond, new_bond, self)
	return angle

    #
    # find an appropriate vector for configuring a new bond to a_start
    #
    ntimes0 = 0
    ntimes1a = 0
    ntimes1b = 0
    ntimes1 = 0
    ntimes1n = 0
    ntimes2 = 0
    ntimes3 = 0
    ntimesv1 = 0
    ntimesv2 = 0
    ncount2 = 0
    ncount3 = 0
    ncount1b = 0

    def nextBondDirection(self, a_start, new_atom, dihedral_list, forcefield):
	bonds = self.bonds_to_atom[id(a_start)]
	num_bonds = len(bonds)
	if num_bonds == 0:	# start in arbitrary direction:
	    return (Vector.Vector(1, 0, 0), None, None, None)
	times1 = clock()
	if new_atom is None:		# adding DanglingBond?
	    new_bond = Bonds.DanglingBond((a_start,))
	    dihedral = None
	elif new_atom and Bonds.isDihedral(new_atom):
	    dihedral = new_atom	# adding DanglingBond, new_atom is a constraint
	    new_atom = None
	    new_bond = Bonds.DanglingBond((a_start,))
	else:
	    new_bond = Bonds.generateBond((a_start, new_atom))
	    if dihedral_list:
		dihedrals = dihedral_list.findConstraints(new_bond)
		dihedral = self._selectDihedral(dihedrals, new_bond, a_start)
	    else:
		dihedral = None
	self.ntimes1 = self.ntimes1 + (clock() - times1)
	#print 'times1',(clock() - times1),self.ntimes1
	far_bond = None
	cmbond = None
	if dihedral:
	    cmbond = dihedral.b2
	    far_bond = dihedral.b1
	    near_bond = dihedral.b3
	    far_ca = dihedral.a2
	    if far_bond == new_bond or Bonds.isDanglingBond(far_bond):
		far_bond = dihedral.b3
		near_bond = dihedral.b1
		far_ca = dihedral.a3
	times0 = clock()
	if num_bonds == 1:
	    #print '2nd bond',hex(id(bonds[0])),hex(id(bonds[0].a1)),hex(id(bonds[0].a2)),a_start.fullName(),bonds[0]
	    angle = self.findBondAngle(bonds[0], new_bond, a_start, forcefield)
	    if dihedral:
		r = self.rotateDihedral(dihedral, new_bond, a_start, angle)
		self.ntimes1a = self.ntimes1a + (clock() - times0)
		print 'nextBondDirection 1a',(clock() - times0),self.ntimes1a
		print '1.usedd',hex(id(dihedral))
		self.last_usedd = dihedral
		self.dihedrals_used[id(dihedral)] = 1
		return r
	    else:			# make arbitrary dihedral
		self.ntimes1n = self.ntimes1n - clock()
		bond0_dir = bonds[0].asVector(a_start)
		if Bonds.isDanglingBond(new_bond):
		    near_atom = None
		else:
		    near_atom = None #self._findNearestAtom(new_atom, new_bond,\
					#	      forcefield)
		self.ntimes1n = self.ntimes1n + clock()
		if near_atom is not None and\
		   (near_atom.position() - a_start.position()).length() > 1.e-6:
		    if 0:
			print 'near_atom',near_atom,a_start,\
			  near_atom.position(),a_start.position(),\
			  (near_atom.position() - a_start.position()).length()
		    axis = bond0_dir.cross(a_start.position() - near_atom.position())
		    while axis.length() < 1.e-6:
			pos = near_atom.position() + 0.01*randomDirection()
			axis = bond0_dir.cross(a_start.position() - pos)
		else:
		    axis1 = bond0_dir.cross(Vector.Vector(1,0,0))
		    while axis1.length() < 1.e-6:
			axis1 = bond0_dir.cross(randomDirection())
			self.ncount1b = self.ncount1b + 1
		    axis = rotateDirection(bond0_dir, axis1.normal(), umath.pi/2)
		r = rotateDirection(bond0_dir, axis.normal(), angle)
		self.ntimes1b = self.ntimes1b + (clock() - times0)
		print 'nextBondDirection 1b',self.ntimes1n,self.ntimes1b,self.ncount1b
		return (r, None, bonds[0], None)
	else:				# num_bonds > 1
	    if num_bonds > 3:
		alist = ''
		for b1 in bonds:
		    other = b1.otherAtom(a_start)
		    if other is None:
			alist = alist + ' <dangling>'
		    else:
			alist = alist + ' ' + other.fullName()
		raise ValueError, "Don't know how to handle more than 4 bonds/atom, at atom " + a_start.fullName() + '; already has bonds to' + alist
	    if num_bonds == 2:
		vlist = self.findPossibleDirections(a_start, new_bond,
						    bonds, forcefield)
		if len(vlist) < 1: # probably should create new exception type for this
		    print 'a_start',a_start,'new_atom',new_atom,'bonds',bonds
		    print vlist
		    raise ValueError, "No intersecting cones found in nextBondDirection"
		if len(vlist) == 1:
		    print 'only one direction to choose from',new_bond,vlist
		    constraint = dihedral
		    if dihedral:
			vlist[0] = self.reconcileVectorWithDihedral(dihedral,\
					    new_atom, new_bond, vlist[0])
		elif dihedral:
		    v1 = far_bond.asVector(far_ca)
		    cv = dihedral.b2.asVector(far_ca)
		    dihedral0 = Bonds.calcDihedralAngle(v1, vlist[0], cv)
		    dihedral1 = Bonds.calcDihedralAngle(v1, vlist[1], cv)
		    wantAngle = dihedral.angle
		    diff1 = abs(dihedral0 - wantAngle)
		    diff2 = abs(dihedral1 - wantAngle)
		    if diff1 > umath.pi:
			diff1 = 2*umath.pi - diff1
		    if diff2 > umath.pi:
			diff2 = 2*umath.pi - diff2
		    if abs(diff1 - umath.pi) < umath.pi/100:# ????
			diff1 = 0
		    if abs(diff2 - umath.pi) < umath.pi/100:
			diff2 = 0
		    if diff1 > diff2:
			(vlist[0], vlist[1]) = (vlist[1], vlist[0])
			diff = diff2
		    else:
			diff = diff1
		    #print 'new_bond',new_bond,cmbond,far_bond,'\ncv',dihedral,far_ca
		    print 'choose from %.2f %.2f want %.2f err %.2f diffs %.3f %.3f'\
			  % (dihedral0/Units.deg, dihedral1/Units.deg,\
			     wantAngle/Units.deg, diff/Units.deg,\
			     diff1/Units.deg, diff2/Units.deg)
		    vlist[0] = self.reconcileVectorWithDihedral(dihedral,\
					    new_atom, new_bond, vlist[0])
		    if diff > umath.pi/8:
			Utility.warning('Nearest available angle %.2f degrees from expected %s; expected %.2f'\
				     % (diff/Units.deg, `dihedral`,\
					wantAngle/Units.deg))
			if 0 and len(self.dihedrals_used.keys()) != len(self.dihedrals):
			    for d in self.dihedrals:
				if not self.dihedrals_used.has_key(id(d)):
				    print 'dihedral not used:',d
			if diff > umath.pi/4:
			    import RasMol
			    ras_session = RasMol.RasMol(self, 1)
			    ras_session.AddPolyLines([far_ca.position()+v1,
						      far_ca.position(),
						      far_ca.position()+cv,
						      far_ca.position()+cv+vlist[0],
						      far_ca.position()+cv,
						      far_ca.position()+cv+vlist[1]
						      ])
			    ras_session.RefreshScreen()
			    ras_session.Interact()
			    raise ConformationError
		    constraint = dihedral
		    #print '4.usedd',hex(id(dihedral))
		    self.last_usedd = dihedral
		    self.dihedrals_used[id(dihedral)] = 1
		else:			# should only do this for bond order==4
		    ca = Bonds.findCommonAtom(bonds[0], bonds[1])
		    dir0 = bonds[0].asVector(ca)
		    dir1 = bonds[1].asVector(ca)
		    angle0 = dir0.angle(vlist[1])
		    angle1 = dir1.angle(vlist[1])
		    orientation = vlist[1] * dir0.cross(dir1)
		    constraint = SwapWithDangling(bonds[0], angle0, bonds[1],\
						  angle1, orientation)
		    print 'make SwapWithDangling',hex(id(constraint)),far_bond,cmbond,bonds
		    #print 'nextBondDirection2 ',vlist[0]
		    if not hasattr(self, 'alt_bond_dir'):
			self.alt_bond_dir = {}
		    self.alt_bond_dir[id(a_start)] = []
		    for v in vlist[1:]:
			angle0 = findRotationAngle(bonds[0].asVector(a_start),\
					       bonds[1].asVector(a_start), v)
			self.alt_bond_dir[id(a_start)].append(angle0)
		self.ntimes2 = self.ntimes2 + (clock() - times0)
		print 'nextBondDirection 2',(clock() - times0),(dihedral is not None),self.ntimes2,len(vlist),hex(id(a_start)),a_start.fullName()
		return (vlist[0], constraint, cmbond, far_bond)
	    times0 = clock()
	    other_atom = bonds[1].otherAtom(a_start)
	    if hasattr(self, 'alt_bond_dir') and \
	       self.alt_bond_dir.has_key(id(a_start)) and\
	       other_atom is not None:
		start_pos = other_atom.position()
		axis1 = Line(a_start.position(), bonds[0].asVector(a_start))
		angle1 = self.alt_bond_dir[id(a_start)][0]
		new_atom_pos = rotatePoint(start_pos, axis1, angle1)
		dir = new_atom_pos - a_start.position()
		angle = dir.angle(axis1.direction)
		self.ncount2 = self.ncount2 + 1
	    else:
		#print 'no alt_bond_dir',hex(id(a_start)),a_start.fullName()
		vlist = self.findPossibleDirections(a_start, new_bond,
#						    bonds, forcefield)
						    bonds[0:2], forcefield)
		if len(vlist) > 0 and verify4thBondDir(vlist[0], a_start, bonds[2]):
		    dir = vlist[0]
		    angle = vlist[0].angle(bonds[0].asVector(a_start))
		elif len(vlist) > 1 and verify4thBondDir(vlist[1], a_start, bonds[2]):
		    dir = vlist[1]
		    angle = vlist[1].angle(bonds[0].asVector(a_start))
		elif len(vlist) >= 2:
		    print 'compute selectVector the hard way',a_start
		    (dir, angle) = selectVector(vlist, a_start, bonds)
		else:
		    raise ValueError, "No intersecting cones found in nextBondDirection"
		self.ncount3 = self.ncount3 + 1
		self.ntimes3 = self.ntimes3 + (clock() - times0)
			
	    print 'nextBondDirection 3',(clock() - times0),self.ntimes3,self.ntimesv1,self.ntimesv2,self.ncount2,self.ncount3,'nearatom',self.ntimes1n
	    return (dir, angle, cmbond, far_bond)

    def addBond(self, b):
	if self.bonds_to_atom[id(b.a1)].count(b) > 0: raise RuntimeError
	if self.bonds_to_atom[id(b.a2)].count(b) > 0: raise RuntimeError
	self._dihedrals_between.addBond(b)
	self.bonds_to_atom[id(b.a1)].append(b)
	self.bonds_to_atom[id(b.a2)].append(b)

    # handle case where 2 atoms have been positioned before knowing they
    # share a bond

    def reconcile2(self, b, forcefield, unpositioned_bonds_to_atom,\
		   bond_to_bend = None):
	print 'cycle found, bond length',\
	      (b.a1.position()-b.a2.position()).length(),b.a1,'<->',b.a2
	d_constraint = None
	d_list = self.dihedrals.findConstraints(b)
	for d in d_list:
	    far_b = d.otherBond(b, d.b2)
	    if self.bonds_to_atom[id(far_b.a1)].count(far_b) > 0:
		#print 'select d',d
		d_constraint = d
		break
	if d_constraint:
	    dist = self.moveToSatisfy(d_constraint, b, forcefield, bond_to_bend)
	    #print 'c.usedd',hex(id(d_constraint))
	    self.last_usedd = d_constraint
	    self.dihedrals_used[id(d_constraint)] = 1
	else:
	    dist = self.RotatePartsToCloseCycle(b.a1, b.a2,\
					forcefield.defaultBondLength(b, self),\
						forcefield, bond_to_bend)
	return dist

    p1times0 = 0
    p1times1 = 0
    p1times2 = 0
    p1times3 = 0
    p1times4 = 0
    p1times5 = 0
    p1times6 = 0
    p1times9 = 0
    p0count = 0
    p1count = 0
    def position1Bond(self, b, unpositioned_bonds_to_atom, forcefield,\
	  configuration, first_bond, dihedrals_by_bond, follow_bonds = None):
	self.p1times9 = self.p1times9 - clock()
	self.p1count = self.p1count + 1
	pos1 = b.a1.position(configuration)
	pos2 = b.a2.position(configuration)
	reconciling = 0
	dir = None
	a_start = b.a1
	if first_bond:
	    first_bond = 0
	    self.p1times0 = self.p1times0 - clock()
	    #print hex(id(b)),hex(id(b.a1)),hex(id(b.a2)),b.a1.fullName()
	    desired_length = forcefield.defaultBondLength(b, self)
	    self.p1times0 = self.p1times0 + clock()
	    def1 = Utility.isDefinedPosition(pos1)
	    def2 = Utility.isDefinedPosition(pos2)
	    if def1 and def2:
		print 'already positioned',b
		return -1
	    if def1 or def2:
		msg = 'unimplemented case, '+`b.a1`+' '+`b.a2`
		print msg
		raise RuntimeError,msg
	    else:			# setting first position in this Group
		self.setPosition(b.a1, Vector.null)
		self.setPosition(b.a2, Vector.Vector(desired_length, 0, 0))
		constraint = None
		cmbond = None
		far_bond = None
	elif Utility.isDefinedPosition(pos1):
	    if Utility.isDefinedPosition(pos2):
		if ((b in self.bonds_to_atom[id(b.a1)])\
		    and (b in self.bonds_to_atom[id(b.a2)])):
		    #print 'repeated bond',b
		    self.p1times9 = self.p1times9 + clock()
		    return 0
		self.p1times1 = self.p1times1 - clock()
		self.reconcile2(b, forcefield, unpositioned_bonds_to_atom, [])
		reconciling = 1
		self.p1times1 = self.p1times1 + clock()
		self.p1times9 = self.p1times9 + clock()
		constraint = 0.0 # awkward way of indicating no underconstrained_dihedral
		cmbond = None
		far_bond = None
	    else:
		self.p1times2 = self.p1times2 - clock()
		d_list = _collectDihedrals([b], dihedrals_by_bond)
		(dir,constraint,cmbond,far_bond) =\
		  self.nextBondDirection(b.a1, b.a2, d_list, forcefield)
		self.setPosition(b.a2,pos1 + dir.normal()*forcefield.defaultBondLength(b,self))
		self.p1times2 = self.p1times2 + clock()
	elif Utility.isDefinedPosition(pos2):
	    self.p1times3 = self.p1times3 - clock()
	    d_list = _collectDihedrals([b], dihedrals_by_bond)
	    (dir,constraint,cmbond,far_bond) =\
		self.nextBondDirection(b.a2, b.a1, d_list, forcefield)
	    how_long = forcefield.defaultBondLength(b,self)
	    self.setPosition(b.a1, pos2 + dir.normal()*how_long)
	    a_start = b.a2
	    self.p1times3 = self.p1times3 + clock()
	else:
	    print 'neither defined',b
	    self.p1times9 = self.p1times9 + clock()
	    return 0
	self.p1times4 = self.p1times4 - clock()
	self._processNewlyPositionedBond(b, a_start, constraint, cmbond,\
					 far_bond, forcefield)
	#pprint.pprint(unpositioned_bonds_to_atom[id(b.a1)])
	#print hex(id(b)),b
	unpositioned_bonds_to_atom[id(b.a1)].remove(b)
	unpositioned_bonds_to_atom[id(b.a2)].remove(b)
	new_atom = b.otherAtom(a_start)
	for alt_a in new_atom.bondedTo():
	    #print 'alt_a',alt_a,self._dihedrals_between.sameGroup(new_atom, alt_a),Utility.isDefinedPosition(alt_a.position())
	    if alt_a is not a_start and not reconciling \
	       and self._dihedrals_between.sameGroup(new_atom, alt_a)\
	       and Utility.isDefinedPosition(alt_a.position()):
		b2 = Bonds.lookUpBond(new_atom, alt_a)
		print 'recurse for cycle'
		self.position1Bond(b2, unpositioned_bonds_to_atom, forcefield,\
			   configuration, 0, dihedrals_by_bond, follow_bonds)
	changes = 1
	newly_pos_atom = b.otherAtom(a_start)
	self.p1times4 = self.p1times4 + clock()
	self.p1times5 = self.p1times5 - clock()
	for b2 in unpositioned_bonds_to_atom[id(newly_pos_atom)][:]:
	    c = self.positionNewlyConstrainedBonds(b2,\
		   unpositioned_bonds_to_atom, forcefield, configuration,\
		   newly_pos_atom, dihedrals_by_bond, follow_bonds)
	    changes = changes + c
	self.p1times5 = self.p1times5 + clock()
	self.p1times9 = self.p1times9 + clock()
	print 'postimes %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f %6.2f %d' %\
	      (self.p1times0,self.p1times1,self.p1times2,self.p1times3,self.p1times4,self.p1times5,self.p1times9,self.p1count)
	return changes

    def positionNewlyConstrainedBonds(self, b2, unpositioned_bonds_to_atom,\
	     forcefield, configuration, newly_pos_atom, dihedrals_by_bond,\
	     follow_bonds):
	if Utility.isDefinedPosition(b2.otherAtom(newly_pos_atom).position()):
	    #self.reconcile2(b2, forcefield, unpositioned_bonds_to_atom)
	    return 0
	else:
	    d_list = _collectDihedrals([b2], dihedrals_by_bond)
	    d = self._selectDihedral(d_list.findConstraints(b2), b2, \
				     newly_pos_atom, 1)
	    if d:
		if follow_bonds is not None:
		    follow_bonds.append(b2)
		    follow_bonds.append(d.b2)
		    return 0
		print 'follow constraint',b2,d,len(dihedrals_by_bond.keys())
		self.p1times5 = self.p1times5 + clock()
		self.p1times9 = self.p1times9 + clock()
					# newly constrained bond:
					# also look at bond maybe closing cycle
		follow_bonds = [b2, d.b2]
		c = 0
		while follow_bonds:
		    follow_next = []
		    for b_follow in follow_bonds:
			c = c + self.position1Bond(b_follow,\
			   unpositioned_bonds_to_atom, forcefield, \
			   configuration, 0, dihedrals_by_bond, follow_next)
		    follow_bonds = follow_next
		print 'done following constraint',b2
		self.p1times9 = self.p1times9 - clock()
		self.p1times5 = self.p1times5 - clock()
		return c
	    else:
		pass #print 'no constraint on',b2
	return 0

    def _processNewlyPositionedBond(self, b, a_start, constraint, cmbond,\
				    far_bond, forcefield):
	print 'add bond',b.a1.name,'to',b.a2.name,'{',
	for btmp in self.bonds_to_atom[id(a_start)]:
	    try:
		v = Bonds.BondAngle(b,btmp,a_start).radians()/Units.deg
		print '%5.2f' % v,
		if v < 45:
		    print '}', btmp, b.a1.position(), b.a2.position()
		    raise ConformationError
	    except ValueError,msg:
		print 'ValueError',msg,btmp,b
	print '}' #,dir
	self.addBond(b)
	if type(constraint) is types.InstanceType\
	     and constraint.__class__ is SwapWithDangling:
	    #existing_d = self.dihedrals.findRotatesAbout(cmbond, self.bonds_to_atom)
	    other_atom_bond_order = 0
	    print 'sw cmbond',cmbond,far_bond
	    if cmbond is None:
		for b2 in self.bonds_to_atom[id(a_start)]:
		    bond_order = forcefield.bondOrder(b2.otherAtom(a_start), self)
		    if not b2 is b and bond_order > other_atom_bond_order:
			print 'bond_order',bond_order,other_atom_bond_order,b2
			cmbond = b2
			other_atom_bond_order = bond_order
	    d = Bonds.DihedralConstraint(far_bond, b, cmbond, constraint)
	    #print 'add underconstrained_dihedral B:',d,constraint
	    self._dihedrals_between.addDihedral(d)
	elif type(constraint) is not types.FloatType\
	   and not Bonds.isDihedral(constraint):
	    #print 'cmbond',cmbond,'far_bond',far_bond
	    (cmbond, far_bond) = _fillEmptyDihedralBonds(cmbond,
					 far_bond, a_start, self.bonds_to_atom)
	    #print 'changed to',cmbond, far_bond
	    prior_dihedral = self.dihedrals.findRotatesAbout(cmbond, self.bonds_to_atom)
	    if not prior_dihedral or not (b.otherAtom(a_start)\
				  in (prior_dihedral.a1, prior_dihedral.a4)):
		d = Bonds.DihedralConstraint(far_bond, b, cmbond, constraint)
		self._dihedrals_between.addDihedral(d)
		#print 'add underconstrained_dihedral',d,constraint
	    else: pass #print 'no underconstrained_dihedral',cmbond,prior_dihedral,constraint,'\n',self.dihedrals
	else: pass #print 'no underconstrained_dihedral',constraint, far_bond, cmbond,\
	      #len(self.bonds_to_atom[id(b.a1)]),len(self.bonds_to_atom[id(b.a2)])

    def kludgeAtomGroup(self, obj):
	r = {}
	for a in self.atoms:
	    if self._dihedrals_between.isIn(a, obj):
		r[id(a)] = 1
	print 'kludgeAtomGroup',r
	return r

    def _collectBondSpans(self):
	dict = DictWithDefault.DictWithDefault([])
	bond_pos = {}
	_MapBondPos(self, bond_pos)
	for b in self.bonds:
	    pair = self._dihedrals_between.groupsSpanned(b, b is self.bonds[0])
	    if pair:
		#print hex(id(b)),'spans',hex(id(pair[0])),hex(id(pair[1])),bond_pos[id(pair[0])],bond_pos[id(pair[1])],b.a1.fullName()[18:],b.a2.fullName()[18:]
		assert(pair[0])
		assert(pair[1])
		if bond_pos[id(pair[0])] < bond_pos[id(pair[1])]:
		    dict[(id(pair[0]), id(pair[1]))].append(b)
		else:
		    dict[(id(pair[1]), id(pair[0]))].append(b)
	self._bonds_spanning = dict

    def _newBondList(self, prior_groups, new_group, parent_group):
	bondlist = []
	for b in []:#parent_group.joining_bonds:
	    for p in prior_groups:
		if self._dihedrals_between.spans(b, p, new_group):
		    bondlist.append(b)
	blist = []
	for p in prior_groups:
	    blist = blist + self._bonds_spanning[(id(p), id(new_group))]
	if 0:
	    from Joiner import fullSublistName
	    print 'newbondlist',len(bondlist),len(parent_group.joining_bonds),len(prior_groups),len(new_group.bonds),fullSublistName(new_group)
	    for p in prior_groups:
		print 'prior_group',fullSublistName(p)
	    pprint.pprint(bondlist)
	    pprint.pprint(blist)
	    for b in parent_group.joining_bonds:
		print hex(id(b)),'span',b.a1.fullName()[8:],b.a2.fullName()[8:]
	return blist

    b1times0 = 0
    b1times1 = 0
    b1times2 = 0
    b1times3 = 0

    def _firstBondBetweenGroups(self, b, dihedral_list, \
				forcefield, configuration, atom_on_new_side):
	print '_firstBondBetweenGroups',b,b.a1.index,b.a2.index,b.a1.position(),b.a2.position()
	dihedrals = dihedral_list.findConstraints(b)
	if len(dihedrals) > 0:
	    dihedral = dihedrals[0] # ??
	    if dihedral.b2.hasAtom(b.a1):
		(atom1, atom2) = (b.a1, b.a2)
	    else:
		print 'swap'
		(atom1, atom2) = (b.a2, b.a1)
	else:
	    print 'no dihedral constraining',b
	    (atom1, atom2) = (b.a1, b.a2)
	(dir, constraint, cmbond, far_bond) =\
	      self.nextBondDirection(atom1, atom2, dihedral_list, forcefield)
	self.b1times1 = self.b1times1 - clock()
	pos1 = atom1.position(configuration)
	pos2 = atom2.position(configuration)
	a_start = atom1
	desired_length = forcefield.defaultBondLength(b, self)
	desired_pos2 = pos1 + dir.normal()*desired_length
	translate_v = desired_pos2 - pos2
	if self._dihedrals_between.closerGroup(atom2, atom_on_new_side, atom1):
	    translateAtomsBy(atom2, translate_v, self.bonds_to_atom, {})
	else:
	    translateAtomsBy(atom1, -translate_v, self.bonds_to_atom, {})
	if 0:
	    for d1 in self.dihedrals:
		try:
		    print 'dih %8.3f %.4s %.4s %.4s %.4s' % (d1.calcAngle()/Units.deg,
							d1.a1.fullName()[-4:],
							d1.a2.fullName()[-4:],
							d1.a3.fullName()[-4:],
							d1.a4.fullName()[-4:])
		except TypeError: pass
	other1 = self.bonds_to_atom[id(atom1)][0].otherAtom(atom1)
	new_angle = (other1.position() - atom1.position()).angle(b.asVector(atom1))
	self.b1times1 = self.b1times1 + clock()
	#print 'new angle', new_angle/Units.deg,'other1',other1,constraint
	print 'join bond',atom2.name,'to',atom1.name,'{',
	for btmp in self.bonds_to_atom[id(atom1)]:
	    v = b.asVector(atom1).angle(btmp.asVector(atom1))/Units.deg
	    print '%s: %5.2f' % (btmp.otherAtom(atom1).name, v),
	    if v < 90:
		print '}', btmp, atom1
		print atom1.position(),btmp.otherAtom(atom1).position()
		print b.otherAtom(atom1).position()
		raise ConformationError
	print '}',dir
	if len(dihedrals) == 1:
	    self.dont_use_last_dihedral = 1
	(dir2, constraint2, cmbond2, far_bond2) =\
	       self.nextBondDirection(atom2, atom1, dihedral_list, forcefield)
	if not (constraint2 is None and cmbond2 is None and far_bond2 is None):
	    self.b1times2 = self.b1times2 - clock()
	    self.dont_use_last_dihedral = 0
	    if cmbond2 is not None and not cmbond2.hasAtom(atom1):
	    	#print 'swap?',atom1,atom2,cmbond2
		(atom1, atom2) = (atom2, atom1)
		dir2 = -dir2
	    if type(constraint2) is not types.FloatType\
	       and not Bonds.isDihedral(constraint2):
		(cmbond2, far_bond2) = _fillEmptyDihedralBonds(cmbond2,\
				   far_bond2, atom2, self.bonds_to_atom)
		if not self.dihedrals.findRotatesAbout(cmbond2):
		    d = Bonds.DihedralConstraint(far_bond2, b, cmbond2, constraint2)
		    self._dihedrals_between.addDihedral(d)
		    print '2add underconstrained_dihedral',d

	    normal = dir.cross(dir2)
	    if normal.length() > 1.e-9:
		axis = Line(atom1.position(), normal)
	    else:		# 0 or 180 degree angle between dir and dir2
				# use arbitrary normal. does this always 
				# mean nextBondDirection(atom2...) incorrectly
				# returned same as nextBondDirection(atom1?
	        print 'bad crossproduct',dir2,dir
		axis = Line(atom1.position(), dir.cross(randomDirection()))
	    rotate_angle = dir.angle(dir2)
	    if cmbond2 is not None and cmbond2.hasAtom(atom1):
	        #print 'cmbond2',cmbond2,atom1,atom2,'\n',self.bonds_to_atom[id(atom1)]
		other2 = cmbond2.otherAtom(atom1)
	    else:
		other2 = self.bonds_to_atom[id(atom1)][0].otherAtom(atom1)
		print 'no cmbond2',other2,cmbond2,atom1
	    want_angle = dir2.angle(other2.position() - atom1.position())
	    if dir2 * (other2.position() - atom1.position()) < 0:
		want_angle = -want_angle
	    #print 'set2',cmbond2, far_bond2, constraint2, other2, self.bonds_to_atom[id(atom1)][0]
	    self.b1times2 = self.b1times2 + clock()
	    self.b1times3 = self.b1times3 - clock()
	    rot_pos = rotatePoint(other2.position(), axis, rotate_angle)
	    angle = (rot_pos-atom1.position()).angle(b.asVector(atom1))
	    if (rot_pos-atom1.position()) * b.asVector(atom1) < 0:
		angle = -angle
	    #print 'want',want_angle/Units.deg, angle/Units.deg,rotate_angle/Units.deg,atom1, atom_on_new_side, atom2
	    if abs(want_angle - angle) > 1.e-8 and\
	       abs(want_angle - angle - umath.pi) > 1.e-8:
		rotate_angle = -rotate_angle
	    if self._dihedrals_between.closerGroup(atom1, atom_on_new_side, atom2):
		self.RotateAtomsOnOneSide(atom1, None, axis, rotate_angle)
	    else:
		self.setPosition(atom2, rotatePoint(atom2.position(), axis, -rotate_angle))
		self.RotateAtomsOnOneSide(atom2, None, axis, -rotate_angle)
	    self.b1times3 = self.b1times3 + clock()
	    if 0:
		if far_bond is not None:
		    di = Bonds.Dihedral((far_bond,b,cmbond,None))
		    print '2new dihedral %.2f %s' % (di.setAngle()/Units.deg,di)
	    self._dihedrals_between.removeAbout(cmbond2)
	(b, bond2) = _fillEmptyDihedralBonds(b, None, a_start,\
					     self.bonds_to_atom)
	    #print 'b angle dir',dir.angle(b.asVector(atom1))/Units.deg
	    #print 'b angle dir2',dir2.angle(b.asVector(atom1))/Units.deg
	try:
	    bond1 = self.bonds_to_atom[id(a_start)][0]
	except KeyError:
	    bond1 = None
	#print 'make dihedral',bond1, bond2, b
	if bond2 is not None:
	    d = Bonds.DihedralConstraint(bond1, bond2, b, None)
	    self._dihedrals_between.addDihedral(d)
	    #print '1 add underconstrained_dihedrals',d
	else: print 'no DihedralConstraint',bond1,b
	print 'join bond',atom1,'to',atom2,'{',
	for btmp in self.bonds_to_atom[id(atom2)]:
	    v = b.asVector(atom2).angle(btmp.asVector(atom2))/Units.deg
	    print '%s: %5.2f' % (btmp.otherAtom(atom2).name, v),
	    if v < 90:
		print '}',dir2, btmp, atom2
		raise ConformationError
	print '}' #,dir2
	self._processNewlyPositionedBond(b, a_start, constraint, cmbond,\
					 far_bond, forcefield)

    jtimes0 = 0
    jtimes1 = 0
    jtimes2 = 0
    jtimes3 = 0
    jtimes4 = 0
    def joinGroups(self, newbondlist, dihedrals_by_bond, forcefield,\
		   unpositioned_bonds_to_atom, configuration, \
		   atom_on_new_side):
	self.jtimes0 = self.jtimes0 - clock()
	dihedral_list0 = _collectDihedrals(newbondlist[0:1], dihedrals_by_bond)
	self.jtimes0 = self.jtimes0 + clock()
	self.jtimes1 = self.jtimes1 - clock()
	self._firstBondBetweenGroups(newbondlist[0], dihedral_list0, \
			     forcefield, configuration, atom_on_new_side)
	self.jtimes1 = self.jtimes1 + clock()
	joining_dihedral = dihedral_list0.findRotatesAbout(newbondlist[0])
	if joining_dihedral:
	    self.rotateToMatchDihedral(joining_dihedral)
	    print 'remove joining'
	    self._dihedrals_between.removeAbout(joining_dihedral.b2)
	#else: print 'no joining_dihedral for',newbondlist[0]
	if len(newbondlist) == 1:
	    return
	print 'joinGroups reconcile2'
	b = newbondlist[1]
	dist = self.reconcile2(b, forcefield, unpositioned_bonds_to_atom,\
			       newbondlist[0:1])
	self.addBond(b)
	try:
	    unpositioned_bonds_to_atom[id(b.a1)].remove(b)
	except ValueError: pass
	try:
	    unpositioned_bonds_to_atom[id(b.a2)].remove(b)
	except ValueError: pass
	for i in range(2, len(newbondlist)):
	    b = newbondlist[i]
	    len1 = b.asVector().length()
	    print 'check length %.2f %s %s' % (len1,b.a1.name, b.a2.name)
	    desired_distance = forcefield.defaultBondLength(b, self)
	    if abs(len1 - desired_distance) > 0.01:
		self.jtimes2 = self.jtimes2 - clock()
		exclude_atoms = []
		bend_atoms = []
		pivot = Vector.null
		for b2 in newbondlist[0:i]:
		    if self._dihedrals_between.closerGroup(b2.a1, atom_on_new_side, b2.a2):
			(a0, a1) = (b2.a1, b2.a2)
		    else:
			(a0, a1) = (b2.a2, b2.a1)
		    exclude_atoms.append(a0)
		    bend_atoms.append(a1)
		    pivot = pivot + (a0.position() + a1.position())/2
		#print 'newbondlist[0:',i,']'
		#pprint.pprint(newbondlist[0:i])
		#print 'exclude_atoms',exclude_atoms
		pivot = pivot / i
		if self._dihedrals_between.closerGroup(b.a2, atom_on_new_side, b.a1):
		    pos1 = b.a1.position()
		    pos2 = b.a2.position()
		else:
		    pos1 = b.a2.position()
		    pos2 = b.a1.position()
		    print 'reverse pos'
		self.jtimes2 = self.jtimes2 + clock()
		self.jtimes3 = self.jtimes3 - clock()
		desired_angle = _findDesiredAngle(pivot, pos1, pos2,\
						  desired_distance)
		self.jtimes3 = self.jtimes3 + clock()
		self.jtimes4 = self.jtimes4 - clock()
		rotate_angle = desired_angle - (pos1 - pivot).angle(pos2 - pivot)
		axis = Line(pivot, (pos1 - pivot).cross(pos2 - pivot))
		pos_plus = rotatePoint(pos1, axis, rotate_angle)
		pos_minus = rotatePoint(pos1, axis, -rotate_angle)
		#print rotate_angle/Units.deg,(pos_plus - pos2).length(),\
		#      (pos_minus - pos2).length(), b
		#print 'bend_atoms',bend_atoms
		if abs((pos_plus - pos2).length() - desired_distance)\
		       > abs((pos_minus - pos2).length() - desired_distance):
		    rotate_angle = -rotate_angle
		#for a in bend_atoms:
		self.RotateAtom(bend_atoms[0], axis, rotate_angle, exclude_atoms)
		len1 = b.asVector().length()
		print 'rotate_angle',rotate_angle/Units.deg,'recheck length %.2f' % (len1,)
		if len1 > 0.5:
		    import RasMol
		    ras_session = RasMol.RasMol(self, 1)
		    ras_session.RefreshScreen()
		    ras_session.Interact()
		    raise ConformationError
		    
		self.jtimes4 = self.jtimes4 + clock()
	    self.addBond(b)
	print 'joinGroups times',self.jtimes0,self.jtimes1,self.jtimes2,self.jtimes3,self.jtimes4,'  ',self.b1times0,self.b1times1,self.b1times2,self.b1times3


    positionGrouptime = 0
    positionGrouptime1 = 0
    positionGrouptime2 = 0
    positionGrouptime3 = 0
    def positionGroup(self, grp, unpositionedBondsToAtom, forcefield,\
		      configuration = None, max_tries = 1):
	self.positionGrouptime = self.positionGrouptime - clock()
	if hasattr(grp, 'subgroups') and grp.subgroups:
	    bondlist = Bonds.BondList([])	# all bonds from sublists
	    dihedral_list = Bonds.DihedralAngleList([])
	    i = 0
	    subgroup_bonds = Bonds.BondList([])
	    for subgrp in grp.subgroups:
		print 'positionGroup',i,'of',len(grp.subgroups),'depth',self.groupdepth,len(subgrp.joining_bonds),len(subgrp.bonds),hex(id(subgrp)),subgrp.name
		#for b1 in subgrp.joining_bonds:
		#   print hex(id(b1)),b1.a1.fullName()[8:],b1.a2.fullName()[8:]
		self.groupdepth = self.groupdepth + 1
		self.positionGrouptime = self.positionGrouptime + clock()
		(blist, dmap, dlist1) = self.positionGroup(subgrp,\
						   unpositionedBondsToAtom,\
						   forcefield, configuration)
		self.positionGrouptime = self.positionGrouptime - clock()
		if not i:
		    dihedrals_by_bond = dmap
		else:
		    for k in dmap.keys():
			dihedrals_by_bond[k] = self.dihedrals_by_bond[k]
		subgroup_bonds = subgroup_bonds + subgrp.joining_bonds
		bondlist = bondlist + blist
		dihedral_list = dihedral_list + dlist1
		self.groupdepth = self.groupdepth - 1
		if i:
		    self.positionGrouptime1 = self.positionGrouptime1 - clock()
		    newbondlist = self._newBondList(subgrp.attach_subgroups, \
						    subgrp, grp)
		    #pprint.pprint(newbondlist)
		    bondlist = bondlist + newbondlist
		    for b in newbondlist:
			dihedrals_by_bond[id(b)] = self.dihedrals_by_bond[id(b)]
		    self.positionGrouptime1 = self.positionGrouptime1 + clock()
		    self.positionGrouptime2 = self.positionGrouptime2 - clock()
		    new_dlist = _collectDihedrals(newbondlist, dihedrals_by_bond)
		    dihedral_list = dihedral_list + new_dlist
		    self.positionGrouptime2 = self.positionGrouptime2 + clock()
		    self.positionGrouptime3 = self.positionGrouptime3 - clock()
		    if not newbondlist:
			Utility.warning('Warning - subgroup %s not bonded'\
					% subgrp.name)
		    elif not newbondlist[0] in self.bonds_to_atom[id(newbondlist[0].a1)]:
			atom_on_new_side = self._dihedrals_between.atomIn(b, subgrp)
			#print len(bondlist),len(blist),'<>',atom_on_new_side,subgrp,newbondlist[0]
			self.joinGroups(newbondlist, dihedrals_by_bond, \
			    forcefield, unpositionedBondsToAtom,\
			    configuration, atom_on_new_side)
		    else: pass
			#print 'already positioned',newbondlist
		    self.positionGrouptime3 = self.positionGrouptime3 + clock()
		print 'positionGroup done',i,'depth',self.groupdepth,len(subgrp.joining_bonds)
		i = i + 1
	    first_bond = 0
	else:
	    first_bond = 1
	    dihedrals_by_bond = {}
	    for b in grp.bonds[:]:
		#print hex(id(b)),b,hex(id(grp)),grp.name
		if Bonds.isDanglingBond(b):		# ?? ugly hack
		    grp.bonds.remove(b)
		    continue
		dihedrals_by_bond[id(b)] = self.dihedrals_by_bond[id(b)]
		#print 'posbond',b,b.a1.position(),b.a2.position()
		if (not Utility.isDefinedPosition(b.a1.position()) \
		    or not Utility.isDefinedPosition(b.a2.position())):
		    #print 'add unpositionedBondsToAtom',b,unpositionedBondsToAtom[id(b.a1)].count(b),unpositionedBondsToAtom[id(b.a2)].count(b)
		    unpositionedBondsToAtom[id(b.a1)].append(b)
		    unpositionedBondsToAtom[id(b.a2)].append(b)
	    dihedral_list = _collectDihedrals(grp.bonds, dihedrals_by_bond)
	changes = 1
	#p = regsub.gsub('Atom diamond111diamond111.diamond111.','alt.',`self.dihedrals`)
	#print regsub.gsub('Atom diamond111.','',p)
	if grp.atoms and Utility.isDefinedPosition(grp.atoms[0].position()):
	    changes = 0
	tries = 0
	while(changes > 0 and tries < max_tries):
	    changes = 0
	    max_tries = max_tries + 1
	    if grp.atoms and not grp.bonds:	# must be single atom group
		if not Utility.isDefinedPosition(grp.atoms[0].position()):
		    from RandomStuff import randomPointInBox
		    grp.atoms[0].setPosition(randomPointInBox(1.0))
		    print 'randomPointInBox',grp.atoms[0].fullName(),grp.atoms[0].position()
		break
	    for b in grp.bonds:
		#print 'position1bond',changes,'depth',self.groupdepth,first_bond,b,len(dihedrals_by_bond.keys())
		self.positionGrouptime = self.positionGrouptime + clock()
		c = self.position1Bond(b, unpositionedBondsToAtom,\
		      forcefield, configuration, first_bond, dihedrals_by_bond)
		self.positionGrouptime = self.positionGrouptime - clock()
		changes = changes + c
		if c == -1:
		    break
		first_bond = 0
	self.positionGrouptime = self.positionGrouptime + clock()
	print 'positionGroup times',self.positionGrouptime,self.positionGrouptime1,self.positionGrouptime2,self.positionGrouptime3
	return (grp.bonds, dihedrals_by_bond, dihedral_list)
	
    def clearBondAttributes(self):
	for a in self.atoms:
	    a.clearBondAttribute()

    def setBondAttributes(self):
	for b in self.bonds:
	    #print 'bond',hex(id(b)),b.a1,b.a2
	    b.a1.setBondAttribute(b.a2)
	    b.a2.setBondAttribute(b.a1)
	    #print 'setBondAttribute',hex(id(b)),b
	    #print hex(id(self.atoms.index(b.a1.bonded_to__[-1]))),\
		#  self.atoms.index(b.a1.bonded_to__[-1]),\
		#  self.atoms.index(b.a2.bonded_to__[-1])
	try:
	    dbonds = self.danglingbonds
	except AttributeError:
	    pass
	else:
	    for b in dbonds:
		try:
		    b.a1.dangling_bonds_to__.append(b)
		except AttributeError:
		    b.a1.dangling_bonds_to__ = [b]

    def setPositions(self, forcefield, configuration = None):
	t0 = clock()
	#for d1 in self.dihedrals:
	#    print 'di:',hex(id(d1)),d1
	print 'setPositions',hex(id(self))
	self.clearBondAttributes()
	self.setBondAttributes()
	for a in self.atoms:
	    print 'ai:',hex(id(a)),a.fullName()
	    if 0:
		try:
		    ndb = len(a.numDanglingBondsTo())
		except: ndb = 0
		try:
		    print len(a.bondedTo()),ndb,'db',len(self.bonds_to_atom[id(a)])
		except: print 
	self.dihedrals.generateMap()
	t1 = clock()
	self.dihedrals_used = {}
	self.groupdepth = 0
	self.alt_bond_dir = {}
	if not hasattr(self, 'bonds_to_atom'):
	    self.bonds_to_atom = {}
	unpositionedBondsToAtom = {}
	for i in range(len(self.atoms)):
	    a = self.atoms[i]
	    idx = a.index
	    if idx is None:
		idx = i
	    self.setIndex(a, idx) # avoid slowness in getAtomProperty; why no atom_references here?
	    unpositionedBondsToAtom[id(a)] = []
	    if not self.bonds_to_atom.has_key(id(a)):
		self.bonds_to_atom[id(a)] = []
		#print 'init bonds_to_atom',hex(id(a)),a.fullName(),hex(id(self))
	self.findAtomTypes(forcefield)
	t2 = clock()
	self._dihedrals_between = MapDihedralsBetween(self.bonds_to_atom)
	self.dihedrals_by_bond = {}
	for b in self.bonds:
	    self.dihedrals_by_bond[id(b)] = []
	    if 0 and Utility.isDefinedPosition(b.a1.position()) and\
	       Utility.isDefinedPosition(b.a2.position()):
		if b not in self.bonds_to_atom[id(b.a1)]:
		    self.bonds_to_atom[id(b.a1)].append(b)
		if b not in self.bonds_to_atom[id(b.a2)]:
		    self.bonds_to_atom[id(b.a2)].append(b)
	    #print hex(id(b)),hex(id(b.a1)),hex(id(b.a2)),b.a1.fullName()[15:],b.a2.fullName()[15:]
	    #print 'a1 index',self.atoms.index(b.a1)
	    #print 'a2 index',self.atoms.index(b.a2)
	for d in self.dihedrals:
	    for b in (d.b1, d.b2, d.b3):
		try:
		    self.dihedrals_by_bond[id(b)].append(d)
		except KeyError: pass
	self._dihedrals_between.mark(self)
	t4 = clock()
	self._collectBondSpans()
	t5 = clock()
	self.positionGroup(self, unpositionedBondsToAtom,
			   forcefield, configuration, 100)
	t6 = clock()
	self.translateTo(Vector.null)
	for a in self.atomList():
	    try:
		expected_order = forcefield.bondOrder(a, self)
	    except:
		print a.name,' has no bond order ',a.type.symbol
		continue
	    if len(self.bonds_to_atom[id(a)]) != expected_order:
		   print 'Warning: ',a.fullName(),'has',len(self.bonds_to_atom[id(a)]),\
			 ' bonds; expected ',expected_order
	if 0 and len(self.dihedrals_used.keys()) != len(self.dihedrals):
	    for d in self.dihedrals:
		if not self.dihedrals_used.has_key(d):
		    print 'dihedral not used:',d
	t9 = clock()
	print 'settimes',t1-t0,t2-t1,t4-t2,t5-t4, t6-t5, t9-t6
	return 1

    def findAtomTypes(self, forcefield):
	if not hasattr(self.atoms[0], 'bonded_to__'):
	    self.setBondAttributes()
	for a in self.atoms:
	    try:
		type1 = self.getAtomProperty(a, forcefield.atom_type_property)
	    except (AttributeError, TypeError):
		try:
		    type1 = forcefield.getOrSetAtomType(a, self)
		except (AttributeError, TypeError):
		    type1 = 'unknown'
	    print a.symbol, type1, a.fullName(), len(self.bonds_to_atom[id(a)])

    def printDihedrals(self):
	for b in self.bonds:
	    for b1 in self.bonds_to_atom[id(b.a1)]:
		if not b1.hasAtom(b.otherAtom(b.a1)):
		    for b3 in self.bonds_to_atom[id(b.a2)]:
			if not b3.hasAtom(b.otherAtom(b.a2)):
			    d = Bonds.Dihedral((b1, b3, b, None))
			    d.setAngle()
			    print d

    def printBadDihedrals(self):
	n = -6
	bonds_in_bad_dihedrals = {}
	for d in self.dihedrals:
	    if d.a1 is None or d.a4 is None:
		continue
	    u = self.dihedrals_used.has_key(id(d))
	    try:
		diff = (d.calcAngle() - d.angle) % (2*umath.pi)
	    except TypeError:
		print 'Bad angle,',d.angle/Units.deg,d.calcAngle()
	    else:
		if diff > umath.pi:
		    diff = diff - 2*umath.pi
		if abs(diff) < 0.5*Units.deg:
		    continue
		print 'dihedral err%6.1f' % (diff/Units.deg,), '%5.1f' %\
		      (d.angle/Units.deg,), hex(id(d)),
	    print ('    ', 'used')[u], d.a1.fullName()[n:], d.a2.fullName()[n:],\
		  d.a3.fullName()[n:], d.a4.fullName()[n:]
	    for b in (d.b1, d.b2, d.b3):
		bonds_in_bad_dihedrals[(b.a1.index, b.a2.index)] = 1
	return bonds_in_bad_dihedrals

    def calcDanglingBonds(self, forcefield):
	if not hasattr(self, 'bonds_to_atom'):
	    self.setPositions(forcefield)
	if not hasattr(self, 'danglingbonds'):
	    self.danglingbonds = Bonds.BondList([])
	num_dangle = DictWithDefault.DictWithDefault(0)
	for db in self.danglingbonds:
	    num_dangle[db.a1] = num_dangle[db.a1] + 1
	for a in self.atomList():
	    try:
		expected_order = forcefield.bondOrder(a, self)
	    except:
		print a.name,'has no bond order',a.type.symbol
		continue
	    num_normal_bonds = len(self.bonds_to_atom[id(a)])
	    num_dangling_bonds = num_dangle[a]
	    while (num_normal_bonds + num_dangling_bonds) < expected_order:
		(dir,constraint,cmbond,far_bond) =\
		   self.nextBondDirection(a, None, self.dihedrals, forcefield)
		print 'gen dangling bond',a.name,constraint
		(cmbond, far_bond) = _fillEmptyDihedralBonds(cmbond,
				     far_bond, a, self.bonds_to_atom)
		#print 'far_bond',far_bond,'cmbond',cmbond, len(self.bonds_to_atom[id(a)])
		if far_bond is None and cmbond is not None:
		    (cmbond, far_bond) = _fillEmptyDihedralBonds(cmbond,
			 far_bond, cmbond.otherAtom(a), self.bonds_to_atom)
		#print 'far_bond',far_bond,'cmbond',cmbond
		    
		dangling = Bonds.DanglingBond((a, constraint, dir))
		if far_bond is not None and cmbond is not None:
		    dihedral = Bonds.Dihedral((far_bond, dangling,
					       cmbond, None))
		    print 'create Dihedral',far_bond.asVector().angle(dangling.direction)/Units.deg,\
			  dihedral.a1,dihedral.b1
		else:
		    dihedral = self._findDihedralTo(a)
		    if dihedral is None:
			Utility.warning("Can't generate dihedral for " + a.fullName())
			num_dangling_bonds = num_dangling_bonds + 1
			#raise ConformationError
			continue
		if dihedral.setAngle() is not None:
		    print 'add dihedral',dihedral
		    dangling.setConstraint(dihedral)
		self.danglingbonds.append(dangling)
		if self.bonds_to_atom[id(a)].count(dangling) > 0: raise RuntimeError
		self.bonds_to_atom[id(a)].append(dangling)
		num_dangling_bonds = num_dangling_bonds + 1

    def _findDihedralTo(self, atom):
	atom_list = [atom]
	while len(atom_list) != 4:
	    for b in self.bonds_to_atom[id(atom)]:
		a = b.otherAtom(atom)
		print '_findDihedralTo',a.name, atom_list
		if not a in atom_list:
		    atom = a
		    atom_list.append(a)
		    break
	    else:
		print len(self.bonds_to_atom[id(atom)]),'bonds to ',atom
		return None
	atom_list.append(None)		# angle
	return Bonds.Dihedral(tuple(atom_list))

    def printDanglingBonds(self, verbosity = 0, gen_hooks = 0):
	d_list = []
	for a in self.atomList():
	    i = 0
	    for b in self.bonds_to_atom[id(a)]:
		if not Bonds.isDanglingBond(b): continue
		id_str = 'd%s%s' % (a.name, string.lowercase[i])
		dangling_str = `b`
		if not verbosity:
		    dangling_str = regsub.gsub('Atom '+self.name+'\.','',dangling_str)
		print id_str,'=',dangling_str
		d_list.append(id_str)
		i = i + 1
		if gen_hooks:
		    h = Hook((self, [b]))
		    self.hooks.append(h)
		    setattr(self, 'hook_' + id_str, h)

    def removeDanglingBond(self, db):
	self.danglingbonds.remove(db)
	try:
	    self.bonds_to_atom[id(db.a1)].remove(db)
	except ValueError: pass

    def positionDanglingBonds(self, forcefield):
	#??self._dihedrals_between = MapDihedralsBetween(self.bonds_to_atom)
	#self._dihedrals_between.mark(self)
	for db in self.danglingbonds:
	    pos1 = db.a1.position()
	    if Utility.isDefinedPosition(pos1):
		#print 'positionDanglingBonds',db,db.constraint.angle/Units.deg
		if Vector.isVector(db.direction):
		    dir = db.direction
		else:
		    (dir, constraint, cmbond, far_bond) =\
			  self.nextBondDirection(db.a1, db.constraint,\
						 self.dihedrals, forcefield)
		length = 0.15 #forcefield.defaultBondLength(db, self)
		db.setDirection(length * dir.normal())
		if not db in self.bonds_to_atom[id(db.a1)]:
		    self.bonds_to_atom[id(db.a1)].append(db)
		#print 'positionDanglingBonds pos for',hex(id(db)),db.a1.fullName(),db.DanglingPosition()

    def labelDanglingBonds(self):
	for attr_name in vars(self).keys():
	    attr_val = getattr(self, attr_name)
	    if type(attr_val) == types.InstanceType \
	       and (attr_val.__class__ is Hook or attr_val.__class__ is HookList):
		attr_val.Label(attr_name)
	if hasattr(self, 'hooks'):
	    _labelHooks(self.hooks)


    #def printDanglingBondAngles(self):
	

    def generateNewAtom(self, symbol):
	try:
	    count = self.atom_type_count[symbol]
	except (AttributeError, KeyError):
	    if not hasattr(self, 'atom_type_count'):
		self.atom_type_count = {}
	    count = 0
	    sym_len = len(symbol)
	    for a in self.atomList():
		if a.name[0:sym_len] == symbol:
		    n = string.atoi(a.name[sym_len:])
		    if n > count:
			count = n
	from ChemicalObjects import Atom  # can't import at top, circular ref
	a = Atom(Database.BlueprintAtom(symbol))
	self.atom_type_count[symbol] = count + 1
	a.name = '%s%d' % (symbol, count)
	a.symbol = symbol		# ?
	return a

    def autoType(self, a, forcefield, bond_to_list = None):
	if forcefield.atom_type_property == 'amber_atom_type':
	    if a.symbol == 'H':
		htype = 'HC' # arbitrary?
	    else:
		htype = symbol
	elif hasattr(forcefield, 'deduceAtomType'):
	    htype = forcefield.deduceAtomType(a, bond_to_list, 0)
	else:
	    raise ValueError, 'forcefield provides no default H type'
	self.setType(a, forcefield, htype)

    def setType(self, a, forcefield, type):
	type_dict = getattr(self, forcefield.atom_type_property)
	type_dict[self.getReference(a)] = type
	try:
	    charge = getattr(self, forcefield.charge_property)
	except AttributeError:
	    charge = 0
	#charge_dict[self.getReference(a)] = 0 # arbitrary?

    def hydrogenateDanglingBonds(self, forcefield):
	self.fillDanglingBonds(forcefield, 'H')

    def fillDanglingBonds(self, forcefield, atom_symbol, atom_type = None,\
			  bond_order = 1):
	#bond_order = forcefield.bondOrder(atom)
	if bond_order >= 2:
	    #from DelaunayTriangulation import DelaunayTriangulation
	    #dt = DelaunayTriangulation(self, danglingPositionsOnly = 1)
	    #dbond_list = dt.listPositionsCloserThan(0.075, 3)
	    db_map = DictWithDefault.DictWithDefault([])
	    for i in range(len(self.danglingbonds)):
		db1 = self.danglingbonds[i]
		pos1 = db1.DanglingPosition()
		if not pos1:
		    continue
		db_map[id(db1)] = [db1]
		for j in range(i+1, len(self.danglingbonds)):
		    db2 = self.danglingbonds[j]
		    pos2 = db2.DanglingPosition()
		    if not pos2 or (pos2 - pos1).length() > 0.075:
			continue
		    print db1
		    db_map[id(db1)].append(db2)
	    dbond_list = []
	    for dblist in db_map.values():
		if len(dblist) == bond_order:
		    dbond_list.append(dblist)
	    #print 'fillDanglingBonds'
	    #pprint.pprint(dbond_list)
	else:
	    dbond_list = self.danglingbonds[:]
	times0 = 0
	times1 = 0
	atoms_created = []
	for db_group in dbond_list:
	    if bond_order == 1:
		db_group = [db_group]
	    times0 = times0 - clock()
	    a = self.generateNewAtom(atom_symbol)
	    atoms_created.append(a)
	    a.parent = self
	    self.addAtom(a)
	    if atom_type:
		print 'set atom type',a,atom_type
		self.setType(a, forcefield, atom_type)
	    else:
		bonded_to = []
		for db in db_group:
		    bonded_to.append(db.a1)
		self.autoType(a, forcefield, bonded_to)
	    self.bonds_to_atom[id(a)] = []
	    last_pos = None
	    for db in db_group:
		b = Bonds.Bond((db.a1, a))
		pos1 = db.a1.position()
		#print 'filldb',hex(id(db)),db.a1.fullName()
		try:
		    self.removeDanglingBond(db)
		except ValueError:
		    continue   # db.a1 removed; need cleaner way to handle this
		print 'fillDanglingBonds',hex(id(db)),hex(id(db.a1)),db.a1.fullName(),pos1, b
		if Utility.isDefinedPosition(pos1):
		    times1 = times1 - clock()
		    if Vector.isVector(db.direction):
			dir = db.direction
		    else:
			(dir, constraint, cmbond, far_bond) =\
			  self.nextBondDirection(db.a1, a, self.dihedrals, forcefield)
		    times1 = times1 + clock()
		    length = forcefield.defaultBondLength(b, self)
		    a.setPosition(pos1 + length * dir.normal())
		    self.bonds.append(b)
		    self.bonds_to_atom[id(a)].append(b)
		    if self.bonds_to_atom[id(db.a1)].count(b) > 0: raise RuntimeError
		    self.bonds_to_atom[id(db.a1)].append(b)
		    #print hex(id(a)),a.fullName(),'pos set to',a.position()
		    if last_pos and (last_pos - a.position()).length() > 0.25*length:
			print 'warning - position mismatch', a.fullName(),\
			      (last_pos - a.position()).length(),length, a.position(),last_pos
			print last_dir.angle(dir)/Units.deg
			import RasMol
			ras_session = RasMol.RasMol(self, 1)
			ras_session.Interact()
			raise ConformationError
		    last_pos = a.position()
		    last_dir = dir
		    try:
			del self.alt_bond_dir[id(db.a1)]
		    except (AttributeError, KeyError): pass
		else: print 'no pos',db,pos1
		times0 = times0 + clock()
		#print 'H`nate times',times0,times1
		c = db.constraint
		if c is not None:
		    c.fillAtom(a)
	    more_bonds_needed = forcefield.bondOrder(a, self) - bond_order
	    print 'positionDanglingBonds more_bonds_needed',more_bonds_needed
	    for i in range(more_bonds_needed):
		db = Bonds.DanglingBond((a,))
		self.danglingbonds.append(db)
	self.clearBondAttributes()
	self.setBondAttributes()
	return atoms_created

    def attachBondBetween(self, atom1, atom2):
	db1 = None
	db2 = None
	print 'attachBondBetween',hex(id(atom1)),hex(id(atom2)),atom1.fullName(), atom2.fullName()
	assert(atom1 in self.atoms)
	assert(atom2 in self.atoms)
	for db in self.danglingbonds:
	    if db.a1 is atom1 and not db1:
		db1 = db
	    if db.a1 is atom2 and not db2:
		db2 = db
	assert(db1 and db2)
	self.removeDanglingBond(db1)
	self.removeDanglingBond(db2)
	b = Bonds.lookUpBond(atom1, atom2)
	if b is not None:
	    if b.order < 2:
		b.setOrder(2)
	    else:
		b.setOrder(b.order+1)
	    print 'bond_order increased to',b.order
	    return
	b = Bonds.Bond((atom1, atom2))
	self.bonds.append(b)
	self.bonds_to_atom[id(atom1)].append(b)	# ?? check except
	self.bonds_to_atom[id(atom2)].append(b)
	self.clearBondAttributes()
	self.setBondAttributes()

    def describeBentAngles(self, forcefield):
	from BondedInteractions import angle_energy
	from ForceField import ForceFieldData
	desc_list = []
	global_data = ForceFieldData()
	data = ForceFieldData()
	forcefield.collectAtomTypesAndIndices([self], global_data)
	forcefield.collectCharges([self], global_data)
	for angle in self.bonds.bondAngles():
	    #print 'add bond term',angle
	    forcefield.bonded.addBondAngleTerm(data, angle, self, global_data)
	for a in data.get('angles'):
	    #print 'calc angle',a
	    energy = angle_energy(self.atoms[a[0]].position(),
				  self.atoms[a[1]].position(),
				  self.atoms[a[2]].position(), a[3], a[4])
	    desc_list.append('%7.3f %s %s %s\n' %\
			     (energy, self.atoms[a[0]].fullName(),\
			      self.atoms[a[1]].fullName(),\
			      self.atoms[a[2]].fullName()))
	desc_list.sort()
	return desc_list	

    def _description(self, tag, index_map, toplevel):
	letter, kwargs = self._descriptionSpec()
	s = letter + '(' + `self.name` + ',['
	for o in self._subunits():
	    s = s + o._description(tag, index_map, 0) + ','
	for a in self.atoms:
	    if not hasattr(a, tag):
		s = s + a._description(tag, index_map, 0) + ','
	s = s + ']'
	if toplevel:
	    s = s + ',' + `self._typeName()`
	if kwargs is not None:
	    s = s + ',' + kwargs
	return s + ')'

    def _typeName(self):
	return self.type.name

    def _graphics(self, conf, distance_fn, model, module, options):
	lists = []
	for bu in self.bondedUnits():
	    for a in bu.atomList():
		lists.append(a._graphics(conf, distance_fn, model,
					 module, options))
	    if hasattr(bu, 'bonds'):
		for b in bu.bonds:
		    lists.append(b._graphics(conf, distance_fn, model,
					     module, options))
	return reduce(operator.add, lists)
    is_CompositeChemicalObject = 1

def isCompositeChemicalObject(obj):
    return hasattr(obj, 'is_CompositeChemicalObject')

def verify4thBondDir(dir, a_start, bond3):
    return dir.angle(bond3.asVector(a_start)) > umath.pi/2  # crude rule of thumb

#
#Angle about pos0 needed for distance between pos1, pos2 to be desired_distance
#
def _findDesiredAngle(pos0, pos1, pos2, desired_distance):
    if abs(desired_distance) < 0.0001:
	return 0
    normal = (pos1 - pos0).cross(pos2 - pos0)
    circle1 = Circle(pos1, normal, desired_distance)
    circle2 = Circle(pos0, normal, (pos2 - pos0).length())
    result = circle1.intersectWith(circle2)
    try:
	(intersect1, intersect2) = result
    except TypeError:
	print '_findDesiredAngle no intersect',circle1.radius,circle2.radius,(pos1 - pos0).length(), result
	new_pos = pos0 + (pos1 - pos0).length()/2*(pos1 - pos0)
    else:
	if (intersect2 - pos2).length() < (intersect1 - pos2).length():
	    new_pos = intersect2
	else: new_pos = intersect1
    return (new_pos - pos0).angle(pos1 - pos0)


def _fillEmptyDihedralBonds(cmbond, far_bond, a, bonds_to_atom):
    if cmbond is None:
	for b in bonds_to_atom[id(a)]:
	    other_atom = b.otherAtom(a)
	    if other_atom is not None and len(bonds_to_atom[id(other_atom)]) > 0:
		for b2 in bonds_to_atom[id(other_atom)]:
		    if b2 is not b and b2.a2 is not None:
			far_bond = b2
			cmbond = b
			break
		    if cmbond is not None: break
    elif far_bond is None and cmbond.hasAtom(a):
	other_atom = cmbond.otherAtom(a)
	if other_atom is not None and len(bonds_to_atom[id(other_atom)]) > 0:
	    for b2 in bonds_to_atom[id(other_atom)]:
		if b2 is not cmbond and b2.a2 is not None:
		    far_bond = b2
    return (cmbond, far_bond)

#if string.atof(sys.version) >= 1.5:
#    class ConformationError(Exception):
#	def __init__(self, msg):
#	    Exception.__init__(self, msg)
#else:
ConformationError = 'ConformationError'

def _variationsOnDidhedral(da):
    if (da.angle is None or isSwapWithDangling(da.angle)) and da.b2 is not None:
	if da.a2 is None or da.a3 is None:
	    #print '_variationsOnDidhedral missing middle As',da
	    return []
	if da.a1 is None:
	    if da.a4 is None:
		raise ValueError
	    a = da.a2
	    atuple = (da.a2, da.a3, da.a4, da.angle)
	elif da.a4 is None:
	    a = da.a3
	    index = 3
	    atuple = (da.a3, da.a2, da.a1, da.angle)
	else:
	    #print '_variationsOnDidhedral already complete',da
	    return [da]
	d_list = []
	for other_a in a.bondedTo():
	    if other_a is not atuple[0] and other_a is not atuple[1]:
		d_list.append(Bonds.Dihedral((other_a,) + atuple))# really ought to be DihedralConstraint
	#print 'expanded',da,'to',d_list
	return d_list
    else:
	#print '_variationsOnDidhedral wrong type',da.angle
	return []
#
# create a vector from a list of 6+ candidate vectors that represents
# an average of a subgroup of vectors that are close enough to be
# treated as the same bond direction
#
def _cmpVTuples(t1,t2):
    return cmp(t1[2],t2[2])

def closest_angle(v, a_start, bvlist):
    r = 2*umath.pi
    for bv in bvlist:
	a = v.angle(bv)
	if a < r:
	    r = a
    return r

def selectVector(vlist, a_start, bonds_to_atom):
    assert(len(vlist) >= 2)
    if len(vlist) == 2:
	angle = vlist[0].angle(vlist[1])
	print 'selectVector2',angle/Units.deg
	return ((vlist[0] + vlist[1]).normal(), angle)
    alist = []
    bvlist = []
    for b in bonds_to_atom:
	bvlist.append(b.asVector(a_start))
    n = len(vlist)
    for i in range(n):
	for j in range(i+1, n):
	    closest = closest_angle(vlist[i], a_start, bvlist)
	    alist.append((i, j, vlist[j].angle(vlist[i]), closest))
    
    alist.sort(_cmpVTuples)
		# find gap between good group of vectors and rest
    biggest_jump = 0
    best_index = -1	# where cluster of equivalent vectors ends
    best_angle = 0	# farthest from existing bonds
    small_angle = umath.pi/10
    for j in range(0, len(alist)-1):
	#print alist[j][0],alist[j][1],alist[j+1][2]/Units.deg,alist[j][2]/Units.deg,'biggest',biggest_jump/Units.deg
	if alist[j+1][2] - alist[j][2] > biggest_jump:
	    best_index = j+1
	    biggest_jump = alist[j+1][2] - alist[j][2]
	if alist[j+1][3] < best_angle - small_angle:
	    break
	if alist[j+1][3] > best_angle:
	    best_angle = alist[j+1][3]
    bestv = Vector.null
    best_angle = None
    for i in range(0, best_index):	# add up good vectors
	bestv = bestv + vlist[alist[i][0]] + vlist[alist[i][1]]
	best_angle = alist[i][2]
	#print alist[i][0],alist[i][1],closest_angle(bestv,a_start,bvlist)/Units.deg,
	#print 'selectVector ',i,' add ',alist[i][2]/Units.deg
    return (bestv.normal(), best_angle)

def translateAtomsBy(a, translate_v, bonds_to_atom, xlated_atoms):
    if not Utility.isDefinedPosition(a.position()):
	return
    if not xlated_atoms.has_key(id(a)):
	a.translateBy(translate_v)
	#print 'translate',a.fullName()[-6:]
	xlated_atoms[id(a)] = 1
	for b in bonds_to_atom[id(a)]:
	    translateAtomsBy(b.otherAtom(a), translate_v, bonds_to_atom, xlated_atoms)

class MapDihedralsBetween:
    def __init__(self, bonds_to_atom):
	self.bonds_to_atom = bonds_to_atom
	self._group = DictWithDefault.DictWithDefault(None)
	self._dihedrals = DictWithDefault.DictWithDefault([])
	self._dihedral_keys = DictWithDefault.DictWithDefault([])
	self.atom_pairs_per_uc_dihedral = DictWithDefault.DictWithDefault([])
	self._dihedrals_about = DictWithDefault.DictWithDefault([])

    def closerGroup(self, atom1, atom2, otheratom):
	obj1 = self._group[id(atom1)]
	obj2 = self._group[id(atom2)]
	obj3 = self._group[id(otheratom)]
	list1 = []
	list2 = []
	list3 = []
	while obj1:
	    list1.append(obj1)
	    obj1 = self._group[id(obj1)]
	while obj2:
	    list2.append(obj2)
	    obj2 = self._group[id(obj2)]
	while obj3:
	    list3.append(obj3)
	    obj3 = self._group[id(obj3)]
	list1.reverse()
	list2.reverse()
	list3.reverse()
	for i in range(min(len(list1), min(len(list2), len(list3)))):
	    if not (list1[i] is list2[i]):
		return 0
	    if not (list2[i] is list3[i]):
		return 1
	return 0

    def groupsSpanned(self, b, dbug):
	obj1 = self._group[id(b.a1)]
	obj2 = self._group[id(b.a2)]
	list1 = []
	list2 = []
	while obj1:
	    list1.append(obj1)
	    obj1 = self._group[id(obj1)]
	while obj2:
	    list2.append(obj2)
	    obj2 = self._group[id(obj2)]
	list1.reverse()
	list2.reverse()
	if dbug:
	    for o in list1: print hex(id(o)),o.name
	    print
	    for o in list2: print hex(id(o)),o.name
	for i in range(min(len(list1), len(list2))):
	    if not (list1[i] is list2[i]):
		return (list1[i], list2[i])
	return None

    def atomIn(self, bond, grp):
	obj1 = self._group[id(bond.a1)]
	while obj1:
	    if obj1 is grp:
		return bond.a1
	    obj1 = self._group[id(obj1)]
	obj2 = self._group[id(bond.a2)]
	while obj2:
	    if obj2 is grp:
		return bond.a2
	    obj2 = self._group[id(obj2)]
	return None

    def spans(self, b, prior_group, new_group):
	obj1 = self._group[id(b.a1)]
	obj2 = self._group[id(b.a2)]
	while obj1 and obj1 is not prior_group and obj1 is not new_group:
	    obj1 = self._group[id(obj1)]
	while obj2 and obj2 is not prior_group and obj2 is not new_group:
	    obj2 = self._group[id(obj2)]
	if 0:
	    print 'spans?',b
	    print (obj1 is obj2),obj1,obj2,hex(id(prior_group)),hex(id(new_group))
	if not obj1:
	    return 0
	if not obj2:
	    return 0
	return not (obj1 is obj2)

    def sameGroup(self, obj1, obj2):
	return self._group[id(obj1)] is self._group[id(obj2)]

    def addDihedral(self, d):
	#print 'addDihedral',d
	self._dihedrals_about[id(d.b2)].append(d)
	mark_used1 = { id(d.a3) : 1}
	mark_used2 = { id(d.a2) : 1}
	a1 = d.a1
	a2 = d.a4
	if a1 is None:
	    if d.a2 is None:
		pass #print 'cant use uc',d
	    else:
		for dvar in _variationsOnDidhedral(d):
		    self.addDihedralBetween(dvar, dvar.a1, dvar.a4,\
					    mark_used1, mark_used2)
		    self._dihedrals_about[id(d.b2)].append(dvar)
	    return
	if a2 is None:
	    if d.a3 is None:
		pass # print 'cant use uc',d
	    else:
		for dvar in _variationsOnDidhedral(d):
		    self.addDihedralBetween(dvar, dvar.a1, dvar.a4,\
					    mark_used1, mark_used2)
		    self._dihedrals_about[id(d.b2)].append(dvar)
	    return
	self.addDihedralBetween(d, a1, a2, mark_used1, mark_used2)

    def addDihedralBetween(self, d, a1, a2, mark_used1, mark_used2):
	#pred_map = ShortestPath.Dijkstra_Algorithm(a1, )
	list1 = self._listAtomsFrom(a1, mark_used1)
	list2 = self._listAtomsFrom(a2, mark_used2)
	#print 'alists:',hex(id(a1)),hex(id(a2)),a1.fullName(), a2.fullName()
	#pprint.pprint(list1)
	#pprint.pprint(list2)
	grp1 = self._group[id(a1)]
	grp2 = self._group[id(a2)]
	if grp2 is None:
	    o1 = a2.parent_subgroup
	    while o1 is not None:
		print hex(id(o1)),o1
		o1 = o1.parent
	    print a2.fullName()
	if grp1 is None:
	    o1 = a1.parent_subgroup
	    while o1 is not None:
		print hex(id(o1)),o1
		o1 = o1.parent
	    print a1.fullName()
	assert(grp1 is not None)
	assert(grp2 is not None)
	dlist = [d]
	if grp1 is grp2:
	    for atom1 in list1:
		for atom2 in list2:
		    self._add1(dlist, id(atom1), id(atom2))
	else:
	    for atom1 in list1:
		self._add1(dlist, id(atom1), id(grp1))
	    for atom2 in list2:
		self._add1(dlist, id(grp2), id(atom2))

    def _add1(self, dlist, id1, id2):
	key = (id1, id2)
	#print 'adducpath',hex(id1),hex(id2),key,hex(id(self._dihedrals[`key`])),'self',hex(id(self)),hex(id(self._dihedrals)),hex(id(dlist[0])),dlist
	assert(id2 != id(None))
	for d in dlist:
	    self._dihedrals[`key`].append(d)
	    if self.atom_pairs_per_uc_dihedral[id(d)].count(key) == 0:
		self.atom_pairs_per_uc_dihedral[id(d)].append(key)
		#print 'add1key',hex(id1),hex(id2),hex(id(d)),d
	    else:
		pass
		#print 'addU dup key',hex(id1),hex(id2),d
		#raise RuntimeError
	self._dihedral_keys[id1].append(key)
	self._dihedral_keys[id2].append(key)
	#print '_dihedral_keys lens',len(self._dihedral_keys[id1]),\
	#len(self._dihedral_keys[id2]),obj1.fullName()

    def _listAtomsFrom(self, a, mark_used):
	try:
	    list1 = self.bonds_to_atom[id(a)]
	except KeyError: list1 = []
	rlist = []
	if not mark_used.has_key(id(a)):
	    rlist.append(a)
	    mark_used[id(a)] = 1
	#print '_listAtomsFrom',a.name,list1
	for b in list1:
	    new_atom = b.otherAtom(a)
	    #print 'list?',new_atom,mark_used.has_key(id(new_atom)),hex(id(self._group[id(new_atom)])),hex(id(self._group[id(a)]))
	    if not mark_used.has_key(id(new_atom)) and\
	       self._group[id(new_atom)] is self._group[id(a)]:
		mark_used[id(new_atom)] = 1
		rlist.append(new_atom)
		rlist = rlist + self._listAtomsFrom(new_atom, mark_used)
	return rlist

    def addBond(self, b):
	#print 'addBond',hex(id(b.a1)),hex(id(b.a2)),b
	if not self._dihedrals.has_key((id(b.a1),id(b.a2)))\
	   and not self._dihedrals.has_key((id(b.a2),id(b.a1))):
	    marked = {}
	    list1 = self._dihedral_keys[id(b.a1)][:]
	    list2 = self._dihedral_keys[id(b.a2)][:]
	    for key in list1:
		#print 'list1',hex(key[0]),hex(key[1])
		self._addDihedralKey(key, b.a1, b.a2, marked)
	    for key in list2:
		#print 'list2',hex(key[0]),hex(key[1])
		self._addDihedralKey(key, b.a2, b.a1, marked)

    def _addDihedralKey(self, key, a1, a2, marked):
	if key[0] == id(a1):
	    id1 = id(a2)
	    id2 = o = key[1]
	else:
	    assert(key[1] == id(a1))
	    id1 = o = key[0]
	    id2 = id(a2)
	#print 'addu',hex(id(a1)),hex(id(a2)),'new_key',hex(id1),hex(id2),a2.name
	if 0:
	    for k in (id1,id2):
		group = self._group[k]
		depth = 0
		while group:
		    print depth,'group',hex(id(group))
		    group = self._group[id(group)]
		    depth = depth + 1
	while 1:
	    if not marked.has_key((id1, id2)):
		self._add1(self._dihedrals[`key`], id1, id2)
		marked[(id1, id2)] = 1
	    #print hex(o),'==>',hex(id(self._group[o])),hex(id1),hex(id2)
	    o = self._group[o]
	    if not o:
		break
	    dir = self._orientation(self._dihedrals[`key`][0], o)
	    if not dir:
		break
	    if dir > 0:
		#print 'merge1',hex(id(o)), hex(id(a2))
		id1 = id(o)
		id2 = id(a2)
	    else:
		#print 'merge2',hex(id(a2)), hex(id(o))
		id1 = id(a2)
		id2 = id(o)
	    o = id(o)
	if len(self._dihedrals[`key`]) > 20000:
	    print a1, a2
	    raise RuntimeError
	
    def findAbout(self, bond):
	return self._dihedrals_about[id(bond)]

    def removeAbout(self, bond):
	for d in self._dihedrals_about[id(bond)]:
	    self.removeDihedral(d)
	    #print '1 underconstrained_dihedrals remove',hex(id(d)),d

    def removeDihedral(self, d):
	for key in self.atom_pairs_per_uc_dihedral[id(d)]:
	    #print 'removeDihedral',hex(id(d)),hex(key[0]),hex(key[1])
	    self._dihedrals[`key`].remove(d)
	    if len(self._dihedrals[`key`]) == 0:
		del self._dihedrals[`key`]
	    try:
		self._dihedral_keys[key[0]].remove(key)
	    except ValueError:
		print 'warning - key missing'
	    try:
		self._dihedral_keys[key[1]].remove(key)
	    except ValueError:
		print 'warning - key missing'
	del self.atom_pairs_per_uc_dihedral[id(d)]
	#print 'removeDihedral',hex(id(d)),d

    def mark(self, obj, depth = 0):
	try:
	    attach = obj.attach_subgroups
	except AttributeError:
	    attach = []
	print '  '*depth,'mark obj',hex(id(obj)),
	for a1 in attach:
	    print hex(id(a1)),
	print
	if hasattr(obj, 'subgroups') and obj.subgroups:
	    for grp in obj.subgroups:
		print '  '*depth,'mark grp',hex(id(grp))
		self.mark(grp, depth + 1)
		self._group[id(grp)] = obj
	else:
	    for a in obj.atoms:
		print '  '*depth,'M',hex(id(a)),a.fullName()
		self._group[id(a)] = obj

    def isIn(self, a, obj):
	print 'isIn',hex(id(a)), hex(id(self._group[id(a)])), hex(id(obj)), a.fullName(), obj
	return self._group[id(a)] is obj

    def _orientation(self, d, grp):
	grp1 = self._group[id(d.a1)]
	grp2 = self._group[id(d.a4)]
	while grp1 and grp2 and not (grp1 is grp2):
	    if grp1 is grp:
		return 1
	    if grp2 is grp:
		return -1
	    grp1 = self._group[id(grp1)]
	    grp2 = self._group[id(grp2)]
	#print 'left',hex(id(grp1)),hex(id(grp2))
	return 0

    def findDihedralsOnPath(self, atom1, atom2):
	group1 = self._group[id(atom1)]
	group2 = self._group[id(atom2)]
	super1 = group1
	path1 = [group1]
	super2 = group2
	path2 = [group2]
	if group1 is group2:
	    r = []
	    for key2 in ((id(atom1), id(atom2)), \
			 (id(self._group[id(atom1)]), id(atom2)),\
			 (id(atom1), id(self._group[id(atom2)]))):
		key1 = (key2[1], key2[0])
		#print 'findDihedralsOnPath group1 is group2', hex(key1[0]), hex(key1[1]), atom1.fullName(), atom2.fullName()
		#print self._dihedrals[`key1`]
		#print self._dihedrals[`key2`]
		r.append((self._dihedrals[`key1`], self._dihedrals[`key2`]))
	    return r
	super1_depth = self._findDepth(group1)
	super2_depth = self._findDepth(group2)
	while not (super1 is super2):
	    if super1_depth >= super2_depth:
		if self._group[id(super1)]:
		    super1 = self._group[id(super1)]
		    path1.append(super1)
		    #print 'ascend1',super1_depth,hex(id(super1)),super1
		    super1_depth = super1_depth - 1
		    if super1_depth < 0:
			raise RuntimeError, 'internal error detected in MapDihedralsBetween.findDihedralsOnPath'
	    if super2_depth > super1_depth:
		if self._group[id(super2)]:
		    super2 = self._group[id(super2)]
		    path2.append(super2)
		    #print 'ascend2',super2_depth,hex(id(super2)),super2
		    super2_depth = super2_depth - 1
		    if super2_depth < 0:
			raise RuntimeError, 'internal error detected in MapDihedralsBetween.findDihedralsOnPath'
	#print 'findDihedralsOnPath',hex(id(atom1)), hex(id(atom2)), atom1, atom2
	return self._doPath(path1, super2, atom1, 0) +\
	       self._doPath(path2, super1, atom2, 1)

    def _doPath(self, group_list, other, a, reverseTuples):
	if len(group_list) == 1:
	    if 0:
		key1 = (id(group_list[0]), id(a))
		key2 = (id(a), id(group_list[0]))
		if not self._dihedrals[`key1`] and not self._dihedrals[`key2`]:
		    print '_doPath',hex(id(group_list[0])),hex(id(a)),\
			  key1,hex(id(self._dihedrals[`key1`])),\
			  key2,hex(id(self._dihedrals[`key2`])),\
			  group_list[0],a
		else: print '_doPath',hex(id(group_list[0])),hex(id(a))
	    if reverseTuples:
		r = (self._dihedrals[`(id(a), id(group_list[0]))`],\
		      self._dihedrals[`(id(group_list[0]), id(a))`])
	    else:
		r = (self._dihedrals[`(id(group_list[0]), id(a))`],\
		      self._dihedrals[`(id(a), id(group_list[0]))`])
	    if len(r[0]) + len(r[1]) == 0:
		return []
	    return [r]
	if reverseTuples:
	    r = (self._dihedrals[`(id(other), id(group_list[-1]))`],\
		  self._dihedrals[`(id(group_list[-1]), id(other))`])
	else:
	    r = (self._dihedrals[`(id(group_list[-1]), id(other))`],\
		  self._dihedrals[`(id(other), id(group_list[-1]))`])
	r2 = self._doPath(group_list[:-1], group_list[-1], a, reverseTuples)
	if not r2:
	    ret = [r]
	elif len(r[0]) + len(r[1]) == 0:
	    ret = r2
	else:
	    ret = [r] + r2
	#print '_doPath',len(group_list),len(ret),len(r),hex(id(group_list[-1])), hex(id(other)),a,other,ret
	return ret

    def _findDepth(self, group):
	depth = 0
	while group:
	    group = self._group[id(group)]
	    depth = depth + 1
	return depth

class SwapWithDangling:
    def __init__(self, b1, angle1, b2, angle2, orientation):
	self.bond1 = b1
	self.bond2 = b2
	self.angle1 = angle1
	self.angle2 = angle2
	self.orientation = orientation

    is_SwapWithDangling = 1

    def __repr__(self):
	return 'SwapWithDangling(' + `self.bond1` + ',' + `self.angle1` +\
	       ',' + `self.bond2` + ',' + `self.angle2` + ')'
    __str__ = __repr__

    def hasAtom(self, a):
	return self.bond1.hasAtom(a) or self.bond2.hasAtom(a)

    def altDirection(self):
	ca = Bonds.findCommonAtom(self.bond1, self.bond2)
	cone1 = Cone(ca.position(), self.bond1.asVector(ca), self.angle1)
	cone2 = Cone(ca.position(), self.bond2.asVector(ca), self.angle2)
	result = _intersectConeCone(cone1,cone2)
	if type(result) is type(()) and len(result) == 2:
	    x = cone1.axis.cross(cone2.axis)
	    if x * result[0] * self.orientation > 0:
		return (result[0], ca)
	    return (result[1], ca)
	raise RuntimeError,'SwapWithDangling.altDirection failed for ' + `self`

def _MapBondPos(grp, bond_pos, pos = 0):
    if not hasattr(grp, 'subgroups'):
	return pos
    for subgrp in grp.subgroups:
	pos = _MapBondPos(subgrp, bond_pos, pos)
    bond_pos[id(grp)] = pos
    return pos + 1

def isSwapWithDangling(obj):
    return hasattr(obj, 'is_SwapWithDangling')

def _labelHooks(hlist, basename = 'hooks['):
    print ' '*len(basename),'_labelHooks',len(hlist),basename
    for i in range(len(hlist)):
	hook = hlist[i]
	if type(hook) == type([]):
	    _labelHooks(hook, basename + `i` + '][')
	else:
	    hook.Label(basename + `i` + ']')

def _replaceBond(grp, b, new_b):
    if hasattr(grp, 'subgroups'):
	for subgrp in grp.subgroups:
	    _replaceBond(subgrp, b, new_b)
    try:
	pos = grp.bonds.index(b)
    except ValueError:
	return
    #print 'replacebb',hex(id(b)),'with',hex(id(new_b))
    #print grp.bonds[pos]
    #print new_b
    grp.bonds[pos] = new_b

def _removeBonds(grp, to_delete):
    if hasattr(grp, 'subgroups'):
	for subgrp in grp.subgroups:
	    _removeBonds(subgrp, to_delete)
    new_blist = []
    for b in grp.bonds:
	if not to_delete.has_key(b.a1) and\
	   not to_delete.has_key(b.a2) and not to_delete.has_key(id(b)):
	    new_blist.append(b)
	    #print '_removeBonds keep',hex(id(grp)),hex(id(b)),b
	#else: print '_removeBonds',hex(id(grp)),hex(id(b)),b
    grp.bonds = Bonds.BondList(new_blist)


def _collectDihedrals(bondlist, dihedrals_by_bond):
    dihedral_list = []
    for b in bondlist:
	for d in dihedrals_by_bond[id(b)]:
	    try:
		if d in dihedrals_by_bond[id(d.b1)] and\
		   d in dihedrals_by_bond[id(d.b2)] and\
		   d in dihedrals_by_bond[id(d.b3)]:
		    dihedral_list.append(d)
	    except KeyError: pass
    return Bonds.DihedralAngleList(dihedral_list)

class RotationParms:

    def __init__(self, dihedral, error = 1.e9, angle = 0.0, axis = None,\
		 atom3 = None, atoms_rotating_other_way = None, warn_msg=None):
	self.dihedral = dihedral
	self.error = error
	self.angle = angle
	self.axis = axis
	self.atom3 = atom3
	self.atoms_to_reverse = atoms_rotating_other_way
	self.warn_msg = warn_msg
