# This module deals with input and output of configurations.
#
# Written by Konrad Hinsen
# last revision: 1998-2-25
#

import ChemicalObjects, Collection, Database, Units, Universe, Utility
from Geometry import Sphere, Cone, Plane, Line
from Vector import Vector
import PDB, VRML
import copy, os

#
# Error type
#
IOError = 'IOError'

#
# This class represents a PDB file from which
# arbitrary objects can be read.
#
class PDBFile:

    def __init__(self, filename, map = 'pdbmap', alt = 'pdb_alternative'):
	filename = Database.PDBPath(filename)
	if not os.path.exists(filename):
	    raise IOError, 'File ' + filename + ' not found'
	self.filename = filename
	self.map = map
	self.alt = alt

    substitute = 1

    def readConfiguration(self):
	file = PDB.PDBFile(self.filename)
	conf = file.readConfiguration()
	file.close()
	return conf

    def readSequence(self):
	file = PDB.PDBFile(self.filename)
	seq = file.readSequence()
	file.close()
	return seq

    def readSequenceWithConfiguration(self):
	file = PDB.PDBFile(self.filename)
	seq = file.readSequenceWithConfiguration()
	file.close()
	return seq

    def applyTo(self, object):
	conf = self.readConfiguration()
	PDBConfiguration(conf, self.map, self.alt).applyTo(object)

#
# This class represents a PDB configuration as read from a PDB file.
#
class PDBConfiguration:

    def __init__(self, conf, map = 'pdbmap', alt = 'pdb_alternative'):
	self._conf = conf
	self._map = map
	self._alt = alt

    def applyTo(self, object, initial = 0):
	if self._conf.hasChains() and hasattr(object, 'is_protein'):
	    for o, c in map(None, object, self._conf):
		self._applyTo(o, initial, c)
	else:
	    self._applyTo(object, initial, self._conf)

    def _applyTo(self, object, initial, conf):
	total = len(object.atomList())
	nres = len(conf)
	alt = None
	if hasattr(object, self._map):
	    map = getattr(object, self._map)
	    try: alt = getattr(object, self._alt)
	    except AttributeError: pass
	    if len(map) != nres:
		raise IOError, 'PDB configuration does not match object ' + \
		      object.fullName()
	    for i in range(nres):
		total = total - self._applyResidue(conf[i], map[i], object, alt)
	elif Collection.isCollection(object) or Universe.isUniverse(object):
	    if len(object) != nres:
		raise IOError, 'PDB configuration does not match object ' + \
		      object.fullName()
	    for i in range(nres):
		o = object[i]
		map = getattr(o, self._map)
		try: alt = getattr(o, self._alt)
		except AttributeError: pass
		total = total - self._applyResidue(conf[i], map[0], o, alt)
	else:
	    if len(object.groups) != nres:
		raise IOError, 'PDB configuration does not match object ' + \
		      object.fullName()
	    for i in range(nres):
		g = object.groups[i]
		map = getattr(g, self._map)
		try: alt = getattr(g, self._alt)
		except AttributeError: pass
		total = total - self._applyResidue(conf[i], map[0], g, alt)
	if total > 0 and not initial:
	    name = '[unnamed object]'
	    try:
		name = object.fullName()
	    except: pass
	    if name: name = ' in ' + name
	    Utility.warning(`total` + ' atom(s)' + name +
			    ' were not assigned (new) positions.')

    def _applyResidue(self, res, map, object, alt):
	resnameo = map[0]
	map = map[1]
	defined = 0
	for atom in res:
	    name = atom.name
	    if alt:
		try: name = alt[name]
		except KeyError: pass
	    if map.has_key(name):
		object.setPosition(map[name], atom.position*Units.Ang)
		try:
		    object.setIndex(map[name], atom.number-1)
		except (AttributeError, ValueError):
		    pass
		defined = defined + 1
	    elif object.isSubsetModel():
		pass
	    else:
		raise IOError, 'Atom ' + atom.name + ' of PDB residue ' + \
		      res.name + ' not found in residue ' + resnameo + \
		      ' of object ' + object.fullName()
	return defined

#
# This class represents a Z-Matrix. Z-Matrix data consists of a list
# with one element for each atom being defined. Each entry is a
# list containing the data defining the atom.
#
class ZMatrix:

    def __init__(self, data):
	self.data = data
	self.coordinates = {}

    substitute = 1

    def findPositions(self):
	# First atom at origin
	self.coordinates[self.data[0][0]] = Vector(0,0,0)
	# Second atom along x-axis
	self.coordinates[self.data[1][0]] = Vector(self.data[1][2],0,0)
	# Third atom in xy-plane
	pos1 = self.coordinates[self.data[2][1]]
	pos2 = self.coordinates[self.data[2][3]]
	sphere = Sphere(pos1, self.data[2][2])
	cone = Cone(pos1, pos2-pos1, self.data[2][4])
	plane = Plane(Vector(0,0,0), Vector(0,0,1))
	points = sphere.intersectWith(cone).intersectWith(plane)
	self.coordinates[self.data[2][0]] = points[0]
	# All following atoms defined by distance + angle + dihedral
	for entry in self.data[3:]:
	    pos1 = self.coordinates[entry[1]]
	    pos2 = self.coordinates[entry[3]]
	    pos3 = self.coordinates[entry[5]]
	    distance = entry[2]
	    angle = entry[4]
	    dihedral = entry[6]
	    sphere = Sphere(pos1, distance)
	    cone = Cone(pos1, pos2-pos1, angle)
	    plane123 = Plane(pos3, pos2, pos1)
	    plane234 = plane123.rotate(Line(pos1, pos2), dihedral)
	    points = sphere.intersectWith(cone).intersectWith(plane234)
	    for p in points:
		if (pos3-pos2).cross(p-pos3)*(plane234.normal) > 0:
		    self.coordinates[entry[0]] = p
		    break

    def applyTo(self, object):
	if not len(self.coordinates):
	    self.findPositions()
	for entry in self.coordinates.items():
	    object.setPosition(entry[0], entry[1])
	object.normalizePosition()

#
# This class represents a dictionary of Cartesian positions
#
class Cartesian:

    def __init__(self, dict):
	self.dict = dict

    substitute = 1

    def applyTo(self, object):
	for a, r in self.dict.items():
	    object.setPosition(a, Vector(r[0], r[1], r[2]))

#
# This object represents a PDB file for output.
#
class PDBOutputFile:

    def __init__(self, filename):
	self.file = PDB.PDBFile(filename, 'w')
	self.warning = 0
	self.atom_no = {}
	self.atom_sequence = []

    def write(self, object, configuration = None, tag = None):
	if not ChemicalObjects.isChemicalObject(object):
	    for o in object:
		self.write(o, configuration)
	else:
	    toplevel = tag is None
	    if toplevel:
		tag = Utility.uniqueAttribute()
	    if hasattr(object, 'pdbmap'):
		for residue in object.pdbmap:
		    self.file.nextResidue(residue[0])
		    for atom_name, atom in residue[1].items():
			atom = object.getAtom(atom)
			p = atom.position(configuration)
			if Utility.isDefinedPosition(p):
			    try: occ = atom.occupancy
			    except AttributeError: occ = 0.
			    try: temp = atom.temperature_factor
			    except AttributeError: temp = 0.
			    self.file.writeAtom(atom_name, p/Units.Ang,
						occ, temp)
			    self.atom_sequence.append(atom)
			    self.atom_no[id(atom)] = self.file.atom_num
			else:
			    self.warning = 1
			setattr(atom, tag, None)
	    else:
		if hasattr(object, 'is_protein'):
		    for chain in object:		    
			self.file.nextChain()
			self.write(chain, configuration, tag)
		    self.file.noChain()
		elif hasattr(object, 'molecules'):
		    for m in object.molecules:
			self.write(m, configuration, tag)
		elif hasattr(object, 'groups'):
		    for g in object.groups:
			self.write(g, configuration, tag)
	    if toplevel:
		for a in object.atomList():
		    if not hasattr(a, tag):
			self.write(a, configuration, tag)
		    delattr(a, tag)

    def close(self):
	self.file.close()
	if self.warning:
	    Utility.warning('Some atoms are missing in the output file ' + \
			    'because their positions are undefined.')
	    self.warning = 0

class PDBConnectOutputFile(PDBOutputFile):
    def __init__(self, filename):
	PDBOutputFile.__init__(self, filename)
	self.top_level = 1

    def write(self, object, configuration = None, tag = None):
	PDBOutputFile.write(self, object, configuration, tag)
	if tag is None:
	    if hasattr(object, 'bonds_to_atom'):
		id_list = object.bonds_to_atom.keys()
		id_list.sort()
		for id_a in id_list:
		    blist = object.bonds_to_atom[id_a]
		    connect_list = []
		    double_bonds = []
		    triple_bonds = []
		    for b in blist:
			try:
			    a2 = b.otherAtomFromId(id_a)
			except ValueError:
			    pass # ignore DanglingBond.otherAtomName complaints
			else:
			    a_no = self.atom_no[id(a2)]
			    connect_list.append(a_no)
			    if b.order >= 2:
				double_bonds.append(a_no)
			    if b.order >= 3:
				triple_bonds.append(a_no)
		    try:
			a_no = self.atom_no[id_a]
		    except KeyError: pass
		    else:
			self.file.writeBond(a_no, connect_list)
			if double_bonds:
			    self.file.writeBond(a_no, double_bonds)
			if triple_bonds:
			    self.file.writeBond(a_no, triple_bonds)
	    else:
		Utility.warning('No CONECT lines written because no '
				+ 'bonds_to_atom list found.')
#
# VRML output
#
class VRMLWireframeFile(VRML.VRMLFile):

    def __init__(self, filename):
	VRML.VRMLFile.__init__(self, filename, 'w')
	self.warning = 0

    def write(self, object, configuration = None, distance = None):
	if distance is None:
	    try:
		distance = object.universe().distanceVector
	    except AttributeError:
		distance = Universe.InfiniteUniverse().distanceVector
	if not ChemicalObjects.isChemicalObject(object):
	    for o in object:
		self.write(o, configuration, distance)
	else:
	    for bu in object.bondedUnits():
		for a in bu.atomList():
		    self.writeAtom(a, configuration)
		if hasattr(bu, 'bonds'):
		    for b in bu.bonds:
			self.writeBond(b, configuration, distance)

    def close(self):
	VRML.VRMLFile.close(self)
	if self.warning:
	    Utility.warning('Some atoms are missing in the output file ' + \
			    'because their positions are undefined.')
	    self.warning = 0

    def writeAtom(self, atom, configuration):
	pass

    def writeBond(self, bond, configuration, distance):
	p1 = bond.a1.position(configuration)
	p2 = bond.a2.position(configuration)
	if p1 is not None and p2 is not None:
	    bond_vector = 0.5*distance(bond.a1, bond.a2, configuration)
	    cut = bond_vector != 0.5*(p2-p1)
	    material1 = material = VRML.DiffuseMaterial(bond.a1.color)
	    material2 = material = VRML.DiffuseMaterial(bond.a2.color)
	    if material1 == material2 and not cut:
		c = VRML.Line(p1, p2, material = material1)
		c.writeToFile(self)
	    else:
		c = VRML.Line(p1, p1+bond_vector, material = material1)
		c.writeToFile(self)
		c = VRML.Line(p2, p2-bond_vector, material = material2)
		c.writeToFile(self)


class VRMLHighlight(VRMLWireframeFile):

    def writeAtom(self, atom, configuration):
	try:
	    highlight = atom.highlight
	except AttributeError:
	    highlight = 0
	if highlight:
	    p = atom.position(configuration)
	    if p is None:
		self.warning = 1
	    else:
		s = VRML.Sphere(p, 0.1*Units.Ang,
				material = VRML.DiffuseMaterial(atom.color),
				reuse = 1)
		s.writeToFile(self)


class VRMLBallAndStickFile(VRMLWireframeFile):

    def writeAtom(self, atom, configuration):
	p = atom.position(configuration)
	if p is None:
	    self.warning = 1
	else:
	    s = VRML.Sphere(p, 0.1*Units.Ang,
			    material = VRML.DiffuseMaterial(atom.color),
			    reuse = 1)
	    s.writeToFile(self)

    def writeBond(self, bond, configuration, distance):
	p1 = bond.a1.position(configuration)
	p2 = bond.a2.position(configuration)
	if p1 is not None and p2 is not None:
	    bond_vector = 0.5*distance(bond.a1, bond.a2, configuration)
	    cut = bond_vector != 0.5*(p2-p1)
	    material1 = material = VRML.DiffuseMaterial(bond.a1.color)
	    material2 = material = VRML.DiffuseMaterial(bond.a2.color)
	    if material1 == material2 and not cut:
		c = VRML.Cylinder(p1, p2, 0.03*Units.Ang,
				  material = material1)
		c.writeToFile(self)
	    else:
		c = VRML.Cylinder(p1, p1+bond_vector, 0.03*Units.Ang,
				  material = material1)
		c.writeToFile(self)
		c = VRML.Cylinder(p2, p2-bond_vector, 0.03*Units.Ang,
				  material = material2)
		c.writeToFile(self)


class VRMLChargeFile(VRMLWireframeFile):

    color_scale = VRML.SymmetricColorScale(1.)

    def writeAtom(self, atom, configuration):
	p = atom.position(configuration)
	c = atom.charge()
	c = max(min(c, 1.), -1.)
	if p is None:
	    self.warning = 1
	else:
	    s = VRML.Sphere(p, 0.1*Units.Ang,
			    material = VRML.Material(diffuse_color =
						     self.color_scale(c)))
	    s.writeToFile(self)

    bond_material = VRML.DiffuseMaterial('black')

    def writeBond(self, bond, configuration, distance):
	p1 = bond.a1.position(configuration)
	p2 = bond.a2.position(configuration)
	if p1 is not None and p2 is not None:
	    bond_vector = 0.5*distance(bond.a1, bond.a2, configuration)
	    cut = bond_vector != 0.5*(p2-p1)
	    if not cut:
		c = VRML.Line(p1, p2, material = self.bond_material)
		c.writeToFile(self)
	    else:
		c = VRML.Line(p1, p1+bond_vector, material = self.bond_material)
		c.writeToFile(self)
		c = VRML.Line(p2, p2-bond_vector, material = self.bond_material)
		c.writeToFile(self)

VRMLFile = VRMLWireframeFile

#
# Recognize some standard file types by their extensions
#
def fileFormatFromExtension(filename):
    filename, ext = os.path.splitext(filename)
    if ext in _file_compressions:
	filename, ext = os.path.splitext(filename)
    try:
	return _file_formats[ext]
    except KeyError:
	raise IOError, 'Unknown file format'

_file_formats = {'.pdb': 'pdb', '.wrl': 'vrml'}
_file_compressions = ['.gz', '.Z']

#
# Output file for a specified format
#
def OutputFile(filename, format = None):
    if format is None:
	format = fileFormatFromExtension(filename)
    return _file_types[format](filename)

_file_types = {'pdb': PDBOutputFile, 'pdb.connect': PDBConnectOutputFile,
	       'vrml': VRMLFile,
	       'vrml.wireframe': VRMLWireframeFile,
	       'vrml.highlight': VRMLHighlight,
	       'vrml.ball_and_stick': VRMLBallAndStickFile,
	       'vrml.charge': VRMLChargeFile}
