# Interface to DL_POLY 2.x.
#
# Written by Konrad Hinsen
# last revision: 1997-8-28
#

# Attention: at the moment this code can handle only the Amber force field.
# It produces input files for a modified version of DLPOLY to which a
# three-term dihedral potential has been added. Without this addition,
# DLPOLY cannot handle the Amber potential.

from ForceField import EnergyEvaluator
from FortranFormat import FortranFormat, FortranLine
import Units
from Numeric import sqrt
import copy, operator

from AmberData import Amber94


# Choose energy unit for output file
energy_unit = Units.kJ/Units.mol
energy_unit_name = 'kJ'
length_unit = Units.Ang
angle_unit = Units.deg

# Conversion factors
energy_factor = 1./energy_unit
length_factor = 1./length_unit
angle_factor = 1./angle_unit

# Class that keeps track of excluded pairs
class ExclusionList:

    def __init__(self):
	self.dict = {}

    def add(self, pair):
	i1, i2 = pair
	if i1 > i2:
	    pair = (i2, i1)
	self.dict[pair] = 1

    def checkPair(self, pair):
	i1, i2 = pair
	if i1 > i2:
	    pair = (i2, i1)
	return self.dict.has_key(pair)

# Enhanced energy evaluator class that writes DLPOLY files
class DLPOLY(EnergyEvaluator):

    def __init__(self, universe, force_field, title = ' ',
		 conf_file = 'CONFIG', field_file = 'FIELD'):
	EnergyEvaluator.__init__(self, universe, force_field)
	self.bonded = self.ff.bondedForceFields()
	if len(self.bonded) > 1:
	    raise ValueError, "Can't handle this force field"
	else:
	    self.bonded = self.bonded[0]
	objects = copy.copy(self.universe.objectList())
	objects.sort(lambda a, b: cmp(a.type, b.type))

	# this information should be obtained properly from the force field
	self.atom_type_property = 'amber_atom_type'
	self.atom_charge_property = 'amber_charge'
	self.lj_1_4 = 1./2.
	self.electrostatic_1_4 = 1./1.2

	index_dict = {}
	atom_types = {}
	j = 0
	for o in objects:
	    for a in o.atomList():
		i = a.index
		j = j + 1
		index_dict[i] = j
		atom_types[self._atomType(o, a)] = 1
	atom_types = atom_types.keys()

	base = self.universe.baseVectors()
	config = open(conf_file, 'w')
	config.write(title+'\n')
	if base is None:
	    config.write(FortranLine([0,0],'2I10').text+'\n')
	else:
	    config.write(FortranLine([0,2],'2I10').text+'\n')
	    format = FortranFormat('3F20.10')
	    for vector in base:
		v = length_factor*vector
		config.write(str(FortranLine([v[0], v[1], v[2]], format))+'\n')
	format1 = FortranFormat('A8,I5')
	format2 = FortranFormat('3F20.10')
	for o in objects:
	    for a in o.atomList():
		i = a.index
		c = length_factor*self.configuration[i]
		config.write(FortranLine([self._atomType(o, a), index_dict[i]],
					 format1).text+'\n')
		config.write(FortranLine([c[0],c[1],c[2]], format2).text+'\n')
	config.close()

	molecules = []
	last_type = None
	n = 1
	sample = None
	for o in objects:
	    t = o.type
	    if t is None or t != last_type:
		if sample is not None:
		    molecules.append((sample, n))
		sample = o
		n = 1
	    else:
		n = n + 1
	    last_type = t
	if sample is not None:
	    molecules.append((sample, n))

	field = open(field_file, 'w')
	field.write(title+'\nUNITS ' + energy_unit_name + '\n\n')
	field.write('MOLECULES ' + `len(molecules)` + '\n')
	for m, n in molecules:
	    self._molecularField(m, n, field, index_dict)
	ntypes = len(atom_types)
	field.write('VDW ' + `(ntypes*(ntypes+1))/2` + '\n')
	format = FortranFormat('2A8,A4,2F12.5')
	for i1 in xrange(ntypes):
	    a1 = atom_types[i1]
	    eps1, sig1 = Amber94.ljParameters(a1)
	    for i2 in xrange(i1, ntypes):
		a2 = atom_types[i2]
		eps2, sig2 = Amber94.ljParameters(a2)
		field.write(FortranLine([a1, a2, 'lj',
					 energy_factor*sqrt(eps1*eps2),
					 0.5*length_factor*(sig1+sig2)],
					format).text+'\n')
	field.write('CLOSE\n')
	field.close()

    def _molecularField(self, o, n, field, index_dict):
	atoms = o.atomList()
	atom_indices = map(lambda a: a.index, atoms)
	field.write(o.name+'\n')
	field.write('NUMMOLS ' + `n` + '\n')
	field.write('ATOMS ' + `len(atoms)` + '\n')
	offset = index_dict[atoms[0].index]-1
	format = FortranFormat('A8,2F12.5,I5')
	for a in atoms:
	    field.write(FortranLine([self._atomType(o, a),
				     self._atomMass(o, a),
				     self._atomCharge(o, a),
				     1], format).text+'\n')
	bonds = self.bonded.bonds(self.global_data)
	bonds = filter(lambda b, ai=atom_indices: b[0] in ai, bonds)
	if bonds:
	    field.write('BONDS ' + `len(bonds)` + '\n')
	    format = FortranFormat('A4,2I5,2F12.5')
	    for b in bonds:
		field.write(FortranLine(['harm',
					 index_dict[b[0]]-offset,
					 index_dict[b[1]]-offset,
					 b[2]*length_factor,
					 2.*b[3]*energy_factor/length_factor**2],
					format).text+'\n')
	angles = self.bonded.angles(self.global_data)
	angles = filter(lambda a, ai=atom_indices: a[0] in ai, angles)
	excluded = ExclusionList()
	if angles:
	    field.write('ANGLES ' + `len(angles)` + '\n')
	    format = FortranFormat('A4,3I5,2F12.5')
	    for a in angles:
		field.write(FortranLine(['harm',
					 index_dict[a[0]]-offset,
					 index_dict[a[1]]-offset,
					 index_dict[a[2]]-offset,
					 a[3]*angle_factor,
					 2.*a[4]*energy_factor/angle_factor**2],
					format).text+'\n')
		excluded.add((a[0], a[2]))
	dihedrals = self.bonded.dihedrals(self.global_data)
	dihedrals = filter(lambda d, ai=atom_indices: d[0] in ai, dihedrals)
	if dihedrals:
	    combined = []
	    last_atoms = [None, None, None, None]
	    for d in dihedrals:
		atoms = d[:4]
		if atoms != last_atoms:
		    combined.append([d])
		else:
		    combined[-1].append(d)
		last_atoms = atoms
	    dihedrals = combined
	    field.write('DIHEDRALS ' + `len(dihedrals)` + '\n')
	    format1 = FortranFormat('A4,4I5,5F12.5')
	    format2 = FortranFormat('A4,4I5,11F12.5')
	    for d in dihedrals:
		one_four = (d[0][0], d[0][3])
		if excluded.checkPair(one_four):
		    el_1_4 = 0.
		    lj_1_4 = 0.
		else:
		    el_1_4 = self.electrostatic_1_4
		    lj_1_4 = self.lj_1_4
		    excluded.add(one_four)
		if len(d) == 1:
		    d = d[0]
		    field.write(FortranLine(['cos',
					     index_dict[d[0]]-offset,
					     index_dict[d[1]]-offset,
					     index_dict[d[2]]-offset,
					     index_dict[d[3]]-offset,
					     d[6], d[5]*angle_factor,
					     d[4]*energy_factor,
					     el_1_4, lj_1_4],
					    format1).text+'\n')
		elif len(d) < 4:
		    atoms = [index_dict[d[0][0]]-offset,
			     index_dict[d[0][1]]-offset,
			     index_dict[d[0][2]]-offset,
			     index_dict[d[0][3]]-offset]
		    parameters = map(lambda t: [t[6], t[5]*angle_factor,
						t[4]*energy_factor], d)
		    if len(d) == 2:
			parameters.append([1., 0., 0.])
		    parameters = reduce(operator.add, parameters)
		    field.write(FortranLine(['four'] + atoms + parameters[:3] +
					    [el_1_4, lj_1_4] + parameters[3:],
					    format2).text+'\n')
		else:
		    raise ValueError, "dihedral term not supported by DLPOLY"
	field.write('FINISH\n')

    def _atomType(self, o, a):
	return getattr(self.global_data, self.atom_type_property)[a]

    def _atomMass(self, o, a):
	return a.mass()

    def _atomCharge(self, o, a):
	return o.getAtomProperty(a, self.atom_charge_property)
