# This module implements MD integrators
#
# Written by Konrad Hinsen
# last revision: 1998-2-4
#

import Trajectory, Units
import _dynamics, _forcefield
import Numeric, operator

#
# Integrator base class
#
class Integrator(Trajectory.TrajectoryGenerator):

    def __init__(self, universe, options):
	Trajectory.TrajectoryGenerator.__init__(self, universe, options)

    default_options = {'trajectory': None, 'log': 10,
		       'steps': 100, 'delta_t': 1.*Units.fs,
		       'options': []}

    available_data = ['configuration', 'velocities', 'gradients',
		      'energy', 'temperature', 'time']

    def __call__(self, options):
	raise AttributeError

#
# Velocity-Verlet integrator
#
class VelocityVerletIntegrator(Integrator):

    def __init__(self, universe, **options):
	Integrator.__init__(self, universe, options)

    def __call__(self, **options):
	self.setCallOptions(options)
	configuration = self.universe.configuration()
	velocities = self.universe.velocities()
	if velocities is None:
	    raise ValueError, "no velocities"
	masses = self.universe.masses()
	fixed = self.universe.getAtomBooleanArray('fixed')
	f, p = self.universe.CcorrectionFunction()
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = _forcefield.TotalEvaluator(evaluator.allTerms())
	_dynamics.integrateVV(self.universe, f, p,
			      configuration.array, velocities.array,
			      masses.array, fixed.array, evaluator,
			      self.getOption('delta_t'),
			      self.getOption('steps'),
			      [self.getOption('trajectory'),
			       self.getOption('log')] +
			      self.getOption('options'),
			      'Dynamics trajectory with ' +
			      self.optionString(['delta_t', 'steps']))

#
# Velocity scaling, removal of global translation/rotation
#
class VelocityScaler(Trajectory.TrajectoryFunction):

    def __init__(self, temperature, window=0.):
	self.parameters = Numeric.array([temperature, window], Numeric.Float)
	self.Cfunction = _dynamics.scaleVelocities

class Heater(Trajectory.TrajectoryFunction):

    def __init__(self, temp1, temp2, gradient):
	self.parameters = Numeric.array([temp1, temp2, gradient], Numeric.Float)
	self.Cfunction = _dynamics.heat

class TranslationRemover(Trajectory.TrajectoryFunction):

    def __init__(self):
	self.parameters = None
	self.Cfunction = _dynamics.removeTranslation

class RotationRemover(Trajectory.TrajectoryFunction):

    def __init__(self):
	self.parameters = None
	self.Cfunction = _dynamics.removeRotation

