# This module defines tensor fields in molecular systems
#
# Written by Konrad Hinsen
# last revision: 1997-11-25
#

import Collection, Variables, Visualization
import Color, TensorAnalysis
from Vector import Vector
import Numeric

class AtomicField:

    def __init__(self, system, grid_size, values):
	self.system = system
	if type(grid_size) == type(()):
	    self.box, self.field, self.points, self.colors = grid_size
	else:
	    if values.data_rank != 1:
		raise TypeError, "data not a particle variable"
	    self.box = Collection.PartitionedAtomCollection(grid_size, system)
	    self._setField(values)

    def _setField(self, data):
	points = []
	values = []
	colors = []
	default_color = Color.ColorByName('black')
	for min, max, atoms in self.box.partitions():
	    center = 0.5*(min+max)
	    points.append(center.array)
	    v = data.zero()
	    total_weight = 0.
	    color = default_color
	    for a in atoms:
		d = a.position()-center
		weight = Numeric.exp(-d*d/self.box.partition_size**2)
		v = v + weight*data[a]
		total_weight = total_weight + weight
		c = default_color
		try: c = a.color
		except AttributeError: pass
		if type(c) == type(''):
		    c = Color.ColorByName(c)
		color = color + c
	    values.append(v/total_weight)
	    colors.append(color)
	min = Numeric.minimum.reduce(points)
	max = Numeric.maximum.reduce(points)
	axes = (Numeric.arange(min[0], max[0]+1, self.box.partition_size),
		Numeric.arange(min[1], max[1]+1, self.box.partition_size),
		Numeric.arange(min[2], max[2]+1, self.box.partition_size))
	array = Numeric.zeros(tuple(map(len, axes)) + data.value_rank*(3,),
			      Numeric.Float)
	inside = Numeric.zeros(tuple(map(len, axes)), Numeric.Float)
	for p, v in map(None, points, values):
	    indices = Numeric.floor((p-min)/self.box.partition_size+0.5)
	    indices = indices.astype(Numeric.Int)
	    array[indices] = v
	    inside[indices] = 1.
	self.field = self.field_class(axes, array, data.zero())
	inside = TensorAnalysis.ScalarField(axes, inside).gradient().length()
	self.points = []
	self.colors = []
	for i in range(len(points)):
	    p = points[i]
	    test = 0
	    try:
		test = apply(inside, tuple(p)) > 1.e-10
	    except ValueError: pass
	    if test:
		self.points.append(p)
		self.colors.append(colors[i])

    def __call__(self, point):
	return self.field(point)

    def particleVariable(self):
	universe = self.system.universe()
	rank = self.field.rank
	if rank == 0:
	    v = Variables.ParticleScalar(universe)
	elif rank == 1:
	    v = Variables.ParticleVector(universe)
	else:
	    raise ValueError, "no appropriate return type"
	for a in self.system.atomList():
	    v[a] = self.field(a.position())
	return v

    def writeToFile(self, filename, scale = 1., color = None):
	import VRML
	objects = self.graphicsObjects(scale=scale, color=color,
				       graphics_module=VRML)
	VRML.Scene(objects).writeToFile(filename)

    def view(self, scale = 1., color = None):
	import VRML
	objects = self.graphicsObjects(scale=scale, color=color,
				       graphics_module=VRML)
	VRML.Scene(objects).view()


class AtomicScalarField(AtomicField, Visualization.Viewable):

    def __init__(self, system, grid_size, values):
	if values is not None and values.value_rank != 0:
	    raise TypeError, "data not a vector field"
	self.field_class = TensorAnalysis.ScalarField
	AtomicField.__init__(self, system, grid_size, values)

    def gradient(self):
	field = self.field.gradient()
	return AtomicVectorField(self.system, (self.box, field,
					       self.points, self.colors), None)

    def laplacian(self):
	field = self.field.laplacian()
	return AtomicScalarrField(self.system, (self.box, field,
						self.points, self.colors), None)

    def _graphics(self, conf, distance_fn, model, module, options):
	scale = 1.
	try: scale = options['scale']
	except KeyError: pass
	range = (None, None)
	try: range = options['range']
	except KeyError: pass

	lower, upper = range
	size = self.box.partition_size/10.
	objects = []
	color_scale = module.SymmetricColorScale(1.)
	for p, c in map(None, self.points, self.colors):
	    p = tuple(p)
	    v = apply(self.field, p)
	    if (lower is None or v > lower) and (upper is None or v < upper):
		p = apply(Vector, p)
		v = scale*v
		if v < -1. or v > 1.:
		    m = module.DiffuseMaterial('black')
		else:
		    m = module.Material(diffuse_color = color_scale(scale*v))
		objects.append(module.Cube(p, size, material = m))
	return objects


class AtomicVectorField(AtomicField, Visualization.Viewable):

    def __init__(self, system, grid_size, values):
	if values is not None and values.value_rank != 1:
	    raise TypeError, "data not a vector field"
	self.field_class = TensorAnalysis.VectorField
	AtomicField.__init__(self, system, grid_size, values)

    def length(self):
	field = self.field.length()
	return AtomicScalarField(self.system, (self.box, field,
					       self.points, self.colors), None)

    def divergence(self):
	field = self.field.divergence()
	return AtomicScalarField(self.system, (self.box, field,
					       self.points, self.colors), None)

    def curl(self):
	field = self.field.curl()
	return AtomicVectorField(self.system, (self.box, field,
					       self.points, self.colors), None)

    def laplacian(self):
	field = self.field.curl()
	return AtomicVectorField(self.system, (self.box, field,
					       self.points, self.colors), None)

    def _graphics(self, conf, distance_fn, model, module, options):
	scale = 1.
	try: scale = options['scale']
	except KeyError: pass
	diameter = 1.
	try: diameter = options['diameter']
	except KeyError: pass
	color = None
	try: color = options['color']
	except KeyError: pass
	range = (None, None)
	try: range = options['range']
	except KeyError: pass

	lower, upper = range
	size = diameter*self.box.partition_size/50.
	if color is not None:
	    color = module.ColorByName(color)
	objects = []
	materials = {}
	for p, c in map(None, self.points, self.colors):
	    p = tuple(p)
	    v = apply(self.field, p)
	    lv = v.length()
	    if (lower is None or lv > lower) and (upper is None or lv < upper):
		p = apply(Vector, p)
		if color is not None:
		    c = color
		try: m = materials[c]
		except KeyError: m = module.Material(diffuse_color = c)
		materials[c] = m
		objects.append(module.Arrow(p, p+scale*v, size, material = m))
	return objects

