# This module implements classes that represent general force fields
# and force field evaluators.
#
# Written by Konrad Hinsen
# last revision: 1997-11-25
#

import Universe, Utility, Variables
import copy, Numeric, operator

# Class definitions

#
# The base class ForceField contains common operations for all force fields
#
class ForceField:

    def __init__(self, name):
	self.name = name
	self.type = None

    is_force_field = 1

    def __getstate__(self):
	return None

    def __setstate__(self, state):
	if state is not None:
	    raise ValueError, "Illegal state value"

    def __getinitargs__(self):
	return self.arguments

    def evaluator(self, universe, subset1, subset2, global_data):
	# must be defined by derived classes
	raise AttributeError

    def __add__(self, other):
	return CompoundForceField(self, other)

    def getOptions(self, object, options):
	attr = self.type + '_options'
	if hasattr(object, attr):
	    for key, value in getattr(object, attr).items():
		options[key] = value
	attr = self.name + '_options'
	if hasattr(object, attr):
	    for key, value in getattr(object, attr).items():
		options[key] = value

    def ready(self, global_data):
	return 1

    def bondedForceFields(self):
	return []

#
# The base class ForceFieldEvaluator contains common operations
# for all force field evaluators
#
class ForceFieldEvaluator:

    def CEvaluator(self):
	from _forcefield import PythonEvaluator
	return PythonEvaluator(self)

    def lastEnergyTerms(self):
	dict = {}
	dict[self.name] = self.last_energy
	return dict

#
# A CompoundForceField represents the sum of its component force fields
#
class CompoundForceField(ForceField):

    def __init__(self, *args):
	self.fflist = []
	for ff in args:
	    if isCompoundForceField(ff):
		self.fflist.append(ff.fflist)
	    else:
		self.fflist.append(ff)
	self.name = '*'
	self.type = 'compound'

    is_compound_force_field = 1

    def evaluator(self, system, subset1, subset2, global_data):
	eval_objects = []
	remaining = copy.copy(self.fflist)
	while remaining:
	    done = []
	    for ff in remaining:
		if ff.ready(global_data):
		    eval_objects.append(ff.evaluator(system, subset1, subset2,
						     global_data))
		    done.append(ff)
	    if not done:
		raise TypeError, "Cyclic force field dependence"
	    for ff in done:
		remaining.remove(ff)
	return CompoundEvaluator(eval_objects)

    def bondedForceFields(self):
	return reduce(operator.add,
		      map(lambda f: f.bondedForceFields(), self.fflist))

class CompoundEvaluator(ForceFieldEvaluator):

    def __init__(self, eval_objects):
	self.eval_objects = eval_objects

    def __call__(self, coordinates, gradients, force_constants, small_change=0):
	energy = 0.
	for eo in self.eval_objects:
	    energy_term = eo(coordinates, gradients, force_constants, small_change)
	    energy = energy + energy_term
	return energy

    def CEvaluator(self):
	from _forcefield import CombinedEvaluator
	return CombinedEvaluator(Numeric.array(map(lambda e: e.CEvaluator(),
						   self.eval_objects)))

    def lastEnergyTerms(self):
	dict = {}
	for eo in self.eval_objects:
	    d = eo.lastEnergyTerms()
	    for name, value in d.items():
		dict[name] = value
	return dict

#
# This class serves to define data containers used in force field initialization
#
class ForceFieldData:

    def __init__(self):
	self.dict = {}

    def add(self, tag, value):
	try:
	    self.dict[tag].append(value)
	except KeyError:
	    self.dict[tag] = [value]

    def get(self, tag):
	try:
	    return self.dict[tag]
	except KeyError:
	    return []

    def set(self, tag, value):
	self.dict[tag] = value


# Type check functions

def isForceField(x):
    return hasattr(x, 'is_force_field')

def isCompoundForceField(x):
    return hasattr(x, 'is_compound_force_field')

#
# Force field evaluator support functions.
# These functions are meant to be used with force field evaluators
# implemented in Python using automatic derivatives. They put the
# first and second derivative information into the right places.
#
def addToGradients(coordinates, indices, vectors):
    for index, vector in map(None, indices, vectors):
	coordinates[index] = coordinates[index] + vector.array

def addToForceConstants(total_fc, indices, small_fc):
    indices = map(None, indices, range(len(indices)))
    for i1, i2 in map(lambda i: (i,i), indices) + Utility.pairs(indices):
	ii1, jj1 = i1
	ii2, jj2 = i2
	if ii1 > ii2:
	    ii1, ii2 = ii2, ii1
	    jj1, jj2 = jj2, jj1
	jj1 = 3*jj1
	jj2 = 3*jj2
	total_fc[ii1,:,ii2,:] = total_fc[ii1,:,ii2,:] + \
				small_fc[jj1:jj1+3, jj2:jj2+3]

#
# High-level energy evaluator (i.e. the Python interface)
#
class EnergyEvaluator:

    def __init__(self, universe, force_field, subset1=None, subset2=None):
	if not Universe.isUniverse(universe):
	    raise TypeError, "energy evaluator defined only for universes"
	self.universe = universe
	self.universe_version = self.universe._version
	self.ff = force_field
	self.configuration = self.universe.configuration()
	self.global_data = ForceFieldData()
	if subset1 is not None and subset2 is None:
	    subset2 = subset1
	self.evaluator = self.ff.evaluator(self.universe,
					   subset1, subset2,
					   self.global_data)
	    
    def checkUniverseVersion(self):
	if self.universe_version != self.universe._version:
	    raise ValueError, 'the universe has been modified'

    def __call__(self, gradients = None, force_constants = None,
		 small_change=0):
	self.checkUniverseVersion()
	args = [self.configuration.array]

	if Variables.isUniverseVariable(gradients):
	    args.append(gradients.array)
	elif type(gradients) == Numeric.arraytype:
	    gradients = Variables.ParticleVector(self.universe, gradients)
	    args.append(gradients.array)
	elif gradients:
	    gradients = Variables.ParticleVector(self.universe, None)
	    args.append(gradients.array)
	else:
	    args.append(None)

	if Variables.isUniverseVariable(force_constants):
	    args.append(force_constants.array)
	elif type(force_constants) == Numeric.arraytype:
	    force_constants = Variables.SymmetricPairTensor(self.universe,
							    force_constants)
	    args.append(force_constants.array)
	else:
	    sparse_type = None
	    try:
		from _forcefield import SparseForceConstants
		sparse_type = type(SparseForceConstants(2, 2))
	    except ImportError: pass
	    if type(force_constants) == sparse_type:
		args.append(force_constants)
	    elif force_constants:
		force_constants = Variables.SymmetricPairTensor(self.universe,
								None)
		args.append(force_constants.array)
	    else:
		args.append(None)

	args.append(small_change)
	try:
	    from _forcefield import TotalEvaluator
	    evaluator = self.evaluator.CEvaluator()
	    evaluator = TotalEvaluator(Numeric.array([evaluator]))
	except ImportError:
	    evaluator = self.evaluator
	print 'evaluator',evaluator
	energy = apply(evaluator, tuple(args))
	if force_constants:
	    return energy, gradients, force_constants
	elif gradients:
	    return energy, gradients
	else:
	    return energy

    def lastEnergyTerms(self):
	return self.evaluator.lastEnergyTerms()
