# This module implements energy minimizers.
#
# Written by Konrad Hinsen
# last revision: 1998-2-4
#

import Trajectory, Units
from _minimization import conjugateGradient, steepestDescent
from _forcefield import TotalEvaluator
import Numeric, operator

#
# Minimizer base class
#
class Minimizer(Trajectory.TrajectoryGenerator):

    def __init__(self, universe, options):
	Trajectory.TrajectoryGenerator.__init__(self, universe, options)

    default_options = {'trajectory': None, 'log': None,
		       'steps': 100, 'step_size': 0.02*Units.Ang,
		       'convergence': 0.01*Units.kJ/(Units.mol*Units.nm)}

    available_data = ['energy', 'configuration', 'gradients']

    def __call__(self, options):
	raise AttributeError

#
# Steepest descent minimizer
#
class SteepestDescentMinimizer(Minimizer):

    def __init__(self, universe, **options):
	Minimizer.__init__(self, universe, options)

    def __call__(self, **options):
	self.setCallOptions(options)
	configuration = self.universe.configuration()
	fixed = self.universe.getAtomBooleanArray('fixed')
	f, p = self.universe.CcorrectionFunction()
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = TotalEvaluator(Numeric.array([evaluator]))
	steepestDescent(self.universe, f, p,
			configuration.array, fixed.array, evaluator,
			self.getOption('steps'), self.getOption('step_size'),
			self.getOption('convergence'),
			[self.getOption('trajectory'), self.getOption('log')],
			'Steepest descent minimization with ' +
			self.optionString(['convergence', 'step_size',
					   'steps']))


#
# Conjugate gradient minimizer
#
class ConjugateGradientMinimizer(Minimizer):

    def __init__(self, universe, **options):
	Minimizer.__init__(self, universe, options)

    def __call__(self, **options):
	self.setCallOptions(options)
	configuration = self.universe.configuration()
	fixed = self.universe.getAtomBooleanArray('fixed')
	f, p = self.universe.CcorrectionFunction()
	evaluator = self.universe.energyEvaluator().evaluator.CEvaluator()
	evaluator = TotalEvaluator(Numeric.array([evaluator]))
	conjugateGradient(self.universe, f, p,
			  configuration.array, fixed.array, evaluator,
			  self.getOption('steps'), self.getOption('step_size'),
			  self.getOption('convergence'),
			  [self.getOption('trajectory'), self.getOption('log')],
			  'Conjugate gradient minimization with ' +
			  self.optionString(['convergence', 'step_size',
					     'steps']))
