# This module implements normal mode calculations
#
# Written by Konrad Hinsen
# last revision: 1998-4-23
#

import Units, Variables, Visualization
import Numeric, copy

#
# Class for a single mode
#
class Mode(Variables.ParticleVector):

    def __init__(self, universe, n, frequency, mode):
	self.number = n
	self.frequency = frequency
	Variables.ParticleVector.__init__(self, universe, mode)

    def __str__(self):
	return 'Mode ' + `self.number` + ' with frequency ' + `self.frequency`

    __repr__ = __str__

    return_class = Variables.ParticleVector

    def effectiveMassAndForceConstant(self):
	m = self.massWeightedProjection(self)/self.projection(self)
	k = m*(2.*Units.pi*self.frequency)**2
	return m, k

    def view(self, factor=1., subset=None):
	Visualization.viewMode(self, factor, subset)

#
# Class for a full set of normal modes
#
class NormalModes:

    def __init__(self, universe=None, temperature = 300):
	if universe == None:
	    return
	self.universe = universe
	self.sqrt_mass = Numeric.sqrt(self.universe.masses().array)
	self.sqrt_mass = self.sqrt_mass[:, Numeric.NewAxis]

	self._forceConstantMatrix()
	ev = self._diagonalize()

	self.imaginary = Numeric.less(ev, 0.)
	self.frequencies = Numeric.sqrt(Numeric.fabs(ev)) / (2.*Units.pi)
	self.sort_index = Numeric.argsort(self.frequencies)

	self._scale(temperature)
	del self.sqrt_mass

    def __len__(self):
	return self.nmodes

    def __getitem__(self, item):
	index = self.sort_index[item]
	f = self.frequencies[index]
	if self.imaginary[index]:
	    f = f*1.j
	return Mode(self.universe, item, f, self.array[index])

    def __getslice__(self, first, last):
	return map(lambda i, s=self: s[i], range(first, last))

    def reduceToRange(self, first, last):
	junk1 = list(self.sort_index[:first])
	junk2 = list(self.sort_index[last:])
	junk1.sort()
	junk2.sort()
	if junk1 == range(0, first) and \
	   junk2 == range(last, len(self.sort_index)):
	    self.array = self.array[first:last]
	    self.frequencies = self.frequencies[first:last]
	    self.imaginary = self.imaginary[first:last]
	    self.sort_index = self.sort_index[first:last]-first
	else:
	    keep = self.sort_index[first:last]
	    self.array = Numeric.take(self.array, keep)
	    self.frequencies = Numeric.take(self.frequencies, keep)
	    self.imaginary = Numeric.take(self.imaginary, keep)
	    self.sort_index = Numeric.arange(0, last-first)
	self.nmodes = last-first

    def _forceConstantMatrix(self):
	energy, force_constants = self.universe.energyAndForceConstants()
	self.array = force_constants.array
	self.natoms = self.array.shape[0]
	self.nmodes = 3*self.natoms
	Numeric.divide(self.array,
		       self.sqrt_mass[Numeric.NewAxis, Numeric.NewAxis, :, :],
		       self.array)
	Numeric.divide(self.array,
		       self.sqrt_mass[:, :, Numeric.NewAxis, Numeric.NewAxis],
		       self.array)
	self.array.shape = 2*(self.nmodes,)

    def _diagonalize(self):
	dsyev = None
	try:
	    from lapack_dsy import dsyev
	except ImportError: pass
	if dsyev is None:
	    try:
		from lapack_dsyev import dsyev
	    except ImportError: pass
	if dsyev is None:
	    from LinearAlgebra import eigenvectors
	    _symmetrize(self.array)
	    ev, modes = eigenvectors(self.array)
	    self.array = modes
	    if ev.typecode() == Numeric.Complex:
		ev = ev.real
	else:
	    ev = Numeric.zeros((self.nmodes,), Numeric.Float)
	    lwork = 3*self.nmodes
	    work = Numeric.zeros((lwork,), Numeric.Float)
	    results = dsyev('V', 'L', self.nmodes, self.array, self.nmodes, ev,
			    work, lwork, 0)
	    if results['info'] > 0:
		raise ValueError, 'Eigenvalue calculation did not converge'
	return ev

    def _scale(self, temperature):
	if temperature is not None:
	    factor = Numeric.sqrt(2.*temperature*Units.k_B/Units.amu) / \
		     (2.*Units.pi)
	self.array.shape = (self.nmodes, self.natoms, 3)
	for i in range(self.nmodes):
	    index = self.sort_index[i]
	    if index >= 6 and temperature is not None:
		amplitude = factor/self.frequencies[index]
	    else:
		amplitude = 1.
	    self.array[index] = amplitude*self.array[index] / self.sqrt_mass

#
# Classes for normal modes with respect to a given basis
#
class SubspaceNormalModes(NormalModes):

    def __init__(self, universe = None, basis = None, temperature = 300):
	if universe is None:
	    return
	self.basis = basis
	NormalModes.__init__(self, universe, temperature)
	del self.basis

    def _setupBasis(self):
	try:
	    from lapack_dge import dgesvd
	except ImportError:
	    from lapack_lite import dgesvd
	if type(self.basis) is type(()):
	    excluded, basis = self.basis
	else:
	    excluded = []
	    basis = self.basis
	nexcluded = len(excluded)
	nmodes = len(basis)
	ntotal = nexcluded + nmodes
	natoms = len(basis[0])
	sv = Numeric.zeros((min(ntotal, 3*natoms),), Numeric.Float)
	work = Numeric.zeros((max(3*min(3*natoms,ntotal)+max(3*natoms,ntotal),
				  5*min(3*natoms,ntotal)-4),), Numeric.Float)
	dummy = Numeric.zeros((1,), Numeric.Float)
	if nexcluded > 0:
	    self.basis = Numeric.zeros((ntotal, 3*natoms), Numeric.Float)
	    for i in range(nexcluded):
		self.basis[i] = Numeric.ravel(excluded[i].array*self.sqrt_mass)
	    result = dgesvd('O', 'N', 3*natoms, nexcluded, self.basis,
			    3*natoms, sv, dummy, 1, dummy, 1,
			    work, work.shape[0], 0)
	    if result['info'] != 0:
		raise ValueError, 'Lapack SVD: ' + `result['info']`
	    svmax = Numeric.maximum.reduce(sv)
	    nexcluded = Numeric.add.reduce(Numeric.greater(sv, 1.e-10*svmax))
	    ntotal = nexcluded + nmodes
	    for i in range(nmodes):
		self.basis[i+nexcluded] = \
			Numeric.ravel(basis[i].array*self.sqrt_mass)
	    result = dgesvd('O', 'N', 3*natoms, ntotal, self.basis,
			    3*natoms, sv, dummy, 1, dummy, 1,
			    work, work.shape[0], 0)
	    if result['info'] != 0:
		raise ValueError, 'Lapack SVD: ' + `result['info']`
	    svmax = Numeric.maximum.reduce(sv)
	    ntotal = Numeric.add.reduce(Numeric.greater(sv, 1.e-10*svmax))
	    nmodes = ntotal - nexcluded
	else:
	    if hasattr(self.basis, 'may_modify') and \
	       hasattr(self.basis, 'array'):
		self.basis = self.basis.array
	    else:
		self.basis = Numeric.array(map(lambda v: v.array, basis))
	    Numeric.multiply(self.basis, self.sqrt_mass, self.basis)
	    self.basis.shape = (nmodes, 3*natoms)
	    result = dgesvd('O', 'N', 3*natoms, nmodes, self.basis, 3*natoms,
			    sv, dummy, 1, dummy, 1, work, work.shape[0], 0)
	    if result['info'] != 0:
		raise ValueError, 'Lapack SVD: ' + `result['info']`
	    svmax = Numeric.maximum.reduce(sv)
	    nmodes = Numeric.add.reduce(Numeric.greater(sv, 1.e-10*svmax))
	    ntotal = nmodes
	self.basis = self.basis[nexcluded:ntotal, :]

    def _forceConstantMatrix(self):
	self._setupBasis()
	NormalModes._forceConstantMatrix(self)
	_symmetrize(self.array)
	self.array = Numeric.dot(Numeric.dot(self.basis, self.array),
				 Numeric.transpose(self.basis))
	self.nmodes = self.array.shape[0]

    def _diagonalize(self):
	ev = NormalModes._diagonalize(self)
	self.array = Numeric.dot(self.array, self.basis)
	return ev


class FiniteDifferenceSubspaceNormalModes(SubspaceNormalModes):

    def __init__(self, universe = None, basis = None,
		 delta = 0.0001*Units.nm*Numeric.sqrt(Units.amu),
		 temperature = 300.):
	self.delta = delta
	SubspaceNormalModes.__init__(self, universe, basis, temperature)

    def _forceConstantMatrix(self):
	self._setupBasis()
	self.nmodes = len(self.basis)
	self.array = Numeric.zeros((self.nmodes, self.nmodes), Numeric.Float)
	conf = copy.copy(self.universe.configuration())
	conf_array = conf.array
	self.natoms = len(conf_array)
	small_change = 0
	for i in range(self.nmodes):
	    d = self.delta*Numeric.reshape(self.basis[i], (self.natoms, 3)) / \
		self.sqrt_mass
	    conf_array[:] = conf_array+d
	    self.universe.setConfiguration(conf)
	    energy, gradients1 = self.universe.energyAndGradients(None, None,
								  small_change)
	    small_change = 1
	    conf_array[:] = conf_array-2.*d
	    self.universe.setConfiguration(conf)
	    energy, gradients2 = self.universe.energyAndGradients(None, None,
								  small_change)
	    conf_array[:] = conf_array+d
	    v = (gradients1.array-gradients2.array) / \
		(2.*self.delta*self.sqrt_mass)
	    self.array[i] = Numeric.dot(self.basis, Numeric.ravel(v))
	self.universe.setConfiguration(conf)
	self.array = 0.5*(self.array+Numeric.transpose(self.array))


class SparseMatrixSubspaceNormalModes(SubspaceNormalModes):

    def _forceConstantMatrix(self):
	from _forcefield import SparseForceConstants
	self._setupBasis()
	nmodes, natoms = self.basis.shape
	natoms = natoms/3
	self.nmodes = nmodes
	self.natoms = natoms
	fc = SparseForceConstants(natoms, 5*natoms)
	eval = self.universe.energyEvaluator()
	energy, g, fc = eval(0, fc, 0)
	self.array = Numeric.zeros((nmodes, nmodes), Numeric.Float)
	for i in range(nmodes):
	    v = Numeric.reshape(self.basis[i], (natoms, 3))/self.sqrt_mass
	    v = fc.multiplyVector(v)
	    v = v/self.sqrt_mass
	    v.shape = (3*natoms,)
	    self.array[i, :] = Numeric.dot(self.basis, v)

#
# Helper functions
#

# Fill in the lower triangle of an upper-triangle symmetric matrix

def _symmetrize(a):
    n = a.shape[0]
    for i in range(n):
	for j in range(i+1, n):
	    a[j,i] = a[i,j]
