# ReadDCD: Transform DCD trajectories into Trajectory objects
# 
# $Id: ReadDCD.py,v 1.1.1.1 1998/03/14 05:14:45 pcm Exp $
#
"""
DCDReader: This class serves to read in trajectories in the DCD 
format into trajectory objects. The aim is to use MMTK's analysis and navigation
facilities for trajectories generated by other MD programs (XPLOR in my case).

Remarks
------------------------------------------------------------
Various assumptions underly this module:

o The DCD trajectory contains only 'free' atoms. That is, every single atom's 
  coordinates are written out *every*single*timestep*. No constrained atoms
  please, as the the module right now can't cope with the 'freeIndex' feature
  of the DCD format

o The order of atom coordinates in the DCD file is the same as in the PDB
  structure file that was used to generate world.protein. Otherwise, you'll
   have the coordinates of your atoms all mixed up.


Usage
------------------------------------------------------------
from mmtk import *

# Construct system
world = InfiniteUniverse(AmberForceField())
world.protein = Protein('barnase_barstar')

# DCD
import ReadDCD

# generate the trajectory object to hold the DCD trajectory
traj = Trajectory(world, "testDCD.nc", "w","DCD trajectory")

# instantiate a DCD reader for a given world
dcdReader = ReadDCD.ReadDCD(world)

# The arguments below are vital and MUST BE GIVEN WITH THEIR VARIABLE NAMES.
# Arg1) 'trajectory' arg is a tuple with the following structure:
#   (starting step of DCD traj, last step of DCD traj, stepsize)
# Arg2) 'dcdFileName' gives the file name of the DCD trajectory file 
#       (difficult, is it?) ;-)
dcdReader(trajectory = (0, None, 1, traj),
          dcdFileName = 'DCD/test.dcd')
traj.close()

"""


import Trajectory, Units
from _readDCD import readDCD
from _forcefield import TotalEvaluator
import Numeric, operator


class ReadDCD(Trajectory.TrajectoryGenerator):
    default_options = {'trajectory': None, 'log': None}
    available_data = ['configuration', 'time']

    def __init__(self, universe, **options):
	Trajectory.TrajectoryGenerator.__init__(self, universe,
						options, self.default_options)

    def __call__(self, **options):
	self.setCallOptions(options)
	configuration = self.universe.configuration()
	readDCD(self.universe, configuration.array,
		[self.getOption('trajectory'), self.getOption('log')],
		self.getOption('dcdFileName'))
