# This module reads force field data from a file of the format found
# in Tinker's params directory
#
# Written by Peter McCluskey
#

#
# Not supported:
# ECHO
# MULTIPOLE
#

import Units
import string
from string import atoi, atof
from Bonds import lookUpBond
from Database import ParameterPath
from TextFile import TextFile
from operator import add
from DictWithDefault import DictWithDefault

tinker_energy_unit = Units.kcal/Units.mol
tinker_length_unit = Units.Ang
tinker_angle_unit = Units.deg
tinker_fc_angle_unit = Units.rad

class TinkerParamData:
    def __init__(self, short_filename):
	self.short_filename = short_filename	# e.g "mm3"
	filename = ParameterPath(short_filename + ".prm")
	file = TextFile(filename)
	for name in ('atom_class', 'symbol', 'description', 'atomic_number',\
		     'atomic_weight', 'ligands', 'radius', 'eps', 'reduct',\
		     'bonds', 'bonds5', 'bonds4', 'bonds3', 'bond_angles_average',\
		     'bond_angle_const', 'bond_angle_energy',\
		     'bond_angle_const5', 'bond_angle_energy5',\
		     'bond_angle_const4', 'bond_angle_energy4',\
		     'bond_angle_const3', 'bond_angle_energy3', 'stdn',\
		     'outOfPlaneBend', 'urey_const', 'dst', 'angleAngle',\
		     'dihedral_const', 'tdi', 'improperTorsion',\
		     'dihedrals', 'dihedrals5', 'dihedrals4', 'stretchTorsion',\
		     'charge', 'dipole', 'pos',  'dipole5', 'pos5', \
		     'dipole4', 'pos4',  'dipole3', 'pos3', 'polarization',\
		     'electron', 'ionize', 'repulse', 'sslope', 'tslope',\
		     'sslope5', 'tslope5', 'sslope4', 'tslope4'):
	    setattr(self, name, {})
	line_no = 0
	while 1:
	    line_no = line_no + 1
	    raw_line = file.readline()
	    if len(raw_line) == 0:
		break
	    line = string.upper(raw_line[:-1])
	    self.parseLine(line, line_no, filename)

	self.atom_type_map = {}
	for key, sym in self.symbol.items():
	    key2 = (sym, self.ligands[key])
	    if not self.atom_type_map.has_key(key2):
		self.atom_type_map[key2] = [self.atom_class[key]]
	    else:
		self.atom_type_map[key2].append(self.atom_class[key])

	self.setDeducer()
	self.setAvgAngleParameters()

    def setDeducer(self):
	self._atomTypeDeducer = {}
	if self.short_filename == 'mm3':
	    self._atomTypeDeducer[('C',4)] = mm3deduceC4Type
	    self._atomTypeDeducer[('C',3)] = mm3deduceC3Type
	    self._atomTypeDeducer[('C',2)] = mm3deduceC2Type
	    self._atomTypeDeducer[('N',3)] = mm3deduceN3Type
	    self._atomTypeDeducer[('N',2)] = mm3deduceN2Type
	    self._atomTypeDeducer[('N',1)] = mm3deduceN1Type
	    self._atomTypeDeducer[('O',2)] = mm3deduceO2Type
	    self._atomTypeDeducer[('O',1)] = mm3deduceO1Type
	    self._atomTypeDeducer[('S',4)] = mm3deduceS4Type
	    self._atomTypeDeducer[('S',2)] = mm3deduceS2Type
	    self._atomTypeDeducer[('P',4)] = mm3deduceP4Type
	    self._atomTypeDeducer[('H',1)] = mm3deduceH1Type

    def deduceAtomType(self, a, blist, num_dangling):
	#print 'deduceAtomType',a.name,len(blist),num_dangling,'dangle'
	n = len(blist) + num_dangling
	try:
	    func = self._atomTypeDeducer[(a.symbol, n)]
	except KeyError:
	    msg = 'No method found to deduce ' + self.short_filename + ' type for atom ' + a.name
	    print msg
	    raise KeyError, msg
	t = func(a, blist, self)
	#print 'deduced type',t, a.name, a.symbol, n, hex(id(a))
	return t

    def parseLine(self, line, line_no, filename):
	tokens = string.split(line)
	try:
	    keyword = tokens[0]
	except IndexError:
	    return
	if keyword[0] == '#':
	    return
	keyword_list = {\
	    'ANGANGTERM' : 1,'ANGLETERM' : 1,'BONDTERM' : 1,\
	    'CHGDPLTERM' : 1,'CHARGETERM' : 1,'DIPOLETERM' : 1,\
	    'EXTRATERM' : 1,'IMPROPERTERM' : 1,'IMPTORSTERM' : 1,\
	    'MPOLETERM' : 1,'OPBENDTERM' : 1,'POLARIZETERM' : 1,\
	    'RESTRAINTERM' : 1,'RXNFIELDTERM' : 1,'SOLVATETERM' : 1,\
	    'STRBNDTERM' : 1,'STRTORTERM' : 1,'TORSIONTERM' : 1,\
	    'TORTORTERM' : 1,'UREYTERM' : 1,'VDWTERM' : 1,\
	    'FORCEFIELD' : 1,'PARAMETERS' : 1,'BOND-CUBIC' : 1,\
	    'BOND-QUARTIC' : 1,'ANGLE-CUBIC' : 1,'ANGLE-QUARTIC' : 1,\
	    'ANGLE-PENTIC' : 1,'ANGLE-SEXTIC' : 1,'A-EXPTERM' : 1,\
	    'B-EXPTERM' : 1,'C-EXPTERM' : 1,'DIELECTRIC ' : 1,\
	    'ANGANGUNIT' : 1,'ANGLETYPE' : 1,'ANGLEUNIT' : 1,\
	    'BONDTYPE' : 1,'BONDUNIT' : 1,'CHG-12-USE' : 1,\
	    'CHG-13-USE' : 1,'CHG-14-USE' : 1,'CHG-14-SCALE' : 1,\
	    'VDW-12-USE' : 1,'VDW-13-USE' : 1,'VDW-14-USE' : 1,\
	    'VDW-14-SCALE' : 1,'VDW-GAUSSIANS' : 1,'PISYSTEM' : 1,\
	    'STRBNDUNIT' : 1,'UREYUNIT' : 1,'STRTORUNIT' : 1,\
	    'TORSIONUNIT' : 1,'VDWTYPE' : 1,'EPSILONRULE' : 1,\
	    'RADIUSRULE' : 1,'RADIUSTYPE' : 1,'RADIUSSIZE' : 1,\
	    'POLARIZATION' : 1,'POLAR-DAMP' : 1,'POLAR-EPS' : 1,\
	    'REACTIONFIELD' : 1}
	keyword_defaults = {'FORCEFIELD' : self.short_filename}
	if keyword_list.has_key(keyword):
	    try:
		setattr(self, keyword, tokens[1])
	    except IndexError:
		if keyword_defaults.has_key(keyword):
		    forcefield_name = keyword_defaults[keyword]
		else:
		    msg = 'Error in forcefield file ' + filename + ': ' + keyword + ' command has no argument' 
		    raise RuntimeError, msg
	elif keyword == 'ATOM':
	    ia = atoi(tokens[1])
	    try:
		atom_class = atoi(tokens[2])
	    except ValueError:
		atom_class = 0
	    if atom_class == 0:
		atom_class = ia
		offset = 0
	    else:
		offset = 1
	    self.atom_class[ia] = atom_class
	    self.symbol[ia] = tokens[2+offset]
	    describe = tokens[3+offset]
	    if tokens[3+offset][0] == '"':
		while tokens[3+offset][-1] != '"':
		    describe = describe + ' ' + tokens[3+offset]
		    offset = offset + 1
		describe = describe[1:]
	    self.description[ia] = describe
	    self.atomic_number[ia] = atoi(tokens[4+offset])
	    self.atomic_number[self.symbol[ia]] = atoi(tokens[4+offset])
	    self.atomic_weight[ia] = atof(tokens[5+offset])
	    self.ligands[ia] = atoi(tokens[6+offset])
	elif keyword == 'VDW':
	    ia = atoi(tokens[1])
	    self.radius[ia] = atof(tokens[2])
	    self.eps[ia] = atof(tokens[3])
	    try:
		self.reduct[ia] = atof(tokens[4])
	    except IndexError:
		self.reduct[ia] = 0.0
	elif keyword == 'VDWPR':
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.radius[(ia,ib)] = atof(tokens[3])
	    self.eps[(ia,ib)] = atof(tokens[4])
	elif keyword == 'BOND':
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    if ia > ia:
		msg = 'file %s line %d indices out of order %d %d'\
		      % (filename, line_no, ia, ib)
		raise RuntimeError, msg
	    self.bonds[(ia,ib)] = (atof(tokens[4])*tinker_length_unit,\
				   atof(tokens[3])*tinker_energy_unit/tinker_length_unit**2)
	elif keyword == 'BOND5':
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.bonds5[(ia,ib)] = (atof(tokens[4])*tinker_length_unit,\
				    atof(tokens[3])*tinker_energy_unit/tinker_length_unit**2)
	elif keyword == 'BOND4':
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.bonds4[(ia,ib)] = (atof(tokens[4])*tinker_length_unit,\
				    atof(tokens[3])*tinker_energy_unit/tinker_length_unit**2)
	elif keyword == 'BOND3':
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.bonds3[(ia,ib)] = (atof(tokens[4])*tinker_length_unit,\
				    atof(tokens[3])*tinker_energy_unit/tinker_length_unit**2)
	elif keyword == 'ANGLE':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = (atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]))
	    self.bond_angle_energy[key] = \
			atof(tokens[4])*tinker_energy_unit/tinker_fc_angle_unit
	    t = [atof(tokens[5])*tinker_angle_unit]
	    for i in (6,7):
		if len(tokens) > i and atof(tokens[i]) != 0:
		    t.append(atof(tokens[i])*tinker_angle_unit)
	    self.bond_angle_const[key] = t
	elif keyword == 'ANGLE5':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = (atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]))
	    self.bond_angle_energy5[key] = \
			atof(tokens[4])*tinker_energy_unit/tinker_fc_angle_unit
	    t = [atof(tokens[5])*tinker_angle_unit]
	    for i in (6,7):
		if len(tokens) > i and atof(tokens[i]) != 0:
		    t.append(atof(tokens[i])*tinker_angle_unit)
	    self.bond_angle_const5[key] = t
	elif keyword == 'ANGLE4':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = (atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]))
	    self.bond_angle_energy4[key] = \
			atof(tokens[4])*tinker_energy_unit/tinker_fc_angle_unit
	    t = [atof(tokens[5])*tinker_angle_unit]
	    for i in (6,7):
		if len(tokens) > i and atof(tokens[i]) != 0:
		    t.append(atof(tokens[i])*tinker_angle_unit)
	    self.bond_angle_const4[key] = t
	elif keyword == 'ANGLE3':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = (atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]))
	    self.bond_angle_energy3[key] = \
			atof(tokens[4])*tinker_energy_unit/tinker_fc_angle_unit
	    t = [atof(tokens[5])*tinker_angle_unit]
	    for i in (6,7):
		if len(tokens) > i and atof(tokens[i]) != 0:
		    t.append(atof(tokens[i])*tinker_angle_unit)
	    self.bond_angle_const3[key] = t
	elif keyword == 'STRBND':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    self.stdn[ia] = [atof(tokens[2]), atof(tokens[3]), atof(tokens[4])]
	elif keyword == 'OPBEND':
	    if len(tokens) < 4:
		msg = 'file %s line %d expected 4 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.outOfPlaneBend[(ia, ib)] = atof(tokens[3])
	elif keyword == 'UREYBRAD':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = (atoi(tokens[1]), atoi(tokens[2]), atoi(tokens[3]))
	    self.urey_const[key] = atof(tokens[4])
	    self.dst[key] = atof(tokens[5])
	elif keyword == 'ANGANG':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    self.angleAngle[ia] = [atof(tokens[2]), atof(tokens[3]), atof(tokens[4])]
	elif keyword == 'IMPROPER':
	    if len(tokens) < 7:
		msg = 'file %s line %d expected 7 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = key4(tokens[1:5])
	    self.dihedral_const[key] = atof(tokens[5])
	    self.tdi[key] = atof(tokens[6])
	elif keyword == 'IMPTORS':
	    if len(tokens) < 7:
		msg = 'file %s line %d expected 7 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = key4(tokens[1:5])
	    self.improperTorsion[key] = self.torsionPhase(tokens[5:])
	elif keyword == 'TORSION':
	    if len(tokens) < 7:
		msg = 'file %s line %d expected 7 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = key4(tokens[1:5])
	    self.dihedrals[key] = self.torsionPhase(tokens[5:])
	elif keyword == 'TORSION5':
	    if len(tokens) < 7:
		msg = 'file %s line %d expected 7 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = key4(tokens[1:5])
	    self.dihedrals5[key] = self.torsionPhase(tokens[5:])
	elif keyword == 'TORSION4':
	    if len(tokens) < 7:
		msg = 'file %s line %d expected 7 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    key = key4(tokens[1:5])
	    self.dihedrals4[key] = self.torsionPhase(tokens[5:])
	elif keyword == 'STRTORS':
	    if len(tokens) < 6:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.stretchTorsion[(ia,ib)] = (atof(tokens[3]), atof(tokens[4]), atof(tokens[5]))
	elif keyword == 'CHARGE':
	    if len(tokens) < 3:
		msg = 'file %s line %d expected 6 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    self.charge[ia] = atof(tokens[2])
	elif keyword == 'DIPOLE':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.dipole[(ia,ib)] = atof(tokens[3])
	    self.pos[(ia,ib)] = atof(tokens[4])
	elif keyword == 'DIPOLE5':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.dipole5[(ia,ib)] = atof(tokens[3])
	    self.pos5[(ia,ib)] = atof(tokens[4])
	elif keyword == 'DIPOLE4':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.dipole4[(ia,ib)] = atof(tokens[3])
	    self.pos4[(ia,ib)] = atof(tokens[4])
	elif keyword == 'DIPOLE3':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.dipole3[(ia,ib)] = atof(tokens[3])
	    self.pos3[(ia,ib)] = atof(tokens[4])
	elif keyword == 'POLARIZE':
	    if len(tokens) < 3:
		msg = 'file %s line %d expected 3 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    self.polarization[ia] = atof(tokens[2])
	elif keyword == 'PIATOM':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    self.electron[ia] = atof(tokens[2])
	    self.ionize[ia] = atof(tokens[3])
	    self.repulse[ia] = atof(tokens[4])
	elif keyword == 'PIBOND':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.sslope[(ia,ib)] = atof(tokens[3])
	    self.tslope[(ia,ib)] = atof(tokens[4])
	elif keyword == 'PIBOND5':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.sslope5[(ia,ib)] = atof(tokens[3])
	    self.tslope5[(ia,ib)] = atof(tokens[4])
	elif keyword == 'PIBOND4':
	    if len(tokens) < 5:
		msg = 'file %s line %d expected 5 tokens, got %d'\
		      % (filename, line_no, len(tokens))
		raise RuntimeError, msg
	    ia = atoi(tokens[1])
	    ib = atoi(tokens[2])
	    self.sslope4[(ia,ib)] = atof(tokens[3])
	    self.tslope4[(ia,ib)] = atof(tokens[4])
#	else:
#	    Utility.warning('file %s line %d unknown keyword: %s'\
#			    % (filename, line_no, keyword))

    def torsionPhase(self, tokens):
	r = []
	while tokens:
	    n = atoi(tokens[1])
	    phase = (180*tinker_angle_unit, 0)[n >= 0]
	    r.append((abs(n), phase, atof(tokens[0]) * tinker_energy_unit))
	    tokens = tokens[2:]
	return r

    def setAvgAngleParameters(self):
	bond_angle_sum = {}
	for (key, angle_const) in self.bond_angle_const.items():
	    a = reduce(add, angle_const)/len(angle_const)
	    key2 = key[0:2]
	    try:
		(asum, ksum, count) = bond_angle_sum[key2]
	    except:
		asum = 0
		ksum = 0
		count = 0
	    k = self.bond_angle_energy[key]
	    bond_angle_sum[key2] = (asum + a, ksum + k, count + 1)
	    try:
		(asum, ksum, count) = bond_angle_sum[key[2]]
	    except:
		asum = 0
		ksum = 0
		count = 0
	    bond_angle_sum[key[2]] = (asum + a, ksum + k, count + 1)
	for key in bond_angle_sum.keys():
	    (angle_sum, k_sum, n) = bond_angle_sum[key]
	    try:
		(k1, k2) = key
	    except:
		k1 = key
		k2 = None
	    self.bond_angles_average[key] = (angle_sum/n, k_sum/n)

    def bondParameters(self, type_index1, type_index2):
	if type_index1 > type_index2:
	    (type_index1, type_index2) = (type_index2, type_index1)
	return self.bonds[(type_index1, type_index2)]

    def bondAngleParameters(self, type_index1, type_index2, type_index3, num_h):
	if type_index1 > type_index3:
	    (type_index1, type_index3) = (type_index3, type_index1)
	key = (type_index1, type_index2, type_index3)
	return (self.bond_angle_const[key][num_h], self.bond_angle_energy[key])

    def bondAngleAvgParameters(self, type_index1, type_index2):
	if type_index1 is None:
	    p = self.bond_angles_average[type_index2]
	elif type_index2 is None:
	    p = self.bond_angles_average[type_index1]
	else:
	    p = self.bond_angles_average[(type_index1, type_index2)]
	return (p[0], p[1])

    def dihedralParameters(self, type_index1, type_index2, type_index3, type_index4):
	if type_index2 > type_index3:
	    key = (type_index4, type_index3, type_index2, type_index1)
	else:
	    key = (type_index1, type_index2, type_index3, type_index4)
	try:
	    return self.dihedrals[key]
	except KeyError: return None

    def improperParameters(self, type_index1, type_index2, type_index3, type_index4):
	if type_index2 > type_index3:
	    key = (type_index4, type_index3, type_index2, type_index1)
	else:
	    key = (type_index1, type_index2, type_index3, type_index4)
	try:
	    return self.improperTorsion[key]
	except KeyError:
	    return None
#	    for k2 in [(key[1], key[2]), (key[1], key[3]), (key[2], key[3])]:
#		try:
#		    p = self.impropers_1[(key[0],) + k2]
#		except KeyError: pass
#	    if p is None:
#		for k in [key[1], key[2], key[3]]:
#		    try:
#			p = self.impropers_2[(key[0], k)]
#		    except KeyError: pass

    def bondOrder(self, type1):
	return self.ligands[type1]

def key4(tokens):
    ib = atoi(tokens[1])
    ic = atoi(tokens[2])
    if ib > ic:
	return (atoi(tokens[3]), ic, ib, atoi(tokens[0]))
    return (atoi(tokens[0]), ib, ic, atoi(tokens[3]))

def mm3deduceC4Type(a, bonds_to_atom, dataset):
    return 1
#atom     22    C     "CYCLOPROPANE"            6     12.000     4
#atom     56    C     "CSP3 CYCLOBUTANE"        6     12.000     4

def mm3deduceC3Type(a, bonds_to_atom, dataset):
    try:
	atno = [dataset.atomic_number[string.upper(bonds_to_atom[0].symbol)],\
		dataset.atomic_number[string.upper(bonds_to_atom[1].symbol)],\
		dataset.atomic_number[string.upper(bonds_to_atom[2].symbol)]]
    except IndexError:
	atno = []
    #if benzene:
    #   return 50
    if atno.count(8) == 1:
	return 3			# CSP2 CARBONYL
    return 2				# CSP2 ALKENE
#atom     38    C     "CSP2 CYCLOPROPENE"       6     12.000     3
#atom     50    C     "BENZENE (LOCALIZED)"     6     12.000     3
#atom     57    C     "CSP2 CYCLOBUTENE"        6     12.000     3
#atom     58    C     "C=O CYCLOBUTANONE"       6     12.000     3
#atom     67    C     "C=O CYCLOPROPANONE"      6     12.000     3
#atom    156    C     "C=C NUCLEIC ACID"        6     12.000     3
#atom    157    C     "C=C NUCLEIC ACID"        6     12.000     3
#atom    158    C     "C=O NUCLEIC ACID"        6     12.000     3


def mm3deduceC2Type(a, bonded_to_list, dataset):
    try:
	a0 = bonded_to_list[0]
	a1 = bonded_to_list[1]
    except IndexError:
	return 4
    atno0 = dataset.atomic_number[string.upper(a0.symbol)]
    atno1 = dataset.atomic_number[string.upper(a1.symbol)]
    if atno0 == 6 and atno1 == 6:
	return 68			# CUMULENE CSP =C=
    if atno0 == 8 or atno1 == 8:
	return 106			# =C=O KETENE
    # when to use 71 "KETONIUM CARBON"?
    return 4				# CSP ALKYNE

def mm3deduceN3Type(a, bonded_to_list, dataset):
    return 8

def mm3deduceN2Type(a, bonded_to_list, dataset):
    return 37				# temp hack
#atom     37    N     "-N=C-/PYR (DELOCLZD)"    7     14.003     2
#atom     43    N     "=N-O AZOXY (LOCAL)"      7     14.003     2
#atom     45    N     "AZIDE (CENTER N)"        7     14.003     2
#atom     72    N     "=N- IMINE (LOCALZD)"     7     14.003     2
#atom    107    N     "-N=N- AZO (LOCAL)"       7     14.003     2
#atom    108    N     "=N-OH OXIME"             7     14.003     2
#atom    109    N     "-N= AZOXY (LOCAL)"       7     14.003     2
#atom    143    N     "=N-O AXOXY (DELOC)"      7     14.003     2
#atom    144    N     "-N= AZOXY (DELOC)"       7     14.003     2


def mm3deduceN1Type(a, bonded_to_list, dataset):
    return 10

def mm3deduceO2Type(a, bonded_to_list, dataset):
    a0 = bonded_to_list[0]
    a1 = bonded_to_list[1]
    atno0 = dataset.atomic_number[string.upper(a0.symbol)]
    atno1 = dataset.atomic_number[string.upper(a1.symbol)]
    if atno1 < atno0:
	(atno0, atno1) = (atno1, atno0)
	(a0, a1) = (a1, a0)
    if atno0 == 6 and atno1 == 6:
	if lookUpBond(a0, a1):
	    return 49			# EPOXY
    if atno0 == 1 and atno1 == 7:
	return 145			# >N-OH HYDROXYAMINE
    if atno0 == 1 and atno1 == 6:
	if isCarboxyl(a1, a, dataset.atomic_number):
	    return 75			# O-H, O-C (CARBOXYL)
    return 6				# C-O-H, C-O-C, O-O
# ignoring:
#atom     41    O     "OSP2 FURAN"              8     15.995     2
#atom    148    O     "-O- ANHYDRIDE (LOCL)"    8     15.995     2
#atom    149    O     "-O- ANHYDRIDE (DELO)"    8     15.995     2


def mm3deduceO1Type(a, bonded_to_list, dataset):
    return 7

def mm3deduceP4Type(a, bonded_to_list, dataset):
    for a2 in bonded_to_list:
	if dataset.atomic_number[a2] == 8\
	   and len(a2.bondedTo()) + a.numDanglingBondsTo() == 1:
	    return 153			# >P=O PHOSPHATE
    return 60				# >P- PHOSPHINE

def mm3deduceS2Type(a, bonded_to_list, dataset):
    return 15
#atom     15    S     "-S- SULFIDE"            16     31.972     2
#atom     42    S     "SSP2 THIOPHENE"         16     31.972     2
#atom    104    S     "-S-S- DISULFIDE"        16     31.972     2
#atom    105    S     "-S- POLYSULFIDE"        16     31.972     2
def mm3deduceS4Type(a, bonded_to_list, dataset):
    return 18
#atom     18    S     ">SO2 SULFONE"           16     31.972     4
#atom    154    S     ">SO2 SULFONAMIDE"       16     31.972     4

def mm3deduceH1Type(a, bonded_to_list, dataset):
    a0 = bonded_to_list[0]
    atno0 = dataset.atomic_number[string.upper(a0.symbol)]
    if atno0 == 8:			# hydroxyl
	a01 = otherAtom(a0.bondedTo(), a)
	if a01 is not None and dataset.atomic_number[string.upper(a01.symbol)] == 6: # C-O-H
	    if isCarboxyl(a01, a0, dataset.atomic_number):
		return 24		# COOH
	    # confused about whether 73 is H in "H-O ENOL/PHENOL" or a nitrogen
	    return 21
	if a01 is not None and a01.symbol == 'H':
	    return 21			# uninformed guess
	raise KeyError, 'mm3 only understands hydroxyls attached to C, got one attached to ' + a01.name
    if atno0 == 7:			# amine
	other_atomic_nos = atomsTypesBondedOtherThan(a, a0, dataset.atomic_number)
	if other_atomic_nos.count(1) > 2:
	    return 48			# ammonium
	if other_atomic_nos.count(6) == 0:
	    raise KeyError, 'mm3 cant handle H-N-(not C)'
	if 0:
	    return 28			# H-N-C=O AMIDE
	return 23			# NH AMINE/IMINE
    if atno0 == 16:
	return 44			# SH THIOL
    if atno0 == 6:
	print a0.fullName(), a0.bondedTo()
	a01 = otherAtom(a0.bondedTo(), a)
	#if atom_type[a01] == 2:
	#    return 124
    return 5

def isCarboxyl(c, o_of_oh, atomic_number):
    blist = c.bondedTo()
    if len(blist) + c.numDanglingBondsTo() != 2:
	return 0
    for a in blist:
	if atomic_number[string.upper(a.symbol)] == 8 \
	   and len(a.bondedTo() + a.numDanglingBondsTo()) == 1:
	    return 1
    return 0

def atomsTypesBondedOtherThan(exclude_a, central_a, atomic_number):
    r = []
    for a in central_a.bondedTo():
	if a is not exclude_a:
	    r.append(atomic_number[string.upper(a.symbol)])
    return r
	
def otherAtom(alist, a):
    try:
	return alist[alist[1] is a]
    except IndexError:
	if len(alist) == 1 and alist[0] is not a:
	    return alist[0]
	return None
