# This module implements the various types of universes
# (infinite, periodic etc.). A universe defines the
# geometry of space, the force field, and external interactions
# (boundary conditions, external fields, etc.)
#
# Written by Konrad Hinsen
# last revision: 1998-2-4
#

import ChemicalObjects, Collection, ForceField, RandomStuff, Utility, \
       Variables, Visualization
import Transformation
from Vector import Vector, isVector, ex, ey, ez
import copy, Numeric, operator, string

#
# The base class for all universes.
#
class Universe(Collection.GroupOfAtoms, Visualization.Viewable):

    def __init__(self, forcefield, properties):
	self._forcefield = forcefield
	self._evaluator = {}
	self.name = ''
	if properties.has_key('name'):
	    self.name = properties['name']
	    del properties['name']
	self._objects = Collection.Collection()
	self._configuration = None
	self.atom_properties = {}
	self._version = 0
	self._ncoor = None

    is_universe = 1
    is_periodic = 0

    def __getstate__(self):
	state = copy.copy(self.__dict__)
	state['atom_properties'] = {}
	state['_evaluator'] = {}
	state['_configuration'] = None
	return state

    def __setstate__(self, state):
	for attr, value in state.items():
	    self.__dict__[attr] = value
	self._evaluator = {}

    def __len__(self):
	return len(self._objects)

    def __getitem__(self, item):
	return self._objects[item]

    def __setattr__(self, attr, value):
	self.__dict__[attr] = value
	if attr[0] != '_':
	    self.addObject(value)

    def __delattr__(self, attr):
	self.removeObject(self.__dict__[attr])
	del self.__dict__[attr]

    def __repr__(self):
	return self.__class__.__name__ + ' ' + self.name
    __str__ = __repr__

    def __copy__(self):
	return copy.deepcopy(self)

    def objectList(self):
	return self._objects.objectList()

    def atomList(self):
	return self._objects.atomList()

    def bondList(self):
	return self._objects.bondList()

    def danglingBondList(self):
	return self._objects.danglingBondList()

    def universe(self):
	return self

    def addObject(self, object):
	if Collection.isCollection(object):
	    for o in object:
		self.addObject(o)
	elif ChemicalObjects.isChemicalObject(object):
	    if object.parent is not None:
		if isUniverse(object.parent):
		    raise ValueError, `object` + \
			  ' is already in another universe'
		else:
		    raise ValueError, `object` + ' is part of another object'
	    object.parent = self
	    self._objects.addObject(object)
	    self._changed()

    def removeObject(self, object):
	if Collection.isCollection(object):
	    for o in object:
		self.removeObject(o)
	elif ChemicalObjects.isChemicalObject(object):
	    if object.parent != self:
		raise ValueError, `object` + ' is not in this universe.'
	    object.parent = None
	    self._objects.removeObject(object)
	    self._changed()

    def selectShell(self, point, r1, r2=0.):
	return self._objects.selectShell(point, r1, r2)

    def selectBox(self, p1, p2):
	return self._objects.selectBox(p1, p2)

    def _changed(self):
	self._evaluator = {}
	self._configuration = None
	self.atom_properties = {}
	self._ncoor = None
	self._version = self._version + 1

    def setForceField(self, forcefield):
	self._forcefield = forcefield
	self._evaluator = {}

    def position(self, object, conf):
	if ChemicalObjects.isChemicalObject(object):
	    return object.position(conf)
	elif isVector(object):
	    return object
	else:
	    return Vector(object)

    def numberOfCartesianCoordinates(self):
	if self._ncoor is None:
	    self._ncoor = 0
	    for o in self._objects.objectList():
		self._ncoor = self._ncoor + o.numberOfCartesianCoordinates()
	return self._ncoor

    def degreesOfFreedom(self):
	return 3*self.numberOfCartesianCoordinates()

    def configuration(self):
	if self._configuration is None:
	    ncoor = self.numberOfCartesianCoordinates()
	    coordinates = Numeric.zeros((ncoor, 3), Numeric.Float)
	    indices = range(ncoor)
	    redef = []
	    for a in self._objects.atomList():
		if not a.setArray(coordinates, indices):
		    redef.append(a)
	    for a in redef:
		a.setArray(coordinates, indices[0:1])
		del indices[0]
	    self._configuration = 1
	    self._configuration = Variables.ParticleVector(self, coordinates)
	return self._configuration

    def atomNames(self):
	self.configuration()
	names = self.numberOfCartesianCoordinates()*[None]
	for a in self._objects.atomList():
	    names[a.index] = a.fullName()
	return names

    def setConfiguration(self, configuration):
	conf = self.configuration()
	conf.assign(configuration)

    def _setConfiguration(self, coordinates):
	self._configuration = Variables.ParticleVector(self, coordinates)

    def addToConfiguration(self, displacement):
	conf = self.configuration()
	conf.assign(conf+displacement)

    def getAtomScalarArray(self, name, datatype = Numeric.Float):
	conf = self.configuration()
	array = Numeric.zeros((len(conf),), datatype)
	for a in self._objects.atomList():
	    array[a.index] = getattr(a, name)
	return Variables.ParticleScalar(self, array)

    def getAtomBooleanArray(self, name):
	conf = self.configuration()
	array = Numeric.zeros((len(conf),), Numeric.Int8)
	for a in self._objects.atomList():
	    try:
		array[a.index] = getattr(a, name)
	    except AttributeError: pass
	return Variables.ParticleScalar(self, array)

    def renumberAtoms(self, indices):
	conf = self.configuration()
	for a in self.atomList():
	    a.index = indices[a.index]
	indices = Numeric.argsort(indices)
	conf.array[:] = Numeric.take(conf.array, indices)
	self._changed()

    def masses(self):
	return self.getAtomScalarArray('_mass')

    def velocities(self):
	try:
	    return self.atom_properties['velocity']
	except KeyError:
	    return None

    def initializeVelocitiesToTemperature(self, temperature):
	self.configuration()
	masses = self.masses()
	if self.atom_properties.has_key('velocity'):
	    del self.atom_properties['velocity']
	fixed = self.getAtomBooleanArray('fixed')
	ncoor = self.numberOfCartesianCoordinates()
	velocities = Numeric.zeros((ncoor, 3), Numeric.Float)
	for i in xrange(self.numberOfCartesianCoordinates()):
	    m = masses[i]
	    if m > 0. and not fixed[i]:
		velocities[i] = RandomStuff.randomVelocity(temperature, m).array
	self.atom_properties['velocity'] = \
			  Variables.ParticleVector(self, velocities)

    def scaleVelocitiesToTemperature(self, temperature):
	velocities = self.velocities()
	factor = Numeric.sqrt(temperature/self.temperature())
	velocities.scaleBy(factor)

    def forcefield(self):
	return self._forcefield

    def energyEvaluator(self, subset1 = None, subset2 = None):
	if self._forcefield is None:
	    raise ValueError, "no force field defined"
	try:
	    eval = self._evaluator[(subset1, subset2)]
	except KeyError:
	    eval = ForceField.EnergyEvaluator(self, self._forcefield,
					      subset1, subset2)
	    self._evaluator[(subset1, subset2)] = eval
	return eval

    def energy(self, subset1 = None, subset2 = None, small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(0, 0, small_change)

    def energyAndGradients(self, subset1 = None, subset2 = None,
			   small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(1, 0, small_change)

    def energyAndForceConstants(self, subset1 = None, subset2 = None,
				small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	e, g, fc =  eval(0, 1, small_change)
	return e, fc

    def energyGradientsAndForceConstants(self, subset1 = None, subset2 = None,
					 small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	return eval(1, 1, small_change)

    def energyTerms(self, subset1 = None, subset2 = None, small_change=0):
	eval = self.energyEvaluator(subset1, subset2)
	eval(0, 0, small_change)
	return eval.lastEnergyTerms()

    def distance(self, p1, p2, conf = None):
	return self.distanceVector(p1, p2, conf).length()

    def angle(self, p1, p2, p3, conf = None):
	v1 = self.distanceVector(p2, p1, conf)
	v2 = self.distanceVector(p2, p3, conf)
	return v1.angle(v2)

    def dihedral(self, p1, p2, p3, p4, conf = None):
	v1 = self.distanceVector(p2, p1, conf)
	v2 = self.distanceVector(p3, p2, conf)
	v3 = self.distanceVector(p3, p4, conf)
	a = v1.cross(v2).normal()
	b = v3.cross(v2).normal()
	cos = a*b
	sin = b.cross(a)*v2/v2.length()
	return Transformation.angleFromSineAndCosine(sin, cos)

    def _deleteAtom(self, atom):
	pass

    def _replaceAtom(self, atom, new_atom):
	pass

    def baseVectors(self):
	return None

    def cellVolume(self):
	return None

    def randomPoint(self):
	raise TypeError, "undefined operation"

    def map(self, function):
	return self._objects.map(function)

    def description(self, objects = None, index_map = None):
	if objects is None:
	    objects = self
	return 'c(' + \
	       `self.__class__.__name__ + self._descriptionArguments()` + \
	       ',[' + string.join(map(lambda o, m=index_map: o.description(m),
				      objects.objectList()), ',') + ',])'

    def _graphics(self, conf, distance_fn, model, module, options):
	return self._objects._graphics(conf, distance_fn, model,
				       module, options)

    def setFromTrajectory(self, trajectory, step):
	conf = self.configuration()
	conf.assign(trajectory.configuration[step])
	vel = self.velocities()
	try:
	    vel_tr = trajectory.velocities[step]
	except AttributeError:
	    if vel is not None:
		Utility.warning("velocities were not modified because the " +
				"trajectory does not contain velocity data.")
	    return
	if vel is None:
	    self.atom_properties['velocity'] = vel_tr
	else:
	    vel.assign(vel_tr)

#
# Infinite universes
#
class InfiniteUniverse(Universe):

    def __init__(self, forcefield=None, **properties):
	Universe.__init__(self, forcefield, properties)

    def distanceVector(self, p1, p2, conf = None):
	p1 = self.position(p1, conf)
	p2 = self.position(p2, conf)
	return p2-p1

    def CdistanceFunction(self):
	from _forcefield import infinite_universe_distance_function
	return infinite_universe_distance_function, Numeric.array([0.])

    def CcorrectionFunction(self):
	from _forcefield import infinite_universe_correction_function
	return infinite_universe_correction_function, Numeric.array([0.])

    def _descriptionArguments(self):
	return '()'

#
# Orthorhombic universe with periodic boundary conditions
#
class OrthorhombicPeriodicUniverse(Universe):

    def __init__(self, size = None, forcefield = None, **properties):
	Universe.__init__(self, forcefield, properties)
	self.data = Numeric.zeros((2, 3), Numeric.Float)
	self.edges = self.data[0]
	if size is not None:
	    self.setSize(size)

    is_periodic = 1

    def setSize(self, size):
	self.edges[0], self.edges[1], self.edges[2] = size
	self.data[1] = 0.5*self.data[0]

    def scaleSize(self, factor):
	self.data[:] = factor*self.data

    def distanceVector(self, p1, p2, conf = None):
	p1 = self.position(p1, conf)
	p2 = self.position(p2, conf)
	d = p2-p1
	edge = self.edges[0]
	half_edge = 0.5*edge
	if d[0] < 0.: half_edge = -half_edge
	x = Numeric.fmod(d[0]+half_edge, edge) - half_edge
	edge = self.edges[1]
	half_edge = 0.5*edge
	if d[1] < 0.: half_edge = -half_edge
	y = Numeric.fmod(d[1]+half_edge, edge) - half_edge
	edge = self.edges[2]
	half_edge = 0.5*edge
	if d[2] < 0.: half_edge = -half_edge
	z = Numeric.fmod(d[2]+half_edge, edge) - half_edge
	return Vector(x,y,z)

    def CdistanceFunction(self):
	from _forcefield import orthorhombic_universe_distance_function
	return orthorhombic_universe_distance_function, self.data

    def CcorrectionFunction(self):
	from _forcefield import orthorhombic_universe_correction_function
	return orthorhombic_universe_correction_function, self.data

    def baseVectors(self):
	return [self.edges[0]*ex, self.edges[1]*ey, self.edges[2]*ez]

    def cellVolume(self):
	return Numeric.multiply.reduce(self.edges)

    def randomPoint(self):
	return RandomStuff.randomPointInBox(self.edges[0], self.edges[1],
					    self.edges[2])

    def _descriptionArguments(self):
	return `((self.edges[0], self.edges[1], self.edges[2]),)`

#
# Cubic universe with periodic boundary conditions
#
class CubicPeriodicUniverse(OrthorhombicPeriodicUniverse):

    def setSize(self, size):
	OrthorhombicPeriodicUniverse.setSize(self, 3*(size,))

    def _descriptionArguments(self):
	return `(self.edges[0],)`

#
# Recognition functions
#
def isUniverse(object):
    return hasattr(object, 'is_universe')
