# This module implements classes that represent the variables in a
# simulation, i.e. configurations, force vectors, etc.
#
# Written by Konrad Hinsen
# last revision: 1998-3-30
#

from Vector import Vector, isVector
from Tensor import Tensor, isTensor
import copy, Numeric

#
# Base class for all variables defined for a universe.
#
class UniverseVariable:

    def __init__(self, universe, data_rank, value_rank):
	universe.configuration()
	self.universe = universe
	self.version =  universe._version
	self.n = universe.numberOfCartesianCoordinates()
	self.data_rank = data_rank
	self.value_rank = value_rank

    is_universe_variable = 1

    def __getinitargs__(self):
	return (None, None)

    def _checkCompatibility(self, other):
	if isUniverseVariable(other):
	    if other.data_rank != self.data_rank:
		raise TypeError, 'Incompatible types'
	    if self.universe != other.universe:
		raise ValueError, 'Variables are for different universes'
	    if self.version != other.version:
		raise ValueError, "Version numbers don't agree"
	    if self.value_rank < other.value_rank:
		return other.return_class, other.array
	    else:
		return self.return_class, other.array
	elif isVector(other) or isTensor(other):
	    if len(other.array.shape) > self.value_rank:
		raise TypeError, 'Incompatible types'
	    return self.return_class, other.array
	else:
	    return self.return_class, other

    def _arithmetic(self, other, op):
	return_class, other = self._checkCompatibility(other)
	return return_class(self.universe, op(self.array, other))

    def __add__(self, other):
	return self._arithmetic(other, Numeric.add)

    __radd__ = __add__

    def __sub__(self, other):
	return self._arithmetic(other, Numeric.subtract)

    def __rsub__(self, other):
	return self._arithmetic(other, _rsub)

    def __mul__(self, other):
	return self._arithmetic(other, Numeric.multiply)

    __rmul__ = __mul__

    def __div__(self, other):
	return self._arithmetic(other, Numeric.divide)

    def __rdiv__(self, other):
	return self._arithmetic(other, _rdiv)

    def __neg__(self):
	return self.return_class(self.universe, -self.array)

    def __copy__(self, memo = None):
	return self.__class__(self.universe, copy.copy(self.array))
    __deepcopy__ = __copy__

    def assign(self, other):
	self._checkCompatibility(other)
	self.array[:] = other.array[:]

    def scaleBy(self, factor):
	self.array[:] = self.array[:]*factor

UniverseVariable.return_class = UniverseVariable

def _rsub(a , b):
    return Numeric.subtract(b, a)

def _rdiv(a , b):
    return Numeric.divide(b, a)


#
# One scalar per particle.
#
class ParticleScalar(UniverseVariable):

    def __init__(self, universe, data_array=None):
	if universe is not None:
	    UniverseVariable.__init__(self, universe, 1, 0)
	    if data_array is None:
		self.array = Numeric.zeros((self.n,), Numeric.Float)
	    else:
		self.array = data_array
		if data_array.shape[0] != self.n:
		    raise ValueError, 'Data incompatible with universe'

    def __len__(self):
	return self.n

    def __getitem__(self, item):
	if type(item) != type(0):
	    item = item.index
	return self.array[item]

    def __setitem__(self, item, value):
	if type(item) != type(0):
	    item = item.index
	self.array[item] = value

    def zero(self):
	return 0.

    def maximum(self):
	return Numeric.maximum.reduce(self.array)

    def sumOverParticles(self):
	return Numeric.add.reduce(self.array)

ParticleScalar.return_class = ParticleScalar

#
# One vector per particle.
#
class ParticleVector(UniverseVariable):

    def __init__(self, universe, data_array=None):
	if universe is not None:
	    UniverseVariable.__init__(self, universe, 1, 1)
	    if data_array is None:
		self.array = Numeric.zeros((self.n, 3), Numeric.Float)
	    else:
		self.array = data_array
		if data_array.shape[0] != self.n:
		    raise ValueError, 'Data incompatible with universe'

    def __len__(self):
	return self.n

    def __getitem__(self, item):
	if type(item) != type(0):
	    item = item.index
	return Vector(self.array[item])

    def __setitem__(self, item, value):
	if type(item) != type(0):
	    item = item.index
	self.array[item] = value.array

    def __mul__(self, other):
	if isUniverseVariable(other):
	    if other.value_rank == 0:
		return ParticleVector(self.universe,
				      self.array*other.array[:,Numeric.NewAxis])
	    elif other.value_rank == 1:
		return ParticleScalar(self.universe,
				      Numeric.add.reduce(self.array * \
							 other.array, -1))
	    else:
		raise TypeError, 'not yet implemented'
	elif isVector(other):
	    return ParticleScalar(self.universe,
				  self.array*other.array[Numeric.NewAxis, :])
	elif isTensor(other):
	    raise TypeError, 'not yet implemented'
	else:
	    return ParticleVector(self.universe, self.array*other)

    __rmul__ = __mul__

    def zero(self):
	return Vector(0., 0., 0.)

    def length(self):
	return ParticleScalar(self.universe,
			      Numeric.sqrt(Numeric.add.reduce(self.array**2,
							      -1)))

    def sumOverParticles(self):
	return Vector(Numeric.add.reduce(self.array, 0))

    def totalNorm(self):
	return Numeric.sqrt(Numeric.add.reduce(Numeric.ravel(self.array**2)))
    norm = totalNorm

    def scaledToNorm(self, norm):
	f = norm/self.totalNorm()
	return ParticleVector(self.universe, f*self.array)

    def projection(self, other):
	return Numeric.add.reduce(Numeric.ravel(self.array * other.array))

    def massWeightedProjection(self, other):
	m = self.universe.masses().array
	return Numeric.add.reduce(Numeric.ravel(self.array * other.array * \
						m[:, Numeric.NewAxis]))

ParticleVector.return_class = ParticleVector

#
# One tensor per pair, symmetric.
#
class SymmetricPairTensor(UniverseVariable):

    def __init__(self, universe, data_array=None):
	if universe is not None:
	    UniverseVariable.__init__(self, universe, 2, 2)
	    if data_array is None:
		self.array = Numeric.zeros((self.n,3, self.n,3), Numeric.Float)
	    else:
		self.array = data_array
		if data_array.shape[0] != self.n or \
		   data_array.shape[2] != self.n:
		    raise ValueError, 'Data incompatible with universe'

    def __getitem__(self, item):
	i1, i2 = item
	if type(i1) != type(0):
	    i1 = i1.index
	if type(i2) != type(0):
	    i2 = i2.index
	if i1 > i2:
	    i1, i2 = i2, i1
	    return Tensor(Numeric.transpose(self.array[i1,:,i2,:]))
	else:
	    return Tensor(self.array[i1,:,i2,:])

    def __setitem__(self, item, value):
	i1, i2 = item
	if type(i1) != type(0):
	    i1 = i1.index
	if type(i2) != type(0):
	    i2 = i2.index
	if i1 > i2:
	    i1, i2 = i2, i1
	    self.array[i1,:,i2,:] = values.transpose().array
	else:
	    self.array[i1,:,i2,:] = values.array

    def zero(self):
	return Tensor(3*[[0., 0., 0.]])

SymmetricPairTensor.return_class = SymmetricPairTensor

#
# Type check function
#
def isUniverseVariable(x):
    return hasattr(x, 'is_universe_variable')
