/* Low-level dynamics integrators
 *
 * Written by Konrad Hinsen
 * last revision: 1998-2-4
 */

#include "forcefield.h"
#include "trajectory.h"


/* Global variables */

double temperature_factor;

/* Allocate and initialize Output variable descriptors */

static PyTrajectoryVariable *
get_data_descriptors(PyArrayObject *configuration, PyArrayObject *velocities,
		     PyArrayObject *gradients, PyArrayObject *masses,
		     double *time, double *p_energy, double *k_energy,
		     double *temperature)
{
  PyTrajectoryVariable *vars = (PyTrajectoryVariable *)
                               malloc(9*sizeof(PyTrajectoryVariable));
  if (vars != NULL) {
    vars[0].name = "time";
    vars[0].text = "Time: %lf\n";
    vars[0].unit = time_unit_name;
    vars[0].type = PyTrajectory_Scalar;
    vars[0].class = PyTrajectory_Time;
    vars[0].value.dp = time;
    vars[1].name = "potential_energy";
    vars[1].text = "Potential energy: %lf, ";
    vars[1].unit = energy_unit_name;
    vars[1].type = PyTrajectory_Scalar;
    vars[1].class = PyTrajectory_Energy;
    vars[1].value.dp = p_energy;
    vars[2].name = "kinetic_energy";
    vars[2].text = "Kinetic energy: %lf\n";
    vars[2].unit = energy_unit_name;
    vars[2].type = PyTrajectory_Scalar;
    vars[2].class = PyTrajectory_Energy;
    vars[2].value.dp = k_energy;
    vars[3].name = "temperature";
    vars[3].text = "Temperature: %lf\n";
    vars[3].unit = temperature_unit_name;
    vars[3].type = PyTrajectory_Scalar;
    vars[3].class = PyTrajectory_Thermodynamic;
    vars[3].value.dp = temperature;
    vars[4].name = "configuration";
    vars[4].text = "Configuration:\n";
    vars[4].unit = length_unit_name;
    vars[4].type = PyTrajectory_ParticleVector;
    vars[4].class = PyTrajectory_Configuration;
    vars[4].value.array = configuration;
    vars[5].name = "velocities";
    vars[5].text = "Velocities:\n";
    vars[5].unit = velocity_unit_name;
    vars[5].type = PyTrajectory_ParticleVector;
    vars[5].class = PyTrajectory_Velocities;
    vars[5].value.array = velocities;
    vars[6].name = "gradients";
    vars[6].text = "Energy gradients:\n";
    vars[6].unit = energy_gradient_unit_name;
    vars[6].type = PyTrajectory_ParticleVector;
    vars[6].class = PyTrajectory_Gradients;
    vars[6].value.array = gradients;
    vars[7].name = "masses";
    vars[7].text = "Masses:\n";
    vars[7].unit = mass_unit_name;
    vars[7].type = PyTrajectory_ParticleScalar;
    vars[7].class = PyTrajectory_Masses;
    vars[7].value.array = masses;
    vars[8].name = NULL;
  }
  return vars;
}

/* Velocity-Verlet integrator */

static PyObject *
integrateVV(PyObject *dummy, PyObject *args)
{
  PyObject *universe, *correction_function, *correction_parameters;
  PyArrayObject *configuration;
  PyArrayObject *velocities;
  PyArrayObject *masses;
  PyArrayObject *fixed;
  PyArrayObject *gradients;
  PyListObject *spec_list;
  PyFFEvaluatorObject *evaluator;
  PyTrajectoryOutputSpec *output;
  PyTrajectoryVariable *data_descriptors;
  double delta_t;
  int steps;
  char *description;
  vector3 *x, *v, *f;
  double *m;
  char *fix;
  double factor;
  correction_fn *c_fn;
  double *c_data;
  double p_energy, k_energy, temperature, time;
  int atoms, moving_atoms;
  int i, j;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "OOOO!O!O!O!OdiO!s", &universe,
			&correction_function, &correction_parameters,
			&PyArray_Type, &configuration,
			&PyArray_Type, &velocities,
			&PyArray_Type, &masses,
			&PyArray_Type, &fixed,
			&evaluator,
			&delta_t, &steps,
			&PyList_Type, &spec_list,
			&description))
    return NULL;
  if (!PyFFEvaluator_Check(evaluator)) {
    PyErr_SetString(PyExc_TypeError, "evaluator has incorrect type");
    return NULL;
  }

  /* Create gradient array */
  gradients = (PyArrayObject *)PyArray_FromDims(configuration->nd,
						configuration->dimensions,
						PyArray_DOUBLE);
  if (gradients == NULL)
    return NULL;

  /* Set some convenient variables */
  atoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  v = (vector3 *)velocities->data;
  f = (vector3 *)gradients->data;
  m = (double *)masses->data;
  fix = (char *)fixed->data;
  c_fn = (correction_fn *)PyCObject_AsVoidPtr(correction_function);
  c_data = (double *)((PyArrayObject *)correction_parameters)->data;

  moving_atoms = atoms;
  for (j = 0; j < atoms; j++)
    if (fix[j])
      moving_atoms--;

  /* Prepare output data descriptors */
  data_descriptors = get_data_descriptors(configuration, velocities, gradients,
					  masses,
					  &time, &p_energy, &k_energy,
					  &temperature);
  if (data_descriptors == NULL)
    goto error2;

  /* Initialize output */
  output = PyTrajectory_OutputSpecification(universe, spec_list,
					    description,
					    data_descriptors);
  if (output == NULL)
    goto error2;

  /* Initial force calculation */
  p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms,
				     configuration, (PyObject *)gradients,
				     NULL, NULL, NULL, 0);
  time  = 0.;
  /* Main integration loop */
  for (i = 0; i < steps; i++) {
    k_energy = 0.;
    for (j = 0; j < atoms; j++)
      if (!fix[j])
	k_energy += m[j]*(v[j][0]*v[j][0]+v[j][1]*v[j][1]+v[j][2]*v[j][2]);
    k_energy *= 0.5;
    temperature = k_energy*temperature_factor/(1.5*moving_atoms);
    if (!PyTrajectory_Output(output, i, data_descriptors))
      goto error;
    if (data_descriptors[4].modified) {
      (*c_fn)(x, atoms, c_data);
      p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms,
					 configuration, (PyObject *)gradients,
					 NULL, NULL, NULL, 1);
      data_descriptors[4].modified = 0;
    }
    for (j = 0; j < atoms; j++)
      if (!fix[j]) {
	factor = -0.5*delta_t/(m[j]);
	v[j][0] += factor*f[j][0];
	v[j][1] += factor*f[j][1];
	v[j][2] += factor*f[j][2];
	x[j][0] += delta_t*v[j][0];
	x[j][1] += delta_t*v[j][1];
	x[j][2] += delta_t*v[j][2];
      }
    (*c_fn)(x, atoms, c_data);
    p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms,
				       configuration, (PyObject *)gradients,
				       NULL, NULL, NULL, 1);
    for (j = 0; j < atoms; j++)
      if (!fix[j]) {
	factor = -0.5*delta_t/(m[j]);
	v[j][0] += factor*f[j][0];
	v[j][1] += factor*f[j][1];
	v[j][2] += factor*f[j][2];
      }
    time += delta_t;
  }
  k_energy = 0.;
  for (j = 0; j < atoms; j++)
    if (!fix[j])
      k_energy += m[j]*(v[j][0]*v[j][0]+v[j][1]*v[j][1]+v[j][2]*v[j][2]);
  k_energy *= 0.5;
  temperature = k_energy*temperature_factor/(1.5*moving_atoms);
  if (!PyTrajectory_Output(output, i, data_descriptors))
    goto error;

  PyTrajectory_OutputFinish(output, 0, data_descriptors);
  free(data_descriptors);
  Py_DECREF(gradients);
  Py_INCREF(Py_None);
  return Py_None;

error:
  PyTrajectory_OutputFinish(output, 1, data_descriptors);
  free(data_descriptors);
error2:
  Py_DECREF(gradients);
  return NULL;
}


/* Temperature scaling option */

static int
getMassesAndVelocities(PyTrajectoryVariable *dynamic_data,
		       int step, void **scratch)
{
  if (step == -1) {  /* Initialization */
    PyTrajectoryVariable *var = dynamic_data;
    int found = 0;
    *scratch = malloc(3*sizeof(void *));
    if (*scratch == NULL) {
      PyErr_NoMemory();
      return 0;
    }
    while (var->name != NULL) {
      if (strcmp(var->name, "masses") == 0) {
	((PyArrayObject **)*scratch)[0] = var->value.array;
	found++;
      }
      if (strcmp(var->name, "velocities") == 0) {
	((PyArrayObject **)*scratch)[1] = var->value.array;
	found++;
      }
      var++;
    }
    if (found != 2) {
      PyErr_SetString(PyExc_ValueError,
		      "trajectory function needs masses and velocities");
      return 0;
    }
  }
  else if (step == -2) {  /* Clean up */
    free(*scratch);
  }
  return 1;
}

static int
scaleVelocities(PyTrajectoryVariable *dynamic_data, PyObject *parameters,
		int step, void **scratch)
{
  PyArrayObject *p_array = (PyArrayObject *)parameters;
  double temperature = ((double *)p_array->data)[0];
  double window = ((double *)p_array->data)[1];
  if (step < 0) {
    if (!getMassesAndVelocities(dynamic_data, step, scratch))
      return 0;
  }
  else {  /* Scale velocities */
    PyArrayObject *masses = ((PyArrayObject **)*scratch)[0];
    PyArrayObject *velocities = ((PyArrayObject **)*scratch)[1];
    vector3 *v = (vector3 *)velocities->data;
    int atoms = velocities->dimensions[0];
    double *m = (double *)masses->data;
    double k_energy = 0.;
    double t;
    int j;
    for (j = 0; j < atoms; j++)
      k_energy += m[j]*(v[j][0]*v[j][0]+v[j][1]*v[j][1]+v[j][2]*v[j][2]);
    t = k_energy*temperature_factor/(3.*atoms);
    if (t > 0. && fabs(t-temperature) > window) {
      double f = sqrt(temperature/t);
      for (j = 0; j < atoms; j++) {
	v[j][0] *= f;
	v[j][1] *= f;
	v[j][2] *= f;
      }
    }
  }
  return 1;
}

static int
heat(PyTrajectoryVariable *dynamic_data, PyObject *parameters,
     int step, void **scratch)
{
  PyArrayObject *p_array = (PyArrayObject *)parameters;
  double temp1 = ((double *)p_array->data)[0];
  double temp2 = ((double *)p_array->data)[1];
  double gradient = ((double *)p_array->data)[2];
  if (step == -1) {  /* Initialization */
    PyTrajectoryVariable *var = dynamic_data;
    int found = 0;
    *scratch = malloc(3*sizeof(void *));
    if (*scratch == NULL) {
      PyErr_NoMemory();
      return 0;
    }
    while (var->name != NULL) {
      if (strcmp(var->name, "masses") == 0) {
	((PyArrayObject **)*scratch)[0] = var->value.array;
	found++;
      }
      if (strcmp(var->name, "velocities") == 0) {
	((PyArrayObject **)*scratch)[1] = var->value.array;
	found++;
      }
      if (strcmp(var->name, "time") == 0) {
	((double **)*scratch)[2] = var->value.dp;
	found++;
      }
      var++;
    }
    if (found != 3) {
      PyErr_SetString(PyExc_ValueError,
		      "trajectory function has insufficient data");
      return 0;
    }
  }
  else if (step == -2) {  /* Clean up */
    free(*scratch);
  }
  else {  /* Scale velocities */
    PyArrayObject *masses = ((PyArrayObject **)*scratch)[0];
    PyArrayObject *velocities = ((PyArrayObject **)*scratch)[1];
    double *time = ((double **)*scratch)[2];
    vector3 *v = (vector3 *)velocities->data;
    int atoms = velocities->dimensions[0];
    double *m = (double *)masses->data;
    double temperature = temp1 + (*time) * gradient;
    double k_energy = 0.;
    double t, f;
    int j;
    if (temperature > temp2)
      temperature = temp2;
    for (j = 0; j < atoms; j++)
      k_energy += m[j]*(v[j][0]*v[j][0]+v[j][1]*v[j][1]+v[j][2]*v[j][2]);
    t = k_energy*temperature_factor/(3.*atoms);
    if (t > 0.){
      f = sqrt(temperature/t);
      for (j = 0; j < atoms; j++) {
	v[j][0] *= f;
	v[j][1] *= f;
	v[j][2] *= f;
      }
    }
  }
  return 1;
}

/* Eliminitation of global translation */

static int
removeTranslation(PyTrajectoryVariable *dynamic_data, PyObject *parameters,
		  int step, void **scratch)
{
  if (step < 0) {
    if (!getMassesAndVelocities(dynamic_data, step, scratch))
      return 0;
  }
  else {  /* Remove translation */
    PyArrayObject *masses = ((PyArrayObject **)*scratch)[0];
    PyArrayObject *velocities = ((PyArrayObject **)*scratch)[1];
    vector3 *v = (vector3 *)velocities->data;
    int atoms = velocities->dimensions[0];
    double *m = (double *)masses->data;
    double total_mass, momentum;
    int i, j;
    total_mass = 0.;
    for (j = 0; j < atoms; j++)
      total_mass += m[j];
    for (i = 0; i < 3; i++) {
      momentum = 0.;
      for (j = 0; j < atoms; j++)
	momentum += m[j]*v[j][i];
      momentum /= total_mass;
      for (j = 0; j < atoms; j++)
	v[j][i] -= momentum;
    }
  }
  return 1;
}

/* Eliminitation of global rotation */

static void
solve_3x3(tensor3 A, vector3 B, vector3 X)
{
  double a = A[0][0];
  double b = A[1][1];
  double c = A[2][2];
  double d = A[0][1];
  double e = A[0][2];
  double f = A[1][2];
  double o = B[0];
  double p = B[1];
  double q = B[2];

  double af_de = a*f-d*e;
  double aq_eo = a*q-e*o;
  double ab_dd = a*b-d*d;
  double ac_ee = a*c-e*e;

  double z = (af_de*(a*p-d*o)-ab_dd*aq_eo) / (af_de*af_de-ab_dd*ac_ee);
  double y = (aq_eo - z*ac_ee)/af_de;
  double x = (o - d*y - e*z)/a;

  X[0] = x;
  X[1] = y;
  X[2] = z;
}

static int
removeRotation(PyTrajectoryVariable *dynamic_data, PyObject *parameters,
	       int step, void **scratch)
{
  if (step == -1) {  /* Initialization */
    PyTrajectoryVariable *var = dynamic_data;
    int found = 0;
    *scratch = malloc(3*sizeof(PyArrayObject *));
    if (*scratch == NULL) {
      PyErr_NoMemory();
      return 0;
    }
    while (var->name != NULL) {
      if (strcmp(var->name, "masses") == 0) {
	((PyArrayObject **)*scratch)[0] = var->value.array;
	found++;
      }
      if (strcmp(var->name, "velocities") == 0) {
	((PyArrayObject **)*scratch)[1] = var->value.array;
	found++;
      }
      if (strcmp(var->name, "configuration") == 0) {
	((PyArrayObject **)*scratch)[2] = var->value.array;
	found++;
      }
      var++;
    }
    if (found != 3) {
      PyErr_SetString(PyExc_ValueError,
		      "trajectory function needs configuration, "
		      "masses and velocities");
      return 0;
    }
  }
  else if (step == -2) {  /* Clean up */
    free(*scratch);
  }
  else {  /* Remove rotation */
    PyArrayObject *masses = ((PyArrayObject **)*scratch)[0];
    PyArrayObject *velocities = ((PyArrayObject **)*scratch)[1];
    PyArrayObject *configuration = ((PyArrayObject **)*scratch)[2];
    vector3 *v = (vector3 *)velocities->data;
    vector3 *x = (vector3 *)configuration->data;
    double *m = (double *)masses->data;
    int atoms = masses->dimensions[0];

    double total_mass, trace;
    vector3 cm, l, o;
    tensor3 inertia;
    int i, j, k;

    total_mass = 0.;
    cm[0] = cm[1] = cm[2] = 0.;
    for (i = 0; i < atoms; i++) {
      total_mass += m[i];
      cm[0] += x[i][0]; cm[1] += x[i][1]; cm[2] += x[i][2];
    }
    cm[0] /= total_mass; cm[1] /= total_mass; cm[2] /= total_mass;

    for (j = 0; j < 3; j++)
      for (k = 0; k < 3; k++)
	inertia[j][k] = 0.;
    l[0] = l[1] = l[2] = 0.;
    for (i = 0; i < atoms; i++) {
      vector3 temp, l1;
      tensor3 i1;
      temp[0] = x[i][0]-cm[0];
      temp[1] = x[i][1]-cm[1];
      temp[2] = x[i][2]-cm[2];
      cross(l1, temp, v[i]);
      vector_scale(l1, m[i]);
      l[0] += l1[0]; l[1] += l1[1]; l[2] += l1[2];
      tensor_product(i1, temp, temp, m[i]);
      for (j = 0; j < 3; j++)
	for (k = 0; k < 3; k++)
	  inertia[j][k] -= i1[j][k];
    }
    trace = inertia[0][0] + inertia[1][1] + inertia[2][2];
    inertia[0][0] += trace;
    inertia[1][1] += trace;
    inertia[2][2] += trace;

    solve_3x3(inertia, l, o);

    for (i = 0; i < atoms; i++) {
      vector3 temp, v1;
      temp[0] = x[i][0]-cm[0];
      temp[1] = x[i][1]-cm[1];
      temp[2] = x[i][2]-cm[2];
      cross(v1, o, temp);
      v[i][0] += v1[0];
      v[i][1] += v1[1];
      v[i][2] += v1[2];
    }

  }
  return 1;
}

/*
 * List of functions defined in the module
 */

static PyMethodDef dynamics_methods[] = {
  {"integrateVV", integrateVV, 1},
  {NULL, NULL}		/* sentinel */
};

/* Initialization function for the module */

void
init_dynamics()
{
  PyObject *m, *dict;
  PyObject *trajectory, *forcefield, *units;

  /* Create the module and add the functions */
  m = Py_InitModule("_dynamics", dynamics_methods);
  dict = PyModule_GetDict(m);

  /* Import the array module */
#ifdef import_array
  import_array();
#endif

  /* Import _forcefield and _trajectory and retrieve C API address arrays */
  forcefield = PyImport_ImportModule("_forcefield");
  if (forcefield != NULL) {
    PyObject *module_dict = PyModule_GetDict(forcefield);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyFF_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }
  trajectory = PyImport_ImportModule("_trajectory");
  if (trajectory != NULL) {
    PyObject *module_dict = PyModule_GetDict(trajectory);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyTrajectory_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }

  /* Get the temperature conversion factor from Units */
  units = PyImport_ImportModule("Units");
  if (units != NULL) {
    PyObject *module_dict = PyModule_GetDict(units);
    PyObject *factor = PyDict_GetItemString(module_dict, "k_B");
    temperature_factor = 1./PyFloat_AsDouble(factor);
  }

  /* Add addresses of C functions to module dictionary */
  PyDict_SetItemString(dict, "scaleVelocities",
		       PyCObject_FromVoidPtr((void *)scaleVelocities, NULL));
  PyDict_SetItemString(dict, "heat",
		       PyCObject_FromVoidPtr((void *)heat, NULL));
  PyDict_SetItemString(dict, "removeTranslation",
		       PyCObject_FromVoidPtr((void *)removeTranslation, NULL));
  PyDict_SetItemString(dict, "removeRotation",
		       PyCObject_FromVoidPtr((void *)removeRotation, NULL));

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module _dynamics");
}
