/* Low-level force field calculations
 *
 * Written by Konrad Hinsen
 * last revision: 1998-7-15
 */

#define _FORCEFIELD_MODULE
#include "forcefield.h"
#include "forcefield_private.h"

/* PMTA definitions */

#ifdef WITH_DPMTA
#include "dpmta.h"
#endif

/* Global variables */

double electrostatic_energy_factor;

/* Type "evaluator"
 *
 * Objects of this type are callable, expecting a coordinate array,
 * a gradient array (None if no gradients are to be calculated),
 * and a force constant array (None if no force constants are to be
 * calculated). They return a float object containing the energy.
 * For efficient access from C code, these objects have an attribute
 * cfunc, which is a pointer to an equivalent C function.
 *
 * The structure declaration is in forcefield.h.
 */

/* Allocation and deallocation */

PyFFEvaluatorObject *
PyFFEvaluator_New(void)
{
  PyFFEvaluatorObject *self;
  self = PyObject_NEW(PyFFEvaluatorObject, &PyFFEvaluator_Type);
  if (self == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  self->distance_vector_function = NULL;
  self->distance_vector_data = NULL;
  self->scratch = NULL;
  self->data1 = NULL;
  self->data2 = NULL;
  self->data3 = NULL;
  self->data4 = NULL;
  self->energy = 0.;
  self->virial = 0.;
  self->out1 = 0.;
  self->out2 = 0.;
  self->out3 = 0.;
  self->out4 = 0.;
  self->add_energy = default_add_energy;
  self->error = 0;
  self->virial_available = 0;
  return self;
}

static void
evaluator_dealloc(PyFFEvaluatorObject *self)
{
  Py_XDECREF(self->distance_vector_function);
  Py_XDECREF(self->distance_vector_data);
  Py_XDECREF(self->data1);
  Py_XDECREF(self->data2);
  Py_XDECREF(self->data3);
  Py_XDECREF(self->data4);
  if (self->scratch != NULL)
    free(self->scratch);
  PyMem_DEL(self);
}

/* Call */

#if 0
int
test_gradient_function(struct ffeval *evaluator,
		       PyObject *data, int i, vector3 gradient)
{
  PyArrayObject *array = (PyArrayObject *)data;
  vector3 *f = (vector3 *)array->data;
  int natoms = array->dimensions[0];
  if (i < 0) {
    int i;
#if 0
    printf("Zeroing gradient array\n");
#endif
    for (i = 0; i < natoms; i++) {
      f[i][0] = 0.;
      f[i][1] = 0.;
      f[i][2] = 0.;
    }
  }
  else {
#if 0
    printf("Gradient for atom %d from %s\n", i, evaluator->name);
    printf("  %lf, %lf, %lf\n", gradient[0], gradient[1], gradient[2]);
#endif
    f[i][0] += gradient[0];
    f[i][1] += gradient[1];
    f[i][2] += gradient[2];
  }
  return 1;
}

int
test_fc_function(struct ffeval *evaluator, PyObject *data,
	    int i, int j, tensor3 fc, double r_sq)
{
  PyArrayObject *array = (PyArrayObject *)data;
  double *sd = (double *)array->data;
  int natoms = array->dimensions[0];
  if (i < 0) {
    int i;
#if 0
    printf("Zeroing force constant array\n");
#endif
    for (i = 0; i < 9*natoms*natoms; i++)
      sd[i] = 0.;
  }
  else if (fc != NULL) {
    double *fcij = sd + 9*natoms*i + 3*j;
    int l, m;
#if 0
    printf("Force constants for atom pair (%d, %d) with distance %lf from %s\n",
	   i, j, sqrt(r_sq), evaluator->name);
#endif
    if (i > j)
      printf("Illegal index pair: %d > %d\n", i, j);
    for (l = 0; l < 3; l++) {
      for (m = 0; m < 3; m++)
	fcij[m] += fc[l][m];
      fcij += 3*natoms;
    }
  }
  return 1;
}
#endif

static PyObject *
evaluator_call(PyFFEvaluatorObject *self, PyObject *args)
{
  PyObject *coordinates = NULL;
  PyObject *gradients = NULL;
  PyObject *force_constants = NULL;
  int small_change = 0;
  gradient_function *gf = NULL;
  fc_function *fcf = NULL;
  double energy;
  int natoms;
  if (!PyArg_ParseTuple(args, "O|OOi",
			&coordinates, &gradients, &force_constants,
			&small_change))
    return NULL;
  if (gradients == Py_None)
    gradients = NULL;
  if (force_constants == Py_None)
    force_constants = NULL;
  if (!PyArray_Check(coordinates)) {
    PyErr_SetString(PyExc_TypeError, "argument must be an array");
    return NULL;
  }
  if (gradients != NULL && !PyArray_Check(gradients)) {
    PyObject *fnptr = PyObject_CallMethod(gradients, "accessFunction", NULL);
    if (fnptr == NULL)
      return NULL;
    gf = (gradient_function *)PyCObject_AsVoidPtr(fnptr);
  }
  if (force_constants != NULL && !PyArray_Check(force_constants)) {
    PyObject *fnptr = PyObject_CallMethod(force_constants,
					  "accessFunction", NULL);
    if (fnptr == NULL)
      return NULL;
    fcf = (fc_function *)PyCObject_AsVoidPtr(fnptr);
  }
  natoms = ((PyArrayObject *)coordinates)->dimensions[0];
  energy = (*self->eval_func)((PyObject *)self, natoms,
			      (PyArrayObject *)coordinates,
			      gradients,
			      force_constants,
			      gf, fcf, small_change);
  return PyFloat_FromDouble(energy);
}

/* Return name */

static PyObject *
get_name(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  return PyString_FromString(((PyFFEvaluatorObject *)self)->name);
}

/* Return self  */

static PyObject *
C_evaluator(PyObject *self, PyObject *args)
{
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  Py_INCREF(self);
  return self;
}

/* Return list of all elementary terms */

staticforward ff_eval_function combined_evaluator;
staticforward ff_eval_function total_evaluator;

static int
count_terms(PyFFEvaluatorObject *e)
{
  if (e->eval_func == combined_evaluator || e->eval_func == total_evaluator) {
    int i;
    int n = 0;
    for (i = 0; i < e->n; i++) {
      PyFFEvaluatorObject *eval =
	((PyFFEvaluatorObject **)((PyArrayObject *)e->data1)->data)[i];
      n += count_terms(eval);
    }
    return n;
  }
  else
    return 1;
}

static int
set_terms(PyFFEvaluatorObject *e, PyFFEvaluatorObject **list, int offset)
{
  if (e->eval_func == combined_evaluator || e->eval_func == total_evaluator) {
    int i;
    int n = 0;
    for (i = 0; i < e->n; i++) {
      PyFFEvaluatorObject *eval =
	((PyFFEvaluatorObject **)((PyArrayObject *)e->data1)->data)[i];
      n += set_terms(eval, list, offset+n);
    }
    return n;
  }
  else {
    list[offset] = e;
    Py_INCREF(e);
    return 1;
  }
}

static PyObject *
get_terms(PyObject *self, PyObject *args)
{
  PyFFEvaluatorObject *e = (PyFFEvaluatorObject *)self;
  PyArrayObject *array;
  int n;
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  if (e->eval_func == combined_evaluator || e->eval_func == total_evaluator) {
    n = count_terms(e);
    array = (PyArrayObject *)PyArray_FromDims(1, &n, PyArray_OBJECT);
    if (array != NULL)
      set_terms(e, (PyFFEvaluatorObject **)array->data, 0);
  }
  else {
    n = 1;
    array = (PyArrayObject *)PyArray_FromDims(1, &n, PyArray_OBJECT);
    if (array != NULL)
      *(PyObject **)array->data = self;
  }
  return (PyObject *)array;
}

/* Return dictionary containing the terms of the last energy evaluation */

static PyObject *
get_last(PyObject *self, PyObject *args)
{
  PyFFEvaluatorObject *e = (PyFFEvaluatorObject *)self;
  PyObject *dict;
  if (!PyArg_ParseTuple(args, ""))
    return NULL;
  dict = PyDict_New();
  if (dict != NULL)
    e->add_energy(self, dict);
  return dict;
}

void
default_add_energy(PyObject *ev, PyObject *dict)
{
  PyFFEvaluatorObject *e = (PyFFEvaluatorObject *)ev;
  PyObject *energy = PyFloat_FromDouble(e->energy);
  PyDict_SetItemString(dict, e->name, energy);
  Py_DECREF(energy);
}

void
pseudo_add_energy(PyObject *ev, PyObject *dict)
{
}

void
combined_add_energy(PyObject *ev, PyObject *dict)
{
  PyFFEvaluatorObject *e = (PyFFEvaluatorObject *)ev;
  int i;
  for (i = 0; i < e->n; i++) {
    PyFFEvaluatorObject *eval =
      ((PyFFEvaluatorObject **)((PyArrayObject *)e->data1)->data)[i];
    eval->add_energy((PyObject *)eval, dict);
  }
}

void
python_add_energy(PyObject *ev, PyObject *dict)
{
  /* Attention: there is dangerously little error checking in this function! */
  PyFFEvaluatorObject *e = (PyFFEvaluatorObject *)ev;
  PyObject *py_eval = e->data1;
  PyObject *result = PyObject_CallMethod(py_eval, "lastEnergyTerms", NULL);
  if (result != NULL) {
    PyObject *list = PyDict_Items(result);
    if (list != NULL) {
      int n = PyList_Size(list);
      int i;
      for (i = 0; i < n; i++) {
	PyObject *item = PyList_GetItem(list, i);
	PyDict_SetItem(dict, PyTuple_GetItem(item, 0),
		       PyTuple_GetItem(item, 1));
      }
      Py_DECREF(list);
    }
    Py_DECREF(result);
  }
}

/* Documentation string */

static char PyFFEvaluator_Type__doc__[] = 
  "force field evaluator";

/* Method table */

statichere struct PyMethodDef evaluator_methods[] = {
  {"name", get_name, 1},
  {"CEvaluator", C_evaluator, 1},
  {"lastEnergyTerms", get_last, 1},
  {"allTerms", get_terms, 1},
  {NULL, NULL} /* sentinel */
};

/* Attribute access. */

static PyObject *
evaluator_getattr(PyFFEvaluatorObject *self, char *name)
{
  return Py_FindMethod(evaluator_methods, (PyObject *)self, name);
}

/* Type object */

statichere PyTypeObject PyFFEvaluator_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,			          /*ob_size*/
  "FFEvaluator",	          /*tp_name*/
  sizeof(PyFFEvaluatorObject),    /*tp_basicsize*/
  0,			          /*tp_itemsize*/
  /* methods */
  (destructor)evaluator_dealloc,  /*tp_dealloc*/
  0,			          /*tp_print*/
  (getattrfunc)evaluator_getattr, /*tp_getattr*/
  0, 			          /*tp_setattr*/
  0,			          /*tp_compare*/
  0,                              /*tp_repr*/
  0,                              /*tp_as_number*/
  0,			          /*tp_as_sequence*/
  0,			          /*tp_as_mapping*/
  0,			          /*tp_hash*/
  (ternaryfunc)evaluator_call,	  /*tp_call*/
  0,                              /*tp_str*/
  0,                              /*tp_getattro*/
  0,                              /*tp_setattro*/
  /* Space for future expansion */
  0L,0L,
  /* Documentation string */
  PyFFEvaluator_Type__doc__
};


/* Type "non-bonded list"
 *
 * Objects of this type provide a list of atoms pairs whose distance
 * is below a certain cutoff and which do not appear in the excluded
 * pair list. Pairs that are in the 1-4 list are marked.
 *
 * The structure declaration is in forcefield.h.
 */

/* Allocation and deallocation */

static PyNonbondedListObject *
nblist_new(void)
{
  PyNonbondedListObject *self;
  self = PyObject_NEW(PyNonbondedListObject, &PyNonbondedList_Type);
  if (self == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  self->excluded_pairs = NULL;
  self->one_four_pairs = NULL;
  self->atom_subset = NULL;
  self->distance_vector_function = NULL;
  self->distance_vector_data = NULL;
  self->pairs = NULL;
  self->allocated_pairs = 0;
  self->npairs = 0;
  self->remake = 1;
  return self;
}

static void
nblist_dealloc(PyNonbondedListObject *self)
{
  Py_XDECREF(self->excluded_pairs);
  Py_XDECREF(self->one_four_pairs);
  Py_XDECREF(self->atom_subset);
  Py_XDECREF(self->distance_vector_function);
  Py_XDECREF(self->distance_vector_data);
  if (self->pairs != NULL)
    free(self->pairs);
  PyMem_DEL(self);
}

/* Sequence protocol */

static int
nblist_length(PyNonbondedListObject *self)
{
  return self->npairs;
}

static PyObject *
nblist_item(PyNonbondedListObject *self, int i)
{
  if (i < 0 || i >= self->npairs) {
    PyErr_SetString(PyExc_IndexError,"index out of bounds");
    return NULL;
  }
  else
    return Py_BuildValue("iidi", self->pairs[i].i1, self->pairs[i].i2,
			 self->pairs[i].r_sq, self->pairs[i].one_four);
}

/* Methods */

static PyObject *
nblist_update_py(PyObject *self, PyObject *args)
{
  PyArrayObject *array;
  if (!PyArg_ParseTuple(args, "O", &array))
    return NULL;
  if (!PyArray_Check(array)) {
    PyErr_SetString(PyExc_TypeError, "argument must be an array");
    return NULL;
  }
  ((PyNonbondedListObject *)self)->remake = 1;
  nblist_update((PyNonbondedListObject *)self,
		array->dimensions[0], (double *)array->data);
  Py_INCREF(Py_None);
  return Py_None;
}

static PyObject *
nblist_set_cutoff(PyObject *self, PyObject *args)
{
  PyObject *cutoff_ob;
  if (!PyArg_ParseTuple(args, "O", &cutoff_ob))
    return NULL;
  if (cutoff_ob == Py_None)
    ((PyNonbondedListObject *)self)->cutoff = 0.;
  else if (PyNumber_Check(cutoff_ob)) {
    cutoff_ob = PyNumber_Float(cutoff_ob);
    ((PyNonbondedListObject *)self)->cutoff = PyFloat_AsDouble(cutoff_ob);
  }
  else {
    PyErr_SetString(PyExc_TypeError, "cutoff must be a number or None");
    return NULL;
  }
  Py_INCREF(Py_None);
  return Py_None;
}

/* Documentation string */

static char PyNonbondedList_Type__doc__[] = 
  "nonbonded list";

/* Method table */

statichere struct PyMethodDef nblist_methods[] = {
  {"update", nblist_update_py, 1},
  {"setCutoff", nblist_set_cutoff, 1},
  {NULL, NULL} /* sentinel */
};

/* Attribute access. */

static PyObject *
nblist_getattr(PyNonbondedListObject *self, char *name)
{
  return Py_FindMethod(nblist_methods, (PyObject *)self, name);
}

/* Type object */

static PySequenceMethods nblist_as_sequence = {
  (inquiry)nblist_length,   /*sq_length*/
  0,                        /*nb_add, concat is numeric add*/
  0,                        /*nb_multiply, repeat is numeric multiply*/
  (intargfunc)nblist_item,  /* sq_item*/
  0,                        /*sq_slice*/
  0,                        /*sq_ass_item*/
  0,                        /*sq_ass_slice*/
};

statichere PyTypeObject PyNonbondedList_Type = {
  PyObject_HEAD_INIT(&PyType_Type)
  0,			          /*ob_size*/
  "NonbondedList",	          /*tp_name*/
  sizeof(PyNonbondedListObject),  /*tp_basicsize*/
  0,			          /*tp_itemsize*/
  /* methods */
  (destructor)nblist_dealloc,     /*tp_dealloc*/
  0,			          /*tp_print*/
  (getattrfunc)nblist_getattr,    /*tp_getattr*/
  0, 			          /*tp_setattr*/
  0,			          /*tp_compare*/
  0,                              /*tp_repr*/
  0,			          /*tp_as_number*/
  &nblist_as_sequence,            /*tp_as_sequence*/
  0,			          /*tp_as_mapping*/
  0,			          /*tp_hash*/
  0,                              /*tp_call*/
  0,                              /*tp_str*/
  0,                              /*tp_getattro*/
  0,                              /*tp_setattro*/
  /* Space for future expansion */
  0L,0L,
  /* Documentation string */
  PyNonbondedList_Type__doc__
};

/*
 * Force field evaluation function for combined force fields
 */

static double
combined_evaluator(PyObject *object, int natoms, PyArrayObject *coordinates,
		   PyObject *gradients, PyObject *force_constants,
		   gradient_function *gradient_fn, fc_function *fc_fn,
		   int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  PyFFEvaluatorObject *term;
  int i;
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  for (i = 0; i < self->n; i++) {
    object = ((PyObject **)((PyArrayObject *)self->data1)->data)[i];
    term = (PyFFEvaluatorObject *)object;
    (*term->eval_func)(object, natoms, coordinates, gradients, force_constants,
		       gradient_fn, fc_fn, small_change);
    self->energy += term->energy;
    self->virial += term->virial;
    self->virial_available &= term->virial_available;
  }
  return self->energy;
}

static double
total_evaluator(PyObject *object, int natoms, PyArrayObject *coordinates,
		PyObject *gradients, PyObject *force_constants,
		gradient_function *gradient_fn, fc_function *fc_fn,
		int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  PyFFEvaluatorObject *term;
  int i;
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  if (gradients != NULL)
    if (gradient_fn != NULL)
      (*gradient_fn)(self, gradients, -1, NULL);
    else {
      double *data = (double *)((PyArrayObject *)gradients)->data;
      for (i = 0; i < 3*natoms; i++)
	data[i] = 0.;
    }
  if (force_constants != NULL)
    if (fc_fn != NULL)
      (*fc_fn)(self, force_constants, -1, -1, NULL, 0.);
    else {
      double *data = (double *)((PyArrayObject *)force_constants)->data;
      for (i = 0; i < 9*natoms*natoms; i++)
	data[i] = 0.;
    }
  for (i = 0; i < self->n; i++) {
    object = ((PyObject **)((PyArrayObject *)self->data1)->data)[i];
    term = (PyFFEvaluatorObject *)object;
    (*term->eval_func)(object, natoms, coordinates, gradients, force_constants,
		       gradient_fn, fc_fn, small_change);
    self->energy += term->energy;
    self->virial += term->virial;
    self->virial_available &= term->virial_available;
  }
  return self->energy;
}

/*
 * Force field evaluation function for Python force fields
 */

static double
python_evaluator(PyObject *object, int natoms, PyArrayObject *coordinates,
		 PyObject *gradients, PyObject *force_constants,
		 gradient_function *gradient_fn, fc_function *fc_fn,
		 int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  PyObject *py_eval = self->data1;
  PyObject *args, *result;

  self->error = 0;
  if (gradients == NULL)
    gradients = Py_None;
  if (force_constants == NULL)
    force_constants = Py_None;
  args = PyTuple_New(3);
  Py_INCREF(coordinates);
  Py_INCREF(gradients);
  Py_INCREF(force_constants);
  PyTuple_SetItem(args, 0, (PyObject *)coordinates);
  PyTuple_SetItem(args, 1, gradients);
  PyTuple_SetItem(args, 2, force_constants);    
  result = PyObject_CallObject(py_eval, args);
  Py_DECREF(args);
  if (result == NULL) {
    self->error = 1;
    return 0.;
  }
  else {
    self->energy = PyFloat_AsDouble(result);
    self->virial_available = 0;
    return self->energy;
  }
}

/*
 * Support functions for pair force fields
 */

void
add_pair_fc(PyFFEvaluatorObject *eval, PyObject *fc, fc_function *fc_fn,
	    int i, int j, vector3 dr, double r_sq, double f1, double f2)
{
  if (fc_fn != NULL) {
    if ((*fc_fn)(eval, fc, i, j, NULL, r_sq)) {
      tensor3 fij;
      int k, l;
      for (k = 0; k < 3; k++) {
	for (l = 0; l < 3; l++)
	  fij[k][l] = (f2-f1)*dr[k]*dr[l]/r_sq;
	fij[k][k] += f1;
      }
      (*fc_fn)(eval, fc, i, i, fij, r_sq);
      (*fc_fn)(eval, fc, j, j, fij, r_sq);
      tensor_changesign(fij);
      if (i > j)
	(*fc_fn)(eval, fc, j, i, fij, r_sq);
      else
	(*fc_fn)(eval, fc, i, j, fij, r_sq);
    }
  }
  else {
    double *data = (double *)((PyArrayObject *)fc)->data;
    int n = ((PyArrayObject *)fc)->dimensions[0];
    double *fcii = data + 9*n*i+3*i;
    double *fcjj = data + 9*n*j+3*j;
    double *fcij;
    int k, l;
    if (i > j) {
      int temp = i;  i = j;  j = temp;
    }
    fcij = data + 9*n*i+3*j;
    for (k = 0; k < 3; k++) {
      for (l = 0; l < 3; l++) {
	double f = (f2-f1)*dr[k]*dr[l]/r_sq;
	fcii[3*n*k+l] += f;
	fcjj[3*n*k+l] += f;
	fcij[3*n*k+l] -= f;
      }
      fcii[k*(3*n+1)] += f1;
      fcjj[k*(3*n+1)] += f1;
      fcij[k*(3*n+1)] -= f1;
    }
  }
}


/*
 * Constructor functions for the various evaluators
 */

static PyObject *
ListOfNParticleTermsEvaluator(PyObject *args, ff_eval_function f)
{
  PyFFEvaluatorObject *self;
  self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOOO",
			&self->distance_vector_function,
			&self->distance_vector_data,
			&self->data1, &self->data2))
    return NULL;
  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  Py_INCREF(self->data1);  /* indices */
  Py_INCREF(self->data2);  /* parameters */
  self->n = ((PyArrayObject *)self->data1)->dimensions[0];
  self->eval_func = f;
  return (PyObject *)self;
}

static PyObject *
HarmonicBondEvaluator(PyObject *dummy, PyObject *args)
{
  PyObject *ev = ListOfNParticleTermsEvaluator(args, harmonic_bond_evaluator);
  ((PyFFEvaluatorObject *)ev)->name = "harmonic bond";
  return ev;
}

static PyObject *
HarmonicAngleEvaluator(PyObject *dummy, PyObject *args)
{
  PyObject *ev = ListOfNParticleTermsEvaluator(args, harmonic_angle_evaluator);
  ((PyFFEvaluatorObject *)ev)->name = "harmonic bond angle";
  return ev;
}

static PyObject *
CosineDihedralEvaluator(PyObject *dummy, PyObject *args)
{
  PyObject *ev = ListOfNParticleTermsEvaluator(args, cosine_dihedral_evaluator);
  ((PyFFEvaluatorObject *)ev)->name = "cosine dihedral angle";
  return ev;
}

static PyObject *
LennardJonesEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOO!O!O!dd",
			&self->distance_vector_function,
			&self->distance_vector_data,
			&PyNonbondedList_Type, &self->data1,
			&PyArray_Type, &self->data2,
			&PyArray_Type, &self->data3,
			&self->param1, &self->param2))
    return NULL;
  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  Py_INCREF(self->data1);
  Py_INCREF(self->data2);
  Py_INCREF(self->data3);
  self->eval_func = lennard_jones_evaluator;
  self->name = "Lennard-Jones";
  return (PyObject *)self;
}

static PyObject *
ElectrostaticEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOO!O!dd",
			&self->distance_vector_function,
			&self->distance_vector_data,
			&PyNonbondedList_Type, &self->data1,
			&PyArray_Type, &self->data2,
			&self->param1, &self->param2))
    return NULL;
  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  Py_INCREF(self->data1);
  Py_INCREF(self->data2);
  self->eval_func = electrostatic_evaluator;
  self->name = "electrostatic";
  return (PyObject *)self;
}

#ifdef WITH_DPMTA
static PyObject *
EsMPEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  PyArrayObject *array_subset, *box_shape;
  PmtaInitData *initdata;
  int nlevels, mp, fft, fftblock, kterm;
  double theta;
  int natoms;
  size_t scratch_size;
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOO!O!O!diiiiid",
			&self->distance_vector_function,
			&self->distance_vector_data,
			&PyArray_Type, &box_shape,
			&PyList_Type, &self->data1,
			&PyArray_Type, &self->data2,
			&self->param1,
			&nlevels, &mp, &fft, &fftblock, &kterm, &theta))
    return NULL;

  array_subset = (PyArrayObject *)PyList_GetItem(self->data1, 2);
  natoms = array_subset->dimensions[0];
  if (natoms == 0)
    natoms = ((PyArrayObject *)self->data2)->dimensions[0];
  scratch_size = sizeof(PmtaInitData) + 
                 natoms*(sizeof(PmtaParticle)+sizeof(PmtaPartInfo));
  self->scratch = malloc(scratch_size);
  if (self->scratch == NULL) {
    PyErr_NoMemory();
    return NULL;
  }
  initdata = (PmtaInitData *)self->scratch;
  initdata->nlevels = nlevels;
  initdata->mp = mp;
  initdata->fft = fft;
  initdata->fftblock = fftblock;
  initdata->kterm = kterm;
  initdata->theta = theta;
  if (box_shape->nd == 0) {
    initdata->pbc = 0;
    self->param2 = 0.;
  }
  else {
    double *data = (double *)box_shape->data;
    initdata->pbc = 1;
    self->param2 = 1.;
    initdata->v1.x = *data++;
    initdata->v1.y = *data++;
    initdata->v1.z = *data++;
    initdata->v2.x = *data++;
    initdata->v2.y = *data++;
    initdata->v2.z = *data++;
    initdata->v3.x = *data++;
    initdata->v3.y = *data++;
    initdata->v3.z = *data++;
    initdata->cellctr.x = 0.;
    initdata->cellctr.y = 0.;
    initdata->cellctr.z = 0.;
  }
  initdata->nprocs = 1;
  initdata->calling_num = 0;
  initdata->calling_tids = NULL;

  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  Py_INCREF(self->data1);
  Py_INCREF(self->data2);
  self->eval_func = es_mp_evaluator;
  self->name = "electrostatic";
  return (PyObject *)self;
}
#endif

static PyObject *
LjEsEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOO!O!O!O!dd",
			&self->distance_vector_function,
			&self->distance_vector_data,
			&PyList_Type, &self->data1,
			&PyArray_Type, &self->data2,
			&PyArray_Type, &self->data3,
			&PyArray_Type, &self->data4,
			&self->param1, &self->param2))
    return NULL;
  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  Py_INCREF(self->data1);
  Py_INCREF(self->data2);
  Py_INCREF(self->data3);
  Py_INCREF(self->data4);
  self->eval_func = lj_es_evaluator;
  self->name = "Lennard-Jones + electrostatic";
  return (PyObject *)self;
}

static PyObject *
NonbondedListEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O", &self->data1))
    return NULL;
  Py_INCREF(self->data1);
  self->eval_func = nonbonded_evaluator;
  self->name = "nonbonded update";
  self->add_energy = pseudo_add_energy;
  return (PyObject *)self;
}

static PyObject *
CombinedEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O", &self->data1))
    return NULL;
  Py_INCREF(self->data1);  /* evaluators */
  self->n = ((PyArrayObject *)self->data1)->dimensions[0];
  self->eval_func = combined_evaluator;
  self->add_energy = combined_add_energy;
  self->name = "combined";
  return (PyObject *)self;
}

static PyObject *
PythonEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O", &self->data1))
    return NULL;
  Py_INCREF(self->data1);  /* Python evaluator object */
  self->eval_func = python_evaluator;
  self->add_energy = python_add_energy;
  self->name = "Python";
  return (PyObject *)self;
}

static PyObject *
TotalEvaluator(PyObject *dummy, PyObject *args)
{
  PyFFEvaluatorObject *self = PyFFEvaluator_New();
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "O", &self->data1))
    return NULL;
  Py_INCREF(self->data1);  /* evaluators */
  self->n = ((PyArrayObject *)self->data1)->dimensions[0];
  self->eval_func = total_evaluator;
  self->add_energy = combined_add_energy;
  self->name = "total";
  return (PyObject *)self;
}

/*
 * Constructor function for non-bonded lists
 */

static PyObject *
NonbondedList(PyObject *dummy, PyObject *args)
{
  PyNonbondedListObject *self = nblist_new();
  PyObject *cutoff_ob = NULL;
  if (self == NULL)
    return NULL;
  if (!PyArg_ParseTuple(args, "OOOOOO",
			&self->excluded_pairs, &self->one_four_pairs,
			&self->atom_subset,
			&self->distance_vector_function,
			&self->distance_vector_data,
			&cutoff_ob)) {
    nblist_dealloc(self);
    return NULL;
  }
  if (cutoff_ob == Py_None)
    self->cutoff = 0.;
  else if (PyNumber_Check(cutoff_ob)) {
    cutoff_ob = PyNumber_Float(cutoff_ob);
    self->cutoff = PyFloat_AsDouble(cutoff_ob);
  }
  else {
    PyErr_SetString(PyExc_TypeError, "cutoff must be a number or None");
    nblist_dealloc(self);
    return NULL;
  }
  Py_INCREF(self->excluded_pairs);
  Py_INCREF(self->one_four_pairs);
  Py_INCREF(self->atom_subset);
  Py_INCREF(self->distance_vector_function);
  Py_INCREF(self->distance_vector_data);
  return (PyObject *)self;
}

/*
 * Distance vector and coordinate correction functions
 */

void
distance_vector(vector3 d, vector3 r1, vector3 r2, double *data)
{
  distance_vector_1(d, r1, r2, data)
}

void
orthorhombic_distance_vector(vector3 d, vector3 r1, vector3 r2, double *data)
{
  distance_vector_2(d, r1, r2, data)
}

void
no_correction(vector3 *x, int natoms, double *data)
{
}

void
orthorhombic_correction(vector3 *x, int natoms, double *data)
{
  double a = data[0];
  double b = data[1];
  double c = data[2];
  double ah = data[3];
  double bh = data[4];
  double ch = data[5];
  int i;
#if 0
  if (ah+ah != a || bh+bh != b || ch+ch != c)
    printf("Inconsistent data: %f/%f, %f/%f, %f/%f\n",
	   a, ah, b, bh, c, ch);
#endif
  for (i = 0; i < natoms; i++) {
    if (x[i][0] > ah) x[i][0] -= a;
    else if (x[i][0] < -ah) x[i][0] += a;
    if (x[i][1] > bh) x[i][1] -= b;
    else if (x[i][1] < -bh) x[i][1] += b;
    if (x[i][2] > ch) x[i][2] -= c;
    else if (x[i][2] < -ch) x[i][2] += c;
  }
}

/*
 * List of functions defined in the module
 */

static PyMethodDef forcefield_methods[] = {
  {"HarmonicBondEvaluator", HarmonicBondEvaluator, 1},
  {"HarmonicAngleEvaluator", HarmonicAngleEvaluator, 1},
  {"CosineDihedralEvaluator", CosineDihedralEvaluator, 1},
  {"LennardJonesEvaluator", LennardJonesEvaluator, 1},
  {"ElectrostaticEvaluator", ElectrostaticEvaluator, 1},
#ifdef WITH_DPMTA
  {"EsMPEvaluator",  EsMPEvaluator, 1},
#endif
  {"LjEsEvaluator",  LjEsEvaluator, 1},
  {"NonbondedListEvaluator", NonbondedListEvaluator, 1},
  {"CombinedEvaluator", CombinedEvaluator, 1},
  {"PythonEvaluator", PythonEvaluator, 1},
  {"TotalEvaluator", TotalEvaluator, 1},
  {"NonbondedList", NonbondedList, 1},
  {"SparseForceConstants", SparseForceConstants, 1},
  {NULL, NULL}		/* sentinel */
};


/* Initialization function for the module */

void
init_forcefield()
{
  PyObject *m, *d, *units;
  static void *PyFF_API[PyFF_API_pointers];

  /* Create the module and add the functions */
  m = Py_InitModule("_forcefield", forcefield_methods);
  
  /* Import the array module */
#ifdef import_array
  import_array();
#endif

  /* Add function pointer objects */
  d = PyModule_GetDict(m);
  PyDict_SetItemString(d, "infinite_universe_distance_function",
		       PyCObject_FromVoidPtr((void *) distance_vector, NULL));
  PyDict_SetItemString(d, "infinite_universe_correction_function",
		       PyCObject_FromVoidPtr((void *) no_correction, NULL));
  PyDict_SetItemString(d, "orthorhombic_universe_distance_function",
		       PyCObject_FromVoidPtr((void *)
					     orthorhombic_distance_vector,
					     NULL));
  PyDict_SetItemString(d, "orthorhombic_universe_correction_function",
		       PyCObject_FromVoidPtr((void *) orthorhombic_correction,
					     NULL));

  /* Add C API pointer array */ 
  PyFF_API[PyFFEvaluator_Type_NUM] = (void *)&PyFFEvaluator_Type;
  PyFF_API[PyNonbondedList_Type_NUM] = (void *)&PyNonbondedList_Type;
  PyFF_API[PySparseFC_New_NUM] = (void *)&PySparseFC_New;
  PyFF_API[PySparseFC_Type_NUM] = (void *)&PySparseFC_Type;
  PyFF_API[PySparseFC_Zero_NUM] = (void *)&PySparseFC_Zero;
  PyFF_API[PySparseFC_Find_NUM] = (void *)&PySparseFC_Find;
  PyFF_API[PySparseFC_AddTerm_NUM] = (void *)&PySparseFC_AddTerm;
  PyFF_API[PySparseFC_CopyToArray_NUM] = (void *)&PySparseFC_CopyToArray;
  PyFF_API[PySparseFC_AsArray_NUM] = (void *)&PySparseFC_AsArray;
  PyFF_API[PySparseFC_VectorMultiply_NUM] = (void *)&PySparseFC_VectorMultiply;
  PyFF_API[PySparseFC_Scale_NUM] = (void *)&PySparseFC_Scale;
  PyFF_API[PyFFEvaluator_New_NUM] = (void *)&PyFFEvaluator_New;
  PyDict_SetItemString(d, "_C_API",
		       PyCObject_FromVoidPtr((void *)PyFF_API, NULL));

  /* Get the energy conversion factor from Units */
  units = PyImport_ImportModule("Units");
  if (units != NULL) {
    PyObject *module_dict = PyModule_GetDict(units);
    PyObject *factor = PyDict_GetItemString(module_dict,
					    "electrostatic_energy");
    electrostatic_energy_factor = PyFloat_AsDouble(factor);
  }

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module _forcefield");
}
