/* Low-level minimization
 *
 * Written by Konrad Hinsen
 * last revision: 1998-2-4
 */

#include "forcefield.h"
#include "trajectory.h"

/* Utility functions */

#define max(a, b) ((a) > (b) ? (a) : (b))
#define min(a, b) ((a) < (b) ? (a) : (b))

/* Operations on vector arrays */

static void
copy_vectors(vector3 *s, vector3 *d, int n)
{
  double *src = (double *)s;
  double *dest = (double *)d;
  n *= 3;
  while (n--)
    *dest++ = *src++;
}

static void
add_vectors(vector3 *s, vector3 *d, int n)
{
  double *src = (double *)s;
  double *dest = (double *)d;
  n *= 3;
  while (n--)
    *dest++ += *src++;
}

static void
scale_vectors(vector3 *v, double f, int n)
{
  double *vec = (double *)v;
  n *= 3;
  while (n--)
    *vec++ *= f;
}

/* Allocate and initialize Output variable descriptors */

static PyTrajectoryVariable *
get_data_descriptors(PyArrayObject *configuration, PyArrayObject *gradients,
		     double *p_energy, double *norm)
{
  static PyTrajectoryVariable vars[5];
  if (vars != NULL) {
    vars[0].name = "potential_energy";
    vars[0].text = "Potential energy: %lf, ";
    vars[0].unit = energy_unit_name;
    vars[0].type = PyTrajectory_Scalar;
    vars[0].class = PyTrajectory_Energy;
    vars[0].value.dp = p_energy;
    vars[1].name = "gradient_norm";
    vars[1].text = "Gradient norm: %lf\n";
    vars[1].unit = energy_gradient_unit_name;
    vars[1].type = PyTrajectory_Scalar;
    vars[1].class = PyTrajectory_Energy;
    vars[1].value.dp = norm;
    vars[2].name = "configuration";
    vars[2].text = "Configuration:\n";
    vars[2].unit = length_unit_name;
    vars[2].type = PyTrajectory_ParticleVector;
    vars[2].class = PyTrajectory_Configuration;
    vars[2].value.array = configuration;
    vars[3].name = "gradients";
    vars[3].text = "Energy gradients:\n";
    vars[3].unit = energy_gradient_unit_name;
    vars[3].type = PyTrajectory_ParticleVector;
    vars[3].class = PyTrajectory_Gradients;
    vars[3].value.array = gradients;
    vars[4].name = NULL;
  }
  return vars;
}

/* Steepest descent minimizer */

static PyObject *
steepestDescent(PyObject *dummy, PyObject *args)
{
  PyObject *universe, *correction_function, *correction_parameters;
  PyArrayObject *configuration;
  PyArrayObject *fixed;
  PyListObject *spec_list;
  PyFFEvaluatorObject *evaluator;
  PyTrajectoryOutputSpec *output;
  vector3 *x, *f;
  char *fix;
  int atoms, moving_atoms;
  int steps;
  double step_size, gradient_convergence;
  char *description;
  correction_fn *c_fn;
  double *c_data;

  PyArrayObject *gradients;
  PyTrajectoryVariable *data_descriptors;
  double p_energy, norm, factor;
  double min_energy, min_norm;
  vector3 *min_configuration = NULL, *min_gradients = NULL;
  int i, j;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "OOOO!O!OiddO!s", &universe,
			&correction_function, &correction_parameters,
			&PyArray_Type, &configuration,
			&PyArray_Type, &fixed,
			&evaluator,
			&steps, &step_size, &gradient_convergence,
			&PyList_Type, &spec_list, &description))
    return NULL;
  if (!PyFFEvaluator_Check(evaluator)) {
    PyErr_SetString(PyExc_TypeError, "evaluator has incorrect type");
    return NULL;
  }
  /* Create gradient array */
  gradients = (PyArrayObject *)PyArray_FromDims(configuration->nd,
						configuration->dimensions,
						PyArray_DOUBLE);
  if (gradients == NULL)
    return NULL;
  /* Set some convenient variables */
  atoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  f = (vector3 *)gradients->data;
  fix = (char *)fixed->data;
  c_fn = (correction_fn *)PyCObject_AsVoidPtr(correction_function);
  c_data = (double *)((PyArrayObject *)correction_parameters)->data;

  moving_atoms = atoms;
  for (j = 0; j < atoms; j++)
    if (fix[j])
      moving_atoms--;

  /* Prepare output data descriptors */
  data_descriptors = get_data_descriptors(configuration, gradients,
					  &p_energy, &norm);

  /* Allocate arrays to keep track of current best point */
  min_configuration = (vector3 *)malloc(atoms*sizeof(vector3));
  min_gradients = (vector3 *)malloc(atoms*sizeof(vector3));
  if (min_configuration == NULL || min_gradients == NULL) {
    PyErr_SetString(PyExc_MemoryError, "");
    goto error2;
  }

  /* Initialize output */
  output = PyTrajectory_OutputSpecification(universe, spec_list,
					    description,
					    data_descriptors);
  if (output == NULL)
    goto error2;

  /* Minimization main loop */
  for (i = 0; i < steps; i++) {
    p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms,
				       configuration, (PyObject *)gradients,
				       NULL, NULL, NULL, i > 0);
    if (evaluator->error)
      goto error;
    norm = 0.;
    for (j = 0; j < atoms; j++)
      if (!fix[j])
	norm += f[j][0]*f[j][0] + f[j][1]*f[j][1] + f[j][2]*f[j][2];
    norm = sqrt(norm/moving_atoms);
    if (i == 0 || p_energy < min_energy) {
      min_energy = p_energy;
      min_norm = norm;
      copy_vectors(x, min_configuration, atoms);
      copy_vectors(f, min_gradients, atoms);
      step_size *= 1.1;
    }
    else {
      p_energy = min_energy;
      norm = min_norm;
      copy_vectors(min_configuration, x, atoms);
      copy_vectors(min_gradients, f, atoms);
      step_size *= 0.5;
    }
    if (norm < gradient_convergence)
      break;
    if (!PyTrajectory_Output(output, i, data_descriptors))
      goto error;
    factor = step_size/norm;
    for (j = 0; j < atoms; j++)
      if (!fix[j]) {
	x[j][0] -= factor*f[j][0];
	x[j][1] -= factor*f[j][1];
	x[j][2] -= factor*f[j][2];
      }
    (*c_fn)(x, atoms, c_data);
  }

  /* Restore minimum and do output */
  p_energy = min_energy;
  norm = min_norm;
  copy_vectors(min_configuration, x, atoms);
  copy_vectors(min_gradients, f, atoms);
  if (!PyTrajectory_Output(output, i, data_descriptors))
    goto error;

  /* Clean up and return None */
  PyTrajectory_OutputFinish(output, 0, data_descriptors);
  free(min_configuration);
  free(min_gradients);
  Py_DECREF(gradients);
  Py_INCREF(Py_None);
  return Py_None;

  /* Clean up and return error */
error:
  PyTrajectory_OutputFinish(output, 1, data_descriptors);
error2:
  if (min_configuration != NULL)
    free(min_configuration);
  if (min_gradients != NULL)
    free(min_gradients);
  Py_DECREF(gradients);
  return NULL;
}


/* Conjugate gradient minimizer */

static PyObject *
conjugateGradient(PyObject *dummy, PyObject *args)
{
  PyObject *universe, *correction_function, *correction_parameters;
  PyArrayObject *configuration;
  PyArrayObject *fixed;
  PyListObject *spec_list;
  PyFFEvaluatorObject *evaluator;
  PyTrajectoryOutputSpec *output;
  vector3 *x, *f1, *f2, *h;
  char *fix;
  int atoms, moving_atoms;
  int steps;
  double step_size, gradient_convergence;
  char *description;
  correction_fn *c_fn;
  double *c_data;

  PyArrayObject *gradients1, *gradients2, *direction;
  PyTrajectoryVariable *data_descriptors;
  double p_energy, norm_sq, last_norm_sq, dot, norm;
  double norm_h, line_convergence, sign, step, smallest, deriv;
  double last, a, b, ea, eb, da, db;
  int i, j, reset_count, niter;

  /* Parse and check arguments */
  if (!PyArg_ParseTuple(args, "OOOO!O!OiddO!s", &universe,
			&correction_function, &correction_parameters,
			&PyArray_Type, &configuration,
			&PyArray_Type, &fixed,
			&evaluator,
			&steps, &step_size, &gradient_convergence,
			&PyList_Type, &spec_list, &description))
    return NULL;
  if (!PyFFEvaluator_Check(evaluator)) {
    PyErr_SetString(PyExc_TypeError, "evaluator has incorrect type");
    return NULL;
  }
  /* Create gradient and direction arrays */
  gradients1 = (PyArrayObject *)PyArray_FromDims(configuration->nd,
						 configuration->dimensions,
						 PyArray_DOUBLE);
  if (gradients1 == NULL)
    return NULL;
  gradients2 = (PyArrayObject *)PyArray_FromDims(configuration->nd,
						 configuration->dimensions,
						 PyArray_DOUBLE);
  if (gradients2 == NULL) {
    Py_DECREF(gradients1);
    return NULL;
  }
  direction = (PyArrayObject *)PyArray_FromDims(configuration->nd,
						configuration->dimensions,
						PyArray_DOUBLE);
  if (direction == NULL) {
    Py_DECREF(gradients1);
    Py_DECREF(gradients2);
    return NULL;
  }
  /* Set some convenient variables */
  atoms = configuration->dimensions[0];
  x = (vector3 *)configuration->data;
  f1 = (vector3 *)gradients1->data;
  f2 = (vector3 *)gradients2->data;
  h = (vector3 *)direction->data;
  fix = (char *)fixed->data;
  c_fn = (correction_fn *)PyCObject_AsVoidPtr(correction_function);
  c_data = (double *)((PyArrayObject *)correction_parameters)->data;

  moving_atoms = atoms;
  for (j = 0; j < atoms; j++)
    if (fix[j])
      moving_atoms--;

  /* Prepare output data descriptors */
  data_descriptors = get_data_descriptors(configuration, gradients1,
					  &p_energy, &norm);

  /* Initialize output */
  output = PyTrajectory_OutputSpecification(universe, spec_list,
					    description,
					    data_descriptors);
  if (output == NULL)
    goto error2;

  /* Minimization main loop */
  reset_count = 0;
  p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms,
				     configuration, (PyObject *)gradients1,
				       NULL, NULL, NULL, 0);
  if (evaluator->error)
    goto error;
  norm_sq = 0.;
  for (i = 0; i < steps; i++) {
    last_norm_sq = norm_sq;
    norm_sq = 0.;
    for (j = 0; j < atoms; j++)
      if (!fix[j])
	norm_sq += f1[j][0]*f1[j][0] + f1[j][1]*f1[j][1] + f1[j][2]*f1[j][2];
    norm = sqrt(norm_sq/moving_atoms);
    if (norm < gradient_convergence)
      break;
    if (norm > 50.*gradient_convergence)
      reset_count++;
    if (!PyTrajectory_Output(output, i, data_descriptors))
      goto error;
    if (i == 0)
      copy_vectors(f1, h, atoms);
    else {
      dot = 0.;
      for (j = 0; j < atoms; j++)
	if (!fix[j]) {
	  dot += f1[j][0]*f2[j][0] + f1[j][1]*f2[j][1] + f1[j][2]*f2[j][2];
	  f2[j][0] = f1[j][0];
	  f2[j][1] = f1[j][1];
	  f2[j][2] = f1[j][2];
	}
      if (reset_count == 5*atoms) {
	for (j = 0; j < atoms; j++) {
	  h[j][0] = 0.;
	  h[j][1] = 0.;
	  h[j][2] = 0.;
	}
	reset_count = 0;
      }
      else
	scale_vectors(h, (norm_sq-dot)/last_norm_sq, atoms);
      add_vectors(f1, h, atoms);
    }
    line_convergence = 1.e-3*norm;
    if (line_convergence < gradient_convergence)
      line_convergence = gradient_convergence;
    /* Line minimization */
#define eval(p) \
    { \
      double d = (p)-last; \
      int j; \
      for (j = 0; j < atoms; j++) \
        if (!fix[j]) { \
	  x[j][0] += sign*d*h[j][0]; \
	  x[j][1] += sign*d*h[j][1]; \
	  x[j][2] += sign*d*h[j][2]; \
        } \
      p_energy = (*evaluator->eval_func)((PyObject *)evaluator, atoms, \
					 configuration, (PyObject *)gradients1,\
					 NULL, NULL, NULL, 0); \
      if (evaluator->error) \
        goto error; \
      dot = 0.; \
      for (j = 0; j < atoms; j++) \
        if (!fix[j]) \
	  dot += f1[j][0]*h[j][0] + f1[j][1]*h[j][1] + f1[j][2]*h[j][2]; \
      dot /= (sign*norm_h); \
      last = (p); \
    }

    norm_h = 0.;
    for (j = 0; j < atoms; j++)
      if (!fix[j])
	norm_h += h[j][0]*h[j][0] + h[j][1]*h[j][1] + h[j][2]*h[j][2];
    norm_h = sqrt(norm_h);
    dot = 0.;
    for (j = 0; j < atoms; j++)
      if (!fix[j])
	dot += f1[j][0]*h[j][0] + f1[j][1]*h[j][1] + f1[j][2]*h[j][2];
    sign = (dot > 0.) ? -1. : 1.;
    dot /= (sign*norm_h);
    step = step_size/norm_h;
    last = 0.;
    a = b = 0.;
    ea = eb = p_energy;
    da = db = dot;
    niter = 0;
    while (1) {
      a = b; ea = eb; da = db;
      b += step;
      eval(b); eb = p_energy; db = dot;
      if (db > 0)
	break;
      if (++niter == 100)
	break;
    }
    niter = 0;
    while (1) {
      double new = ((eb-ea)/norm_h+a*da-b*db)/(da-db);
      if (new < a || new > b)
	new = 0.5*(a+b);
      eval(new);
      if (dot < 0) {
	a = new; ea = p_energy; da = dot;
      }
      else {
	b = new; eb = p_energy; db = dot;
      }
      if (db < 0.01*line_convergence)
	break;
      if (++niter == 100)
	break;
    }
#undef eval
  }

  /* Final output */
  if (!PyTrajectory_Output(output, i, data_descriptors))
      goto error;

  /* Clean up and return None */
  PyTrajectory_OutputFinish(output, 0, data_descriptors);
  Py_DECREF(gradients1);
  Py_DECREF(gradients2);
  Py_INCREF(Py_None);
  return Py_None;

  /* Clean up and return error */
error:
  PyTrajectory_OutputFinish(output, 1, data_descriptors);
error2:
  Py_DECREF(gradients1);
  Py_DECREF(gradients2);
  Py_DECREF(direction);
  return NULL;
}

/*
 * List of functions defined in the module
 */

static PyMethodDef minimization_methods[] = {
  {"steepestDescent", steepestDescent, 1},
  {"conjugateGradient", conjugateGradient, 1},
  {NULL, NULL}		/* sentinel */
};

/* Initialization function for the module */

void
init_minimization()
{
  PyObject *m;
  PyObject *trajectory, *forcefield;

  /* Create the module and add the functions */
  m = Py_InitModule("_minimization", minimization_methods);
  
  /* Import the array module */
#ifdef import_array
  import_array();
#endif

  /* Import _forcefield and _trajectory and retrieve C API address arrays */
  forcefield = PyImport_ImportModule("_forcefield");
  if (forcefield != NULL) {
    PyObject *module_dict = PyModule_GetDict(forcefield);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyFF_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }
  trajectory = PyImport_ImportModule("_trajectory");
  if (trajectory != NULL) {
    PyObject *module_dict = PyModule_GetDict(trajectory);
    PyObject *c_api_object = PyDict_GetItemString(module_dict, "_C_API");
    if (PyCObject_Check(c_api_object)) {
      PyTrajectory_API = (void **)PyCObject_AsVoidPtr(c_api_object);
    }
  }

  /* Check for errors */
  if (PyErr_Occurred())
    Py_FatalError("can't initialize module _minimization");
}
