/* Low-level force field calculations: non-bonded interactions
 *
 * Written by Konrad Hinsen
 * last revision: 1998-7-7
 */

#define NO_IMPORT
#define _FORCEFIELD_MODULE
#include "forcefield.h"
#include "forcefield_private.h"

/* PMTA definitions */

#ifdef WITH_DPMTA
#include "dpmta.h"
#endif

/* Factor by which cutoffs are multiplied to provide a boundary
   region in which changes are monitored */
const double safety_factor = 1.08;
const double check_limit = 1.06;

/* Number of iterations before the non-bonded list is always remade */
const int remake_frequency = 50;

/* Nonbonded list update */

static int
nblist_realloc(PyNonbondedListObject *nblist, int npairs)
{
  const min_alloc = 500;
  atom_pair *new;
  int i, alloc;
  if (npairs > nblist->allocated_pairs) {
    alloc = min_alloc*((npairs+min_alloc-1)/min_alloc);
    new = (atom_pair *)malloc(alloc*sizeof(atom_pair));
    if (new == NULL)
      return 0;
    for (i = 0; i < nblist->npairs; i++)
      new[i] = nblist->pairs[i];
    if (nblist->pairs != NULL)
      free(nblist->pairs);
    nblist->pairs = new;
    nblist->allocated_pairs = alloc;
    return 1;
  }
  else
    return 1;
}

int
nblist_update(PyNonbondedListObject *nblist, int natoms, double *coordinates)
{
  vector3 *x = (vector3 *)coordinates;
  distance_fn *d_fn = (distance_fn *)
                        PyCObject_AsVoidPtr(nblist->distance_vector_function);
  double *distance_data = (double *)
                          ((PyArrayObject *)nblist->distance_vector_data)->data;
  vector3 d;

  if (nblist->remake) {

    long *excluded = (long *)((PyArrayObject *)nblist->excluded_pairs)->data;
    int n_ex = 2*((PyArrayObject *)nblist->excluded_pairs)->dimensions[0];
    long *one_four = (long *)((PyArrayObject *)nblist->one_four_pairs)->data;
    int n_14 = 2*((PyArrayObject *)nblist->one_four_pairs)->dimensions[0];
    long *subset = (long *)((PyArrayObject *)nblist->atom_subset)->data;
    int n_sub = ((PyArrayObject *)nblist->atom_subset)->dimensions[0];
    double cutoff_sq = sqr(nblist->cutoff*safety_factor);
    double check_cutoff_sq = sqr(nblist->cutoff*check_limit);
    int i_ex = 0, i_14 = 0;
    int n, i1, i2, j1, j2;
    atom_pair pair;
    double d_sq;

    nblist->npairs = 0;
    nblist_realloc(nblist, 1000);
    if (n_sub == 0)
      n = natoms;
    else
      n = n_sub;
    if (d_fn == distance_vector) {
#     define distance_vector_function distance_vector_1
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
    else if (d_fn == orthorhombic_distance_vector) {
#     define distance_vector_function distance_vector_2
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
    else {
#     define distance_vector_function (*d_fn)
#     include "nonbonded1.i"
#     undef distance_vector_function
    }
    nblist->remake = 0;
    nblist->countdown = remake_frequency;
  }
  else {

    atom_pair *pair = nblist->pairs;
    double real_cutoff_sq = sqr(nblist->cutoff);
    int i;

    if (d_fn == distance_vector) {
#     define distance_vector_function distance_vector_1
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    else if (d_fn == orthorhombic_distance_vector) {
#     define distance_vector_function distance_vector_2
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    else {
#     define distance_vector_function (*d_fn)
#     include "nonbonded2.i"
#     undef distance_vector_function
    }
    if (--nblist->countdown == 0)
      nblist->remake = 1;
  }
  return 1;
}

/* Pseudo forcefield that updates a non-bonded list */

double
nonbonded_evaluator(PyObject *object, int natoms,
		    PyArrayObject *coordinates,
		    PyObject *gradients, PyObject *force_constants,
		    gradient_function *gradient_fn, fc_function *fc_fn,
		    int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  vector3 *x = (vector3 *)coordinates->data;

  if (!small_change)
    nblist->remake = 1;
  nblist_update(nblist, natoms, (double *)x);
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = 1;
  return 0.;
}

/* Pair contribution to nonbonded interactions; written as a macro
   for efficiency */

#define pair_lj_es_energy(eps, sigma, qiqj, distance_vector_function) \
{ \
  vector3 rij; \
  double r_sq; \
  double deriv = 0.; \
  double deriv2 = 0.; \
  distance_vector_function(rij, x[j], x[i], distance_data); \
  r_sq = vector_length_sq(rij); \
  if ((eps) != 0.) { \
    double sr2 = sqr((sigma))/r_sq; \
    double sr6 = cube(sr2); \
    double sr12 = sqr(sr6); \
    double d = -24.*(eps)*(2.*sr12-sr6)/r_sq; \
    self->out1 += 4.*(eps)*(sr12-sr6); \
    self->out3 += d*r_sq; \
    deriv += d; \
    deriv2 += 24.*(eps)*(26.*sr12-7.*sr6)/r_sq; \
  } \
  if ((qiqj) != 0.) { \
    double r = sqrt(r_sq); \
    double d = -(qiqj)/(r*r_sq); \
    self->out2 += (qiqj)/r; \
    self->out4 += d*r_sq; \
    deriv += d; \
    deriv2 += 2.*(qiqj)/(r*r_sq); \
  } \
  if (gradients != NULL) { \
    vector3 grad; \
    grad[0] = deriv*rij[0]; \
    grad[1] = deriv*rij[1]; \
    grad[2] = deriv*rij[2]; \
    if (gradient_fn != NULL) { \
      (*gradient_fn)(self, gradients, i, grad); \
      vector_changesign(grad); \
      (*gradient_fn)(self, gradients, j, grad); \
    } \
    else { \
      vector3 *f = (vector3 *)((PyArrayObject *)gradients)->data; \
      f[i][0] += grad[0]; \
      f[i][1] += grad[1]; \
      f[i][2] += grad[2]; \
      f[j][0] -= grad[0]; \
      f[j][1] -= grad[1]; \
      f[j][2] -= grad[2]; \
    } \
  } \
  if (force_constants != NULL) \
    add_pair_fc(self, force_constants, fc_fn, \
		i, j, rij, r_sq, deriv, deriv2); \
}


/* Combined Lennard-Jones and electrostatic interactions,
   used only for infinite cutoff */

double
lj_es_evaluator(PyObject *object, int natoms,
		PyArrayObject *coordinates,
		PyObject *gradients, PyObject *force_constants,
		gradient_function *gradient_fn, fc_function *fc_fn,
		int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = (distance_fn *)
                        PyCObject_AsVoidPtr(self->distance_vector_function);
  double *distance_data = (double *)
                          ((PyArrayObject *)self->distance_vector_data)->data;

  PyObject *pair_lists = self->data1;
  PyArrayObject *array_ex = (PyArrayObject *)PyList_GetItem(pair_lists, 0);
  PyArrayObject *array_14 = (PyArrayObject *)PyList_GetItem(pair_lists, 1);
  PyArrayObject *array_subset =
                  (PyArrayObject *)PyList_GetItem(pair_lists, 2);
  long *excluded = (long *)array_ex->data;
  long n_ex = 2*array_ex->dimensions[0];
  long *one_four = (long *)array_14->data;
  long n_14 = 2*array_14->dimensions[0];
  long *subset = (long *)array_subset->data;
  long n_sub = array_subset->dimensions[0];
  double *eps_sigma = (double *)((PyArrayObject *)self->data2)->data;
  int ntypes = ((PyArrayObject *)self->data2)->dimensions[0];
  long *lj_type = (long *)((PyArrayObject *)self->data3)->data;
  double *charge = (double *)((PyArrayObject *)self->data4)->data;
  double lj_one_four_factor = self->param1;
  double es_one_four_factor = self->param2;
  int k, l, n;
  
  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = (gradients != NULL);
  self->out1 = self->out2 = self->out3 = self->out4 = 0.;
  if (d_fn == distance_vector) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded3.i"
#   undef distance_vector_function
  }
  self->energy = self->out1 + self->out2;
  self->virial = self->out3 + self->out4;
  return self->energy;
}


/* Lennard-Jones interactions */

double
lennard_jones_evaluator(PyObject *object, int natoms,
			PyArrayObject *coordinates,
			PyObject *gradients, PyObject *force_constants,
			gradient_function *gradient_fn, fc_function *fc_fn,
			int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = (distance_fn *)
                        PyCObject_AsVoidPtr(self->distance_vector_function);
  double *distance_data = (double *)
                          ((PyArrayObject *)self->distance_vector_data)->data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double *eps_sigma = (double *)((PyArrayObject *)self->data2)->data;
  int ntypes = ((PyArrayObject *)self->data2)->dimensions[0];
  long *lj_type = (long *)((PyArrayObject *)self->data3)->data;
  double cutoff_sq = sqr(self->param1);
  double one_four_factor = self->param2;
  double energy = 0.;
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = (gradients != NULL);
  if (d_fn == distance_vector) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded5.i"
#   undef distance_vector_function
  }
  return self->energy;
}


/* Electrostatic interactions */

double
electrostatic_evaluator(PyObject *object, int natoms,
			PyArrayObject *coordinates,
			PyObject *gradients, PyObject *force_constants,
			gradient_function *gradient_fn, fc_function *fc_fn,
			int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = (distance_fn *)
                        PyCObject_AsVoidPtr(self->distance_vector_function);
  double *distance_data = (double *)
                          ((PyArrayObject *)self->distance_vector_data)->data;

  PyNonbondedListObject *nblist = (PyNonbondedListObject *)self->data1;
  double *charge = (double *)((PyArrayObject *)self->data2)->data;
  double cutoff_sq = sqr(self->param1);
  double one_four_factor = self->param2;
  atom_pair *pair = nblist->pairs;
  int k;

  self->energy = 0.;
  self->virial = 0.;
  self->virial_available = (gradients != NULL);
  if (d_fn == distance_vector) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded6.i"
#   undef distance_vector_function
  }
  return self->energy;
}

#ifdef WITH_DPMTA
/* Electrostatic interactions via multipoles */
double
es_mp_evaluator(PyObject *object, int natoms,
		PyArrayObject *coordinates,
		PyObject *gradients, PyObject *force_constants,
		gradient_function *gradient_fn, fc_function *fc_fn,
		int small_change)
{
  PyFFEvaluatorObject *self = (PyFFEvaluatorObject *)object;
  vector3 *x = (vector3 *)coordinates->data;
  distance_fn *d_fn = (distance_fn *)
                        PyCObject_AsVoidPtr(self->distance_vector_function);
  double *distance_data = (double *)
                          ((PyArrayObject *)self->distance_vector_data)->data;
  PmtaInitData *initdata;
  static PmtaInitData *last_initdata = NULL;
  PmtaParticle *particle;
  PmtaPartInfo *force;
  PmtaVector virial;
  double potential;
  PyObject *pair_lists = self->data1;
  PyArrayObject *array_ex = (PyArrayObject *)PyList_GetItem(pair_lists, 0);
  PyArrayObject *array_14 = (PyArrayObject *)PyList_GetItem(pair_lists, 1);
  PyArrayObject *array_subset =
                  (PyArrayObject *)PyList_GetItem(pair_lists, 2);
  long *excluded = (long *)array_ex->data;
  long n_ex = 2*array_ex->dimensions[0];
  long *one_four = (long *)array_14->data;
  long n_14 = 2*array_14->dimensions[0];
  long *subset = (long *)array_subset->data;
  long n_sub = array_subset->dimensions[0];
  double *charge = (double *)((PyArrayObject *)self->data2)->data;
  double es_one_four_factor = self->param1;
  int k, l, n;

  n = (n_sub == 0) ? natoms : n_sub;
  initdata = (PmtaInitData *)self->scratch;
  particle = (PmtaParticle *)(initdata+1);
  force = (PmtaPartInfo *)(particle+n);
  if (initdata != last_initdata) {
    PMTAinit(initdata, NULL);
    last_initdata = initdata;
  }
  for (k = 0; k < n; k++) {
    int i = (n_sub == 0) ? k : subset[k];
    particle[k].p.x = x[i][0];
    particle[k].p.y = x[i][1];
    particle[k].p.z = x[i][2];
    particle[k].q = charge[i];
  }
  if (self->param2 == 0.) {
    PmtaVector v1, v2, v3, center;
    double xmin, xmax, ymin, ymax, zmin, zmax;
    xmin = ymin = zmin = undefined;
    xmax = ymax = zmax = -undefined;
    for (k = 0; k < n; k++) {
      int i = (n_sub == 0) ? k : subset[k];
      if (x[i][0] < xmin) xmin = x[i][0];
      if (x[i][1] < ymin) ymin = x[i][1];
      if (x[i][2] < zmin) zmin = x[i][2];
      if (x[i][0] > xmax) xmax = x[i][0];
      if (x[i][1] > ymax) ymax = x[i][1];
      if (x[i][2] > zmax) zmax = x[i][2];
    }
    v1.x = xmax-xmin; v1.y = 0.; v1.z = 0.;
    v2.x = 0.; v2.y = ymax-ymin; v2.z = 0.;
    v3.x = 0.; v3.y = 0.; v3.z = zmax-zmin;
    center.x = 0.5*(xmin+xmax);
    center.y = 0.5*(ymin+ymax);
    center.z = 0.5*(zmin+zmax);
    PMTAresize(&v1, &v2, &v3, &center);
  }
  else {
    int change = 0;
    if (distance_data[0] != initdata->v1.x) {
      initdata->v1.x = distance_data[0];
      change = 1;
    }
    if (distance_data[1] != initdata->v2.y) {
      initdata->v2.y = distance_data[1];
      change = 1;
    }
    if (distance_data[2] != initdata->v3.z) {
      initdata->v3.z = distance_data[2];
      change = 1;
    }
    if (change)
      PMTAresize(&initdata->v1, &initdata->v2, &initdata->v3,
		 &initdata->cellctr);
  }
  PMTAforce(n, particle, force, NULL);
  PMTAvirial(&potential, &virial, NULL, NULL);
  for (k = 0; k < n; k++) {
    int i = (n_sub == 0) ? k : subset[k];
    vector3 grad;
    grad[0] = -electrostatic_energy_factor*force[k].f.x;
    grad[1] = -electrostatic_energy_factor*force[k].f.y;
    grad[2] = -electrostatic_energy_factor*force[k].f.z;
    if (gradients != NULL) {
      if (gradient_fn != NULL)
	(*gradient_fn)(self, gradients, i, grad);
      else {
	vector3 *f = (vector3 *)((PyArrayObject *)gradients)->data;
	f[i][0] += grad[0];
	f[i][1] += grad[1];
	f[i][2] += grad[2];
      }
    }
  }
  self->virial = electrostatic_energy_factor*(virial.x+virial.y+virial.z);
  self->energy = potential*electrostatic_energy_factor;
  if (force_constants != NULL)
    self->error = 1;
  self->out1 = self->out2 = self->out3 = self->out4 = 0.;
  if (d_fn == distance_vector) {
#   define distance_vector_function distance_vector_1
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  else if (d_fn == orthorhombic_distance_vector) {
#   define distance_vector_function distance_vector_2
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  else {
#   define distance_vector_function (*d_fn)
#   include "nonbonded4.i"
#   undef distance_vector_function
  }
  self->virial_available = 1;
  self->virial += self->out4;
  self->energy += self->out2;
  return self->energy;
}
#endif
