# This file contains some examples for non-standard generation of
# protein objects from PDB files.

from mmtk import *

#
# First problem: construct an all-atom model from a structure without
# hydrogens. This is the standard problem when using an all-atom force
# field with crystallographic structures.
#
# Load the PDB sequence
sequence = PDBFile('insulin.pdb').readSequenceWithConfiguration()
# Construct the peptide chain objects
chains = map(PeptideChain, sequence)
# Construct positions of missing hydrogens
for chain in chains:
    chain.findHydrogenPositions()
# Make protein object
insulin = Protein(chains)

# Write out the structure with hydrogens to a new file - we will use
# it as an input example later on.
insulin.writeToFile('insulin_with_h.pdb')

#
# Second problem: read a file with hydrogens and create a structure
# without them. This is useful for analysis; if you don't need the
# hydrogens, processing is faster without them. Or you might want
# to compare structures with and without hydrogens.
#
# Load the PDB sequence
sequence = PDBFile('./insulin_with_h.pdb').readSequenceWithConfiguration()
# Delete hydrogens from the sequence
sequence.deleteHydrogens()
# Construct the peptide chain objects without hydrogens
chains = []
for chain in sequence:
    chains.append(PeptideChain(chain, hydrogens='none'))
# Make protein object
insulin = Protein(chains)

#
# Third problem: cut off three residues from the start of the second
# chain before constructing the protein. This is useful for comparing
# incomplete structures.
#
# Load the PDB sequence
sequence = PDBFile('insulin.pdb').readSequenceWithConfiguration()
# Cut off the unwanted residues
sequence[2] = sequence[2][3:]
# Mark the second chain as modified
for chain in sequence:
    chain.modified = 0
sequence[2].modified = 1
# Construct the peptide chain objects without hydrogens. For
# modified chains, don't assume the first residue is the N terminus.
chains = []
for chain in sequence:
    if chain.modified:
	chains.append(PeptideChain(chain, model='none', n_terminus=0))
    else:
	chains.append(PeptideChain(chain, model='none'))
# Make protein object
insulin = Protein(chains)


#
# Stop here, since the last "problem" is just an illustration,
# you can't run it directly since there is no "special residue"
# definition.

if 0:

    #
    # Fourth problem: construct a protein with a non-standard residue.
    # The new residue's name is 'XXX' and its definition is stored in the
    # MMTK group database under the name 'special_residue'.
    #
    defineResidue('special_residue', 'xxx')
    sequence = PDBFile('special_protein.pdb').readSequenceWithConfiguration()
    # Make peptide chain and protein objects
    special_protein = Protein(map(PeptideChain, sequence))
