# This module implements classes for peptide chains and proteins.
#
# Written by Konrad Hinsen
# last revision: 1998-7-15
#

import Bonds, ChemicalObjects, Collection, ConfigIO, Database, Geometry
import Units, Universe, Utility
from Vector import Vector
from DictWithDefault import DictWithDefault
import operator, string, umath

#
# Residues are special groups
#
class Residue(ChemicalObjects.Group):

    def __init__(self, name = None, model = 'all'):
	if name is not None:
	    blueprint = _residueBlueprint(name, model)
	    ChemicalObjects.Group.__init__(self, blueprint)
	    self.model = model
	    self._init()

    def __setstate__(self, state):
	for attr, value in state.items():
	    self.__dict__[attr] = value
	try:
	    self.model = self.hydrogens
	except AttributeError:
	    pass
	self._init()

    def _init(self):
	# construct PDB map and alternative names
	type = self.type
	if not hasattr(type, 'pdbmap'):
	    pdb_dict = {}
	    type.pdbmap = [(type.symbol, pdb_dict)]
	    offset = 0
	    for g in type.groups:
		for name, atom in g.pdbmap[0][1].items():
		    pdb_dict[name] = Database.AtomReference(atom.number + \
							    offset)
		offset = offset + len(g.atoms)
	if not hasattr(type, 'pdb_alternative'):
	    alt_dict = {}
	    for g in type.groups:
		if hasattr(g, 'pdb_alternative'):
		    for key, value in g.pdb_alternative.items():
			alt_dict[key] = value
	    setattr(type, 'pdb_alternative', alt_dict)
	# create peptide attribute for calpha model
	if self.model == 'calpha':
	    self.peptide = self
	    self.CBond = self.C_alpha
	    self.NBond = self.C_alpha
	    self.C_alpha._mass = self.type._mass

    def _makeCystine(self):
	if self.model == 'calpha':
	    return self
	if string.lower(self.symbol) != 'cys':
	    raise ValueError, `self` + " is not cysteine."
	new_residue = 'cystine_ss'
	if hasattr(self, 'H_3'):
	    new_residue = new_residue + '_nt'
	elif hasattr(self, 'O_2'):
	    new_residue = new_residue + '_ct'
	new_residue = Residue(new_residue, self.model)
	for g in ['peptide', 'sidechain']:
	    g_old = getattr(self, g)
	    g_new = getattr(new_residue, g)
	    atoms = map(lambda a: a.name, getattr(g_new, 'atoms'))
	    for a in atoms:
		set_method = getattr(getattr(g_new, a), 'setPosition')
		set_method(getattr(getattr(g_old, a), 'position')())
	return new_residue

    def isSubsetModel(self):
	return self.model == 'calpha'

    def phiPsi(self, conf = None):
	universe = self.universe()
	if universe is None:
	    universe = Universe.InfiniteUniverse()
	C = None
	for a in self.peptide.N.bondedTo():
	    if a.parent != self.peptide:
		C = a
	if C is None:
	    phi = None
	else:
	    phi = universe.dihedral(self.peptide.C, self.peptide.C_alpha,
				    self.peptide.N, C, conf)
	N = None
	for a in self.peptide.C.bondedTo():
	    if a.parent != self.peptide:
		N = a
	if N is None:
	    psi = None
	else:
	    psi = universe.dihedral(N, self.peptide.C, self.peptide.C_alpha,
				    self.peptide.N, conf)
	return phi, psi

def _residueBlueprint(name, model):
    try:
	blueprint = _residue_blueprints[(name, model)]
    except KeyError:
	if model == 'polar':
	    name = name + '_uni'
	elif model == 'polar_charmm':
	    name = name + '_uni2'
	elif model == 'none':
	    name = name + '_noh'
	elif model == 'calpha':
	    name = name + '_calpha'
	blueprint = Database.BlueprintGroup(name)
	_residue_blueprints[(name, model)] = blueprint
    return blueprint

_residue_blueprints = {}

#
# Peptide chains are molecules with added features.
#
class PeptideChain(ChemicalObjects.Molecule):

    def __init__(self, sequence, **properties):
	if sequence is not None:
	    hydrogens = self.binaryProperty(properties, 'hydrogens', 'all')
	    if hydrogens == 1:
		hydrogens = 'all'
	    elif hydrogens == 0:
		hydrogens = 'none'
	    n_term = self.binaryProperty(properties, 'n_terminus', 1)
	    c_term = self.binaryProperty(properties, 'c_terminus', 1)
	    try:
		model = string.lower(properties['model'])
	    except KeyError:
		model = hydrogens
	    self.version_spec = {'hydrogens': hydrogens,
				 'n_terminus': n_term,
				 'c_terminus': c_term,
				 'model': model}
	    if type(sequence[0]) == type(''):
		conf = None
	    else:
		conf = sequence
		sequence = map(lambda r: r.name, sequence)
	    sequence = map(_fullName, sequence)
	    if model != 'calpha':
		if n_term:
		    sequence[0] = sequence[0] + '_nt'
		if c_term:
		    sequence[-1] = sequence[-1] + '_ct'
	    self.groups = []
	    n = 0
	    for residue in sequence:
		n = n + 1
		r = Residue(residue, model)
		r.name = r.symbol + `n`
		r.sequence_number = n
		r.parent = self
		self.groups.append(r)
	    self.atoms = []
	    self.bonds = []
	    for g in self.groups:
		self.atoms = self.atoms + g.atoms
		self.bonds = self.bonds + g.bonds
	    for i in range(len(self.groups)-1):
		try:
		    CBond = self.groups[i].peptide.C
		except AttributeError:
		    CBond = self.groups[i].CBond
		try:
		    NBond = self.groups[i+1].peptide.N
		except AttributeError:
		    NBond = self.groups[i+1].NBond
		self.bonds.append(Bonds.Bond((CBond, NBond)))
	    self.bonds = Bonds.BondList(self.bonds)
	    self.parent = None
	    self.type = None
	    self.configurations = {}

	    try:
		self.name = properties['name']
		del properties['name']
	    except KeyError:
		self.name = ''
	    if properties.has_key('configuration'):
		conf = properties['configuration']
		conf = ConfigIO.PDBFile(conf).readSequenceWithConfiguration()
		ConfigIO.PDBConfiguration(conf).applyTo(self, 1)
	    elif conf:
		ConfigIO.PDBConfiguration(conf).applyTo(self, 1)
	    if properties.has_key('position'):
		self.translateTo(properties['position'])
		del properties['position']
	    self.addProperties(properties)

    is_peptide_chain = 1

    def __len__(self):
	return len(self.groups)

    def __getitem__(self, item):
	return self.groups[item]

    def __getslice__(self, first, last):
	return SubChain(self, self.groups[first:last])

    def __setitem__(self, item, value):
	self.replaceResidue(self.groups[item], value)

    def residuesOfType(self, *types):
	types = map(string.lower, types)
	rlist = []
	for r in self.groups:
	    if string.lower(r.type.symbol) in types:
		rlist.append(r)
	return Collection.Collection(rlist)

    def backbone(self):
	return Collection.Collection(map(lambda r: r.peptide, self.groups))

    def sidechains(self):
	return Collection.Collection(map(lambda r: r.sidechain, self.groups))

    def residues(self):
	return Collection.Collection(self.groups)

    def sequence(self):
	return map(lambda r: string.lower(r.type.symbol), self.groups)

    def phiPsi(self, conf = None):
	universe = self.universe()
	if universe is None:
	    universe = Universe.InfiniteUniverse()
	angles = []
	for i in range(len(self)):
	    r = self[i]
	    if i == 0:
		phi = None
	    else:
		phi = universe.dihedral(r.peptide.C, r.peptide.C_alpha,
					r.peptide.N,
					self[i-1].peptide.C, conf)
	    if i == len(self)-1:
		psi = None
	    else:
		psi = universe.dihedral(self[i+1].peptide.N,
					r.peptide.C, r.peptide.C_alpha,
					r.peptide.N, conf)
	    angles.append((phi, psi))
	return angles

    def replaceResidue(self, r_old, r_new):
	n = self.groups.index(r_old)
	for a in r_old.atoms:
	    self.atoms.remove(a)
	for b in r_old.bonds:
	    self.bonds.remove(b)
	self.atoms = self.atoms + r_new.atoms
	self.bonds = self.bonds + r_new.bonds
	r_new.sequence_number = n+1
	r_new.name = r_new.symbol+`n+1`
	r_new.parent = self
	self.groups[n] = r_new
	if n > 0:
	    peptide_old = self.bonds.bondsOf(r_old.peptide.N)
	    self.bonds.remove(peptide_old[0])
	    self.bonds.append(Bonds.Bond((self.groups[n-1].peptide.C,
					  self.groups[n].peptide.N)))
	if n < len(self.groups)-1:
	    peptide_old = self.bonds.bondsOf(r_old.peptide.C)
	    self.bonds.remove(peptide_old[0])
	    self.bonds.append(Bonds.Bond((self.groups[n].peptide.C,
					  self.groups[n+1].peptide.N)))

    def setConfiguration(self, conf):
	ConfigIO.PDBConfiguration(conf).applyTo(self)

    # construct positions of missing hydrogens
    def findHydrogenPositions(self):
	self.setBondAttributes()
	try:
	    unknown = DictWithDefault([])
	    for a in self.atoms:
		if a.position() is None:
		    if a.symbol != 'H':
			raise ValueError, 'position of ' + a.fullName() + \
			      ' is undefined'
		    bonded = a.bondedTo()[0]
		    unknown[bonded].append(a)
	    for a, list in unknown.items():
		bonded = a.bondedTo()
		n = len(bonded)
		known = []
		for b in bonded:
		    if b.position() is not None:
			known.append(b)
		nb = len(list)
		if a.symbol == 'C':
		    if n == 4:
			if nb == 1:
			    self._C4oneH(a, known, list)
			elif nb == 2:
			    self._C4twoH(a, known, list)
			elif nb == 3:
			    self._C4threeH(a, known, list)
		    elif n == 3:
			self._C3oneH(a, known, list)
		    else:
			print a
			raise ValueError, "Can't handle C with "+`n`+" bonds"
		elif a.symbol == 'N':
		    if n == 4:
			if nb == 3:
			    self._N4threeH(a, known, list)
		    elif n == 3:
			if nb == 1:
			    self._N3oneH(a, known, list)
			elif nb == 2:
			    self._N3twoH(a, known, list)
		    else:
			print a
			raise ValueError, "Can't handle N with "+`n`+" bonds"
		elif a.symbol == 'O' and n == 2:
		    self._O2(a, known, list)
		elif a.symbol == 'S' and n == 2:
		    self._S2(a, known, list)
		else:
		    print a
		    raise ValueError, "Can't handle this yet: " + \
			  a.symbol + ' with ' + `n` + ' bonds (' + \
			  a.fullName() + ').'
	finally:
	    self.clearBondAttributes()

    # default C-H bond length and X-C-H angle
    _ch_bond = 1.09*Units.Ang
    _hch_angle = umath.arccos(-1./3.)*Units.rad
    _nh_bond = 1.03*Units.Ang
    _hnh_angle = 120.*Units.deg
    _oh_bond = 0.95*Units.Ang
    _coh_angle = 114.9*Units.deg
    _sh_bond = 1.007*Units.Ang
    _csh_angle = 96.5*Units.deg

    def _C4oneH(self, atom, known, unknown):
	r = atom.position()
	n0 = (known[0].position()-r).normal()
	n1 = (known[1].position()-r).normal()
	n2 = (known[2].position()-r).normal()
	n3 = (n0 + n1 + n2).normal()
	unknown[0].setPosition(r-self._ch_bond*n3)

    def _C4twoH(self, atom, known, unknown):
	r = atom.position()
	r1 = known[0].position()
	r2 = known[1].position()
	plane = Geometry.Plane(r, r1, r2)
	axis = -((r1-r)+(r2-r)).normal()
	plane = plane.rotate(Geometry.Line(r, axis), 90.*Units.deg)
	cone = Geometry.Cone(r, axis, 0.5*self._hch_angle)
	sphere = Geometry.Sphere(r, self._ch_bond)
	circle = sphere.intersectWith(cone)
	points = circle.intersectWith(plane)
	unknown[0].setPosition(points[0])
	unknown[1].setPosition(points[1])

    def _C4threeH(self, atom, known, unknown):
	self._tetrahedralH(atom, known, unknown, self._ch_bond)

    def _C3oneH(self, atom, known, unknown):
	r = atom.position()
	n1 = (known[0].position()-r).normal()
	n2 = (known[1].position()-r).normal()
	n3 = -(n1 + n2).normal()
	unknown[0].setPosition(r+self._ch_bond*n3)

    def _N3oneH(self, atom, known, unknown):
	r = atom.position()
	n1 = (known[0].position()-r).normal()
	n2 = (known[1].position()-r).normal()
	n3 = -(n1 + n2).normal()
	unknown[0].setPosition(r+self._nh_bond*n3)

    def _N3twoH(self, atom, known, unknown):
	r = atom.position()
	r1 = known[0].position()
	others = filter(lambda a: a.symbol != 'H', known[0].bondedTo())
	r2 = others[0].position()
	plane = Geometry.Plane(r, r1, r2)
	axis = (r-r1).normal()
	cone = Geometry.Cone(r, axis, 0.5*self._hnh_angle)
	sphere = Geometry.Sphere(r, self._nh_bond)
	circle = sphere.intersectWith(cone)
	points = circle.intersectWith(plane)
	unknown[0].setPosition(points[0])
	unknown[1].setPosition(points[1])

    def _N4threeH(self, atom, known, unknown):
	self._tetrahedralH(atom, known, unknown, self._nh_bond)

    def _O2(self, atom, known, unknown):
	others = known[0].bondedTo()
	r = others[0].position()
	dihedral = 180.*Units.deg
	self._findPosition(unknown[0], atom.position(), known[0].position(), r,
			   self._oh_bond, self._coh_angle, dihedral)

    def _S2(self, atom, known, unknown):
	c2 = filter(lambda a: a.symbol == 'C', known[0].bondedTo())[0]
	self._findPosition(unknown[0], atom.position(), known[0].position(),
			   c2.position(),
			   self._sh_bond, self._csh_angle,
			   180.*Units.deg)

    def _tetrahedralH(self, atom, known, unknown, bond):
	r = atom.position()
	n = (known[0].position()-r).normal()
	cone = Geometry.Cone(r, n, umath.arccos(-1./3.))
	sphere = Geometry.Sphere(r, bond)
	circle = sphere.intersectWith(cone)
	others = filter(lambda a: a.symbol != 'H', known[0].bondedTo())
	others.remove(atom)
	other = others[0]
	ref = (Geometry.Plane(circle.center, circle.normal) \
	       .projectionOf(other.position())-circle.center).normal()
	p0 = circle.center + circle.radius*ref
	p0 = Geometry.rotatePoint(p0,
				  Geometry.Line(circle.center, circle.normal),
				  60.*Units.deg)
	p1 = Geometry.rotatePoint(p0,
				  Geometry.Line(circle.center, circle.normal),
				  120.*Units.deg)
	p2 = Geometry.rotatePoint(p1,
				  Geometry.Line(circle.center, circle.normal),
				  120.*Units.deg)
	unknown[0].setPosition(p0)
	unknown[1].setPosition(p1)
	unknown[2].setPosition(p2)

    def _findPosition(self, unknown, a1, a2, a3, bond, angle, dihedral):
	sphere = Geometry.Sphere(a1, bond)
	cone = Geometry.Cone(a1, a2-a1, angle)
	plane = Geometry.Plane(a3, a2, a1)
	plane = plane.rotate(Geometry.Line(a1, a2-a1), dihedral)
	points = sphere.intersectWith(cone).intersectWith(plane)
	for p in points:
	    if (a1-a2).cross(p-a1)*(plane.normal) > 0:
		unknown.setPosition(p)
		break

    # add sulfur bridges between cysteine residues
    def _addSSBridges(self, bonds):
	for b in bonds:
	    cys1 = b[0]
	    if string.lower(cys1.symbol) == 'cyx':
		cys_ss1 = cys1
	    else:
		cys_ss1 = cys1._makeCystine()
		self.replaceResidue(cys1, cys_ss1)
	    cys2 = b[1]
	    if string.lower(cys2.symbol) == 'cyx':
		cys_ss2 = cys2
	    else:
		cys_ss2 = cys2._makeCystine()
		self.replaceResidue(cys2, cys_ss2)
	    self.bonds.append(Bonds.Bond((cys_ss1.sidechain.S_gamma,
					  cys_ss2.sidechain.S_gamma)))

    def _descriptionSpec(self):
	kwargs = ''
	for name, value in self.version_spec.items():
	    kwargs = kwargs + name + '=' + `value` + ','
	return "S", kwargs[:-1]

    def _typeName(self):
	return reduce(operator.add, self.sequence())

    def _graphics(self, conf, distance_fn, model, module, options):
	if model != 'backbone':
	    return ChemicalObjects.Molecule._graphics(self, conf,
						      distance_fn, model,
						      module, options)
	color = 'black'
	try: color = options['color']
	except KeyError: pass
	material = module.DiffuseMaterial(color)
	objects = []
	for i in range(len(self.groups)-1):
	    a1 = self.groups[i].peptide.C_alpha
	    a2 = self.groups[i+1].peptide.C_alpha
	    p1 = a1.position(conf)
	    p2 = a2.position(conf)
	    if p1 is not None and p2 is not None:
		bond_vector = 0.5*distance_fn(a1, a2, conf)
		cut = bond_vector != 0.5*(p2-p1)
		if not cut:
		    objects.append(module.Line(p1, p2, material = material))
		else:
		    objects.append(module.Line(p1, p1+bond_vector,
					       material = material))
		    objects.append(module.Line(p2, p2-bond_vector,
					       material = material))
	return objects

#
# Subchains are created by slicing chains or extracting a chain from
# a group of connected chains.
#
class SubChain(PeptideChain):

    def __init__(self, chain, groups, name = ''):
	self.groups = groups
	self.atoms = []
	self.bonds = []
	for g in self.groups:
	    self.atoms = self.atoms + g.atoms
	    self.bonds = self.bonds + g.bonds
	if chain.version_spec['model'] == 'calpha':
	    for i in range(len(self.groups)-1):
		self.bonds.append(Bonds.Bond((self.groups[i].C_alpha,
					      self.groups[i+1].C_alpha)))
	else:
	    for i in range(len(self.groups)-1):
		self.bonds.append(Bonds.Bond((self.groups[i].peptide.C,
					      self.groups[i+1].peptide.N)))
	self.bonds = Bonds.BondList(self.bonds)
	self.name = name
	self.parent = chain.parent
	self.type = None
	self.configurations = {}
	self.part_of = chain

    is_incomplete = 1

    def __repr__(self):
	if self.name == '':
	    return 'SubChain of ' + repr(self.part_of)
	else:
	    return ChemicalObjects.Molecule.__repr__(self)
    __str__ = __repr__


#
# Connected chains are collections of peptide chains connected by s-s bridges.
#
class ConnectedChains(PeptideChain):

    def __init__(self, chains):
	self.chains = []
	self.groups = []
	self.atoms = []
	self.bonds = Bonds.BondList([])
	self.chain_names = []
        self.version_spec = chains[0].version_spec
	for c in chains:
            if c.version_spec['model'] != self.version_spec['model']:
                raise ValueError, "mixing chains of different model: " + \
                      c.version_spec['model']+"/"+self.version_spec['model']
	    ng = len(self.groups)
	    self.chains.append((c.name, ng, ng+len(c.groups)))
	    self.groups = self.groups + c.groups
	    self.atoms = self.atoms + c.atoms
	    self.bonds = self.bonds + c.bonds
	    try: name = c.name
	    except AttributeError: name = ''
	    self.chain_names.append(name)
	for g in self.groups:
	    g.parent = self
	self.name = ''
	self.parent = None
	self.type = None
	self.configurations = {}

    is_connected_chains = 1

    def __len__(self):
	return len(self.chains)

    def __getitem__(self, item):
	c = self.chains[item]
	return SubChain(self, self.groups[c[1]:c[2]], c[0])

    def __getslice__(self, first, last):
	raise TypeError, "Can't slice connected chains"

    def _graphics(self, conf, distance_fn, model, module, options):
	if model != 'backbone':
	    return ChemicalObjects.Molecule._graphics(self, conf,
						      distance_fn, model,
						      module, options)
	objects = []
	for chain in self:
	    objects = objects + chain._graphics(conf, distance_fn,
						model, module, options)
	return objects

#
# Proteins are complexes of peptide chains, connected peptide chains,
# and possibly other things.
#
class Protein(ChemicalObjects.Complex):

    def __init__(self, *items, **properties):
	if items == (None,):
	    return
	self.name = ''
	if len(items) == 1 and type(items[0]) == type(''):
	    blueprint = Database.BlueprintProtein(items[0])
	    items = blueprint.chains
	    for attr, value in vars(blueprint).items():
		if attr not in ['type', 'chains']:
		    setattr(self, attr, value)
	molecules = []
	for i in items:
	    if ChemicalObjects.isChemicalObject(i):
		molecules.append(i)
	    else:
		molecules = molecules + list(i)
	for m, i in map(None, molecules, range(len(molecules))):
	    m._numbers = [i]
	    if not m.name:
		m.name = 'chain'+`i`
	ss = self._findSSBridges(molecules)
	new_mol = {}
	for m in molecules:
	    new_mol[m] = ([m],[])
	for bond in ss:
	    m1 = new_mol[bond[0].topLevelChemicalObject()]
	    m2 = new_mol[bond[1].topLevelChemicalObject()]
	    if m1 == m2:
		m1[1].append(bond)
	    else:
		combined = (m1[0] + m2[0], m1[1] + m2[1] + [bond])
		for m in combined[0]:
		    new_mol[m] = combined
	self.molecules = []
	while new_mol:
	    m = new_mol.values()[0]
	    for i in m[0]:
		del new_mol[i]
	    bonds = m[1]
	    if len(m[0]) == 1:
		m = m[0][0]
	    else:
		numbers = reduce(operator.add, map(lambda i: i._numbers, m[0]))
		m = ConnectedChains(m[0])
		m._numbers = numbers
	    m._addSSBridges(bonds)
	    m.parent = self
	    self.molecules.append(m)

	self.atoms = []
	self.chains = []
	for m in self.molecules:
	    self.atoms = self.atoms + m.atoms
	    if hasattr(m, 'is_connected_chains'):
		for c, name, i in map(None, range(len(m)),
				   m.chain_names, m._numbers):
		    self.chains.append((m, c, name, i))
	    else:
		try: name = m.name
		except AttributeError: name = ''
		self.chains.append((m, None, name, m._numbers[0]))
	self.chains.sort(lambda c1, c2: cmp(c1[3], c2[3]))
	self.chains = map(lambda c: c[:3], self.chains)

	self.parent = None
	self.type = None
	self.configurations = {}
	try:
	    self.name = properties['name']
	    del properties['name']
	except KeyError: pass
	if properties.has_key('position'):
	    self.translateTo(properties['position'])
	    del properties['position']
	self.addProperties(properties)

	undefined = 0
	for a in self.atoms:
	    if a.position() is None:
		undefined = undefined + 1
	if undefined > 0 and undefined != len(self.atoms):
	    Utility.warning('Some atoms in a protein have undefined positions.')

    is_protein = 1

    def __len__(self):
	return len(self.chains)

    def __getitem__(self, item):
	if type(item) == type(0):
	    m, c, name = self.chains[item]
	else:
	    for m, c, name in self.chains:
		if name == item:
		    break
	    if name != item:
		raise ValueError, 'No chain with name ' + item
	if c is None:
	    return m
	else:
	    return m[c]

    def residuesOfType(self, *types):
	rlist = Collection.Collection([])
	for m in self.molecules:
	    if isPeptideChain(m):
		rlist = rlist + apply(m.residuesOfType, types)
	return rlist

    def backbone(self):
	rlist = Collection.Collection([])
	for m in self.molecules:
	    if isPeptideChain(m):
		rlist = rlist + m.backbone()
	return rlist

    def sidechains(self):
	rlist = Collection.Collection([])
	for m in self.molecules:
	    if isPeptideChain(m):
		rlist = rlist + m.sidechains()
	return rlist

    def residues(self):
	rlist = Collection.Collection([])
	for m in self.molecules:
	    if isPeptideChain(m):
		rlist = rlist + m.residues()
	return rlist

    def phiPsi(self, conf = None):
	angles = []
	for chain in self:
	    angles.append(chain.phiPsi(conf))
	return angles

    _ss_bond_max = 2.5*Units.Ang

    def _findSSBridges(self, molecules):
	molecules = filter(lambda m: hasattr(m, 'is_peptide_chain'), molecules)
	cys = Collection.Collection([])
	for m in molecules:
	    if m.version_spec['model'] != 'calpha':
		cys = cys + m.residuesOfType('cys') + m.residuesOfType('cyx')
	s = cys.map(lambda r: r.sidechain.S_gamma)
	ns = len(s)
	ss = []
	for i in xrange(ns-1):
	    for j in xrange(i+1,ns):
		r1 = s[i].position()
		r2 = s[j].position()
		if r1 and r2 and (r1-r2).length() < self._ss_bond_max:
		    ss.append((cys[i], cys[j]))
	return ss

    def setConfiguration(self, conf):
	if type(conf) == type(''):
	    conf = ConfigIO.PDBFile(conf).readSequenceWithConfiguration()
	ConfigIO.PDBConfiguration(conf).applyTo(self)

    def _subunits(self):
	return list(self)

    def _description(self, tag, index_map, toplevel):
	if not toplevel:
	    raise ValueError
	return 'l(' + `self.__class__.__name__` + ',' + `self.name` + ',[' + \
	       reduce(operator.add,
		      map(lambda o, t=tag, m=index_map:
			         o._description(t, m, 1)+',',
			  self)) + '])'

    def _graphics(self, conf, distance_fn, model, module, options):
	if model != 'backbone':
	    return ChemicalObjects.Complex._graphics(self, conf, distance_fn,
						     model, module, options)
	objects = []
	for chain in self:
	    objects = objects + chain._graphics(conf, distance_fn,
						model, module, options)
	return objects

#
# Convenience function for constructing water molecules
# and unknown residues from PDB files.
#
def waterFromPDBSequence(sequence):
    pdb_conf = sequence.waterMolecules()
    water = Collection.Collection()
    for i in range(len(pdb_conf)):
	water.addObject(ChemicalObjects.Molecule('water'))
    ConfigIO.PDBConfiguration(pdb_conf).applyTo(water)
    return water

def unknownResiduesFromPDBSequence(sequence):
    import PDB
    pdb_conf = sequence.unknownResidues()
    atoms = Collection.Collection()
    for r in pdb_conf:
	for a in r:
	    if a.name[:2] in PDB.PDBFile.two_letter_elements:
		element = a.name[:2]
	    else:
		element = a.name[0]
	    atom = ChemicalObjects.Atom(element, name=a.name,
					position=a.position*Units.Ang)
	    try:
		atom.setIndex(a.number-1)
	    except (AttributeError, ValueError):
		pass
	    atoms.addObject(atom)
    return atoms

#
# Type check functions
#
def isPeptideChain(x):
    return hasattr(x, 'is_peptide_chain')

def isProtein(x):
    return hasattr(x, 'is_protein')

#
# Find the full name of a residue
#
def _fullName(residue):
    return _residue_names[string.lower(residue)]

_residue_names = {'ala': 'alanine',          'a': 'alanine',
		  'arg': 'arginine',         'r': 'arginine',
		  'asn': 'asparagine',       'n': 'asparagine',
		  'asp': 'aspartic_acid',    'd': 'aspartic_acid',
		  'cys': 'cysteine',         'c': 'cysteine',
		  'gln': 'glutamine',        'q': 'glutamine',
		  'glu': 'glutamic_acid',    'e': 'glutamic_acid',
		  'gly': 'glycine',          'g': 'glycine',
		  'his': 'histidine',        'h': 'histidine',
		  'ile': 'isoleucine',       'i': 'isoleucine',
		  'leu': 'leucine',          'l': 'leucine',
		  'lys': 'lysine',           'k': 'lysine',
		  'met': 'methionine',       'm': 'methionine',
		  'phe': 'phenylalanine',    'f': 'phenylalanine',
		  'pro': 'proline',          'p': 'proline',
		  'ser': 'serine',           's': 'serine',
		  'thr': 'threonine',        't': 'threonine',
		  'trp': 'tryptophan',       'w': 'tryptophan',
		  'tyr': 'tyrosine',         'y': 'tyrosine',
		  'val': 'valine',           'v': 'valine',
		  'cyx': 'cystine_ss',
		  'hsd': 'histidine_deltah', 'hse': 'histidine_epsilonh',
		  'hsp': 'histidine_plus',
		  'ace': 'ace_beginning',    'nme': 'nmethyl'}

#
# Add a residue to the residue list
#
def defineResidue(full_name, code3, code1 = None):
    if _residue_names.has_key(code3) or _residue_names.has_key(code1):
	raise ValueError, "residue name already defined"
    _residue_names[code3] = full_name
    if code1 is not None:
	_residue_names[code1] = full_name
    try:
	residue = ChemicalObjects.Group(full_name+'_noh')
    except IOError:
	residue = ChemicalObjects.Group(full_name)
    n = len(filter(lambda a: a.symbol != 'H', residue.atomList()))
    import PDB
    PDB.defineResidue(residue.symbol, n)
